<?php


namespace AdminMenu;

use Core\Core;
use Core\Datamodel\Dataset;
use Core\I18N\I18NLoader;
use Core\Router;
use Core\Templates\Template;
use PageModel;


class AdminMenuLauncher {
	private $_pages;
	private $_icons;
	private $_baseurls;
	private $_topLevelUrls;

	public function fetch() : string {
		$pageView = Router::$DefaultView;
		$groups = [];

		$this->_loadPages();

		if(count($this->_pages)) {
			$this->_loadIcons();
			$entries = $this->_processPages();

			foreach($entries as $pageData) {
				if(
					$pageData['parenturl'] &&
					!in_array($pageData['parenturl'], $this->_topLevelUrls)
				) {
					// This is a nested page, but no valid top-level parent located
					continue;
				}

				if(!isset($groups[ $pageData['group'] ])) {
					$groups[ $pageData['group'] ] = [
						'title'    => $pageData['group'],
						'href'     => '',
						'children' => [],
						'active'   => false,
					];
				}

				$group =& $groups[ $pageData['group'] ];
				$parent = $pageData['parenturl'];
				$base = $pageData['baseurl'];

				if($parent) {
					// Nested page; gets grouped onto a parent
					if(!isset($group['children'][ $parent ])) {
						$group['children'][ $parent ] = [
							'page' => null,
							'title' => null,
							'icon' => null,
							'active' => false,
							'children' => [],
						];
					}

					$group['children'][ $parent ]['children'][ $base ] = [
						'page' => $pageData['page'],
						'title' => $pageData['title'],
						'icon' => $pageData['icon'],
						'active' => $pageData['active'],
						'children' => [],
					];

					if($pageData['active']) {
						$group['children'][ $parent ]['active'] = true;
						$group['active'] = true;
					}
				}
				else {
					// Top-level page
					if(!isset($group['children'][ $base ])) {
						$group['children'][ $base ] = [
							'page' => null,
							'title' => null,
							'icon' => null,
							'active' => false,
							'children' => [],
						];
					}

					$group['children'][ $base ]['page'] = $pageData['page'];
					$group['children'][ $base ]['title'] = $pageData['title'];
					$group['children'][ $base ]['icon'] = $pageData['icon'];

					if($pageData['active']) {
						$group['children'][ $base ]['active'] = true;
						$group['active'] = true;
					}
				}
			}

			// Sorted top-level groups will be nice too.
			ksort($groups);
		}


		// Build a list of languages that can be set by the user.
		$locales = I18NLoader::GetLocalesEnabled();
		$selected = I18NLoader::GetUsersLanguage();
		$languages = [];
		if(sizeof($locales) > 1){
			// There is at least 1 language available on the system, YAY!
			foreach($locales as $localeKey => $localeDat){
				if(($pos = strpos($localeKey, '_')) !== false){
					// This locale contains an underscore, that means it has a corresponding country!
					// These are what we want to display to the end user.
					$country = substr($localeKey, $pos+1);

					// Here I am retrieving the language and dialect in the native dialect if at all possible.
					// This is because if you as a user only can read your native language and your browser renders something different,
					// then you want to be able to read what you're switching it to.
					$str1 = new \Core\i18n\I18NString($localeDat['lang']);
					$str1->setLanguage($localeKey);
					$localeTitle = $str1->getTranslation();
					if($localeDat['dialect']){
						$str2 = new \Core\i18n\I18NString($localeDat['dialect']);
						$str2->setLanguage($localeKey);
						$localeTitle .= ' (' . $str2->getTranslation() . ')';
					}

					$languages[] = [
						'key'      => $localeKey,
						'title'    => $localeTitle,
						'country'  => $country,
						'image'    => 'assets/images/iso-country-flags/' . strtolower($country) . '.png',
						'selected' => $localeKey == $selected,
					];
				}
			}
		}

		// Allow components to tap into the admin bar to add additional functionality.
		$additional = \Core\HookHandler::DispatchHook('/core/adminmenu');

		if(!$additional && count($groups) === 0) {
			// No links, no additional content, nothing to render.
			return '';
		}

		$template = Template::Factory('widgets/adminmenu/view');
		$template->assign('groups', $groups);
		$template->assign('languages', $languages);
		$template->assign('additional', $additional);
		$template->assign('controls', $pageView ? $pageView->controls : null);

		return $template->fetch();
	}

	/**
	 * Load the pages available to the user into memory
	 *
	 * @throws \Exception
	 */
	private function _loadPages() {
		$this->_pages = [];
		$this->_baseurls = [];
		$this->_topLevelUrls = [];

		if(\Core\user()) {
			if(isset($_SESSION['user_sudo'])) {
				$p = new PageModel('/user/sudo');
				$p->admin_group = 'SUDO';
				$p->title = 'Exit SUDO Mode';
				$this->_pages[] = $p;
				$this->_baseurls[] = '/user/sudo';
				$this->_topLevelUrls[] = '/user/sudo';
			}

			$pages = PageModel::Find(['admin' => 1]);
			foreach($pages as $page) {
				/** @var PageModel $page */

				if (
					\Core\user()->checkAccess($page->access) &&
					Core::IsComponentAvailable($page->component)
				) {
					$this->_pages[] = $page;
					$this->_baseurls[] = $page->baseurl;

					if(!$page->parenturl) {
						$this->_topLevelUrls[] = $page->baseurl;
					}
				}
			}
		}
	}

	/**
	 * Load all available page icons in bulk
	 */
	private function _loadIcons() {
		$this->_icons = [];

		$i = Dataset::Init()
			->select('baseurl', 'site', 'meta_value_title')
			->table('page_meta')
			->whereEquals('baseurl', $this->_baseurls)
			->whereEquals('meta_key', 'image')
			->executeAndGet();

		foreach($i as $row) {
			if(!isset($this->_icons[ $row['baseurl'] ])) {
				$this->_icons[ $row['baseurl'] ] = [];
			}

			$this->_icons[ $row['baseurl'] ][ $row['site'] ] = $row['meta_value_title'];
		}
	}

	private function _processPages() : array {
		$ret = [];
		$currentURL = Router::$DefaultBaseURL->getBaseURL();

		foreach($this->_pages as $page) {
			/** @var PageModel $page */

			$baseurl = $page->baseurl;
			$title = $page->title;
			// Support i18n here!
			if(strpos($title, 't:') === 0) {
				$title = t(substr($title, 2));
			}

			// Pages can define which sub-menu they get grouped under.
			// The 'Admin' submenu is the default.
			$group = $page->admin_group ?: 't:STRING_ADMIN';

			// Support i18n here!
			if(strpos($group, 't:') === 0) {
				$group = t(substr($group, 2));
			}

			$active = $baseurl === $currentURL;

			if(isset($this->_icons[ $baseurl ]) && isset($this->_icons[ $baseurl ][ $page->site ])) {
				$icon = $this->_icons[ $baseurl ][ $page->site ];
			}
			elseif(isset($this->_icons[ $baseurl ]) && isset($this->_icons[ $baseurl ][ -1 ])) {
				$icon = $this->_icons[ $baseurl ][ -1 ];
			}
			else {
				// Check the Component
				if($page->component) {
					$icon = Core::GetComponent($page->component)->getLogo();
				}
				else {
					$icon = null;
				}
			}

			$ret[] = [
				'baseurl' => $baseurl,
				'title' => $title,
				'active' => $active,
				'icon' => $icon,
				'group' => $group,
				'parenturl' => $page->parenturl,
				'page' => $page,
			];
		}

		// Ensure the results are sorted by their processed title.
		usort($ret, function($a, $b) {
			return $a['title'] > $b['title'];
		});

		return $ret;
	}

	public static function HookPageRender() {
		$pageView = Router::$DefaultView;

		if($pageView && $pageView->mode === \Core\View::MODE_PAGE) {
			// Set the default state if set.
			if(isset($_COOKIE['ADMINBAR']) && $_COOKIE['ADMINBAR'] === 'closed') {
				$pageView->bodyclasses[] = 'adminbar-closed';
			}

			$widget = new self();
			$content = $widget->fetch();

			if($content) {
				$pageView->bodyclasses[] = 'has-adminbar';
				$pageView->appendBodyContent($content);
				$pageView->addStylesheet('assets/css/adminmenu.css');
			}
		}
	}
}