<?php
/**
 * Model for the gallery images themselves.
 *
 * @package Gallery
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2012-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Filestore\Factory;
use Core\Filestore\File;
use Core\Model;
use Core\Parser\ImageParser;


class GalleryImageModel extends Model {
	public static $Schema = array(
		'id' => array(
			'type' => Model::ATT_TYPE_ID,
			'required' => true,
			'null' => false,
		),
		'albumid' => array(
			'type' => Model::ATT_TYPE_INT,
			'required' => true,
			'null' => false,
			'formtype' => 'system'
		),
		'uploaderid' => array(
			'type' => Model::ATT_TYPE_UUID_FK,
			'formtype' => 'disabled',
			'comment' => 'User id of the uploading user',
		),
		'weight' => array(
			'type' => Model::ATT_TYPE_INT,
			'formtype' => 'hidden',
		),
		'file' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'required' => true,
			'form' => array(
				'type' => 'file',
				'basedir' => 'public/galleryalbum', // Defaults, but may be changed in the controller.
				'accept' => 'image/*', // Defaults, but may be changed in the controller.
			),
		),
		'title' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'formtype' => 'disabled',
		),
		'location' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => '64',
			'form' => array(
				'title' => 'Location taken',
				'description' => 'Where was this taken?',
				'type' => 'disabled', // Candidate for removal?
			)
		),
		'latitude' => array(
			'type' => Model::ATT_TYPE_FLOAT,
			'formtype' => 'hidden'
		),
		'longitude' => array(
			'type' => Model::ATT_TYPE_FLOAT,
			'formtype' => 'hidden'
		),
		'datetaken' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => '64',
			'form' => array(
				'title' => 'Date taken',
				'description' => 'When was this taken?',
				'type' => 'disabled', // Candidate for removal?
			)
		),
		'previewfile' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'required' => false,
			'form' => array(
				'title' => 'Preview Image',
				'type' => 'hidden', // Hidden by default, (enabled for non-image uploads)
				'basedir' => 'public/galleryalbum', // Defaults, but may be changed in the controller.
				'accept' => 'image/*',
				'description' => 'Upload an image to use as the preview version of this file.',
			),
			'comment' => 'For videos and non-images, this is the preview image',
		),
		'previewsize' => array(
			'type' => Model::ATT_TYPE_ENUM,
			'options' => array('xs', 'sm', 'med', 'lg', 'xl'),
			'default' => 'med',
			'form' => array(
				'title' => 'Preview Size',
				'description' => 'The preview size of this item on the listing page',
			)
		),
		/*'keywords' => array(
			'type' => Model::ATT_TYPE_TEXT,
		),*/
		/*'description' => array(
			'type' => Model::ATT_TYPE_TEXT,
		),*/
		'rotation' => array(
			'type' => Model::ATT_TYPE_INT,
			'formtype' => 'hidden',
			'default' => 0
		),
		'exifdata' => array(
			'type' => Model::ATT_TYPE_DATA,
			'formtype' => 'disabled',
			'encoding' => Model::ATT_ENCODING_JSON,
			'comment' => 'Exif data from the photo, retrieved automatically'
		),
		'created' => array(
			'type' => Model::ATT_TYPE_CREATED,
			'null' => false,
		),
		'updated' => array(
			'type' => Model::ATT_TYPE_UPDATED,
			'null' => false,
		),
	);

	public static $Indexes = array(
		'primary' => array('id'),
		'unique:albumid_file' => array('albumid', 'file'), // Each image should be in each album at most once.
	);

	/**
	 * Cache of the file object, so consecutive calls to getFile will not re-request the entire file.
	 *
	 * @var null|\Core\Filestore\File
	 */
	private $_file = null;


	public function __construct($key = null) {
		$this->_linked = array(
			'Page' => array(
				'link' => Model::LINK_HASONE,
				'on' => 'baseurl',
			),
			'GalleryAlbum' => array(
				'link' => Model::LINK_BELONGSTOONE,
				'on' => array('id' => 'albumid'),
			)
		);

		parent::__construct($key);
	}

	public function set(string $k, $v){
		switch($k){
			case 'file':
				parent::set($k, $v);
				// Clear out the cache
				$this->_file = null;
				// File was updated... load the exif data too!
				// Note, only do this if it was an image!
				$file = $this->getOriginalFile();
				if($file->isImage() && ($file->getExtension() == 'jpg')){
					parent::set('exifdata', exif_read_data($file->getFilename()));
				}
				else{
					parent::set('exifdata', null);
				}

				// Also if the file is new and it didn't exist... set the uploader id.
				if(!$this->_exists){
					parent::set('uploaderid', \Core\user()->get('id'));
				}
				break;
			case 'exifdata':
				// exif data cannot be changed externally!
				break;
			default:
				parent::set($k, $v);
		}
	}

	public function get($k, $format = null){
		switch($k){
			case 'baseurl':
			case 'rewriteurl':
				return $this->getRewriteURL();
			default:
				return parent::get($k, $format);
		}
	}

	/**
	 * Get the preview file associated to this image/item.
	 *
	 * This is required because albeit images are their own previews, videos, music and other files do not have a useful preview version.
	 * As such, if there was one uploaded, use that image instead of the mimetype image.
	 *
	 * @return File
	 */
	public function getPreviewFile() : File {
		if($this->getOriginalFile()->isImage()){
			return $this->getFile();
		}
		elseif($this->get('previewfile')){
			$file = Factory::File($this->get('previewfile'));
			return $file;
		}
		else{
			return $this->getFile();
		}
	}

	/**
	 * Get the file associated to this image.
	 *
	 * This has an interesting function; it will return the original filename or the rotated version.
	 * This rotated version is stored in a different directory to prevent name conflicts.
	 *
	 * @return File
	 */
	public function getFile() : File {
		if($this->_file !== null) {
			return $this->_file;
		}

		if($this->get('rotation') == 0){
			// Simple enough :p
			$this->_file = $this->getOriginalFile();
		}
		else{
			$filename = $this->get('file');
			$ext = substr($filename, strrpos($filename, '.'));
			$base = substr($filename, 0, 0 -strlen($ext));
			$rotatedfilename = $base . '-deg' . $this->get('rotation') . $ext;

			// Since rotated files are all temporary, (but need to be publicly visible)...
			$tmpfile = Factory::File('public/tmp/galleryalbum/' . $rotatedfilename);

			if(!($tmpfile->exists() && $tmpfile->getFilesize() > 0)){

				// I need physical write access, /tmp provides that!
				$tmptmp = Factory::File('tmp/galleryalbum/' . $rotatedfilename);

				// Write a small amount of data to the file so the subsystem will create the directories and everything.
				// Otherwise, if /tmp/galleryalbum doesn't exist, the image operations won't be able to write to it.
				$tmptmp->putContents('');

				// Rotate it!
				$originallocal = $this->getOriginalFile()->getLocalFilename();

				switch(strtolower($ext)){
					case '.jpg':
					case '.jpeg':
						$imagedat   = imagecreatefromjpeg($originallocal);
						$rotateddat = imagerotate($imagedat, $this->get('rotation'), 0);
						imagejpeg($rotateddat, $tmptmp->getFilename());
						$tmptmp->copyTo($tmpfile, true);
						$tmptmp->delete();
						break;
					case '.png':
						$imagedat   = imagecreatefrompng($originallocal);
						$rotateddat = imagerotate($imagedat, $this->get('rotation'), 0);
						imagepng($rotateddat, $tmptmp->getFilename());
						$tmptmp->copyTo($tmpfile, true);
						$tmptmp->delete();
						break;
					case '.gif':
						$imagedat   = imagecreatefromgif($originallocal);
						$rotateddat = imagerotate($imagedat, $this->get('rotation'), 0);
						imagegif($rotateddat, $tmptmp->getFilename());
						$tmptmp->copyTo($tmpfile, true);
						$tmptmp->delete();
						break;
					default:
						return $this->getOriginalFile();
				}
			}

			$this->_file = $tmpfile;
		}

		return $this->_file;
	}

	/**
	 * Get the original file associated to this image.
	 *
	 * This is critical because the getFile may return the *rotated* image!
	 *
	 * @return File
	 */
	public function getOriginalFile() : File {
		return Factory::File($this->get('file'));
	}

	/**
	 * Get the formatted rewrite URL of this image, with all SEO goodness.
	 *
	 * @return string
	 */
	public function getRewriteURL() : string {
		// Use the Constructor to benefit from caching when calling this multiple times from a listing page.
		$parent = GalleryAlbumModel::Construct($this->albumid);
		if($parent->checkExists()) {
			$link = $parent->get('rewriteurl') . '/';
		}
		else {
			$link = '';
		}


		$link .= $this->get('id');
		if($this->get('title')){
			$link .= '_' . \Core\str_to_url($this->get('title'));
		}

		return $link;
	}

	/**
	 * Get the exif data of this image as an associative array.
	 *
	 * @return ImageParser|null
	 */
	public function getExif() : ?ImageParser {
		$contents = $this->getOriginalFile()->getContentsObject();

		if($contents instanceof \Core\Filestore\Contents\ContentImage) {
			$exif = $contents->getExif();
		}
		elseif($contents instanceof \Core\Filestore\Contents\ContentSVG) {
			$exif = $contents->getExif();
		}
		else {
			$exif = null;
		}


		// This is meant to be publicly visible, so hide some sensitive information.
		if($exif !== null) {
			foreach($exif->fields as $k => $field) {
				if(
					$field instanceof \Core\Parser\JPEGExif\Tags\GPSLongitudeTag ||
					$field instanceof \Core\Parser\JPEGExif\Tags\GPSLatitudeTag
				) {
					unset($exif->fields[$k]);
				}
			}
		}

		return $exif;
	}

	/**
	 * Get an english description for the type of this file.
	 *
	 * @return string image|video|audio|file
	 */
	public function getFileType(){
		$m = $this->getOriginalFile()->getMimetype();
//var_dump($m, $this->getOriginalFile()); die();
		// I just want the beginning part
		$prefix = substr($m, 0, strpos($m, '/'));

		// There are a few exceptions.
		if($m == 'application/ogg') return 'audio';

		switch($prefix){
			case 'audio':
			case 'image':
			case 'video':
				return $prefix;
			default:
				return 'file';
		}
	}

	public function getFileExtension(){
		return $this->getOriginalFile()->getExtension();
	}
}
