<?php
/**
 * File for class MediaManagerNavigatorController definition in the coreplus project
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20130603.1641
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\Controller;
use Core\Filestore\Directory;
use Core\Filestore\Factory;
use Core\Filestore\FileMetaHelper;
use Core\Filestore\File;
use Core\Forms\Form;
use Core\View;
use MediaManager\Navigator;
use MediaManager\Navigator2;
use Core\AccessDeniedException;
use Core\BadRequestException;
use Core\NotFoundException;
use Core\TemporaryFollowRedirectException;


/**
 * Provides a nautilus/explorer UI of the public files currently on the server.
 *
 * These are files that can be managed by super admins anyway.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for MediaManagerNavigatorController
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class MediaManagerNavigatorController extends  Controller {

	/**
	 * Main view for the navigator.
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function index() {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$navigators = Navigator2::GetAccessableNavigators();
		
		if(!count($navigators)){
			throw new AccessDeniedException();
		}
		
		if($request->getParameter('ajax')) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
			$view->mastertemplate = false;
		}
		elseif($request->getParameter('frame')) {
			$view->mastertemplate = 'blank';
		}
		else{
			$view->mastertemplate = 'admin';
		}

		switch($request->getParameter('selectcallback')) {
			case 'tinymce':
				$selectCallback = 'tinymce';
				break;
			default:
				$selectCallback = null;
		}
		
		$view->title = 't:STRING_MEDIA_FILE_BROWSER';
		$view->assign('navigators', $navigators);
		$view->assign('select_callback', $selectCallback);
	}

	/**
	 * Loads only the page templates navigator
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function templates() {
		$view = $this->getView();
		$request = $this->getRequest();

		$navigators = [];
		foreach(Navigator2::GetAccessableNavigators() as $nav) {
			// Only load templates.
			if($nav->getType() === 'page') {
				$navigators[] = $nav;
			}
		}

		if(!count($navigators)) {
			throw new AccessDeniedException();
		}

		if($request->getParameter('ajax')) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
			$view->mastertemplate = false;
		}
		elseif($request->getParameter('frame')) {
			$view->mastertemplate = 'blank';
		}
		else{
			$view->mastertemplate = 'admin';
		}

		$view->templatename = 'pages/mediamanagernavigator/index';
		$view->title = 't:STRING_MEDIAMANAGER_TEMPLATES';
		$view->assign('navigators', $navigators);
		$view->assign('select_callback', null);
	}

	/**
	 * Loads only the site assets navigator
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function assets() {
		$view = $this->getView();
		$request = $this->getRequest();

		$navigators = [];
		foreach(Navigator2::GetAccessableNavigators() as $nav) {
			// Only load templates.
			if($nav->getType() === 'assets') {
				$navigators[] = $nav;
			}
		}

		if(!count($navigators)) {
			throw new AccessDeniedException();
		}

		if($request->getParameter('ajax')) {
			$view->mode = View::MODE_AJAX;
			$view->contenttype = View::CTYPE_JSON;
			$view->mastertemplate = false;
		}
		elseif($request->getParameter('frame')) {
			$view->mastertemplate = 'blank';
		}
		else{
			$view->mastertemplate = 'admin';
		}

		$view->templatename = 'pages/mediamanagernavigator/index';
		$view->title = 't:STRING_MEDIAMANAGER_ASSETS';
		$view->assign('navigators', $navigators);
		$view->assign('select_callback', null);
	}
	
	/**
	 * File editor for a given path
	 * 
	 * @param string $type Type of browser
	 * @param string $path Path of file or directory to query
	 * 
	 * @throws AccessDeniedException
	 */
	public function file_editor(string $type, string $path){
		$view = $this->getView();
		$request = $this->getRequest();
		
		// Load the requested navigator; this will handle the bulk of the heavy lifting.
		$nav = Navigator2::Factory($type);
		
		// Ensure user permissions
		if(!$nav->checkCanUpdate()){
			throw new AccessDeniedException();
		}

		$dir = dirname($path);
		if($dir === '.') {
			$nav->setCWD('');
		}
		else {
			$nav->setCWD($dir . '/');
		}
		
		// Retrieve the File object for the requested path.
		$entry = $nav->getEntry($path);
		
		if($entry === null) {
			throw new NotFoundException();
		}
		
		if(!($entry instanceof File)) {
			throw new BadRequestException();
		}

		//$info = $nav->getInfo([$path]);
		//var_dump($info, $entry, $entry->getMimetype()); die();

		$nav->loadEditor($view, $path);
		

		$view->mastertemplate = 'admin';
		$view->addBreadcrumb($nav->getTitle(), $nav->getURL());
		$view->title = $entry->getFilename('');
	}
	
	
	/**
	 * Get a listing of all files and directories contained inside a given media type
	 *
	 * Also return the base permissions for this navigator
	 * 
	 * @param string $type Type of browser
	 * @throws AccessDeniedException
	 */
	public function files(string $type, string $cwd) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		
		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}

		$nav->setCWD($cwd);
		
		$view->jsondata = [
			'results' => $nav->getListing(),
			'permissions' => $nav->getPermissions(),
		];
	}

	/**
	 * Get a listing of all files and directories contained inside a given media type
	 *
	 * Results are returned as JSON
	 *
	 * @param string $type Type of browser
	 * @throws AccessDeniedException
	 */
	public function permissions(string $type, string $cwd) {
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$nav = Navigator2::Factory($type);

		$nav->setCWD($cwd);

		$view->jsondata = [
			'permissions' => $nav->getPermissions(),
		];
	}
	
	/**
	 * Get additional information on a subset of files/paths, usually in a specific directory.
	 * 
	 * The paths must be a JSON array to prevent Firefox's No Script plugin from freaking out.
	 * 
	 * Results are returned as JSON
	 * 
	 * @param string $type  Type of browser
	 * @param string $paths JSON and HTML encoded string of paths to return
	 * @throws AccessDeniedException
	 */
	public function file_info(string $type, string $paths) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		
		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}
		
		$view->jsondata = [
			'results' => $nav->getInfo(json_decode(html_entity_decode($paths), true))
		];
	}
	
	/**
	 * Get full details for a given path.
	 * 
	 * Results are returned as an HTML string
	 * 
	 * @param string $type Type of browser
	 * @param string $cwd
	 * @param string $path Path of file or directory to query
	 * 
	 * @throws AccessDeniedException
	 */
	public function file_details(string $type, string $cwd, string $path) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->mastertemplate = false;
		
		$nav = Navigator2::Factory($type);
		
		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}

		$nav->setCWD($cwd);
		
		$entry = $nav->getEntry($path);
		if(!$entry) {
			throw new NotFoundException();
		}

		$template = \Core\Templates\Template::Factory('includes/mediamanagernavigator/file_details');
		$template->assign('entry', $entry);
		$template->assign('details', $nav->getEntryDetails($path));

		$view->jsondata = [
			'body' => $template->fetch(),
			'permissions' => $nav->getPermissions($path),
		];
	}
	
	/**
	 * Upload a new file to the given directory
	 * 
	 * @param string $type Type of navigator for the destination
	 * @param string $path Directory to upload file to
	 * 
	 * @throws AccessDeniedException
	 */
	public function file_upload(string $type, string $path) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		$nav->setCWD($path);
		
		if(!$nav->checkCanUpload()){
			throw new AccessDeniedException();
		}
		
		foreach($_FILES as $file){
			$nav->setUploadedFile($file['name'], $file['tmp_name']);
		}
		
		$view->jsondata = ['status' => 1, 'message' => 'Uploaded successfully'];
	}

	/**
	 * Main view for the navigator.
	 *
	 * Handle both list and thumbnail views.
	 *
	 * @return int
	 */
	public function image(){
		$view = $this->getView();
		$request = $this->getRequest();

		$navigator = new Navigator();
		$navigator->setView($view);
		$navigator->setMode($request->getParameter('mode'));
		$navigator->setAccept('image');
		if($request->getParameter('ajax')){
			$navigator->setBaseURL('/mediamanagernavigator/image?ajax=1');
			$view->mastertemplate = false;
			$view->record = false;
			$view->mode = View::MODE_AJAX;
		}
		elseif($request->getParameter('iframe')){
			$navigator->setBaseURL('/mediamanagernavigator/image?iframe=1');
			$view->mastertemplate = 'blank.tpl';
			$view->record = false;
		}
		else{
			$navigator->setBaseURL('/mediamanagernavigator/image');
			$view->mastertemplate = 'admin';
			//$view->mode = View::MODE_PAGE;
		}

		try{
			$navigator->cd($request->getParameter('dir'));
			if($request->getParameter('controls') !== null){
				$navigator->usecontrols = ($request->getParameter('controls') == 1);
			}
			if($request->getParameter('uploader') !== null){
				$navigator->useuploader = ($request->getParameter('uploader') == 1);
			}
		}
		catch(Exception $e){
			\Core\set_message($e->getMessage(), 'error');
		}

		$view->title = 'Images';
		$navigator->render();
		//var_dump($navigator, $navigator->render()); die();
	}


	/**
	 * Helper function for the directory mkdir command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 *
	 * @core-request-method POST
	 *
	 * @throws BadRequestException
	 */
	public function directory_mkdir(string $type, string $cwd, string $path) {
		$view = $this->getView();
		$request = $this->getRequest();
		
		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		
		$nav = Navigator2::Factory($type);
		
		if(!($nav->checkHasAccess() && $nav->checkCanUpload())){
			throw new AccessDeniedException();
		}

		try{
			$nav->setCWD($cwd);
			$newPath = $nav->mkdir($path);
			$view->jsondata = array('status' => 1, 'message' => 'Created directory successfully', 'path' => $newPath);
		}
		catch(Exception $e){
			$view->jsondata = ['status' => 0, 'message' => $e->getMessage()];
		}
	}

	/**
	 * Helper function for the directory rename command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 */
	public function directory_rename(){
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$navigator = new Navigator();
		$navigator->setView($view);

		if(!$request->isPost()){
			$view->jsondata = array('status' => 0, 'message' => 'Invalid request, please ensure it is a POST.');
			return;
		}

		if(!isset($_POST['newdir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Please enter a directory name');
			return;
		}

		if(!isset($_POST['olddir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Previous directory is required!');
			return;
		}

		if(!trim($_POST['newdir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Please enter a directory name');
			return;
		}

		try{
			$navigator->cd($request->getPost('dir'));
			$navigator->rename($_POST['olddir'], $_POST['newdir']);
		}
		catch(Exception $e){
			$view->jsondata = ['status' => 0, 'message' => $e->getMessage()];
			return;
		}

		$view->jsondata = array('status' => 1, 'message' => 'Renamed directory successfully');
	}

	/**
	 * Helper function for the file rename command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 */
	public function file_rename(){
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$navigator = new Navigator();
		$navigator->setView($view);

		if(!$request->isPost()){
			$view->jsondata = array('status' => 0, 'message' => 'Invalid request, please ensure it is a POST.');
			return;
		}

		if(!isset($_POST['newdir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Please enter a file name');
			return;
		}

		if(!isset($_POST['olddir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Previous file is required!');
			return;
		}

		if(!trim($_POST['newdir'])){
			$view->jsondata = array('status' => 0, 'message' => 'Please enter a file name');
			return;
		}

		try{
			$navigator->cd($request->getPost('dir'));
			$navigator->rename($_POST['olddir'], $_POST['newdir']);
		}
		catch(Exception $e){
			$view->jsondata = ['status' => 0, 'message' => $e->getMessage()];
			return;
		}

		$view->jsondata = array('status' => 1, 'message' => 'Renamed file successfully');
	}

	/**
	 * Helper function for the file touch command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 *
	 * @core-request-method POST
	 *
	 * @throws BadRequestException
	 */
	public function file_touch(string $type, string $cwd, string $name) {
		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;

		$nav = Navigator2::Factory($type);

		if(!($nav->checkHasAccess() && $nav->checkCanUpload())){
			throw new AccessDeniedException();
		}

		try{
			$nav->setCWD($cwd);
			$newPath = $nav->touch($name);
			$view->jsondata = array('status' => 1, 'message' => 'Created file successfully', 'path' => $newPath);
		}
		catch(Exception $e){
			$view->jsondata = ['status' => 0, 'message' => $e->getMessage()];
		}
	}

	/**
	 * Helper function for the file delete command.
	 *
	 * Returns JSON data and is expected to be called via ajax.
	 */
	public function file_delete(string $type) {

		$view = $this->getView();
		$request = $this->getRequest();

		$view->mode = View::MODE_AJAX;
		$view->contenttype = View::CTYPE_JSON;
		$view->mastertemplate = false;

		$nav = Navigator2::Factory($type);

		if(!$nav->checkHasAccess()){
			throw new AccessDeniedException();
		}

		if(!$nav->checkCanUpload()) {
			throw new AccessDeniedException();
		}

		if(!$request->isPost()){
			throw new BadRequestException();
		}

		$entry = $nav->getEntry($request->getPost('path'));

		if($entry === null) {
			throw new NotFoundException();
		}

		$entry->delete();

		$view->jsondata = array('status' => 1, 'message' => 'Deleted entry successfully');
	}
}
