<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace MediaManager\Navigators;

use CodeMirror\Utils as CodeMirrorUtils;
use Core\AccessDeniedException;
use Core\BadRequestException;
use Core\Component;
use Core\Filestore\File;
use Core\Forms\Form;
use Core\ModelValidationException;
use Core\NotFoundException;
use Core\View;
use Exception;
use Core\ConfigHandler;
use Core\Core;
use Core\Filestore\Factory as FileFactory;
use Core\Theme;
use MediaManager\Navigator2;

/**
 * Description of AssetsNavigator
 *
 * @author charlie
 */
class AssetsNavigator extends Navigator2 {
	public function __construct() {
		$this->_type      = 'assets';
		$this->_isAdmin   = \Core\user()->checkAccess('g:admin');
		// Disabled for now until we can finish tuning assets on the system.
		$this->_canUpload = false;
		$this->_canUpdate = $this->_isAdmin;
		$this->_canAccess = $this->_isAdmin;
		$this->_basedir   = 'assets/';
	}
	
	public function getTitle() : string {
		return t('STRING_MEDIAMANAGER_ASSETS');
	}

	/**
	 * Get the URL of the respective navigator
	 */
	public function getURL() : string {
		return '/mediamanagernavigator/assets';
	}
	
	/**
	 * Get an array of file and directory listings for this lookup
	 * 
	 * @return array
	 */
	public function getListing() : array {
		if(!$this->_canAccess){
			// No access, no files.
			return [];
		}
		
		$this->_listings = [];
		
		// The source objects to look for assets in.
		// Set initially to all the installed components.
		$assetsources = Core::GetComponents();
		// And add on the current theme.
		$assetsources[] = Theme::GetTheme();
		
		$dirs = [];
		
		foreach($assetsources as $source) {
			/** @var Component $source */
			
			$componentAssets = $source->getAssets();
			
			foreach($componentAssets as $asset) {
				// All assets start with "assets/"
				$rel    = substr($asset->filename, 7);
				$entry  = $asset->getDestination();
				$parent = dirname($rel) . '/';
				
				if(
					\Core\str_ends_with($rel, '.css.map') ||
					\Core\str_ends_with($rel, '.min.css') ||
					\Core\str_ends_with($rel, '.min.js')
				) {
					// Assets that match these should be ignored as they're not useful to the end user.
					continue;
				}
				
				if($parent === './'){
					$parent = '';
				}
				
				if(!isset($dirs[ $parent ])){
					$dirs[ $parent ] = filemtime($entry->getDirectoryName());
				}
				
				$this->_addFileToListings($entry, $rel, $parent);
			}
		}
		
		// dirs contains a pseudo directory structure and the number of files contained.
		foreach($dirs as $rel => $mtime){
			$parent = dirname($rel) . '/';
			
			if($parent === './'){
				$parent = '';
			}
			
			$this->_addDirectoryToListings($rel, $parent, $mtime);
		}
		
		// Sorting would be nice, just because.
		ksort($this->_listings, SORT_STRING | SORT_FLAG_CASE);
		
		return array_values($this->_listings);
	}
	
	/**
	 * Make a directory in the current working directory.
	 *
	 * @param $directory
	 *
	 * @throws \Exception
	 */
	public function mkdir(string $directory) : string {
		// Assets do not support creating new directories.
		throw new Exception('Modify access denied!');
	}

	public function loadEditor(View $view, string $path) {
		$entry = $this->getEntry($path);

		if( $entry->isText() ) {
			return $this->loadCodeEditor($view, $path);
		}
		else {
			$view->templatename = 'pages/mediamanagernavigator/file_editor/assets_other';
		}
	}

	public function loadCodeEditor(View $view, string $path) {
		$entry = $this->getEntry($path);

		switch($entry->getMimetype()) {
			case 'text/css':
				$mode = 'css';
				break;
			case 'text/javascript':
				$mode = 'javascript';
				break;
			default:
				$mode = 'unknown';
				break;
		}

		$form = new Form();
		$form->set('callsmethod', '\\MediaManager\\Navigators\\AssetsNavigator::CodeSaveHandler');

		$form->addElement('system', ['name' => 'type', 'value' => $this->_type]);
		$form->addElement('system', ['name' => 'path', 'value' => $path]);

		$form->addElement(
			'textarea',
			[
				'name' => 'content',
				'value' => $entry->getContents(),
				'rows' => 20,
			]
		);

		// Include the correct type of CodeMirror for this file, only applicable for text files.
		if(Core::IsComponentAvailable('codemirror') && $mode !== 'unknown') {
			CodeMirrorUtils::IncludeMode($mode);
		}


		$form->addElement('submit', ['value' => 't:STRING_SAVE']);

		$view->templatename = 'pages/mediamanagernavigator/file_editor/assets_code';
		$view->assign('form', $form);
		$view->assign('file', $entry);
		$view->assign('mode', $mode);
	}

	public static function CodeSaveHandler(Form $form) {

		$type = $form->getElementValue('type');
		$path = $form->getElementValue('path');
		$contents = $form->getElementValue('content');


		$nav = new self();
		if($nav->getType() !== $type) {
			throw new ModelValidationException('Incorrect navigator type; this can only accept Assets');
		}

		// Ensure user permissions
		if(!$nav->checkCanUpdate()){
			throw new AccessDeniedException();
		}

		// Retrieve the File object for the requested path.
		$entry = $nav->getEntry($path);

		if($entry === null) {
			throw new NotFoundException();
		}

		if(!($entry instanceof File)) {
			throw new BadRequestException();
		}

		// Grab the relative filename and adjust it to fit.
		$relativeFilename = $entry->getFilename(false);
		// Convert "asset" to "assets", this is the directory where it'll reside
		$relativeFilename = 'assets' . substr($relativeFilename, 5);

		// Resolve this to a custom location.
		$custom = FileFactory::File(ROOT_PDIR . 'themes/custom/' . $relativeFilename);
		$custom->putContents($contents);

		// Set the minified version as well.
		// This will not have any minification, but will at least force the site to load the correct version.
		if($entry->getExtension() === 'css') {
			$min = FileFactory::File(ROOT_PDIR . 'themes/custom/' . substr($relativeFilename, 0, -4) . '.min.css');
			$min->putContents($contents);
		}
		elseif($entry->getExtension() === 'js') {
			$min = FileFactory::File(ROOT_PDIR . 'themes/custom/' . substr($relativeFilename, 0, -3) . '.min.js');
			$min->putContents($contents);
		}

		// Force a reinstallation of all assets on the site to resync everything.
		// This would be inconvenient if the admin changed something and didn't see the change immediately.
		foreach(Core::GetComponents() as $component){
			$component->installAssets();
		}
		Theme::GetTheme()->installAssets();

		\Core\set_message('Updated asset successfully', 'success');
		return 'BACK';
	}
}
