<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace MediaManager\Navigators;

use CodeMirror\Utils as CodeMirrorUtils;
use Core\Component;
use Core\Core;
use Core\Date\DateTime;
use Core\Filestore\Contents\ContentImage;
use Core\Filestore\Directory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Core\Filestore\FileMetaHelper;
use Core\Forms\Form;
use Core\ModelValidationException;
use Core\Templates\Backends\Smarty;
use Core\Theme;
use Core\View;
use Exception;
use MediaManager\Navigator2;
use Core\ConfigHandler;

/**
 * Description of MediaNavigator
 *
 * @author charlie
 */
class PageNavigator extends Navigator2 {
	private $_filePaths = null;
	private $_directoryPaths = null;

	public function __construct() {
		$this->_type      = 'page';
		$this->_isAdmin   = \Core\user()->checkAccess('g:admin');
		$this->_canUpload = $this->_isAdmin;
		$this->_canUpdate = $this->_isAdmin;
		$this->_canAccess = $this->_canUpload;
		$this->_basedir   = ROOT_PDIR . 'themes/custom/';
	}

	public function getTitle() : string {
		return t('STRING_MEDIAMANAGER_TEMPLATES');
	}

	/**
	 * Get the URL of the respective navigator
	 */
	public function getURL() : string {
		return '/mediamanagernavigator/templates';
	}

	/**
	 * Get an array of file and directory listings for this lookup
	 *
	 * @return array
	 */
	public function getListing() : array {
		if(!$this->_canAccess) {
			// No access, no files.
			return [];
		}

		$this->_listings = [];

		$this->_filePaths = [];
		$this->_directoryPaths = [];
		$components = Core::GetComponents();

		foreach($components as $c) {
			/** @var $c Component */
			$dir = $c->getBaseDir() . 'templates/pages/';
			if (!is_dir($dir)) {
				continue;
			}

			$dirlen = strlen($dir);
			$componentName = $c->getName();

			$dh = FileFactory::Directory($dir);
			$files = $dh->ls(null, true);

			foreach ($files as $entry) {
				if ($entry instanceof Directory) {
					$relPath = substr($entry->getPath(), $dirlen);
					$this->_directoryPaths[$relPath] = [
						'path' => $entry->getPath(),
						'component' => $componentName,
						'source' => 'installed',
					];
				}
				else {
					$relPath = substr($entry->getFilename(), $dirlen);
					$this->_filePaths[$relPath] = [
						'path' => $entry->getFilename(),
						'component' => $componentName,
						'source' => 'installed',
					];
				}
			}
		}

		try {
			$theme = Theme::GetTheme();
			$dir = $theme->getBaseDir() . 'pages/';
			if (is_dir($dir)) {
				// Only run this logic if the installed theme has page templates
				$themeName = $theme->getName();
				$dirlen = strlen($dir);

				$dh = FileFactory::Directory($dir);
				$files = $dh->ls(null, true);

				foreach ($files as $entry) {
					if ($entry instanceof Directory) {
						$relPath = substr($entry->getPath(), $dirlen);
						$this->_directoryPaths[$relPath] = [
							'path' => $entry->getPath(),
							'theme' => $themeName,
							'source' => 'theme',
						];
					}
					else {
						$relPath = substr($entry->getFilename(), $dirlen);
						$this->_filePaths[$relPath] = [
							'path' => $entry->getFilename(),
							'theme' => $themeName,
							'source' => 'theme',
						];
					}
				}
			}
		} catch (\Exception $ex) {
			// Log and continue
			\Core\log_exception($ex);
		}

		$dir = ROOT_PDIR . 'themes/custom/pages/';
		if (is_dir($dir)) {
			// Only run this logic if the custom has page templates
			$dirlen = strlen($dir);

			$dh = FileFactory::Directory($dir);
			$files = $dh->ls(null, true);

			foreach ($files as $entry) {
				if ($entry instanceof Directory) {
					$relPath = substr($entry->getPath(), $dirlen);
					$this->_directoryPaths[$relPath] = [
						'path' => $entry->getPath(),
						'source' => 'custom',
					];
				}
				else {
					$relPath = substr($entry->getFilename(), $dirlen);
					$this->_filePaths[$relPath] = [
						'path' => $entry->getFilename(),
						'source' => 'custom',
					];
				}
			}
		}

		foreach($this->_directoryPaths as $rel => $data) {
			$path = $data['path'];

			$parent = dirname($rel) . '/';

			if($parent === './') {
				$parent = '';
			}

			$this->_listings[ $rel ] = [
				'type'   => 'directory',
				'name'   => basename($rel),
				'path'   => $rel,
				'parent' => $parent,
				'mtime'  => filemtime($path),
			];
		}

		foreach($this->_filePaths as $rel => $data) {
			$path = $data['path'];

			$file = FileFactory::File($path);

			$parent = dirname($rel) . '/';

			if($parent === './') {
				$parent = '';
			}

			$this->_listings[ $rel ] = [
				'type'   => $file->getMimetype(),
				'name'   => $file->getBasename(),
				'path'   => $rel,
				'parent' => $parent,
				'mtime'  => $file->getMTime(),
				'url'    => null, //$file->getURL(),
			];
		}

		// Sorting would be nice, just because.
		ksort($this->_listings, SORT_STRING | SORT_FLAG_CASE);

		return array_values($this->_listings);
	}

	/**
	 * Get slightly more inclusive information about a subset of files+directories for this navigator.
	 *
	 * This is used to reduce the number of operations required on the initial listing,
	 * because if the user never browses to a subdirectory, there's no need to load the image information.
	 *
	 * @return array
	 */
	public function getInfo(array $paths) : array {
		if(!$this->_canAccess) {
			// No access, no files.
			return [];
		}

		if($this->_directoryPaths === null) {
			// Ensure the listings are loaded, this is required as the filenames are all dynamic.
			$this->getListing();
		}

		$results = [];

		foreach($paths as $rel) {
			if(isset($this->_directoryPaths[ $rel ])) {
				$results[] = $this->_getDirectoryInfo(FileFactory::Directory($this->_directoryPaths[ $rel ]['path']), $rel);
			}
			elseif(isset($this->_filePaths[ $rel ])) {
				$results[] = $this->_getFileInfo(FileFactory::File($this->_filePaths[ $rel ]['path']), $rel);
			}
		}

		return $results;
	}

	/**
	 * Get the entry of a given path as either a File or Directory.
	 *
	 * NULL is returned if the requested path does not match anything.
	 *
	 * @param string $path Requested path to lookup
	 * @return File|Directory|null
	 */
	public function getEntry(string $path) {
		if(!$this->_canAccess) {
			// No access, no files.
			return null;
		}

		if($this->_directoryPaths === null) {
			// Ensure the listings are loaded, this is required as the filenames are all dynamic.
			$this->getListing();
		}

		if(isset($this->_directoryPaths[ $path ])) {
			return FileFactory::Directory($this->_directoryPaths[ $path ]['path']);
		}
		elseif(isset($this->_filePaths[ $path ])) {
			return FileFactory::File($this->_filePaths[ $path ]['path']);
		}
		else {
			return null;
		}
	}

	public function getEntryChildrenCount(string $path) : int {
		if(!$this->_canAccess) {
			// No access, no files.
			return 0;
		}

		if($this->_directoryPaths === null) {
			// Ensure the listings are loaded, this is required as the filenames are all dynamic.
			$this->getListing();
		}

		$c = 0;
		foreach(array_keys($this->_directoryPaths) as $rel) {
			if(strpos($rel, $path) === 0) {
				$c++;
			}
		}
		foreach(array_keys($this->_filePaths) as $rel) {
			if(strpos($rel, $path) === 0) {
				$c++;
			}
		}

		return $c;
	}

	/**
	 * Get the details for an entry, useful for rendering in the details view template
	 *
	 * @param string $path
	 * @return array|null
	 */
	public function getEntryDetails(string $path) : ?array {
		$entry = $this->getEntry($path);
		if(!$entry) {
			return null;
		}

		$details = [
			'preview' => null,
			'name' => $entry->getBasename(),
			'type' => null,
			'attributes' => [],
		];

		if($entry instanceof Directory) {
			$properties = $this->_directoryPaths[$path];

			$details['preview'] = $entry->getMimetypeIconURL('200x200');
			$details['type'] = 'Directory';
			$details['attributes']['Component'] = $properties['component'];
			$details['attributes']['Size'] = t('STRING_MEDIAMANAGER_DIR_N_ITEMS', $this->getEntryChildrenCount($path));
			$details['attributes']['Modified'] = DateTime::FormatString($entry->getMTime(), 'SDT');
		}
		else {
			$properties = $this->_filePaths[$path];
			$contents = $entry->getContentsObject();

			$details['preview'] = $entry->isImage() ? $entry->getPreviewURL('200x200') : $entry->getMimetypeIconURL('200x200');

			try {
				$metas = new FileMetaHelper($entry);
				foreach($metas->getMetas() as $meta) {
					$details['attributes'][ $meta->meta_key ] = $meta->meta_value_title;
				}
			}
			catch(Exception $ex) {
				// Continue.
			}

			$details['type'] = $entry->getMimetype();
			$details['attributes']['Component'] = $properties['component'];
			$details['attributes']['Source'] = $properties['source'];
			$details['attributes']['Size'] = $entry->getFilesize(true);
			$details['attributes']['Modified'] = DateTime::FormatString($entry->getMTime(), 'SDT');

			if($entry->isImageRaster() && $contents instanceof ContentImage) {
				$details['attributes']['Dimensions'] = $contents->getDimensions();

				$exif = $contents->getExif();
				if(count($exif->fields)) {
					$details['attributes']['Image EXIF Metadata'] = '';
					foreach($exif->fields as $field) {
						$details['attributes'][ t($field->title) ] = $field->formatted;
					}
				}
			}
		}

		return $details;
	}

	public function getPermissions(string $path = null) : array {
		if($this->_directoryPaths === null) {
			// Ensure the listings are loaded, this is required as the filenames are all dynamic.
			$this->getListing();
		}

		$ret = [
			'up' => $this->_cwd !== '',
			'mkdir' => $this->checkCanMkdir(),
			'upload' => $this->checkCanUpload(),
			'touch' => $this->checkCanTouch(),
			'rename' => false,
			'rm' => false,
		];

		if($this->_cwd === '') {
			$ret['mkdir'] = false;
			$ret['upload'] = false;
			$ret['touch'] = false;
		}
		elseif(substr_count($this->_cwd, '/') === 2) {
			$ret['mkdir'] = false;
			$ret['upload'] = true;
			$ret['touch'] = true;
		}
		else {
			$ret['mkdir'] = true;
			$ret['upload'] = false;
			$ret['touch'] = false;
		}

		if($path) {
			if(isset($this->_directoryPaths[$path])) {
				$pathData = $this->_directoryPaths[$path];
			}
			elseif(isset($this->_filePaths[$path])) {
				$pathData = $this->_filePaths[$path];
			}
			else {
				return [];
			}

			if($pathData['source'] === 'custom') {
				$ret['rename'] = $this->checkCanUpdate();
				$ret['rm'] = $this->checkCanUpdate();
			}
		}

		return $ret;
	}

	/**
	 * Make a directory in the current working directory.
	 *
	 * @param $directory
	 *
	 * @throws \Exception
	 *
	 * @return string relative path of the newly created directory, useful for browsing immediately to it.
	 */
	public function mkdir(string $directory) : string {
		if(!$this->checkCanUpload()) {
			throw new AccessDeniedException();
		}

		if(!$this->_directoryPaths) {
			$this->getListing();
		}

		// Sanitize the filename
		if(
			strpos($directory, '..') !== false ||
			strpos($directory, '/') !== false
		){
			throw new Exception('Invalid directory name');
		}

		if($this->_cwd === '') {
			throw new \Exception('Template directories cannot be created at the top-level!');
		}

		if(substr_count($this->_cwd, '/') === 2) {
			throw new \Exception('Template directories cannot be created nested inside subtemplate directories.');
		}

		$base = ROOT_PDIR . 'themes/custom/pages/';
		$directory = trim(\Core\str_to_url($directory));

		// Ensure this directory matches a current filename in the same directory, these are *override* directories only.
		$valid = false;
		$mappingMatch = $this->_cwd . $directory;
		foreach(array_keys($this->_filePaths) as $path) {
			if(strpos($path, $mappingMatch) === 0) {
				$baseFilename = pathinfo($path, PATHINFO_FILENAME);
				if($directory === $baseFilename) {
					// Match found, good!
					$valid = true;
					break;
				}
			}
		}
		if(!$valid) {
			throw new \Exception('Template directories must be named the same as an existing file, without the extension.');
		}

		// Directories always get created in custom.
		$dir = FileFactory::Directory($base . $this->_cwd . $directory);

		if($dir->exists()) {
			throw new Exception('Directory ' . $directory . ' already exists!');
		}

		if(!$dir->mkdir()) {
			throw new Exception('Unable to create directory ' . $directory);
		}

		return $this->_cwd . $directory . '/';
	}

	public function touch(string $newFilename) : string {
		if(!$this->checkCanTouch()) {
			throw new AccessDeniedException();
		}

		// Sanitize the filename
		if(
			strpos($newFilename, '..') !== false ||
			strpos($newFilename, '/') !== false
		){
			throw new Exception('Invalid file name');
		}

		$base = ROOT_PDIR . 'themes/custom/pages/';
		$newFilename = trim(\Core\str_to_url($newFilename, true));

		// Check to ensure there's a file extension!
		if(pathinfo($newFilename, PATHINFO_EXTENSION) === '') {
			$newFilename .= '.tpl';
		}

		$file = FileFactory::File($base . $this->_cwd . $newFilename);

		if($file->exists()) {
			touch($file->getFilename());
		}
		else {
			$file->putContents('');
		}

		return $this->_cwd . $newFilename;
	}

	/**
	 * Load the editor for Page templates.
	 *
	 * Will default to the Block editor, but provides a fallback code editor if no blocks are available
	 *
	 * @param View $view
	 * @param string $path
	 */
	public function loadEditor(View $view, string $path) {
		$entry = $this->getEntry($path);
		$contents = $entry->getContents();
		if($contents !== '') {
			// Only do these scans if this file has content, otherwise allow to continue with the block editor.
			$document = new \DOMDocument();
			$document->loadHTML($contents);

			$xpath = new \DOMXPath($document);

			$panels = $xpath->query('//*[@data-block-panel]');
			if($panels->length) {
				return self::loadBlockEditor($view, $path);
			}
			else {
				// No panels are available, in this case, the block editor will cause more damage than good.
				return self::loadCodeEditor($view, $path);
			}
		}
		else {
			return self::loadBlockEditor($view, $path);
		}
	}

	public function loadBlockEditor(View $view, string $path) {
		$blocksAvailable = [];
		$panelsAvailable = [];
		$layoutsAvailable = [];


		$form = new Form();
		$form->set('id', 'block-editor-form');
		$form->set('callsmethod', '\\MediaManager\\Navigators\\PageNavigator::FileSaveHandler');

		$form->addElement('system', ['name' => 'type', 'value' => $this->_type]);
		$form->addElement('system', ['name' => 'path', 'value' => $path]);

		$widgets = \WidgetModel::Find('installable = ');
		foreach($widgets as $widget) {
			$widgetTemplate = $widget->getTemplateContent();
			if($widgetTemplate) {
				$blocksAvailable[] = [
					'name'            => $widget->title,
					'key'             => 'WIDGET',
					'panel'           => '',
					'templateContent' => $widgetTemplate,
					'single'          => 0,
					'image'           => \Core\resolve_asset($widget->getPreviewImage()),
				];
			}
		}


		$entry = $this->getEntry($path);
		$contents = $entry->getContents();
		if($contents !== '') {
			// Only do these scans if this file has content, otherwise allow to continue with the block editor.
			$document = new \DOMDocument();
			$document->loadHTML($contents);

			$xpath = new \DOMXPath($document);

			$htmlRoot = $xpath->query('/html/body');

			$panels = $xpath->query('//*[@data-block-panel]');

			foreach($panels as $panel) {
				/** @var \DOMElement $panel */
				$panelKey = $panelName = $panel->getAttribute('data-block-panel');
				//$panelKey = 'panel-' . \Core\str_to_url($panelName);

				$panelsAvailable[] = [
					'name' => $panelName,
					'key' => $panelKey,
				];

				$blocks = $xpath->query('*[@data-block]', $panel);
				foreach($blocks as $block) {
					$blockName = $block->getAttribute('data-block');
					$blockKey = $block->getAttribute('data-block-key');
					$blockSingle = $block->getAttribute('data-block-single');
					$blockImage = $block->getAttribute('data-block-image');

					if($blockImage) {
						if(strpos($blockImage, '/') === false) {
							$blockImage = \Core\resolve_asset('assets/images/previews/templates/blocks/' . $blockImage);
						}
						else {
							$blockImage = \Core\resolve_asset($blockImage);
						}
					}

					$blockContent = html_entity_decode($block->ownerDocument->saveHTML($block));
					$blocksAvailable[] = [
						'name'            => $blockName,
						'key'             => $blockKey,
						'panel'           => $panelKey,
						'templateContent' => $blockContent,
						'single'          => $blockSingle,
						'image'           => $blockImage,
					];

					$panel->removeChild($block);
				}
			}

			$layoutTemplate = html_entity_decode($document->saveHTML($htmlRoot->item(0)));
			// Skip the body.
			$layoutTemplate = trim(str_replace(['<body>', '</body>'], '', $layoutTemplate));
			$layoutsAvailable[] = [
				'name' => 'Current Layout',
				'templateContent' => $layoutTemplate,
				'panels' => $panelsAvailable,
				'current' => true,
			];
		}
		else {
			$layoutTemplate = '';
		}

		$form->addElement('textarea', ['name' => 'layout', 'value' => $layoutTemplate]);

		// Retrieve all available blocks on the system
		$blockFiles = $this->_getBlocks();
		foreach($blockFiles as $blockFile) {
			/** @var File $blockFile */
			$contents = $blockFile->getContents();

			// @todo Replace DOMDocument with something more flexible.
			// Template files are not *exactly* XML/XHTML; as they contain logic, code, etc.
			// As such, treating them as such causes numerous warnings to be generated.
			// Ideally, this should be switched with something that can ignore them,
			// but still retrieve the necessary tags from the code.
			$document = new \DOMDocument();
			$document->loadHTML($contents);

			$xpath = new \DOMXPath($document);

			$panelKey = '';

			$blocks = $xpath->query('//*[@data-block]');
			foreach($blocks as $block) {
				$blockName = $block->getAttribute('data-block');
				$blockKey = $block->getAttribute('data-block-key');
				$blockSingle = $block->getAttribute('data-block-single');
				$blockImage = $block->getAttribute('data-block-image');

				if($blockImage) {
					$blockImage = \Core\resolve_asset('assets/images/previews/templates/blocks/' . $blockImage);
				}

				//$blockContent = html_entity_decode($block->ownerDocument->saveHTML($block));
				$blocksAvailable[] = [
					'name' => $blockName,
					'key' => $blockKey,
					'panel' => $panelKey,
					//'templateContent' => $blockContent,
					'templateContent' => $contents,
					'single' => $blockSingle,
					'image' => $blockImage,
				];
			}
		}

		// Retrieve all available layouts on the system
		$layoutFiles = $this->_getLayouts();
		foreach($layoutFiles as $layoutFile) {
			$layoutPanels = [];

			/** @var File $layoutFile */
			$contents = $layoutFile->getContents();

			$document = new \DOMDocument();
			$document->loadHTML($contents);

			$xpath = new \DOMXPath($document);

			$panels = $xpath->query('//*[@data-block-panel]');
			if(!$panels->length) {
				continue;
			}

			$layoutName = ucwords(str_replace(['_', '-'], ' ', $layoutFile->getBasename(true)));
			$layoutKey = strtolower($layoutFile->getBasename(true));

			foreach($panels as $panel) {
				/** @var \DOMElement $panel */
				$panelName = $panel->getAttribute('data-block-panel');
				$panelKey = 'panel-' . \Core\str_to_url($panelName);

				$layoutPanels[] = [
					'name' => $panelName,
					'key'  => $panelKey,
				];
			}

			$layoutsAvailable[] = [
				'key' => $layoutKey,
				'name' => $layoutName,
				'templateContent' => $contents,
				'panels' => $layoutPanels,
				'current' => false,
			];
		}

		$form->addElement('textarea', ['name' => 'panels[]']);
//var_dump($layoutsAvailable); die();

		$view->templatename = 'pages/mediamanagernavigator/file_editor/page_block';
		$view->assign('form', $form);
		$view->assign('content', true);
		$view->assign('file', $entry);
		$view->assign('panels', $panelsAvailable); // Current only
		$view->assign('blocks', $blocksAvailable); // Current and available
		$view->assign('layouts', $layoutsAvailable); // Current and available
	}

	public function loadCodeEditor(View $view, string $path) {
		$entry = $this->getEntry($path);

		if( $entry->isText() ) {
			switch($entry->getExtension()) {
				case 'tpl':
					$mode = 'smarty';
					break;
				default:
					$mode = 'htmlmixed';
					break;
			}
		}
		else {
			throw new Exception('Unsupported page template!');
		}

		$form = new Form();
		$form->set('callsmethod', '\\MediaManager\\Navigators\\PageNavigator::CodeSaveHandler');

		$form->addElement('system', ['name' => 'type', 'value' => $this->_type]);
		$form->addElement('system', ['name' => 'path', 'value' => $path]);

		$form->addElement(
			'textarea',
			[
				'name' => 'content',
				'value' => $entry->getContents(),
				'rows' => 20,
			]
		);

		// Include the correct type of CodeMirror for this file, only applicable for text files.
		if(Core::IsComponentAvailable('codemirror') && $mode !== 'unknown') {
			CodeMirrorUtils::IncludeMode($mode);
		}


		$form->addElement('submit', ['value' => 't:STRING_SAVE']);

		$view->templatename = 'pages/mediamanagernavigator/file_editor/page_code';
		$view->assign('form', $form);
		$view->assign('file', $entry);
		$view->assign('mode', $mode);
	}

	public static function FileSaveHandler(Form $form) {
		$layout = $form->getElementValue('layout');
		$panelValues = $form->getElementValue('panels[]');

		$document = new \DOMDocument();
		$document->loadHTML($layout);

		$xpath = new \DOMXPath($document);

		$panels = $xpath->query('//*[@data-block-panel]');
		if(!$panels->length) {
			throw new ModelValidationException('No panels found in the selected template!');
		}


		foreach($panels as $panel) {
			/** @var \DOMElement $panel */
			$panelName = $panel->getAttribute('data-block-panel');
			$panelKey = 'panel-' . \Core\str_to_url($panelName);

			if(isset($panelValues[ $panelName ])) {
				foreach($panelValues[ $panelName ] as $templateSnippet) {
					$fragment = $document->createDocumentFragment();
					$fragment->appendXML(htmlentities($templateSnippet));

					$panel->appendChild($fragment);
				}
			}
		}

		$finalizedTemplate = html_entity_decode($document->saveHTML());

		// Saving HTML from XML nodes always generates fully compliant pages.
		// This is just a snippet, not a full page, so strip those back out.
		$finalizedTemplate = preg_replace('/<!DOCTYPE[^>]*>/', '', $finalizedTemplate);
		$finalizedTemplate = str_replace(['<html>', '<body>', '</body>', '</html>'], '', $finalizedTemplate);

		// All files will get saved to the custom directory so as to not overwrite anything.
		$dest = ROOT_PDIR . 'themes/custom/pages/' . $form->getElementValue('path');
		$file = FileFactory::File($dest);
		$file->putContents($finalizedTemplate);

		Smarty::FlushCache();
		\Core\set_message('Updated page template successfully', 'success');
		return 'BACK';
	}

	public static function CodeSaveHandler(Form $form) {

		// All files will get saved to the custom directory so as to not overwrite anything.
		$dest = ROOT_PDIR . 'themes/custom/pages/' . $form->getElementValue('path');
		$file = FileFactory::File($dest);
		$file->putContents($form->getElementValue('content'));

		Smarty::FlushCache();
		\Core\set_message('Updated page template successfully', 'success');
		return 'BACK';
	}

	/**
	 * Get all block templates available on the system
	 *
	 * @throws \Exception
	 */
	private function _getBlocks() {
		$blocks = [];

		$components = Core::GetComponents();
		foreach($components as $c) {
			/** @var $c Component */
			$dir = FileFactory::Directory($c->getBaseDir() . 'templates/blocks');
			foreach($dir->ls('tpl', false) as $file) {
				if ($file instanceof File) {
					$blocks[$file->getBasename()] = $file;
				}
			}
		}

		// Themes can overwrite components.
		$theme = Theme::GetTheme();
		$dir = FileFactory::Directory($theme->getBaseDir() . 'blocks');
		foreach($dir->ls('tpl', false) as $file) {
			if ($file instanceof File) {
				$blocks[$file->getBasename()] = $file;
			}
		}

		// Custom can overwrite everything.
		$dir = FileFactory::Directory(ROOT_PDIR . 'themes/custom/blocks');
		foreach($dir->ls('tpl', false) as $file) {
			if ($file instanceof File) {
				$blocks[$file->getBasename()] = $file;
			}
		}

		return $blocks;
	}

	/**
	 * Get all layout templates available on the system
	 *
	 * @throws \Exception
	 */
	private function _getLayouts() {
		$blocks = [];

		$components = Core::GetComponents();
		foreach($components as $c) {
			/** @var $c Component */
			$dir = FileFactory::Directory($c->getBaseDir() . 'templates/block_layouts');
			foreach($dir->ls('tpl', false) as $file) {
				if ($file instanceof File) {
					$blocks[$file->getBasename()] = $file;
				}
			}
		}

		// Themes can overwrite components.
		$theme = Theme::GetTheme();
		$dir = FileFactory::Directory($theme->getBaseDir() . 'block_layouts');
		foreach($dir->ls('tpl', false) as $file) {
			if ($file instanceof File) {
				$blocks[$file->getBasename()] = $file;
			}
		}

		// Custom can overwrite everything.
		$dir = FileFactory::Directory(ROOT_PDIR . 'themes/custom/block_layouts');
		foreach($dir->ls('tpl', false) as $file) {
			if ($file instanceof File) {
				$blocks[$file->getBasename()] = $file;
			}
		}

		return $blocks;
	}
}
