/**
 * Collection of Core libraries
 *
 * All these are to be expected to be loaded in the head (as opposed as the foot script)
 *
 */

"use strict";

// Containers used herein

Core.Color = Core.Color || {};
Core.Strings = Core.Strings || {};

/**
 * Just a simple function to reload the current page.
 *
 */
Core.Reload = () => {
	window.location.reload();
}

Core.ModalDialog = function() {
	this.frame = document.createElement('div');
	this.modal = document.createElement('div');
	this.loadingHTML = '<div class="modaldialog-loading"><i class="icon icon-spin icon-spinner"></i></div>';
	this.scripts = [];
	this.inlineScripts = [];
	this.scriptsLoaded = 0;
	// Set to an event when the page and scripts are fully loaded and executed
	this.pageLoaded = null;

	this.frame.classList.add('modaldialog-frame');
	this.modal.classList.add('modaldialog-modal');

	this.modal.addEventListener('click', evt => {
		this.close();
	});

	/**
	 * Event watcher if the user presses escape
	 *
	 * @private
	 * @param evt
	 */
	this.keyWatch = evt => {
		if(evt.key === 'Escape') {
			this.close();
		}
	};

	/**
	 * Create the necessary DOM structure
	 *
	 * @private
	 */
	this.create = () => {
		// Append the items to the DOM so they get registered
		document.body.appendChild(this.frame);
		document.body.appendChild(this.modal);

		document.addEventListener('keydown', this.keyWatch);

		// The frame should have *something* in it.
		if(this.frame.innerHTML === '') {
			this.frame.innerHTML = this.loadingHTML;
		}
	};

	/**
	 * Destroy the created DOM structure and all supplemental nodes
	 *
	 * @private
	 */
	this.destroy = () => {
		document.removeEventListener('keydown', this.keyWatch);

		document.body.removeChild(this.frame);
		document.body.removeChild(this.modal);

		for(let i = 0; i < this.scripts.length; i++) {
			document.body.removeChild(this.scripts[i]);
		}

		this.scripts = [];
		this.frame.innerHTML = '';
	};

	/**
	 * Set any required script for this dialog, usually set from JSON data.scripts
	 *
	 * @param {array} scripts
	 */
	this.setScripts = scripts => {
		let currentDocumentScriptURLs = [],
			toAdd = [];

		// Check the current scripts, we don't want to add duplicates.
		for(let i = 0; i < document.scripts.length; i++) {
			if(document.scripts.item(i).src) {
				currentDocumentScriptURLs.push(document.scripts.item(i).src);
			}
		}

		// Check the incoming scripts and queue only those needed.
		for(let i = 0; i < scripts.length; i++) {
			if(currentDocumentScriptURLs.indexOf(scripts[i]) === -1) {
				toAdd.push(scripts[i]);
			}
		}

		this.scriptsLoaded = 0;

		// We now have a queue of scripts to append.
		for(let i = 0; i < toAdd.length; i++) {
			let scriptNode = document.createElement('script');
			scriptNode.src = toAdd[i];
			scriptNode.addEventListener('load', evt => {
				this.scriptsLoaded += 1;

				if(this.scriptsLoaded === this.scripts.length) {
					// All scripts are loaded.
					this.loadInlineScripts();
				}
			});

			document.body.appendChild(scriptNode);

			// Remember this so we can perform cleanup
			this.scripts.push(scriptNode);
		}
	};

	/**
	 * Set any required script for this dialog, usually set from JSON data.scripts
	 *
	 * These are INLINE scripts, and as such do not have a src to track,
	 * so they better be able to run multiple times on the same page!
	 *
	 * @param {array} scripts
	 */
	this.setInlineScripts = scripts => {
		this.inlineScripts = scripts;
		if(this.scriptsLoaded === this.scripts.length) {
			// All scripts are loaded or no scripts are required.
			this.loadInlineScripts();
		}
	};

	this.loadInlineScripts = () => {
		for(let i = 0; i < this.inlineScripts.length; i++) {
			eval(this.inlineScripts[i]);
		}

		if(typeof(this.pageLoaded) === 'function') {
			// There is a pageloaded function requested to be called after all scripts have loaded, so call it
			this.pageLoaded();
		}
	};

	/**
	 * Create and open a dialog window
	 */
	this.open = () => {
		this.create();
		document.body.style.overflow = 'hidden';

		setTimeout(() => {
			// THEN add the open tags, this way CSS has a "before" and "after" step for transitions.
			this.frame.classList.add('open');
			this.modal.classList.add('open');
		}, 50);
	};

	/**
	 * Close and destroy a dialog window
	 */
	this.close = () => {
		// Remove the classes to indicate it's open
		this.frame.classList.remove('open');
		this.modal.classList.remove('open');

		document.body.style.overflow = 'inherit';

		setTimeout(() => {
			// THEN remove the actual nodes, provides CSS time to run any transitions.
			this.destroy();
		}, 600);
	};

	/**
	 * Convenience function to ajax load content into this Modal.
	 *
	 * Uses Core.Ajax for the backend request.
	 *
	 * @param url
	 * @param parameters
	 *
	 * Supported parameters for the ajaxLoad method of Dialogs
	 * * preSend {function}
	 * * callback {function}
	 * * pageLoaded {function}
	 * * method {string} Default "GET", can be any supported HTTP method
	 * * params {Object|string} Parameters to send as GET string parameters
	 * * accept {string} Content type for acceptable return
	 */
	this.ajaxLoad = (url, parameters) => {
		if(typeof(parameters) === 'undefined') {
			parameters = {};
		}

		let preSend = typeof(parameters.preSend) === 'function' ? parameters.preSend : null,
			callback = typeof(parameters.callback) === 'function' ? parameters.callback : null;

		if(typeof(parameters.pageLoaded) === 'function') {
			// Offset this to the Dialog handler
			this.pageLoaded = parameters.pageLoaded;
		}

		// Scripts are handled internally.
		parameters.loadScripts = false;

		parameters.preSend = () => {
			if(preSend) {
				preSend();
			}
			this.open();
		};

		parameters.callback = (data, contentType) => {
			if(callback) {
				// There's a load available, let them deal with it.
				callback(this, data);
			}
			else if(contentType === 'application/json' && typeof(data.body) !== 'undefined') {
				// Body is expected to be HTML data encoded in the JSON stream.
				this.frame.innerHTML = data.body;
			}
			else if(contentType === 'text/html') {
				this.frame.innerHTML = data;
			}
			else if(contentType === 'text/plain') {
				this.frame.innerHTML = '<pre>' + data + '</pre>';
			}
			else {
				throw new Error('What am I supposed to do with the response data?');
			}

			if(typeof(data.scripts) !== 'undefined') {
				// Everything should be loaded now, include the scripts that may be required.
				this.setScripts(data.scripts);
			}

			if(typeof(data.scripts_inline) !== 'undefined') {
				// Also include any inline scripts
				this.setInlineScripts(data.scripts_inline);
			}

			if(
				typeof(data.scripts) === 'undefined' &&
				typeof(data.scripts_inline) === 'undefined' &&
				typeof(this.pageLoaded) === 'function'
			) {
				// No scripts were provided, but a callback was requested, so call that manually
				this.pageLoaded();
			}
		};

		Core.Ajax(url, parameters).catch(e => {
			this.frame.innerHTML = '<p class="message-error">Sorry, but something unexpected happened.  Press ESCAPE to close.</p>';
			console.error(e);
		});
	};
};

/**
 * Low level AJAX function for Core-compatible data
 *
 * @param {string} url        Fully resolved URL to query
 * @param {Object} parameters Parameters to adjust functionality
 *
 * @returns {Promise<void>}
 *
 * @constructor
 */
Core.Ajax = async function(url, parameters) {
	if(!url) {
		throw new Error('Unable to run Core.Ajax, "url" must be a valid URL');
	}

	if(typeof(parameters) === 'undefined') {
		parameters = {};
	}

	let method = parameters.method || 'GET',
		params = parameters.parameters || null,
		data = parameters.data || null,
		dataType = parameters.dataType || null,
		loadScripts = true,
		loadStyles = true,
		callback = parameters.callback || null,
		preSend = parameters.preSend || null,
		accept = parameters.accept || '*',
		body = null,
		headers = {
			'X-Requested-With': 'XMLHttpRequest'
		};

	accept = accept.toUpperCase();
	method = method.toUpperCase();

	if(typeof(parameters.loadScripts) !== 'undefined') {
		loadScripts = parameters.loadScripts;
	}
	if(typeof(parameters.loadStyles) !== 'undefined') {
		loadStyles = parameters.loadStyles;
	}

	if(accept === '*') {
		headers['Accept'] = 'application/json,text/html,text/plain,*/*';
	}
	else if(accept === 'JSON') {
		headers['Accept'] = 'application/json';
	}
	else if(accept === 'HTML') {
		headers['Accept'] = 'text/html';
	}
	else if(accept === 'TEXT') {
		headers['Accept'] = 'text/plain';
	}
	else {
		throw new Error('Unable to run Core.Ajax, "accept" parameter must be one of: [*, JSON, HTML, TEXT]');
	}

	if(method === 'POST' && data !== null) {
		// POST can have a payload!
		if(dataType === null) {
			dataType = 'application/json';
		}
		else if(dataType.toLowerCase() === 'json') {
			dataType = 'application/json';
		}
		else if(dataType.toLowerCase() === 'form') {
			dataType = 'multipart/form-data';
		}

		if(dataType === 'application/json') {
			body = JSON.stringify(data);
			headers['Content-Type'] = dataType;
		}
		else if(dataType === 'multipart/form-data') {
			body = new FormData();
			Object.keys(data).forEach(key => {
				body.append(key, data[key]);
			});
			// Form data does not require a header; it'll get set automatically
		}
		else {
			throw new Error('Unsupported dataType parameter!  Currently only JSON is supported.');
		}
	}
	else if(method === 'GET' && data !== null) {
		throw new Error('GET requests cannot have data payloads!  Did you mean "parameters"?');
	}


	if(params) {
		// Parameters were requested, this is a convenience option for URL GET parameters.
		if(url.indexOf('?') === -1) {
			url += '?';
		}
		else {
			url += '&';
		}

		url += (new URLSearchParams(params)).toString();
	}

	if(preSend) {
		preSend();
	}

	fetch(
		url,
		{
			method: method,
			body: body,
			headers: new Headers(headers)
		}
	)
		.then(res => {
			let cType = res.headers.get('Content-Type');

			if(cType.includes('application/json')) {
				// Automatically decode JSON data returned by the server.
				res.json().then(data => {
					// There is debug info attached, register that with DevTools.
					// This is beneficial so the request can be tracked by the developer.
					if(typeof(Core.DevTools) !== 'undefined' && typeof(data._debug) !== 'undefined') {
						Core.DevTools.SetPageDetails(data._debug.page);
					}

					if(accept !== '*' && accept !== 'JSON') {
						throw new Error('Server returned application/json when ' + accept + ' was expected');
					}

					if(callback) {
						// Allow the callback to fire with the returned data.
						callback(data, 'application/json');
					}

					if(loadScripts && typeof(data.scripts) !== 'undefined') {
						// Automatically load any scripts requested
						let currentDocumentScriptURLs = [];

						// Check the current scripts, we don't want to add duplicates.
						for(let i = 0; i < document.scripts.length; i++) {
							if(document.scripts.item(i).src) {
								currentDocumentScriptURLs.push(document.scripts.item(i).src);
							}
						}

						// We now have a queue of scripts to append.
						for(let i = 0; i < data.scripts.length; i++) {
							if(currentDocumentScriptURLs.indexOf(data.scripts[i]) === -1) {
								let scriptNode = document.createElement('script');
								scriptNode.src = toAdd[i];
								document.body.appendChild(scriptNode);
							}
						}
					}

					if(loadScripts && typeof(data.scripts_inline) !== 'undefined') {
						// Execute any inline code, if requested
						for(let i = 0; i < data.scripts_inline.length; i++) {
							eval(data.scripts_inline[i]);
						}
					}
				});
			}
			else if(cType.includes('text/html')) {
				if(accept !== '*' && accept !== 'HTML') {
					throw new Error('Server returned text/html when ' + accept + ' was expected');
				}

				res.text().then(html => {
					if(callback) {
						callback(html, 'text/html');
					}
				});
			}
			else if(cType.includes('text/plain')) {
				if(accept !== '*' && accept !== 'TEXT') {
					throw new Error('Server returned text/plain when ' + accept + ' was expected');
				}

				res.text().then(text => {
					if(callback) {
						callback(text, 'text/plain');
					}
				});
			}
			else {
				if(accept !== '*') {
					throw new Error('Server returned ' + cType.toString() + ' when ' + accept + ' was expected');
				}

				res.blob().then(data => {
					if(callback) {
						callback(data, cType.toString());
					}
				});
			}
		});
};

Core.AjaxLoad = async function(target, url, parameters) {
	if(!target) {
		throw new Error('Unable to run Core.AjaxLoad, "target" must be a valid DOMNode');
	}

	if(!url) {
		throw new Error('Unable to run Core.AjaxLoad, "url" must be a valid URL');
	}

	if(typeof(parameters) === 'undefined') {
		parameters = {};
	}

	parameters.callback = (data, contentType) => {
		if(contentType === 'application/json') {
			target.innerHTML = data.body;
		}
		else if(contentType === 'text/html') {
			target.innerHTML = data;
		}
		else if(contentType === 'text/plain') {
			target.innerHTML = '<pre>' + data + '</pre>';
		}
		else {
			target.innerHTML = '<p class="message-error">Unexpected response type, unable to display content!</p>';
		}
	};

	return Core.Ajax(url, parameters);
};



/**
 * Get the brightness value of a given color, between 0 - 255.
 *
 * 0 - 122 is dark
 * 123 - 255 is bright
 *
 * @param {string} color
 *
 * @return {number}
 */
Core.Color.BrightnessLevel = color => {
	// Allow / expect hex values, they may have the hex indicator.
	color = color.replace('#', '');

	let red = parseInt(color.substr(0, 2), 16),
		blue = parseInt(color.substr(4, 2), 16),
		green = parseInt(color.substr(2, 2), 16);


	// Convert each value to their brightness level based on standard color values.
	red *= 0.2126;
	blue *= 0.0722;
	green *= 0.7152;

	return Math.round(red + blue + green);
};

/**
 * Get the tag's background class value based on the colour
 *
 * @param {string} color
 * @returns {string}
 */
Core.Color.BackgroundTag = color => {
	let level = Core.Color.BrightnessLevel(color);
	return level >= 123 ? 'light-background' : 'dark-background';
}




/**
 * Define console as an object so console.logs don't break in IE.
 */
if(typeof console == 'undefined'){
	console = {};
}
if(typeof console.log == 'undefined'){
	console.log = function(){ /* This page purposefully left blank :p */ }
}

// noinspection NonAsciiCharacters
/**
 * Map of character replacements for i18n characters to their latin likeness
 */
Core.Strings.I18NCharacters = {
	"Á":"A",
	"Ă":"A",
	"Ắ":"A",
	"Ặ":"A",
	"Ằ":"A",
	"Ẳ":"A",
	"Ẵ":"A",
	"Ǎ":"A",
	"Â":"A",
	"Ấ":"A",
	"Ậ":"A",
	"Ầ":"A",
	"Ẩ":"A",
	"Ẫ":"A",
	"Ä":"A",
	"Ǟ":"A",
	"Ȧ":"A",
	"Ǡ":"A",
	"Ạ":"A",
	"Ȁ":"A",
	"À":"A",
	"Ả":"A",
	"Ȃ":"A",
	"Ā":"A",
	"Ą":"A",
	"Å":"A",
	"Ǻ":"A",
	"Ḁ":"A",
	"Ⱥ":"A",
	"Ã":"A",
	"Ꜳ":"AA",
	"Æ":"AE",
	"Ǽ":"AE",
	"Ǣ":"AE",
	"Ꜵ":"AO",
	"Ꜷ":"AU",
	"Ꜹ":"AV",
	"Ꜻ":"AV",
	"Ꜽ":"AY",
	"Ḃ":"B",
	"Ḅ":"B",
	"Ɓ":"B",
	"Ḇ":"B",
	"Ƀ":"B",
	"Ƃ":"B",
	"Ć":"C",
	"Č":"C",
	"Ç":"C",
	"Ḉ":"C",
	"Ĉ":"C",
	"Ċ":"C",
	"Ƈ":"C",
	"Ȼ":"C",
	"Ď":"D",
	"Ḑ":"D",
	"Ḓ":"D",
	"Ḋ":"D",
	"Ḍ":"D",
	"Ɗ":"D",
	"Ḏ":"D",
	"ǲ":"D",
	"ǅ":"D",
	"Đ":"D",
	"Ƌ":"D",
	"Ǳ":"DZ",
	"Ǆ":"DZ",
	"É":"E",
	"Ĕ":"E",
	"Ě":"E",
	"Ȩ":"E",
	"Ḝ":"E",
	"Ê":"E",
	"Ế":"E",
	"Ệ":"E",
	"Ề":"E",
	"Ể":"E",
	"Ễ":"E",
	"Ḙ":"E",
	"Ë":"E",
	"Ė":"E",
	"Ẹ":"E",
	"Ȅ":"E",
	"È":"E",
	"Ẻ":"E",
	"Ȇ":"E",
	"Ē":"E",
	"Ḗ":"E",
	"Ḕ":"E",
	"Ę":"E",
	"Ɇ":"E",
	"Ẽ":"E",
	"Ḛ":"E",
	"Ꝫ":"ET",
	"Ḟ":"F",
	"Ƒ":"F",
	"Ǵ":"G",
	"Ğ":"G",
	"Ǧ":"G",
	"Ģ":"G",
	"Ĝ":"G",
	"Ġ":"G",
	"Ɠ":"G",
	"Ḡ":"G",
	"Ǥ":"G",
	"Ḫ":"H",
	"Ȟ":"H",
	"Ḩ":"H",
	"Ĥ":"H",
	"Ⱨ":"H",
	"Ḧ":"H",
	"Ḣ":"H",
	"Ḥ":"H",
	"Ħ":"H",
	"Í":"I",
	"Ĭ":"I",
	"Ǐ":"I",
	"Î":"I",
	"Ï":"I",
	"Ḯ":"I",
	"İ":"I",
	"Ị":"I",
	"Ȉ":"I",
	"Ì":"I",
	"Ỉ":"I",
	"Ȋ":"I",
	"Ī":"I",
	"Į":"I",
	"Ɨ":"I",
	"Ĩ":"I",
	"Ḭ":"I",
	"Ꝺ":"D",
	"Ꝼ":"F",
	"Ᵹ":"G",
	"Ꞃ":"R",
	"Ꞅ":"S",
	"Ꞇ":"T",
	"Ꝭ":"IS",
	"Ĵ":"J",
	"Ɉ":"J",
	"Ḱ":"K",
	"Ǩ":"K",
	"Ķ":"K",
	"Ⱪ":"K",
	"Ꝃ":"K",
	"Ḳ":"K",
	"Ƙ":"K",
	"Ḵ":"K",
	"Ꝁ":"K",
	"Ꝅ":"K",
	"Ĺ":"L",
	"Ƚ":"L",
	"Ľ":"L",
	"Ļ":"L",
	"Ḽ":"L",
	"Ḷ":"L",
	"Ḹ":"L",
	"Ⱡ":"L",
	"Ꝉ":"L",
	"Ḻ":"L",
	"Ŀ":"L",
	"Ɫ":"L",
	"ǈ":"L",
	"Ł":"L",
	"Ǉ":"LJ",
	"Ḿ":"M",
	"Ṁ":"M",
	"Ṃ":"M",
	"Ɱ":"M",
	"Ń":"N",
	"Ň":"N",
	"Ņ":"N",
	"Ṋ":"N",
	"Ṅ":"N",
	"Ṇ":"N",
	"Ǹ":"N",
	"Ɲ":"N",
	"Ṉ":"N",
	"Ƞ":"N",
	"ǋ":"N",
	"Ñ":"N",
	"Ǌ":"NJ",
	"Ó":"O",
	"Ŏ":"O",
	"Ǒ":"O",
	"Ô":"O",
	"Ố":"O",
	"Ộ":"O",
	"Ồ":"O",
	"Ổ":"O",
	"Ỗ":"O",
	"Ö":"O",
	"Ȫ":"O",
	"Ȯ":"O",
	"Ȱ":"O",
	"Ọ":"O",
	"Ő":"O",
	"Ȍ":"O",
	"Ò":"O",
	"Ỏ":"O",
	"Ơ":"O",
	"Ớ":"O",
	"Ợ":"O",
	"Ờ":"O",
	"Ở":"O",
	"Ỡ":"O",
	"Ȏ":"O",
	"Ꝋ":"O",
	"Ꝍ":"O",
	"Ō":"O",
	"Ṓ":"O",
	"Ṑ":"O",
	"Ɵ":"O",
	"Ǫ":"O",
	"Ǭ":"O",
	"Ø":"O",
	"Ǿ":"O",
	"Õ":"O",
	"Ṍ":"O",
	"Ṏ":"O",
	"Ȭ":"O",
	"Ƣ":"OI",
	"Ꝏ":"OO",
	"Ɛ":"E",
	"Ɔ":"O",
	"Ȣ":"OU",
	"Ṕ":"P",
	"Ṗ":"P",
	"Ꝓ":"P",
	"Ƥ":"P",
	"Ꝕ":"P",
	"Ᵽ":"P",
	"Ꝑ":"P",
	"Ꝙ":"Q",
	"Ꝗ":"Q",
	"Ŕ":"R",
	"Ř":"R",
	"Ŗ":"R",
	"Ṙ":"R",
	"Ṛ":"R",
	"Ṝ":"R",
	"Ȑ":"R",
	"Ȓ":"R",
	"Ṟ":"R",
	"Ɍ":"R",
	"Ɽ":"R",
	"Ꜿ":"C",
	"Ǝ":"E",
	"Ś":"S",
	"Ṥ":"S",
	"Š":"S",
	"Ṧ":"S",
	"Ş":"S",
	"Ŝ":"S",
	"Ș":"S",
	"Ṡ":"S",
	"Ṣ":"S",
	"Ṩ":"S",
	"Ť":"T",
	"Ţ":"T",
	"Ṱ":"T",
	"Ț":"T",
	"Ⱦ":"T",
	"Ṫ":"T",
	"Ṭ":"T",
	"Ƭ":"T",
	"Ṯ":"T",
	"Ʈ":"T",
	"Ŧ":"T",
	"Ɐ":"A",
	"Ꞁ":"L",
	"Ɯ":"M",
	"Ʌ":"V",
	"Ꜩ":"TZ",
	"Ú":"U",
	"Ŭ":"U",
	"Ǔ":"U",
	"Û":"U",
	"Ṷ":"U",
	"Ü":"U",
	"Ǘ":"U",
	"Ǚ":"U",
	"Ǜ":"U",
	"Ǖ":"U",
	"Ṳ":"U",
	"Ụ":"U",
	"Ű":"U",
	"Ȕ":"U",
	"Ù":"U",
	"Ủ":"U",
	"Ư":"U",
	"Ứ":"U",
	"Ự":"U",
	"Ừ":"U",
	"Ử":"U",
	"Ữ":"U",
	"Ȗ":"U",
	"Ū":"U",
	"Ṻ":"U",
	"Ų":"U",
	"Ů":"U",
	"Ũ":"U",
	"Ṹ":"U",
	"Ṵ":"U",
	"Ꝟ":"V",
	"Ṿ":"V",
	"Ʋ":"V",
	"Ṽ":"V",
	"Ꝡ":"VY",
	"Ẃ":"W",
	"Ŵ":"W",
	"Ẅ":"W",
	"Ẇ":"W",
	"Ẉ":"W",
	"Ẁ":"W",
	"Ⱳ":"W",
	"Ẍ":"X",
	"Ẋ":"X",
	"Ý":"Y",
	"Ŷ":"Y",
	"Ÿ":"Y",
	"Ẏ":"Y",
	"Ỵ":"Y",
	"Ỳ":"Y",
	"Ƴ":"Y",
	"Ỷ":"Y",
	"Ỿ":"Y",
	"Ȳ":"Y",
	"Ɏ":"Y",
	"Ỹ":"Y",
	"Ź":"Z",
	"Ž":"Z",
	"Ẑ":"Z",
	"Ⱬ":"Z",
	"Ż":"Z",
	"Ẓ":"Z",
	"Ȥ":"Z",
	"Ẕ":"Z",
	"Ƶ":"Z",
	"Ĳ":"IJ",
	"Œ":"OE",
	"ᴀ":"A",
	"ᴁ":"AE",
	"ʙ":"B",
	"ᴃ":"B",
	"ᴄ":"C",
	"ᴅ":"D",
	"ᴇ":"E",
	"ꜰ":"F",
	"ɢ":"G",
	"ʛ":"G",
	"ʜ":"H",
	"ɪ":"I",
	"ʁ":"R",
	"ᴊ":"J",
	"ᴋ":"K",
	"ʟ":"L",
	"ᴌ":"L",
	"ᴍ":"M",
	"ɴ":"N",
	"ᴏ":"O",
	"ɶ":"OE",
	"ᴐ":"O",
	"ᴕ":"OU",
	"ᴘ":"P",
	"ʀ":"R",
	"ᴎ":"N",
	"ᴙ":"R",
	"ꜱ":"S",
	"ᴛ":"T",
	"ⱻ":"E",
	"ᴚ":"R",
	"ᴜ":"U",
	"ᴠ":"V",
	"ᴡ":"W",
	"ʏ":"Y",
	"ᴢ":"Z",
	"á":"a",
	"ă":"a",
	"ắ":"a",
	"ặ":"a",
	"ằ":"a",
	"ẳ":"a",
	"ẵ":"a",
	"ǎ":"a",
	"â":"a",
	"ấ":"a",
	"ậ":"a",
	"ầ":"a",
	"ẩ":"a",
	"ẫ":"a",
	"ä":"a",
	"ǟ":"a",
	"ȧ":"a",
	"ǡ":"a",
	"ạ":"a",
	"ȁ":"a",
	"à":"a",
	"ả":"a",
	"ȃ":"a",
	"ā":"a",
	"ą":"a",
	"ᶏ":"a",
	"ẚ":"a",
	"å":"a",
	"ǻ":"a",
	"ḁ":"a",
	"ⱥ":"a",
	"ã":"a",
	"ꜳ":"aa",
	"æ":"ae",
	"ǽ":"ae",
	"ǣ":"ae",
	"ꜵ":"ao",
	"ꜷ":"au",
	"ꜹ":"av",
	"ꜻ":"av",
	"ꜽ":"ay",
	"ḃ":"b",
	"ḅ":"b",
	"ɓ":"b",
	"ḇ":"b",
	"ᵬ":"b",
	"ᶀ":"b",
	"ƀ":"b",
	"ƃ":"b",
	"ɵ":"o",
	"ć":"c",
	"č":"c",
	"ç":"c",
	"ḉ":"c",
	"ĉ":"c",
	"ɕ":"c",
	"ċ":"c",
	"ƈ":"c",
	"ȼ":"c",
	"ď":"d",
	"ḑ":"d",
	"ḓ":"d",
	"ȡ":"d",
	"ḋ":"d",
	"ḍ":"d",
	"ɗ":"d",
	"ᶑ":"d",
	"ḏ":"d",
	"ᵭ":"d",
	"ᶁ":"d",
	"đ":"d",
	"ɖ":"d",
	"ƌ":"d",
	"ı":"i",
	"ȷ":"j",
	"ɟ":"j",
	"ʄ":"j",
	"ǳ":"dz",
	"ǆ":"dz",
	"é":"e",
	"ĕ":"e",
	"ě":"e",
	"ȩ":"e",
	"ḝ":"e",
	"ê":"e",
	"ế":"e",
	"ệ":"e",
	"ề":"e",
	"ể":"e",
	"ễ":"e",
	"ḙ":"e",
	"ë":"e",
	"ė":"e",
	"ẹ":"e",
	"ȅ":"e",
	"è":"e",
	"ẻ":"e",
	"ȇ":"e",
	"ē":"e",
	"ḗ":"e",
	"ḕ":"e",
	"ⱸ":"e",
	"ę":"e",
	"ᶒ":"e",
	"ɇ":"e",
	"ẽ":"e",
	"ḛ":"e",
	"ꝫ":"et",
	"ḟ":"f",
	"ƒ":"f",
	"ᵮ":"f",
	"ᶂ":"f",
	"ǵ":"g",
	"ğ":"g",
	"ǧ":"g",
	"ģ":"g",
	"ĝ":"g",
	"ġ":"g",
	"ɠ":"g",
	"ḡ":"g",
	"ᶃ":"g",
	"ǥ":"g",
	"ḫ":"h",
	"ȟ":"h",
	"ḩ":"h",
	"ĥ":"h",
	"ⱨ":"h",
	"ḧ":"h",
	"ḣ":"h",
	"ḥ":"h",
	"ɦ":"h",
	"ẖ":"h",
	"ħ":"h",
	"ƕ":"hv",
	"í":"i",
	"ĭ":"i",
	"ǐ":"i",
	"î":"i",
	"ï":"i",
	"ḯ":"i",
	"ị":"i",
	"ȉ":"i",
	"ì":"i",
	"ỉ":"i",
	"ȋ":"i",
	"ī":"i",
	"į":"i",
	"ᶖ":"i",
	"ɨ":"i",
	"ĩ":"i",
	"ḭ":"i",
	"ꝺ":"d",
	"ꝼ":"f",
	"ᵹ":"g",
	"ꞃ":"r",
	"ꞅ":"s",
	"ꞇ":"t",
	"ꝭ":"is",
	"ǰ":"j",
	"ĵ":"j",
	"ʝ":"j",
	"ɉ":"j",
	"ḱ":"k",
	"ǩ":"k",
	"ķ":"k",
	"ⱪ":"k",
	"ꝃ":"k",
	"ḳ":"k",
	"ƙ":"k",
	"ḵ":"k",
	"ᶄ":"k",
	"ꝁ":"k",
	"ꝅ":"k",
	"ĺ":"l",
	"ƚ":"l",
	"ɬ":"l",
	"ľ":"l",
	"ļ":"l",
	"ḽ":"l",
	"ȴ":"l",
	"ḷ":"l",
	"ḹ":"l",
	"ⱡ":"l",
	"ꝉ":"l",
	"ḻ":"l",
	"ŀ":"l",
	"ɫ":"l",
	"ᶅ":"l",
	"ɭ":"l",
	"ł":"l",
	"ǉ":"lj",
	"ſ":"s",
	"ẜ":"s",
	"ẛ":"s",
	"ẝ":"s",
	"ḿ":"m",
	"ṁ":"m",
	"ṃ":"m",
	"ɱ":"m",
	"ᵯ":"m",
	"ᶆ":"m",
	"ń":"n",
	"ň":"n",
	"ņ":"n",
	"ṋ":"n",
	"ȵ":"n",
	"ṅ":"n",
	"ṇ":"n",
	"ǹ":"n",
	"ɲ":"n",
	"ṉ":"n",
	"ƞ":"n",
	"ᵰ":"n",
	"ᶇ":"n",
	"ɳ":"n",
	"ñ":"n",
	"ǌ":"nj",
	"ó":"o",
	"ŏ":"o",
	"ǒ":"o",
	"ô":"o",
	"ố":"o",
	"ộ":"o",
	"ồ":"o",
	"ổ":"o",
	"ỗ":"o",
	"ö":"o",
	"ȫ":"o",
	"ȯ":"o",
	"ȱ":"o",
	"ọ":"o",
	"ő":"o",
	"ȍ":"o",
	"ò":"o",
	"ỏ":"o",
	"ơ":"o",
	"ớ":"o",
	"ợ":"o",
	"ờ":"o",
	"ở":"o",
	"ỡ":"o",
	"ȏ":"o",
	"ꝋ":"o",
	"ꝍ":"o",
	"ⱺ":"o",
	"ō":"o",
	"ṓ":"o",
	"ṑ":"o",
	"ǫ":"o",
	"ǭ":"o",
	"ø":"o",
	"ǿ":"o",
	"õ":"o",
	"ṍ":"o",
	"ṏ":"o",
	"ȭ":"o",
	"ƣ":"oi",
	"ꝏ":"oo",
	"ɛ":"e",
	"ᶓ":"e",
	"ɔ":"o",
	"ᶗ":"o",
	"ȣ":"ou",
	"ṕ":"p",
	"ṗ":"p",
	"ꝓ":"p",
	"ƥ":"p",
	"ᵱ":"p",
	"ᶈ":"p",
	"ꝕ":"p",
	"ᵽ":"p",
	"ꝑ":"p",
	"ꝙ":"q",
	"ʠ":"q",
	"ɋ":"q",
	"ꝗ":"q",
	"ŕ":"r",
	"ř":"r",
	"ŗ":"r",
	"ṙ":"r",
	"ṛ":"r",
	"ṝ":"r",
	"ȑ":"r",
	"ɾ":"r",
	"ᵳ":"r",
	"ȓ":"r",
	"ṟ":"r",
	"ɼ":"r",
	"ᵲ":"r",
	"ᶉ":"r",
	"ɍ":"r",
	"ɽ":"r",
	"ↄ":"c",
	"ꜿ":"c",
	"ɘ":"e",
	"ɿ":"r",
	"ś":"s",
	"ṥ":"s",
	"š":"s",
	"ṧ":"s",
	"ş":"s",
	"ŝ":"s",
	"ș":"s",
	"ṡ":"s",
	"ṣ":"s",
	"ṩ":"s",
	"ʂ":"s",
	"ᵴ":"s",
	"ᶊ":"s",
	"ȿ":"s",
	"ɡ":"g",
	"ᴑ":"o",
	"ᴓ":"o",
	"ᴝ":"u",
	"ť":"t",
	"ţ":"t",
	"ṱ":"t",
	"ț":"t",
	"ȶ":"t",
	"ẗ":"t",
	"ⱦ":"t",
	"ṫ":"t",
	"ṭ":"t",
	"ƭ":"t",
	"ṯ":"t",
	"ᵵ":"t",
	"ƫ":"t",
	"ʈ":"t",
	"ŧ":"t",
	"ᵺ":"th",
	"ɐ":"a",
	"ᴂ":"ae",
	"ǝ":"e",
	"ᵷ":"g",
	"ɥ":"h",
	"ʮ":"h",
	"ʯ":"h",
	"ᴉ":"i",
	"ʞ":"k",
	"ꞁ":"l",
	"ɯ":"m",
	"ɰ":"m",
	"ᴔ":"oe",
	"ɹ":"r",
	"ɻ":"r",
	"ɺ":"r",
	"ⱹ":"r",
	"ʇ":"t",
	"ʌ":"v",
	"ʍ":"w",
	"ʎ":"y",
	"ꜩ":"tz",
	"ú":"u",
	"ŭ":"u",
	"ǔ":"u",
	"û":"u",
	"ṷ":"u",
	"ü":"u",
	"ǘ":"u",
	"ǚ":"u",
	"ǜ":"u",
	"ǖ":"u",
	"ṳ":"u",
	"ụ":"u",
	"ű":"u",
	"ȕ":"u",
	"ù":"u",
	"ủ":"u",
	"ư":"u",
	"ứ":"u",
	"ự":"u",
	"ừ":"u",
	"ử":"u",
	"ữ":"u",
	"ȗ":"u",
	"ū":"u",
	"ṻ":"u",
	"ų":"u",
	"ᶙ":"u",
	"ů":"u",
	"ũ":"u",
	"ṹ":"u",
	"ṵ":"u",
	"ᵫ":"ue",
	"ꝸ":"um",
	"ⱴ":"v",
	"ꝟ":"v",
	"ṿ":"v",
	"ʋ":"v",
	"ᶌ":"v",
	"ⱱ":"v",
	"ṽ":"v",
	"ꝡ":"vy",
	"ẃ":"w",
	"ŵ":"w",
	"ẅ":"w",
	"ẇ":"w",
	"ẉ":"w",
	"ẁ":"w",
	"ⱳ":"w",
	"ẘ":"w",
	"ẍ":"x",
	"ẋ":"x",
	"ᶍ":"x",
	"ý":"y",
	"ŷ":"y",
	"ÿ":"y",
	"ẏ":"y",
	"ỵ":"y",
	"ỳ":"y",
	"ƴ":"y",
	"ỷ":"y",
	"ỿ":"y",
	"ȳ":"y",
	"ẙ":"y",
	"ɏ":"y",
	"ỹ":"y",
	"ź":"z",
	"ž":"z",
	"ẑ":"z",
	"ʑ":"z",
	"ⱬ":"z",
	"ż":"z",
	"ẓ":"z",
	"ȥ":"z",
	"ẕ":"z",
	"ᵶ":"z",
	"ᶎ":"z",
	"ʐ":"z",
	"ƶ":"z",
	"ɀ":"z",
	"ﬀ":"ff",
	"ﬃ":"ffi",
	"ﬄ":"ffl",
	"ﬁ":"fi",
	"ﬂ":"fl",
	"ĳ":"ij",
	"œ":"oe",
	"ﬆ":"st",
	"ₐ":"a",
	"ₑ":"e",
	"ᵢ":"i",
	"ⱼ":"j",
	"ₒ":"o",
	"ᵣ":"r",
	"ᵤ":"u",
	"ᵥ":"v",
	"ₓ":"x",
	'Ð': 'D',
	'Þ': 'Th',
	'ß': 'ss',
	'ð': 'd',
	'þ': 'th',
	'&': 'and'
};

/**
 * Convert a given string's internationalized characters to their latin counters.
 *
 * @param {string} text
 * @return {string}
 */
Core.Strings.toLatin = text => {
	for(let i in Core.Strings.I18NCharacters) {
		text = text.replace(new RegExp(i, 'g'), Core.Strings.I18NCharacters[i]);
	}
	return text;
};


/**
 * Cleanup a string and ensure it can make a valid URL.
 *
 * @param {string} text
 * @return {string}
 */
Core.Strings.toURL = text => {
	var ns, i;
	// URLs should only be in latin.
	text = Core.Strings.toLatin(text);

	if(Core.URLRemoveStopWords) {
		text = text.split(' ');
		ns = '';
		for(i in text){
			if(Core.StopWords.indexOf(text[i]) === -1){
				ns += ' ' + text[i];
			}
		}
		text = ns.trim();
	}

	// Spaces get replaced with a separator
	text = text.replace(/[ ]/g, '-');

	// Anything else I missed?  Get rid of it!
	text = text.replace(/[^a-z0-9\-]/ig, '');

	// Multiple separators should get truncated, along with beginning and trailing ones.
	text = text.replace(/[-]+/g, '-').replace(/^-/, '').replace(/-$/, '');

	// Multiple empty slashes should be truncated to a single one.
	text = text.replace(/\/[\/]*/g, '/');

	// And lowercase it.
	text = text.toLowerCase();

	return text;
};

/**
 * Trim off whitespace from a given string.
 * This is required because IE does not natively support trim() :/
 *
 * @param {string} text
 * @return {string}
 */
Core.Strings.trim = text => {
	return text.replace(/^[\s]*/gm, '').replace(/[\s]*$/gm, '');
};

/**
 * Rot13 a given string
 *
 * @param {string} text
 * @returns {string}
 */
Core.Strings.rot13 = text => {
	return text.replace(/[a-zA-Z]/g, function(c) {
		return String.fromCharCode((c <= 'Z' ? 90 : 122) >= (c = c.charCodeAt(0) + 13) ? c : c - 26);
	});
};

/**
 * Very simple templating system that takes a string and replaces %1, %2, etc with that parameter.
 * This system also supports an Object for the second parameter.  This will perform named replaces.
 *
 * Example:
 * simple_template('Something [%%1%%] goes [%%2%%]', 'red', 'over there');
 * // returns 'Something red goes over there'
 *
 * Example:
 * simple_template('something [%%color%%] goes [%%location%%]', { color: 'blue', location: 'up here' });
 * // returns 'something blue goes up here'
 *
 * @return string
 */
Core.Strings.template = function() {
	var template, args = [], i = 0, regex;

	if(arguments.length < 2){
		throw 'Please provide at least two arguments to simple_format!';
	}

	template = arguments[0];

	// If an object is sent in for the second argument and there are exactly two, then a name replace should be used.
	if(arguments.length === 2 && arguments[1] instanceof Object){

		for(i in arguments[1]){
			regex = new RegExp('\\[%%' + i + '%%\\]', 'g');
			template = template.replace(regex, arguments[1][i]);
		}
		return template;
	}

	for(i=1; i<arguments.length; i++){
		regex = new RegExp('\\[%%' + i + '%%\\]', 'g');
		template = template.replace(regex, arguments[i]);
	}

	return template;
}

Core.Strings.i18n = function() {
	var template, i18ntext;

	if(arguments.length < 1) {
		throw 'Please provide at least one string to run i18n against';
	}

	template = arguments[0];

	if(arguments.length === 2 && arguments[1] === 0) {
		template = template.replace('_N_', '_0_');
	}
	else if(arguments.length === 2 && arguments[1] === 1) {
		template = template.replace('_N_', '_1_');
	}

	if(typeof window.i18n === 'undefined') {
		window.i18n = { strings: { } };
	}

	// This should be present in the window scope, if not complain.
	if(!window.i18n.strings.hasOwnProperty(template)) {
		// I18N data not available, just return the requested string.
		console.error('I18N string ' + template + ' not loaded in the window!');
		return template;
	}

	i18ntext = window.i18n.strings[template];

	for(let i = 1; i < arguments.length; i++) {
		let regex = new RegExp('\\[%' + i + '%\\]', 'g');
		i18ntext = i18ntext.replace(regex, arguments[i]);
	}

	return i18ntext;
}


String.prototype.toLatin = function(){
	return Core.Strings.toLatin(this.toString());
};

String.prototype.toURL = function(){
	return Core.Strings.toURL(this.toString());
};

// Because this is already supported in everything except IE...
if(typeof String.prototype.trim != 'function'){
	String.prototype.trim = function(){
		return Core.Strings.trim(this.toString());
	};
};

String.prototype.template = function(replacements){
	var args = [this.toString()], i;
	for(i in arguments){
		args.push(arguments[i]);
	}

	return Core.Strings.template.apply(this, args);
};

String.prototype.i18n = function() {
	let args = [ this.toString() ];
	for(let i = 0; i < arguments.length; i++) {
		args.push( arguments[i] );
	}
	return Core.Strings.i18n.apply(this, args);
};