"use strict";

(() => {

	/**
	 * Simple function to create a message box with the given message and status
	 *
	 * @param message
	 * @param type
	 *
	 * @return string
	 */
	function create_message_html(message, type) {
		return '<p class="message-' + type + '">' + message + '</p>';
	}

	/**
	 * Simple function to create an icon and optionally apply spin
	 *
	 * @param icon
	 * @param spin
	 * @returns {string}
	 */
	function create_icon_html(icon, spin) {
		let classes = ['icon', 'icon-' + icon];

		if(typeof spin !== 'undefined' && spin === true) {
			classes.push('icon-spin');
		}

		return '<i class="' + classes.join(' ') + '"></i>';
	}

	/**
	 * Create the actual table HTML
	 *
	 * @param table
	 * @param data
	 */
	function render_table_html(table, data) {
		const $data  = table.querySelector('.listing-table-data'),
			$pager   = table.querySelector('.pagination-pager'),
			$totals  = table.querySelector('.total-records'),
			$options = table.querySelector('.listing-display-options');

		let pager = '',
			$optionsInner = $options.querySelector('.fieldset-content');

		$data.innerHTML = data.body;

		if(data.options) {
			$options.style.display = 'block';
			$optionsInner.innerHTML = data.options;
		}
		else {
			$options.style.display = 'none';
		}

		if(data.pager) {
			// Build the pager
			if(data.pages > 1) {
				if (data.page !== 1) {
					// Add the icons before the page number; has to be at least after the first page.

					pager += '<a href="?page=1" title="' + 'STRING_FIRST_PAGE'.i18n() + '">' +
						create_icon_html('fast-backward') +
						'</a> ';

					if (data.page !== 2) {
						pager += '<a href="?page=' + (data.page - 1) + '" title="' + 'STRING_PREVIOUS_PAGE'.i18n() + '">' +
							create_icon_html('backward') +
							'</a> ';
					}
				}
			}

			pager += 'STRING_PAGE_N_OF_N'.i18n(data.page, data.pages);

			if(data.pages > 1) {
				if(data.pages !== data.page) {
					// Add the icons after the page number; has to be before the final page.
					if(data.page + 1 != data.pages) {
						pager += ' <a href="?page=' + (data.page + 1) + '" title="' + 'STRING_NEXT_PAGE'.i18n() + '">' +
							create_icon_html('forward') +
							'</a>';
					}

					pager += ' <a href="?page=' + (data.pages) + '" title="' + 'STRING_LAST_PAGE'.i18n() + '">' +
						create_icon_html('fast-forward') +
						'</a>';
				}
			}

			$pager.innerHTML = pager;
			$totals.innerHTML = 'STRING_TOTAL_RECORDS_N'.i18n(data.total);
		}

		// Sync the parameter data just for reference.
		for(let paramKey in data.parameters) {
			let keyName = paramKey.replace(/_/g, '-'),
				keyValue;

			if(typeof(data.parameters[paramKey]) === 'undefined') {
				keyValue = '';
			}
			else if(data.parameters[paramKey] === null) {
				keyValue = '';
			}
			else {
				keyValue = data.parameters[paramKey];
			}

			table.setAttribute('data-' + keyName, keyValue);
		}

		// Sync the standard data for reference
		table.dataset.page = data.page;
		table.dataset.pages = data.pages;
		table.dataset.total = data.total;

		// Notify any listening system that this table has loaded.
		table.dispatchEvent(new Event('load'));

	}

	/**
	 * Parse a URL and merge it with given parameters
	 *
	 * @param {string}      url        Base URL of this table
	 * @param {string}      tableName  Table name to request, (for multiple tables)
	 * @param {string|null} parameters Additional parameters to send as a GET request
	 * @param {string}      format     Format extension to request, defaults to json
	 *
	 * @returns {string}
	 */
	function parse_url(url, tableName, parameters, format) {
		if(url.indexOf('?') !== -1) {
			// Trim this off the path; it may be provided if the incoming URL had a query string.
			url = url.substr(0, url.indexOf('?'));
		}

		// Prepend the root url on this.
		url = Core.ROOT_URL + url.substr(1);


		// All requests will have an extension, defaults to a JSON request
		url += '.' + (format || 'json');

		// Always add the table name, for reference on multiple tables
		url += '?table=' + tableName;

		if(parameters !== null) {
			if(parameters.indexOf('://') !== -1 && parameters.indexOf('?') !== -1) {
				// Parameters probably are fully resolved
				parameters = parameters.substr(parameters.indexOf('?'));
			}

			// Add the requested parameters to this lookup.
			if(parameters.indexOf('?') === 0) {
				url += '&' + parameters.substr(1);
			}
			else {
				url += '&' + parameters;
			}
		}

		return url;
	}

	/**
	 * Load (or reload) table data
	 *
	 * @param table
	 * @param parameters
	 * @param background
	 */
	function load_table(table, parameters, background) {
		const url    = parse_url(table.dataset.tableBaseurl, table.dataset.tableName, parameters),
			$data    = table.querySelector('.listing-table-data'),
			$pager   = table.querySelector('.pagination-pager'),
			$totals  = table.querySelector('.total-records'),
			$refresh = table.querySelector('.auto-refresh-link i'),
			$options = table.querySelector('.listing-display-options .fieldset-content');

		$refresh.classList.add('icon-spin');

		if(!background) {
			// Set the text as a loading template
			$data.innerHTML = create_message_html(
				[
					'MESSAGE_INFO_LOADING_RESULTS'.i18n(),
					create_icon_html('spinner', true)
				].join(' '),
				'info'
			);

			// Clear the dynamic sections for active loading
			$pager.innerHTML = '';
			$totals.innerHTML = '';
			$options.innerHTML = '';
		}
		
		fetch(url, {method: 'GET'})
			.then(res => {
				/*$data.innerHTML = create_message_html(
					[
						'MESSAGE_INFO_RENDERING_RESULTS'.i18n(),
						create_icon_html('spinner', true)
					].join(' '),
					'info'
				);*/

				return res.json();
			})
			.then(data => {
				if(data.total === 0) {
					$data.innerHTML = create_message_html('MESSAGE_INFO_NO_RESULTS', 'info');
				}
				else {
					render_table_html(table, data);
				}

				$refresh.classList.remove('icon-spin');

				if(typeof(Core.DevTools) !== 'undefined' && typeof(data._debug) !== 'undefined') {
					Core.DevTools.SetPageDetails(data._debug.page);
				}
			})
			.catch(e => {
				$data.innerHTML = create_message_html(e.message, 'error');
				$refresh.classList.remove('icon-spin');
			});
	}

	function matches() {
		if(arguments.length < 2) {
			throw 'matches requires at least 2 arguments, the target followed by the matches';
		}

		const target = arguments[0];

		for(let i = 1; i < arguments.length; i++) {
			let match = target.closest(arguments[i]);
			if(match !== null) {
				// Matched one of the arguments!
				return match;
			}
		}

		return null;
	}

	function register_table_events(table) {
		let refreshTimer = null;

		// Assign a tracker so we know these event has been registered.
		// This prevents double registration.
		table.dataset.eventsRegistered = '1';

		// Assign the events for this table
		table.addEventListener('click', evt => {
			let match;

			if((match = matches(evt.target, '.pagination-pager a')) !== null) {
				load_table(table, match.href, false);
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.print-link')) !== null) {
				window.print();
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.sort-icons .sort-icon')) !== null) {
				load_table(
					table,
					'sortkey=' + match.dataset.key + '&sortdir=' + match.dataset.dir,
					false
				);
			}


			if((match = matches(evt.target, '.apply-filters', '.apply-columns')) !== null) {
				let form = match.closest('form');

				load_table(
					table,
					(new URLSearchParams(new FormData(form))).toString(),
					false
				);
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.reset-filters')) !== null) {
				let form = match.closest('form');

				form.reset();
				load_table(table, match.href,false);
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.table-change-view')) !== null) {

				// Find the view controls that are active and reset them
				table.querySelectorAll('.table-change-view').forEach(link => {
					let li = link.closest('li');
					li.classList.remove('active');
				});


				match.closest('li').classList.add('active');
				load_table(table, match.href,false);
				evt.preventDefault();
			}


			if((match = matches(evt.target, '.auto-refresh-link')) !== null) {
				let li = match.closest('li');

				if(li.classList.contains('active')) {
					li.classList.remove('active');

					if(refreshTimer !== null) {
						clearInterval(refreshTimer);
						refreshTimer = null;
					}
				}
				else {
					li.classList.add('active');

					load_table(table, null,true);

					refreshTimer = setInterval(table => {
						load_table(table, null,true);
					}, 30000, table);
				}

				evt.preventDefault();
			}


			if((match = matches(evt.target, '.bulk-select')) !== null) {
				let $actions = table.querySelector('.bulk-actions');

				if($actions.classList.contains('collapsed')) {
					// If the actions is not already expanded, expand it to provide the user hinting about what checking items controls.
					$actions.querySelector('.fieldset-title').click();
				}
			}


			if((match = matches(evt.target, '.bulk-select-all')) !== null) {
				let $actions = table.querySelector('.bulk-actions'),
					isChecked = match.checked;


				table.querySelectorAll('.bulk-select').forEach(el => {
					el.checked = isChecked;
				});

				if (
					($actions.classList.contains('collapsed') && isChecked) ||
					(!$actions.classList.contains('collapsed') && !isChecked)
				) {
					// If the actions is not already expanded, expand it to provide the user hinting about what checking items controls.
					$actions.querySelector('.fieldset-title').click();
				}
			}


			if((match = matches(evt.target, '.report-view')) !== null) {
				let form = match.closest('form');

				load_table(
					table,
					'report=' + match.dataset.report,
					false
				);
				evt.preventDefault();
			}

			if((match = matches(evt.target, '.report-download')) !== null) {
				let url = parse_url(
					table.dataset.tableBaseurl,
					table.dataset.tableName,
					'report=' + match.dataset.report,
					'pdf'
				);

				window.location.href = url;

				evt.preventDefault();
			}
		});


		// "CHANGE" Event Listener
		table.addEventListener('change', evt => {
			let match;

			if((match = evt.target.closest('.pagination-limit-selector')) !== null) {
				load_table(table, 'limit=' + match.options[match.selectedIndex].value, false);
				evt.preventDefault();
			}
		});


		// "SUBMIT" Event Listener
		table.addEventListener('submit', evt => {
			let match;

			if((match = matches(evt.target, '.bulk-actions form')) !== null) {
				//if((match = evt.target.closest('.bulk-actions form')) !== null) {

				let matchedRecords = false;

				table.querySelectorAll('.bulk-select').forEach(el => {
					if(el.checked) {
						matchedRecords = true;

						match.append(el.cloneNode(false));
					}
				});

				if(!matchedRecords) {
					alert('Please select items first.');
					evt.preventDefault();
				}
			}

			if((match = matches(evt.target, '.listing-filters form')) !== null) {
				load_table(
					table,
					(new URLSearchParams(new FormData(match))).toString(),
					false
				);

				evt.preventDefault();
			}
		});


		// "LOAD" Event Listener
		table.addEventListener('load', evt => {
			if(evt.target.dataset.view === 'collection') {
				// Work tasks support collapsible fields.
				evt.target.querySelectorAll('.collection-item.collapsed').forEach(item => {
					let collapsibles = item.querySelectorAll('.collapsible'),
						expander = item.querySelector('.collection-item-expand-hint'),
						collapser = item.querySelector('.collection-item-collapse-hint'),
						hasCollapsible = false;

					collapsibles.forEach(el => {
						if(el.scrollHeight > el.clientHeight) {
							hasCollapsible = true;
						}
					});

					if(!hasCollapsible) {
						item.classList.remove('collapsed');
					}
					else {
						expander.addEventListener('click', evt => {
							let item = evt.target.closest('.collection-item');
							item.classList.remove('collapsed');
							item.classList.add('expanded');

							evt.preventDefault();
						});

						collapser.addEventListener('click', evt => {
							let item = evt.target.closest('.collection-item');
							item.classList.add('collapsed');
							item.classList.remove('expanded');

							evt.preventDefault();
						});
					}
				});
			}

			evt.target.querySelectorAll('a[data-target="background"]').forEach(el => {
				// This link is being submitted
				el.submit = () => {
					$.notify('Loading...', 'info');
				};

				// The response of this link came back.
				el.load = data => {
					if(data.status === 200) {
						$.notify(data.message, 'success');
						load_table(table, null,true);
					}
					else {
						$.notify(data.message, 'error');
					}
				};
			});
		});
	}

	/**
	 * Check the page for any listing table and ensure it's setup.
	 */
	function check_page_for_tables() {
		let tables = document.querySelectorAll('.listing-table');

		tables.forEach(table => {
			if(table.dataset.eventsRegistered !== '1') {
				// Only register new tables.
				register_table_events(table);

				// Load the table initially with default parameters.
				load_table(table, null, false);
			}
		});
	}


	// Bind to the "change" event for background data
	window.addEventListener('change', evt => {
		check_page_for_tables();
	});

	// First load too!
	check_page_for_tables();
})();
