"use strict";

(() => {

	/**
	 * Add a script to the body, or ignore if it's already added.
	 *
	 * @param {string} src
	 */
	function add_script(src) {
		if(!src) {
			// No source, nothing to do.
			return;
		}

		for(let i = 0; i < document.scripts.length; i++) {
			if(document.scripts[i].src === src) {
				// Script already loaded, not required.
				return;
			}
		}

		// Still here?
		let script = document.createElement('script');
		document.body.appendChild(script);
		script.async = true;
		script.src = src;

		// We don't need to track when it's added, it'll run when it's ready.
	}

	function add_script_body(scriptText, renderArea) {
		let script = document.createElement('script');
		renderArea.appendChild(script);
		script.innerHTML = scriptText;
	}

	function activate_link(link) {
		let pane = link.closest('.pane-viewer'),
			paneName = pane.dataset['pane'],
			links = pane.querySelector('.pane-viewer-links'),
			renderArea = pane.querySelector('.pane-viewer-window');

		renderArea.innerHTML = 'Loading...';

		// Remember this selected link for the next page reload
		localStorage.setItem('core-pane-' + paneName, link.href);

		// Iterate over the set of links contained in the links region and set the correct one to active.
		links.querySelectorAll('a').forEach(el => {
			if(el === link) {
				// The target link gets the active flag.
				link.classList.add('active');
			}
			else {
				el.classList.remove('active');
			}
		});

		fetch(link.href, {method: 'GET'})
			.then(res => {
				if(res.headers.get('Content-Type') === 'application/json') {
					// This will be JSON.
					return res.json();
				}
				else {
					// Regular text or something.
					return res.text();
				}
			})
			.then(response => {
				if(typeof response === 'string') {
					renderArea.innerHTML = response;
				}
				else if(typeof response === 'object' && typeof response.body === 'string') {
					renderArea.innerHTML = response.body;

					// Allow rich JSON data to provide supplemental scripts.
					if(typeof response.scripts === 'object') {
						for(let i = 0; i < response.scripts.length; i++) {
							add_script(response.scripts[i]);
						}
					}

					if(typeof response.scripts_inline === 'object') {
						for(let i = 0; i < response.scripts_inline.length; i++) {
							add_script_body(response.scripts_inline[i], renderArea);
						}
					}

					if(typeof(Core.DevTools) !== 'undefined' && typeof(response._debug) !== 'undefined') {
						Core.DevTools.SetPageDetails(response._debug.page);
					}
				}
				else {
					console.error(response);
					throw new Error('Unsupported response from the server!');
				}

				window.dispatchEvent(new Event('change'));
			})
			.catch(e => {
				renderArea.innerHTML = e.message;
			});
	}

	document.querySelectorAll('.pane-viewer').forEach(pane => {
		let links = pane.querySelector('.pane-viewer-links'),
			paneName = pane.dataset['pane'],
			firstLink = links.querySelector('a'),
			selected = localStorage.getItem('core-pane-' + paneName);

		links.addEventListener('click', evt => {
			let match = evt.target.closest('a');

			if(match !== null) {

				activate_link(match);

				evt.preventDefault();
			}
		});

		// If there's not one selected, then activate the first link.
		if(!selected) {
			activate_link(firstLink);
		}
		else {
			// Find the selected link and try to activate it.
			let selectedLink = null;

			links.querySelectorAll('a').forEach(link => {
				if(link.href === selected) {
					activate_link(link);
					selectedLink = link;
					return false;
				}
			});

			// Fallback
			if(selectedLink === null) {
				activate_link(firstLink);
			}
		}
	});
})();