<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

use Core\AccessDeniedException;
use Core\BadRequestException;
use Core\Component;
use Core\GPG\Key as GPGKey;
use Core\GPG\PrimaryKey as GPGPrimaryKey;
use Core\NotFoundException;
use Core\CLI\CLI;
use Core\Controller;
use Core\Core;
use Core\Dict;
use Core\Installer\InstallHelper;
use Core\Theme;
use Core\I18N\I18NScanner;
use Core\Utilities\Changelog\Parser as ChangelogParser;
use Core\VersionString;
use Core\View;
use Core\Filestore\Factory as FileFactory;

/**
 * Description of DevController
 *
 * @author charlie
 */
class DevController extends Controller {

	private $_packagerSettings;
	
	/**
	 * Used in the documentation URL helper to know what base this file is referenced from.
	 * @var string
	 */
	//private $_docsBase;
	
	public function __common() {
		if(!DEVELOPMENT_MODE){
			throw new AccessDeniedException('Access to development tools requires DEVELOPMENT_MODE to be set to true in your configuration.xml!');
		}
		
		$this->view->mode = View::MODE_NOOUTPUT;
	}
	
	public function docs() {
		$request = $this->getRequest();
		$view    = $this->getView();
		
		// Allow this page to be displayed in both modes.
		if(EXEC_MODE === 'WEB') {
			$view->mode = View::MODE_PAGE;
		}
		
		// Required because Core is often times symlinked to a location that contains other symlinks.
		$realpath = realpath(realpath(ROOT_PDIR) . '/../') . '/';
		
		// Populated with the various document paths
		$docpaths = [];
		
		if(is_dir($realpath . 'docs/')){
			$docpaths['self'] = $realpath . 'docs/';
		}
		
		if(is_dir($realpath . 'submodules/')){
			$dir = opendir($realpath . 'submodules/');
			while($file = readdir($dir)) {
				if($file[0] === '.'){
					continue;
				}
				
				if(is_dir($realpath . 'submodules/' . $file . '/docs/')){
					$docpaths[ $file ] = $realpath . 'submodules/' . $file . '/docs/';
				}
			}
			closedir($dir);
		}
		
		if($request->getParameter(0)){
			$view->addBreadcrumb('Development Documentation', '/dev/docs');
			
			$browser = new MarkdownBrowser\Browser($docpaths[ $request->getParameter(0) ], '/dev/docs/' . $request->getParameter(0) . '/');
			
			if(count($request->getParameters()) > 1 && $request->getParameter(1) !== '') {
				// A subfile was requested.
				$params = $request->getParameters();
				// Drop the first parameter off, this is the source of documentation requested.
				array_splice($params, 0, 1);
				$browser->setPath(implode('/', $params) . '.' . $request->ext);
			}
			else{
				$browser->setPath('');
			}
			
			//var_dump($browser, $request->getParameters()); die();
			
			if($browser->checkIsMarkdown()) {
				foreach($browser->getBreadcrumbs() as $arr) {
					$view->addBreadcrumb($arr['title'], $arr['link']);
				}
				$view->title = $browser->getTitle();
				$view->assign('html', $browser->getHTML());
			}
			else{
				$browser->sendToBrowser();
			}
		}
		else{
			$view->assign('html', null);
		}
	}
	
	/**
	 * Enable development tools on this site.
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 */
	public function enable(){
		
		if(InstallHelper::GetUser() !== 'root'){
			throw new BadRequestException('Please run this enable script as root so permissions can be changed!');
		}
		
		$res = CLI::PromptUser('This will install development utilities and configure the site to operate as a standard user; continue?', 'bool');
		if(!$res){
			CLI::PrintLine('Bye!');
		}
		
		// Caclulate what username the developer is expecting to run the scripts as.
		$gid = stat(ROOT_PDIR)['gid'];
		$group = explode(':', trim(exec('getent group ' . $gid)))[0];
		
		$res = CLI::PromptUser('Is your primary group named ' . $group . '?', 'bool', true);
		if(!$res){
			$group = CLI::PromptUser('Please enter your primary group', 'text-required');
		}
		
		
		// chmod all the necessary files to be accessable by the user.
		$files = [
			ROOT_PDIR . 'config/',
			ROOT_PDIR . 'logs/',
			ROOT_PDIR . 'gnupg/',
			ASSET_PDIR,
			PUBLIC_PDIR,
			PRIVATE_PDIR,
			TMP_DIR
		];
		exec('chgrp -R ' . $group . ' ' . implode(' ', array_map('escapeshellarg', $files)));
		exec('chmod -R g+rwX ' . implode(' ', array_map('escapeshellarg', $files)));
		
		$checks = InstallHelper::GetDevelopmentChecks();
		
		foreach($checks as $c){
			if(!$c->test()){
				CLI::PrintError($c->getTitle() . ' could not be resolved automatically');
				CLI::PrintDebug($c->getFix());
			}
		}
	}
	
	/**
	 * Rebuild the I18N YML database for a given component or theme
	 * 
	 * The i18nscan function should be called if any changes have been performed prior to running this.
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component Component name to scan and rebuild
	 * @param string $theme     Theme name to scan and rebuild
	 * @param string $lang      Language to build the database
	 */
	public function i18n_build( string $component = null, string $theme = null, string $lang = null ) {
		if($component){
			$src = [ Core::GetComponent($component)->getBaseDir() . 'i18n/strings.txt' ];
		}
		elseif($theme){
			$src = [ Theme::GetTheme($theme)->getBaseDir() . 'i18n/strings.txt' ];
		}
		else{
			// Find all strings database in the site.
			$src = [];
			exec('find -L ' . escapeshellarg(ROOT_PDIR) . ' -name strings.txt', $src);
		}
		
		if(!count($src)){
			CLI::PrintError('No strings.txt databases located!');
			CLI::PrintLine('Please run /dev/i18nscan first.');
			return;
		}
		
		if($lang === null){
			// Determine the language based on the user's current session.
			foreach($this->request->acceptLanguages as $check){
				if(strpos($check, '_') === false){
					// Skip the locale for this; if the user did not specify, 
					// then defer to the base language.
					$lang = $check;
					break;
				}
			}
		}
		
		foreach($src as $str){
			I18NScanner::BuildYAMLFile($str, $lang);
			CLI::PrintLine('Compiled ' . $str . ' into ' . $lang);
		}
	}
	
	/**
	 * Merge a source YML file into a dest YML
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 */
	public function i18n_merge( ) {
		
		$src = $this->request->getParameter(0);
		$dst = $this->request->getParameter(1);
		
		if(!($src && $dst)){
			throw new BadRequestException('You must specify exactly two arguments; a source and a destination');
		}
		
		if(!file_exists($src)){
			throw new BadRequestException('Source file ' . $src . ' does not exist!');
		}
		
		if(!file_exists($dst)){
			throw new BadRequestException('Destination file ' . $src . ' does not exist!');
		}
		
		$srcData = yaml_parse_file($src);
		$dstData = yaml_parse_file($dst);
		
		// What to do now depends on the basenames of the yml files; that indicates the version.
		$srcType = basename($src, '.yml');
		$dstType = basename($dst, '.yml');
		
		if($srcType === 'strings' && $dstType === 'strings'){
			// Just recursively merge both nested arraysets.
			$dstData = array_replace_recursive($srcData, $dstData);
		}
		elseif($srcType === 'strings'){
			// Source is the multi-dimensional array; retrieve just the language requested from there.
			foreach($srcData as $k => $dat){
				if(
					isset($dat[ $dstType ]) &&
					$dat[ $dstType ] !== '' &&
					isset($dstData[ $k ]) &&
					$dstData[ $k ] === ''
				){
					// Existing key, but it's empty on the destination and present on the source!
					$dstData[ $k ] = $dat[ $dstType ];
				}
				elseif(
					isset($dat[ $dstType ]) &&
					!isset($dstData[ $k ])
				){
					// New key!
					$dstData[ $k ] = $dat[ $dstType ];
				}
			}
		}
		elseif($dstType === 'strings'){
			// Destination is the multi-dimensional array; expand the source to fit the expected format
			foreach($srcData as $k => $str){
				if(
					isset($dstData[ $k ]) &&
					isset($dstData[ $k ][ $srcType ]) &&
					$str !== '' &&
					$dstData[ $k ][ $srcType ] === ''
				){
					// Flat version has a string for this lang, but destination is empty
					$dstData[ $k ][ $srcType ] = $str;
				}
				elseif(
					isset($dstData[ $k ]) &&
					!isset($dstData[ $k ][ $srcType ])
				){
					// Destination has the key, but not the language
					$dstData[ $k ][ $srcType ] = $str;
				}
				elseif(
					!isset($dstData[ $k ])
				){
					// Destination doesn't even have the key
					$dstData[ $k ] = [
						$srcType => $src,
					];
				}
			}
		}
		else{
			// Neither are multidimensional files!
			$dstData = array_replace($srcData, $dstData);
		}
		
		// Sort the translations!
		ksort($dstData);
		
		yaml_emit_file(
			$dst, 
			$dstData,
			YAML_UTF8_ENCODING, 
			YAML_LN_BREAK
		);
	}
	
	/**
	 * Scan and rebuild the I18N constant database for a given component or theme
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component Component name to scan and rebuild
	 * @param string $theme     Theme name to scan and rebuild
	 */
	public function i18n_scan( string $component = null, string $theme = null ) {
		if(!($component || $theme)){
			throw new BadRequestException('Please include either a component or theme to manage.');
		}
		
		$matches = [];
		
		if($component){
			$src = Core::GetComponent($component);
			$configs = ConfigModel::Find(['component = ' . $src->getKeyName()]);
		}
		elseif($theme){
			$src = Theme::GetTheme($theme);
			$configs = ConfigModel::Find(['component = theme/' . $src->getKeyName()]);
		}
		
		
		array_push($matches, ...I18NScanner::ScanDirectory($src->getBaseDir()));
		
		foreach($configs as $config){
			array_push($matches, ...I18NScanner::ScanConfig($config));
		}
		
		foreach($src->getPermissions() as $perm){
			array_push($matches, ...I18NScanner::ScanPermission($perm));
		}
		
		foreach(array_keys($src->getModelList()) as $model){
			array_push($matches, ...I18NScanner::ScanModel(new $model()));
		}
		
		foreach(array_keys($src->getSupplementalModelList()) as $model){
			array_push($matches, ...I18NScanner::ScanSupplementalModel(new $model()));
		}
		
		// Strip duplicates
		$matches = array_unique($matches);
		
		// Sort the translations!
		sort($matches);
		
		// If this is a component, then load Core's strings so that this script knows which strings are NOT required.
		// Anything in Core is assumed to be present already, as that's the basis for every package!
		// @todo
		if(
			$src->getKeyName() !== 'core' && 
			file_exists(ROOT_PDIR . 'core/i18n/strings.txt')
		){
			$coreTrans = file(ROOT_PDIR . 'core/i18n/strings.txt',  FILE_IGNORE_NEW_LINES );
			
			foreach($matches as $k => $m){
				// For each match, skim through Core's i18n file for this same string.
				// If found, remove from array.
				if(in_array($m, $coreTrans)){
					unset($matches[$k]);
				}
			}
		}
		
		// Save the flat string constants to a plain text file for translators to poll from.
		$flat = $src->getBaseDir() . 'i18n/strings.txt';
		if(!file_exists($flat)){
			touch($flat);
		}
		$existing = md5_file($flat);
		file_put_contents($flat, implode("\n", $matches));
		$new = md5_file($flat);
		
		if($existing !== $new){
			CLI::PrintLine('Updated ' . $flat . ' with ' . count($matches) . ' I18N string constants');
			
			// Call the builder for convenience.
			$this->i18n_build($component, $theme);
		}
		else{
			CLI::PrintLine('No changes detected');
		}
	}
	
	/**
	 * Minfiy JS resources for the given resource
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component Name of the component to minify JS
	 * @param string $theme     Name of the theme to minify JS
	 */
	public function js(string $component = null, string $theme = null) {
		$files = [];
		
		if($component){
			$src = Core::GetComponent($component);
		}
		elseif($theme){
			$src = Theme::GetTheme($theme);
		}
		else{
			throw new BadRequestException('Please include either a component or theme to minify JS.');
		}

		$log = [];
		
		exec('find "' . $src->getBaseDir() . 'assets/js/" -name "[a-z_\-0-9]*.js" | grep -v "\.min\.js" | grep -v jquery', $files);
		$progressEach = '+' . (100 / count($files));
		
		foreach($files as $file){

			$extension = pathinfo($file, PATHINFO_EXTENSION);
			$outfilename = basename($file, $extension);
			$outdirname  = dirname($file) . '/';

			$minfile = $outdirname . $outfilename . 'min.js';
			
			$mtimeOrig = filemtime($file);
			$mtimeMin  = file_exists($minfile) ? filemtime($minfile) : 0;

			if($mtimeOrig !== $mtimeMin) {
				// Only minify changed files.
				//CLI::PrintProgressBar($progressEach, 'Compiling ' . $outfilename . 'js');

				// @todo Fix this for better reliability
				$location = realpath(ROOT_PDIR . '../submodules/core/vendor/google-closure-compiler/');
				if(file_exists($location . '/closure-compiler-v20200719.jar')) {
					$cmd = 'java -jar ' . escapeshellarg($location . '/closure-compiler-v20200719.jar') . ' ' . escapeshellarg($file) . ' > ' . escapeshellarg($minfile);
				}
				else {
					$cmd = 'closure-compiler ' . escapeshellarg($file) . ' > ' . escapeshellarg($minfile);
				}

				$ret = null;
				system($cmd, $ret);

				if($ret !== 0) {
					// The closure command failed!
					// This means the minified file will *NOT* be available.
					// instead of having an empty file, having a non-minified file is at least something.
					echo COLOR_ERROR . 'F' . COLOR_RESET;
					copy($file, $minfile);

					$log[] = COLOR_ERROR . 'Failed to minify ' . $file . ', using unminified version instead' . COLOR_RESET;
				}
				else {
					echo '.';
					touch($minfile, $mtimeOrig);

					$log[] = COLOR_SUCCESS . 'Minified ' . $file . COLOR_RESET;
				}
			}
			else{
				echo 's';

				$log[] = 'Skipped ' . $file . ', minified file timestamp matches';
			}
		}

		echo "\n";
		CLI::PrintSuccess('Done!');
		foreach($log as $line) {
			echo $line . "\n";
		}
	}
	
	/**
	 * Package management for a component or theme
	 *
	 * Optional Arguments:
	 *   --recompile : Include to recompile the code and exit
	 *   --package : Include to package the current version
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component Name of component to manage
	 * @param string $theme     Name of Theme to manage
	 */
	public function package( string $component = null, string $theme = null ){
		if($component){
			$src = Core::GetComponent($component);
		}
		elseif($theme){
			$src = Theme::GetTheme($theme);
		}
		else{
			throw new BadRequestException('Please include either a component or theme to manage.');
		}
		
		if($this->getRequest()->getParameter('recompile')){
			// Recompile and exit without asking the user.
			$this->_saveComponent($src);
			return;
		}

		if($this->getRequest()->getParameter('package')) {
			$this->_performPackage($src);
			return;
		}
		
		$options = new Dict();
		$options->addItem('description', 'Set _D_escription');
		$options->addItem('logo', 'Set _L_ogo');
		$options->addItem('version', 'Set _V_ersion Number');
		$options->addItem('save', '_S_ave and Exit');
		$options->addItem('package', '_P_ackage Theme/Component');
		$options->addItem('quit', '_Q_uit without Saving');
		
		$c = 0;
		while($c < 50){
			++$c; // Anti-infinite loop protection.
			
			CLI::PrintHeader('Packager for ' . $src->getName() . ' ' . $src->getVersion());
			$opt = CLI::PromptUser('What do you want to do?', $options);
			
			switch($opt){
				case 'description':
					$src->setDescription(
						CLI::PromptUser('Enter a description', 'textarea', $src->getDescription())
					);
					break;
				case 'logo':
					$src->setLogo(
						CLI::PromptUser(
							'New Logo', 
							'file', 
							$src->getBaseDir() . $src->getLogo(), 
							'Please select a new logo for ' . $src->getName()
						)
					);
					break;
				case 'version':
					$version = new VersionString($src->getVersion());
					$version->increment();
					$src->setVersion(
						CLI::PromptUser('New Version', 'text-required', (string)$version, 'Please enter the new version for ' . $src->getName())
					);
					break;
				case 'save':
					$this->_saveComponent($src);
					return;
				case 'quit':
					return;
				case 'package':
					$this->_performPackage($src);
					return;
				default:
					CLI::PrintLine('Bye!');
					return;
			}
		}	
	}

	/**
	 * @core-exec-mode CLI
	 * @core-development-mode
	 */
	public function bundle() {
		$version = date('Y.m.d');

		$path = realpath(ROOT_PDIR . '../exports/');
		if(!is_dir($path)) {
			mkdir($path);
		}
		$yml = $path . '/bundles.yml';
		$path .= '/bundles/';
		if(!is_dir($path)) {
			mkdir($path);
		}

		$bundles = yaml_parse_file($yml);

		foreach($bundles as $bundle => $bundleData) {
			CLI::PrintHeader('Processing Bundle ' . $bundle . ' v' . $version);

			$bundlePath = $path . $bundle . '/';
			$bundleFile = $bundle . '-' . $version . '.tar';

			if(!is_dir($bundlePath)) {
				mkdir($bundlePath);
			}

			if(file_exists($bundlePath . $bundleFile)) {
				unlink($bundlePath . $bundleFile);
			}

			CLI::PrintLine('Bundling Core');
			Core::GetComponent('core')->package($bundlePath . $bundleFile, false);

			foreach($bundleData['components'] as $component) {
				CLI::PrintLine('Bundling ' . $component);
				Core::GetComponent($component)->package($bundlePath . $bundleFile, false);
			}

			foreach($bundleData['themes'] as $theme) {
				CLI::PrintLine('Bundling ' . $theme);
				Theme::GetTheme($theme)->package($bundlePath . $bundleFile, false);
			}

			// Create a temp directory to save some various dynamic assets.
			$dir = FileFactory::Directory('tmp/bundle/');
			$dir->mkdir();

			touch($dir->getPath() . '.default-theme');
			file_put_contents($dir->getPath() . '.default-theme', $bundleData['default-theme']);

			exec('tar --append --directory=' . escapeshellarg($dir->getPath()) . ' --file=' . escapeshellarg($bundlePath . $bundleFile) . ' .default-theme');

			$dir->delete();

			if(file_exists($bundlePath . $bundleFile . '.gz')) {
				CLI::PrintLine($bundleFile . '.gz already exists; replacing!');
				unlink($bundlePath . $bundleFile . '.gz');
			}

			CLI::PrintLine('Compressing and Finalizing export');
			exec('gzip ' . escapeshellarg($bundlePath . $bundleFile));
		}
	}
	
	/**
	 * Reinstall the given object
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component Name of component; will only reinstall that one component
	 * @param string $theme     Name of Theme; will only reinstall that one theme
	 * @param string $only      Set to "assets" to only reinstall assets; useful for theme development
	 */
	public function reinstall(string $component = null, string $theme = null, string $only = null) {
		$sources = [];
		$errors = false;
		
		if($component){
			$sources[] = Core::GetComponent($component);
		}
		elseif($theme){
			$sources[] = Theme::GetTheme($theme);
		}
		else{
			$sources = Core::GetComponents();
			$sources[] = Theme::GetTheme();
		}
		
		foreach($sources as $src){
			if($only === 'assets'){
				$changes = $src->installAssets();
			}
			else{
				$changes = $src->reinstall();
			}

			foreach($changes['changes'] as $change) {
				CLI::PrintLine($change);
			}
			foreach($changes['errors'] as $change) {
				CLI::PrintError($change);
				$errors = true;
			}
		}

		if($errors) {
			CLI::PrintWarning('Done with errors');
		}
		else {
			CLI::PrintSuccess('Done!');
		}
	}
	
	/**
	 * Compile SASS/SCSS resources for the given resource
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component Name of the component to compile SASS
	 * @param string $theme     Name of the theme to compile SASS
	 */
	public function sass(string $component = null, string $theme = null) {
		$files = [];
		
		if($component){
			$src = Core::GetComponent($component);
		}
		elseif($theme){
			$src = Theme::GetTheme($theme);
		}
		else{
			throw new BadRequestException('Please include either a component or theme to compile SASS.');
		}
		
		$rootDOM = $src->getXMLLoader()->getRootDOM();
		
		$includePaths = [];
		$includePaths[] = $src->getBaseDir() . 'assets/scss/';
		
		$includes = $rootDOM->getAttribute('sass-includes');
		if($includes){
			$paths = array_map('trim', explode(';', $includes));
			foreach($paths as $path){
				if(strpos($path, 'theme/') === 0){
					$includePaths[] = ROOT_PDIR . 'themes/' . substr($path, 6) . '/assets/scss/';
				}
				elseif($path === 'core'){
					$includePaths[] = ROOT_PDIR . 'core/assets/scss/';
				}
				else{
					$includePaths[] = ROOT_PDIR . 'components/' . $path . '/assets/scss/';
				}
			}
		}
		
		
		exec('find "' . $src->getBaseDir() . 'assets/scss/" -name "[a-z]*.scss"', $files);
		$progressEach = '+' . (100 / count($files) / 2);
		
		foreach($files as $file){

			$extension = pathinfo($file, PATHINFO_EXTENSION);
			$outfilename = basename($file, $extension);
			$outdirname  = dirname($file) . '/';

			// Remap assets/scss/* to assets/css/*
			$outdirname = str_replace('/assets/scss/', '/assets/css/', $outdirname);

			if(!is_dir($outdirname)){
				mkdir($outdirname, 0755, true);
			}

			$cssfile = $outdirname . $outfilename . 'css';
			$minfile = $outdirname . $outfilename . 'min.css';
			
			
			// Allow extends to be used; this retrieve the assets from that base theme.
			if($includePaths){
				$loadPath = '';
				foreach($includePaths as $path){
					$loadPath .= ' -I ' . escapeshellarg($path);
				}
			}
			else{
				$loadPath = '';
			}

			//CLI::PrintProgressBar($progressEach, 'Compiling ' . $outfilename);
			echo '.';
			$cmd = 'sassc' . $loadPath . ' -m -t expanded ' . escapeshellarg($file) . ' ' . escapeshellarg($cssfile);
			system($cmd);

			//CLI::PrintProgressBar($progressEach, 'Compiling ' . $outfilename);
			echo '.';
			$cmd = 'sassc' . $loadPath . ' -M -t compressed ' . escapeshellarg($file) . ' ' . escapeshellarg($minfile);
			system($cmd);
		}
		
		CLI::PrintSuccess('Done!');
	}

	/**
	 * Display a list of all Models registered on the system and links for details on each.
	 *
	 * @core-development-mode
	 * @core-exec-mode BOTH
	 */
	public function models() {
		if(EXEC_MODE === 'WEB' && !\Core\user()->checkAccess('g:admin')) {
			throw new AccessDeniedException();
		}

		$view = $this->getView();
		$components = \Core\Core::GetComponents();
		$models = [];
		$modelSources = [];
		$results = [];

		foreach($components as $component) {
			$classes = $component->getClassList();
			$componentName = $component->getName();

			foreach(array_keys($classes) as $class) {
				if(strrpos($class, 'Model') === strlen($class) - 5 && strpos($class, '\\') === false) {
					$models[] = $class;
					$modelSources[ $class ] = $componentName;
				}
			}
		}

		sort($models);

		// Build the data of everything
		foreach($models as $modelName) {
			$result = [
				'name' => $modelName,
				'source' => $modelSources[ $modelName ],
				'error' => '',
				'has_search' => false,
				'has_updated' => false,
				'has_created' => false,
				'has_deleted' => false,
				'column_count' => 0,
				'export_group' => null,
				'primary_key_count' => 0,
				'index_count' => 0,
			];

			try {
				$ref = new ReflectionClass($modelName);
			}
			catch(ReflectionException $ex) {
				$result['error'] = $ex->getMessage();
				$results[] = $result;
				continue;
			}

			/** @var \Core\ModelSchema $schema */
			$schema = $ref->getMethod('GetSchema')->invoke(null);
			$indexes = $ref->getMethod('GetIndexes')->invoke(null);

			$result['has_search'] = $ref->getProperty('HasSearch')->getValue(null);
			$result['export_group'] = $ref->getProperty('ExportGroup')->getValue(null);
			$result['column_count'] = count($schema->definitions);

			$result['has_updated'] = isset($schema->definitions['updated']) && $schema->definitions['updated'] instanceof \Core\Datamodel\Columns\SchemaColumn___updated;
			$result['has_created'] = isset($schema->definitions['created']) && $schema->definitions['created'] instanceof \Core\Datamodel\Columns\SchemaColumn___created;
			$result['has_deleted'] = isset($schema->definitions['deleted']) && $schema->definitions['deleted'] instanceof \Core\Datamodel\Columns\SchemaColumn___deleted;

			if(isset($indexes['primary'])) {
				$result['primary_key_count'] = is_array($indexes['primary']) ? count($indexes['primary']) : 1;
				$result['index_count'] = count($indexes) - 1;
			}
			else {
				$result['index_count'] = count($indexes);
			}

			if($result['export_group'] && strpos($result['export_group'], 't:') === 0) {
				$result['export_group'] = t(substr($result['export_group'], 2));
			}

			$results[] = $result;
		}


		if(EXEC_MODE === 'CLI') {
			$table = new \Core\CLI\CLITable();

			$table->addNamedColumn('name', 'Model Name');
			$table->addNamedColumn('source', 'Component');
			$table->addNamedColumn('export_group', 'Export');
			$table->addNamedColumn('has_search', 'Srch');
			$table->addNamedColumn('has_updated', 'Updt');
			$table->addNamedColumn('has_created', 'Crtd');
			$table->addNamedColumn('has_deleted', 'Dltd');
			$table->addNamedColumn('column_count', 'Col#', 'right');
			$table->addNamedColumn('primary_key_count', 'PK#', 'right');
			$table->addNamedColumn('index_count', 'Idx#', 'right');

			foreach($results as $row) {
				if($row['error']) {
					CLI::PrintError($row['error']);
				}
				else {
					unset($row['error']);
					$table->addRecord($row);
				}
			}

			$table->render();
		}
		else {
			$view->mastertemplate = 'admin';
			$view->title = 'Model Definitions';
			$view->mode = View::MODE_PAGE;
			$view->assign('models', $results);
		}
	}

	/**
	 * Display details of a specific Model registered on the system
	 *
	 * @core-development-mode
	 * @core-exec-mode BOTH
	 */
	public function model_details() {
		if(EXEC_MODE === 'WEB' && !\Core\user()->checkAccess('g:admin')) {
			throw new AccessDeniedException();
		}

		$view = $this->getView();
		$request = $this->getRequest();

		$modelName = $request->getParameter(0);
		if(!$modelName) {
			throw new \Core\RedirectException('/dev/models');
		}

		try {
			$ref = new ReflectionClass($modelName);
		}
		catch(ReflectionException $ex) {
			throw new NotFoundException();
		}

		/** @var \Core\ModelSchema $schema */
		$schema = $ref->getMethod('GetSchema')->invoke(null);
		$indexes = $ref->getMethod('GetIndexes')->invoke(null);

		var_dump($schema->definitions); die();

		// @todo Finish this.

		$result['has_search'] = $ref->getProperty('HasSearch')->getValue(null);
		$result['export_group'] = $ref->getProperty('ExportGroup')->getValue(null);
		$result['column_count'] = count($schema->definitions);

		$result['has_updated'] = isset($schema->definitions['updated']) && $schema->definitions['updated'] instanceof \Core\Datamodel\Columns\SchemaColumn___updated;
		$result['has_created'] = isset($schema->definitions['created']) && $schema->definitions['created'] instanceof \Core\Datamodel\Columns\SchemaColumn___created;
		$result['has_deleted'] = isset($schema->definitions['deleted']) && $schema->definitions['deleted'] instanceof \Core\Datamodel\Columns\SchemaColumn___deleted;

		if(isset($indexes['primary'])) {
			$result['primary_key_count'] = is_array($indexes['primary']) ? count($indexes['primary']) : 1;
			$result['index_count'] = count($indexes) - 1;
		}
		else {
			$result['index_count'] = count($indexes);
		}

		if($result['export_group'] && strpos($result['export_group'], 't:') === 0) {
			$result['export_group'] = t(substr($result['export_group'], 2));
		}


		if(EXEC_MODE === 'CLI') {
			$table = new \Core\CLI\CLITable();

			$table->addNamedColumn('name', 'Model Name');
			$table->addNamedColumn('source', 'Component');
			$table->addNamedColumn('export_group', 'Export');
			$table->addNamedColumn('has_search', 'Srch');
			$table->addNamedColumn('has_updated', 'Updt');
			$table->addNamedColumn('has_created', 'Crtd');
			$table->addNamedColumn('has_deleted', 'Dltd');
			$table->addNamedColumn('column_count', 'Col#', 'right');
			$table->addNamedColumn('primary_key_count', 'PK#', 'right');
			$table->addNamedColumn('index_count', 'Idx#', 'right');

			foreach($results as $row) {
				if($row['error']) {
					CLI::PrintError($row['error']);
				}
				else {
					unset($row['error']);
					$table->addRecord($row);
				}
			}

			$table->render();
		}
		else {
			$view->mastertemplate = 'admin';
			$view->title = 'Model Definitions';
			$view->mode = View::MODE_PAGE;
			$view->assign('models', $results);
		}
	}
	
	/**
	 * Provide documentation hints for a given Model or component to support Core 6.3 property accessing
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component Name of the component to compile SASS
	 * @param string $model     Name of the theme to compile SASS
	 */
	public function model_documentation(string $component = null, string $model = null) {
		
		if($component){
			$models = array_keys(Core::GetComponent($component)->getModelList());
		}
		elseif($model){
			if(substr($model, -5) !== 'Model' && substr($model, -5) !== 'model'){
				$model .= 'Model';
			}
			
			$models = [ $model ];
		}
		else{
			throw new BadRequestException('Please specify either a component or model to provide documentation');
		}
		
		foreach($models as $m){
			
			$schema = new \Core\ModelSchema($m);
			$width1 = 0;
			$width2 = 0;
			$properties = [];
			
			foreach($schema->definitions as $key => $column){
				
				$type    = $column->getTypecast();
				$prop    = '$' . $key;
				$comment = $column->comment;
				
				$properties[ $key ] = [ $type, $prop, $comment ];
				
				$width1 = max($width1, strlen($type) + 1);
				$width2 = max($width2, strlen($prop) + 1);
			}
			
			foreach($schema->metas as $key => $dat) {
				// Metas do not support automatic typecasting.
				$type    = 'mixed';
				$prop    = '$' . $key;
				
				if(isset($dat['comment']) ){
					$comment = $dat['comment'];
				}
				elseif(isset($dat['formatter']) && is_array($dat['formatter']) ){
					$comment = 'Meta field that calls ' . $dat['formatter'][0] . '->' . $dat['formatter'][1];
				}
				else{
					$comment = 'Meta field';
				}
				
				$properties[ $key ] = [ $type, $prop, $comment ];
				
				$width1 = max($width1, strlen($type) + 1);
				$width2 = max($width2, strlen($prop) + 1);
			}
			
			$doc = '/**' . "\n" . 
				' * Class ' . $m . "\n" .
				' *' . "\n";
			
			ksort($properties);
			foreach($properties as $prop){
				$line = '@property ' . 
					$prop[0] . str_repeat(' ', ($width1 - strlen($prop[0]))) .
					$prop[1];
				
				if($prop[2]){
					$line .= str_repeat(' ', ($width2 - strlen($prop[1]))) . $prop[2];
				}
				
				$doc .= ' * ' . $line . "\n";
			}
			
			$doc .= ' */' . "\n\n";
			
			echo $doc;
		}
	}
	
	/**
	 * Perform the "next" upgrades for a given component or theme.
	 * 
	 * @core-exec-mode CLI
	 * @core-development-mode
	 * 
	 * @param string $component
	 * @param string $theme
	 * @throws BadRequestException
	 */
	public function upgrade_next(string $component = null, string $theme = null) {
		if($component){
			$source = Core::GetComponent($component);
		}
		elseif($theme){
			$source = Theme::GetTheme($theme);
		}
		else{
			throw new BadRequestException('Please include either a component or a theme to process the next upgrades');
		}
		
		$source->upgradeNext();
	}

	private function _loadPackagerSettings() : array {
		if($this->_packagerSettings === null) {
			CLI::PrintDebug('Loading packager settings');
			$this->_packagerSettings = CLI::LoadSettingsFile(
				'packager',
				[
					'packagerKey',
					'packagerName',
					'packagerEmail',
				]
			);

			CLI::PrintDebug('Loading GPG data');
			$gpg = new \Core\GPG\GPG();
			$gpg->homedir = $_SERVER['HOME'] . '/.gnupg';

			if($this->_packagerSettings['packagerKey']) {
				CLI::PrintDebug('Current Packager Key: ' . $gpg->getKey($this->_packagerSettings['packagerKey'])->getLabel('text/plain'));
			}
			else {
				CLI::PrintHeader('Available Packager Keys');
				$keys = $gpg->listSecretKeys();
				foreach($keys as $key) {
					CLI::PrintLine($key->getLabel('text/plain'));
				}
				$this->_packagerSettings['packagerKey'] = CLI::PromptUser(
					'Please provide your GPG key you wish to use for packaging',
					'text-required'
				);
			}

			if($this->_packagerSettings['packagerName']) {
				CLI::PrintDebug('Current Packager Name: ' . $this->_packagerSettings['packagerName']);
			}
			else {
				$this->_packagerSettings['packagerName'] = CLI::PromptUser(
					'Please enter your full name or nickname to use for packaging',
					'text-required'
				);
			}

			if($this->_packagerSettings['packagerEmail']) {
				CLI::PrintDebug('Current Packager Email: ' . $this->_packagerSettings['packagerEmail']);
			}
			else {
				$this->_packagerSettings['packagerEmail'] = CLI::PromptUser(
					'Please enter your email to use for packaging',
					'text-required'
				);
			}

			CLI::SaveSettingsFile('packager', $this->_packagerSettings);
		}

		return $this->_packagerSettings;
	}

	private function _saveComponent(Component $src) {
		$gpg = new \Core\GPG\GPG();
		$gpg->homedir = $_SERVER['HOME'] . '/.gnupg';

		$settings = $this->_loadPackagerSettings();

		// Ensure the latest version of the files are available, just in case something changed.
		$src->rescanFiles();

		// Set the requirement for Core at the current version.
		// This helps to enforce package developer work functions as expected when
		// deployed to the end customers, because the version of Core will need to approximately match.
		if($src->getKeyName() !== 'core') {
			$version = new VersionString(Core::GetComponent('core')->getVersion());
			// We don't need it to be exact, but a close match at least.
			$minVersion = $version->major . '.' . $version->minor . '.0';
			$src->setRequires('core', 'component', $minVersion, 'ge');
		}

		// Save the component.xml or theme.xml back down!
		$src->save();
	}

	/**
	 * Perform the actual packaging of a Component or Theme
	 *
	 * @param $src
	 */
	private function _performPackage(Component $src) {
		$gpg = new \Core\GPG\GPG();
		$gpg->homedir = $_SERVER['HOME'] . '/.gnupg';

		$settings = $this->_loadPackagerSettings();


		// Export the keys into this package, useful for installing to ensure the new install has the necessary keys.
		file_put_contents($src->getBaseDir() . 'LICENSER.gpg', $gpg->getKey($settings['packagerKey'])->getAscii());


		// Ensure there's a CHANGELOG for this package!
		$changelog = new ChangelogParser($src->getName(), $src->getBaseDir() . 'CHANGELOG');
		if(!$changelog->exists()) {
			$changelog->createInitial($src->getVersion());
		}
		else {
			$changelog->parse();
		}

		// Read the current changelog section for this version
		$changelogSection = $changelog->getSection($src->getVersion());

		if($changelogSection->getReleasedDate()) {
			$continue = CLI::PromptUser($src->getVersion() . ' is already marked as released!  Re-package version?', 'bool');
			if(!$continue) {
				CLI::PrintLine('Aborting Package');
				return;
			}
		}

		$changelogText = $changelogSection->fetch();

		// Prepend some useful information to it.
		$changelogHeader = <<<EOT
# Please enter the changelog for {$src->getName()} {$src->getVersion()},
# lines starting with a pound sign ('#') are treated as comments and ignored.
#
# Place each feature or change on a new line with no dashes or asterisks,
# and uses of [BUG], [FEATURE], [SECURITY] are recommended.
#
# For example:
# [BUG] Fix XYZ on some page
# [FEATURE] Implement this awesome feature for users
#
# When done, simply save and close this file. 

EOT;

		// Prompt the user with the ability to change them.
		$changelogText = CLI::PromptUser(
			'Changelog Updates',
			'textarea',
			$changelogHeader . NL . $changelogText
		);

		if($changelogText !== false) {
			// If the user clicks 'Cancel', FALSE is returned; we don't want to wipe out entries.
			// Otherwise, save the new changelog.
			$changelogSection->clearEntries();
			foreach(explode("\n", $changelogText) as $line){
				$line = trim($line);
				if($line === '' || $line[0] === '#') {
					// Skip empty lines and comments.
					continue;
				}
				$changelogSection->addLine($line);
			}
		}

		$changelogSection->markReleased($settings['packagerName'], $settings['packagerEmail']);
		$changelog->save();


		// Save any changes performed on the component to from this process.
		$this->_saveComponent($src);


		if($src->getKeyName() === 'core') {
			$path = 'core/';
		}
		elseif($src instanceof Theme) {
			$path = 'themes/' . $src->getKeyName() . '/';
		}
		else {
			$path = 'components/' . $src->getKeyName() . '/';
		}

		$dest = realpath(ROOT_PDIR . '../') . '/exports/';

		if(!is_dir($dest)) {
			mkdir($dest);
		}

		$dest .= $path;
		$tgz = $src->getKeyName() . '-' . $src->getVersion() . '.tar.gz';

		$dir = FileFactory::Directory($dest);
		$dir->mkdir();

		$src->package($dest . $tgz);

		if(file_exists($dest . $tgz . '.asc')) {
			unlink($dest . $tgz . '.asc');
		}

		$file = $gpg->signFile($dest . $tgz, $settings['packagerKey']);

		CLI::PrintSuccess('Packaged ' . $src->getName() . ' v' . $src->getVersion());
		CLI::PrintLine('Packaged File: ' . $file->getFilename());
		CLI::PrintLine(' Packager Key: ' . $settings['packagerKey']);
		CLI::PrintLine('     Filesize: ' . $file->getFilesize(true));
	}
}
