<?php
/**
 * Core component system, responsible for reading and parsing the component.xml,
 * saving it, and installing all components on the system.
 *
 * @package Core
 * @since 1.9
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;
use DOMElement;
use DOMNode;
use Exception;
use Core\CLI\CLI;
use ComponentModel;
use Core\Datamodel\DMIQueryException;
use Core\Filestore\Backends\FileLocal;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\DirectoryIterator;
use Core\Datamodel\Dataset;
use Core\GPG\GPG;
use mikehaertl\shellcommand\Command;

class Component {
	
	/**
	 * @var string Any error encountered while trying to register this Component, set from the Core registry
	 */
	public $error = null;

	// <editor-fold defaultstate="collapsed" desc="PRIVATE Variable Declaration">
	
	/**
	 * @var int API Version of this component, eg: 24 or 63.
	 */
	protected $_apiVersion = null;

	/**
	 * Array of any authors for the library.
	 * Each element is composed of an array of name, email and url.
	 *
	 * @var array <<array>>
	 */
	protected $_authors = [];
	
	/**
	 * @var array <<string>> List of files to autoload on this component as soon as it's registered
	 */
	protected $_autoLoads;
	
	/**
	 * @var string Fully resolved base directory of this component's root, with a trailing '/'
	 */
	protected $_baseDirectory;

	/**
	 * Array of classes in this component.  This is to reduce the number of lookups required.
	 *
	 * @var array <<string,string>> Hash of classes to fully resolved paths polled from the XML
	 */
	protected $_classList;

	/**
	 * @var array <<string,string>> Hash of controller to fully resolved paths polled from the XML
	 */
	protected $_controllerList;

	/**
	 * @var string Short description as defined in component.xml  (LAZY LOADED)
	 */
	protected $_description;
	
	/**
	 * @var array <<string,string>> Hash of Email Providers available as polled from the XML
	 */
	protected $_emailBackends;

	/**
	 * @var boolean Set to false if this component has been explictly disabled in the GUI
	 */
	protected $_enabled = false;
	
	/**
	 * @var string Filesystem name of this component
	 */
	protected $_keyName;
	
	/**
	 * @var string Logo relative asset/ path for this Component, as defined in the XML
	 */
	protected $_logo;
	
	/**
	 * @var array <<string,string>> Hash of Models to fully resolved paths polled from the XML
	 */
	protected $_modelList;

	/**
	 * @var string Human-friendly name of this component
	 */
	protected $_name;
	
	/**
	 * @var array <<ComponentPageDefinition>> List of pages definitions defined in the XML
	 */
	protected $_pages;

	/**
	 * @var array <<string>> Permissions that are registered in this component (LAZY LOADED)
	 */
	protected $_permissions;
	
	/**
	 * @var array <<ComponentProvide>> List of provides polled from the XML
	 */
	protected $_provides;
	
	/**
	 * @var array <<ComponentRequirement>> List of requirements polled from the XML
	 */
	protected $_requires;
	
	/**
	 * @var array Hash of script libraries and their call as defined in the XML
	 */
	protected $_scriptLibraries;

	/**
	 * The smarty plugin directory cache.  This is to reduce the number of lookups required.
	 *
	 * @var null|boolean|string
	 */
	protected $_smartyPluginDirectory = null;
	
	/**
	 * @var array <<string,string>> Hash of ModelSupplementals to fully resolved paths polled from the XML
	 */
	protected $_supplementalModelList;
	
	/**
	 * @var string Type of component, either "component" or "theme".
	 */
	protected $_type;
	
	/**
	 * Any update sites provided in this library.
	 *
	 * @var array <<string>>
	 */
	protected $_updateSites = [];
	
	/**
	 * @var array 
	 */
	protected $_userAuthDrivers;

	/**
	 * @var string Version as contained on the filesystem
	 */
	protected $_version;
	
	/**
	 * @var XMLLoader Underlying XML Loader object of the component.xml file
	 */
	protected $_xmlLoader;
	
	/**
	 * @var string Filename of the source XML file
	 */
	protected $_xmlFilename;

	/**
	 * Version of the component, as per the database (installed version).
	 *
	 * @var string
	 */
	protected $_versionDB = false;
	
	/**
	 * @var array <<ComponentWidgetDefinition>>
	 */
	protected $_widgets;

	/**
	 * @var array <<string,string>> Hash of widgets to fully resolved paths polled from the XML
	 */
	protected $_widgetList;
	
	/** @var string Raw signed license data from the database, populated by load */
	protected $_licenseDBData = null;
	
	/** @var null|array Result of the LICENSER.php file in this component, populated by getLicenseData as-needed. */
	protected $_licenserFileData = null;

	/** @var string[] List of app-specific skins for this component, as loaded from the XML */
	protected $_appSkins;
	
	// </editor-fold>


	// <editor-fold defaultstate="collapsed" desc="PUBLIC Load and Init Methods">

	public function __construct($filename = null) {
		if($filename) {
			$this->initFromFile($filename);
		}
	}

	/**
	 * @throws Exception
	 */
	public function initFromFile(string $filename) {
		$this->_xmlFilename = $filename;
		$this->_xmlLoader = new XMLLoader();
		$this->_xmlLoader->setRootName('component');
		$this->_xmlLoader->loadFromFilename($filename);
		$this->_baseDirectory = dirname($filename) . '/';

		$this->_init();
	}

	/**
	 * @throws Exception
	 */
	public function initFromString(string $xmlContents) {
		$this->_xmlLoader = new XMLLoader();
		$this->_xmlLoader->setRootName('component');
		$this->_xmlLoader->loadFromString($xmlContents);
		$this->_baseDirectory = '';

		$this->_init();
	}
	
	/**
	 * Retrieve every applicable bit of information from component.xml to be cached locally
	 * in a private variable
	 * 
	 * Not really used anywhere, but useful for debugging and Core development.
	 */
	public function loadAllFromXML(){
		if($this->_classList === null){
			$this->_loadFilesDataFromXML();
		}
		if($this->_provides === null){
			$this->_loadProvidesFromXML();
		}
		
		$this->getLogo();
		$this->getSmartyPluginDirectory();
		$this->getDescription();
		$this->getEmailBackends();
		$this->getPermissions();
		$this->getRequires();
		$this->getPagesDefined();
		$this->getScriptLibraryList();
		$this->getUserAuthDrivers();
	}

	/**
	 * Retrieve all the necessary information from the database regarding this Component.
	 * 
	 * Will throw an exception if the Core has not been loaded before calling!
	 * 
	 * @throws \Exception
	 */
	public function loadAllFromDB(){
		// Here we use Find to retrieve all records because it gets cached and only produces one query
		// in the database as opposed to however many are required for each component.
		$models = ComponentModel::Find();
		
		foreach($models as $m){
			/** @var $m ComponentModel */
			if($m->get('name') === $this->_keyName){
				$this->_versionDB     = $m->get('version');
				$this->_enabled       = $m->get('enabled');
				$this->_licenseDBData = $m->get('license');
				
				return;
			}
		}
	}
	
	public function registerHooks(){
		// Register any hooks that may be present.
		foreach ($this->_xmlLoader->getElementsByTagName('hookregister') as $h) {
			$hook = new Hook($h->getAttribute('name'));
			$hook->description = $h->getAttribute('description');
			if($h->getAttribute('return')){
				$hook->returnType = $h->getAttribute('return');
			}
		}

		// Register any events that may be present.
		foreach ($this->_xmlLoader->getElementsByTagName('hook') as $h) {
			$event = $h->getAttribute('name');
			$call  = $h->getAttribute('call');
			$type  = $h->getAttribute('type');
			HookHandler::AttachToHook($event, $call, $type);
		}
		
		foreach(array_map('strtolower', array_keys($this->getModelList())) as $name){
			$hook = new Hook('/core/controllinks/' . $name);
			$hook->description = 'Dynamic hook for external components to set control links on ' . $name;
			$hook->returnType = 'array';
		}
	}
	
	
	// </editor-fold>
	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC get* Methods">
	
	/**
	 * 
	 * @return ComponentAsset[]
	 */
	public function getAssets() : array {
		$ret = [];
		$dir = $this->getBaseDir();
		
		foreach($this->_xmlLoader->getElements('/assets/file') as $node){
			$ret[] = new ComponentAsset($dir, $node);
		}
		
		return $ret;
	}
	
	/**
	 * Return an array of every author in this component.
	 * 
	 * @return array List of name <<string>> and email <<string>>
	 */
	public function getAuthors() : array {
		$ret = [];
		foreach ($this->_xmlloader->getElementsByTagName('author') as $el) {
			$ret[] = [
				'name'  => $el->getAttribute('name'),
				'email' => $el->getAttribute('email'),
			];
		}
		return $ret;
	}
	
	/**
	 * Get any "autoload" files registered in this component XML
	 * 
	 * These files are expected to be loaded as soon as the component is registered.
	 * 
	 * @return array
	 */
	public function getAutoLoads() : array {
		if($this->_autoLoads === null){
			$this->_loadFilesDataFromXML();
		}
		
		return $this->_autoLoads;
	}
	
	/**
	 * Get the fully resolved base directory of this component with a trailing slash
	 *
	 * Generally /home/foo/public_html/components/componentname/
	 *
	 * @return string Component's full base path
	 */
	public function getBaseDir() : string {
		return $this->_baseDirectory;
	}
	
	/**
	 * Function to get any changed files in this component.
	 * A changed file is any file whose md5 doesn't match what's in the component.xml metafile.
	 *
	 * @return array
	 */
	public function getChangedFiles() : array {
		$changes = [];
		
		foreach($this->_xmlLoader->getElements('/component/files/file|/component/view/file|/component/assets/file') as $file){
			// <file filename="CHANGELOG" md5="e55e5457abe2f0a3a8edcb6a927c92e3"/>
			/** @var $md5 string */
			$md5 = $file->getAttribute('md5');
			/** @var $filename string */
			$filename = $file->getAttribute('filename');

			// Skip the changelog.
			if($filename == 'CHANGELOG'){
				continue;
			}
			
			if(
				is_file($this->_baseDirectory . $filename) && 
				md5_file($this->_baseDirectory . $filename) === $md5
			){
				// File matches, continue!
				continue;
			}

			$changes[] = $filename;
		}
		return $changes;
	}
	
	/**
	 * Get the list of classes provided in this component, (and their filenames)
	 *
	 * @return array <<string,string>>
	 */
	public function getClassList() {
		if($this->_classList === null){
			$this->_loadFilesDataFromXML();
		}
		
		return $this->_classList;
	}
	
	/**
	 * Get the configs defined on this component as an array of ComponentConfig objects.
	 * 
	 * @return ComponentConfig[]
	 */
	public function getConfigs() : array {
		$ret = [];
		
		$node = $this->_xmlLoader->getElement('configs');
		$componentName = $this->getKeyName();

		// Now, get every table under this node.
		foreach ($node->getElementsByTagName('config') as $confignode) {
			/** @var \DOMElement $confignode */
			$ret[] = new ComponentConfig($confignode, $componentName);
		}
		
		return $ret;
	}
	
	/**
	 * Get the list of controllers in this component.
	 *
	 * @return array <<string,string>>
	 */
	public function getControllerList() : array {
		if($this->_controllerList === null){
			$this->_loadFilesDataFromXML();
		}
		
		return $this->_controllerList;
	}
	
	/**
	 * Get the description for this component
	 * 
	 * @return string
	 */
	public function getDescription() : string {
		if ($this->_description === null) {
			$this->_description = trim($this->_xmlLoader->getElement('/description')->nodeValue);
		}

		return $this->_description;
	}
	
	/**
	 * Get all email provider classes defined in this component
	 *
	 * @return array <<string,string>
	 */
	public function getEmailBackends() : array {
		if($this->_emailBackends === null){
			$this->_emailBackends = [];
			// I need to get the schema definitions first.
			$node = $this->_xmlLoader->getElement('provides');

			// Now, get every table under this node.
			foreach ($node->getElementsByTagName('emailbackend') as $subnode) {
				/** @var DOMElement $subnode */

				$name  = $subnode->getAttribute('name');
				$class = $subnode->getAttribute('class');

				$this->_emailBackends[$class] = $name;
			}
		}
		
		return $this->_emailBackends;
	}
	
	/**
	 * Get all form elements defined in this Component, with their keyname to resolved classname.
	 * 
	 * @return array
	 */
	public function getFormElements() : array {
		$ret = [];
		
		foreach($this->_xmlLoader->getElements('forms/formelement') as $el) {
			$ret[ $el->getAttribute('name') ] = $el->getAttribute('class');
		}
		
		return $ret;
	}
	
	/**
	 * Get this component's "key" name.
	 * 
	 * This is acceptable to use for identification and programmatic uses.
	 *
	 * @return string
	 */
	public function getKeyName() : string {
		return $this->_keyName;
	}
	
	/**
	 * Get the list of "Libraries" defined in this Component.
	 * 
	 * A Library is a loose term for some bit of functionality.
	 * 
	 * Each record returned will have its name as the key and the version as the value.
	 * 
	 * @return array <<string,string>>
	 */
	public function getLibraryList() {
		if($this->_provides === null){
			$this->_loadProvidesFromXML();
		}
		
		$ret = [];
		foreach($this->_provides as $p){
			/* @var $p ComponentProvide */
			if($p->type === 'library'){
				$ret[ $p->name ] = $p->version;
			}
		}
		
		return $ret;
	}
	
	/**
	 * Return the fully populated array with the licensed data and the values from the license
	 * 
	 * @return array
	 */
	public function getLicenseData(){
		// This feature relies on a valid server id.
		if(!defined('SERVER_ID')){
			return [];
		}
		if(strlen(SERVER_ID) != 32){
			return [];
		}
		
		// See if I need to lookup the contents; this is cached internally to save on lookups.
		if($this->_licenserFileData === null){
			$f = $this->_baseDirectory . 'LICENSER.php';

			if(file_exists($f)){
				$licenser = include($f);

				if(!isset($licenser)){
					$this->_licenserFileData = [];
					return [];
				}
				elseif(is_array($licenser)){
					$this->_licenserFileData = $licenser;
				}
				else{
					// Just a blank array.
					$this->_licenserFileData = [];
					return [];
				}
			}
			else{
				$this->_licenserFileData = [];
				return [];
			}

			$features = [];
			$status = false;
			$message = 'No license data present';
			$expires = null;
			if($this->_licenseDBData){
				// Only populate feature values if there is a license included.
				// Otherwise, they will get set to FALSE.
				
				// First through, try to pull them from cache to save on a decryption operation.
				// Lookup the cache for this licensed key value.
				$cacheKey = md5('LICENSER:' . SERVER_ID . $this->getKeyName());
				$cached = \Core\Cache::Get($cacheKey);
				if($cached){
					$features = $cached['features'];
					$status = $cached['status'];
					$message = $cached['message'];
					$expires = $cached['expires'];
				}
				else{
					try{
						$gpg = new \Core\GPG\GPG();

						$data = $gpg->decryptData($this->_licenseDBData);
						if($data && ($decoded = json_decode($data, true))){
							if($decoded['status'] && isset($decoded['features'])){
								$features = $decoded['features'];
							}
							
							$status = $decoded['status'];
							if(isset($decoded['message'])){
								$message = $decoded['message'];
							}
							elseif($status){
								$message = 'Valid license'; // Good licenses may not have a message.
							}
							
							if(isset($decoded['expires'])){
								$expires = $decoded['expires'];
							}
						}
						
						// Save these back to cache so that this try/catch can be skipped for a little while again.
						\Core\Cache::Set($cacheKey, ['features' => $features, 'status' => $status, 'message' => $message, 'expires' => $expires], 7200);
					}
					catch(Exception $e){
						// GPG cannot be instantiated; silently skip pulling license features.
					}
				}
			}

			// Pull the information from the cache or database for the keys!
			$newFeatures = [];
			foreach($this->_licenserFileData['features'] as $f){
				if(isset($features[$f])){
					$newFeatures[$f] = $features[$f];
				}
				else{
					$newFeatures[$f] = false;
				}
			}
			
			// Remap the stored key with the version containing the values.
			$this->_licenserFileData['features'] = $newFeatures;
			$this->_licenserFileData['status'] = $status;
			$this->_licenserFileData['message'] = $message;
			$this->_licenserFileData['component'] = $this->getName();
			$this->_licenserFileData['expires'] = $expires;
		}
		
		return $this->_licenserFileData;
	}
	
	/**
	 * Return an array of every license, (and its URL), in this component.
	 * 
	 * @return array List containing title <<string>> and url <<string>>
	 */
	public function getLicenses() : array {
		$ret = [];
		foreach ($this->_xmlloader->getElementsByTagName('license') as $el) {
			$url   = $el->getAttribute('url');
			$ret[] = [
				'title' => $el->nodeValue,
				'url'   => $url
			];
		}
		return $ret;
	}
	
	/**
	 * Get the logo path for this component as-per defined in the XML.
	 * 
	 * This will be prefixed with "asests/" and should be resolvable via the automatic asset resolution system.
	 * 
	 * If this Component does not have a logo or it isn't in assets/, NULL is returned.
	 * 
	 * @return null|string
	 */
	public function getLogo(){
		if($this->_logo === null){
			$icon = $this->_xmlLoader->getRootDOM()->getAttribute('logo');
		
			if($icon === null || $icon === '' || strpos($icon, 'assets/') !== 0){
				$this->_logo = false;
			}
			else{
				$this->_logo = $icon;
			}
		}
		
		return $this->_logo === false ? null : $this->_logo;
	}
	
	/**
	 * Get the list of Models provided in this component, (and their filenames)
	 *
	 * @return array <<string,string>>
	 */
	public function getModelList() {
		if($this->_modelList === null){
			$this->_loadFilesDataFromXML();
		}
		
		return $this->_modelList;
	}
	
	/**
	 * Get this component's Human-friendly name
	 *
	 * @return string
	 */
	public function getName() : string {
		return $this->_name;
	}
	
	/**
	 * Get an array of Page Creates that are defined in the XML for this component.
	 * 
	 * @return array
	 */
	public function getPageCreatesDefined(){
		$pages = [];

		// I need to get the schema definitions first.
		$node = $this->_xmlLoader->getElement('pages');

		// Now, get every table under this node.
		foreach ($node->getElementsByTagName('pagecreate') as $subnode) {
			/** @var DOMElement $subnode */

			$title = $subnode->getAttribute('title');
			$baseurl = $subnode->getAttribute('baseurl');
			$description = $subnode->getAttribute('description');

			// Toss this page onto the stack :)
			$pages[] = [
				'title'      => $title,
				'baseurl'    => $baseurl,
				'description' => $description,
			];
		}

		return $pages;
	}
	
	/**
	 * Get the pages defined in this component.
	 *
	 * These are usually admin-only pages, (but may not be).
	 *
	 * @return array <<ComponentPageDefinition>>
	 */
	public function getPagesDefined() : array{
		if($this->_pages === null){
			$this->_pages = [];
			
			// I need to get the schema definitions first.
			$node = $this->_xmlLoader->getElement('pages');
			
			// Now, get every table under this node.
			foreach ($node->getElementsByTagName('page') as $subnode) {
				/** @var DOMElement $subnode */
				
				$this->_pages[] = new ComponentPageDefinition($subnode, $this->_keyName);
			}
		}
		
		return $this->_pages;
	}
	
	/**
	 * Get the registered permissions for this component.
	 *
	 * @return array <<string>>
	 */
	public function getPermissions() : array {
		if($this->_permissions === null){
			$this->_permissions = [];
			foreach($this->_xmlLoader->getElements('/permissions/permission') as $el){
				/** @var $el DOMElement */
				$this->_permissions[] = $el->getAttribute('key');
			}
		}
		return $this->_permissions;
	}
	
	
	public function getProvides() : array {
		if($this->_provides === null){
			$this->_loadProvidesFromXML();
		}
		
		return $this->_provides;
	}
	
	/**
	 * Get a list of ScriptLibraries defined in this component XML
	 * 
	 * @return array <<string,string>>
	 */
	public function getScriptLibraryList() {
		if($this->_scriptLibraries === null){
			$this->_scriptLibraries = [];
			foreach ($this->_xmlLoader->getElements('/provides/scriptlibrary') as $s) {
				$name = strtolower($s->getAttribute('name'));
				$call = $s->getAttribute('call');
				$this->_scriptLibraries[$name] = $call;
			}
		}
		
		return $this->_scriptLibraries;
	}
	
	
	/**
	 * Return the fully resolved name of the smarty plugin directory for
	 * this component (if there is one).
	 *
	 * Not many templates will use this function, but it is there for when needed.
	 * 
	 * @return string|null
	 */
	public function getSmartyPluginDirectory() {
		if($this->_smartyPluginDirectory === null){
			$this->_smartyPluginDirectory = 
				is_dir($this->_baseDirectory . 'libs/smarty-plugins/') ?
				$this->_baseDirectory . 'libs/smarty-plugins/' :
				false;
		}

		return $this->_smartyPluginDirectory;
	}
	
	/**
	 * Similar to getModelList, only it returns any supplemental models in this Component.
	 * 
	 * @return array <<string,string>>
	 */
	public function getSupplementalModelList() : array {
		if($this->_supplementalModelList === null){
			$this->_loadFilesDataFromXML();
		}
		
		return $this->_supplementalModelList;
	}
	
	/**
	 * Get the raw XML of this component, useful for debugging.
	 *
	 * @return string (XML)
	 */
	public function getRawXML() : string {
		return $this->_xmlLoader->asPrettyXML();
	}
	
	/**
	 * Get the ComponentRequirements for this component
	 * 
	 * @return ComponentRequirement[]
	 */
	public function getRequires() : array {
		if($this->_requires === null){
			$this->_requires = [];
			foreach ($this->_xmlLoader->getElements('/requires/require') as $r) {
				$t  = $r->getAttribute('type');     // Typically required.
				$n  = $r->getAttribute('name');     // Typically required.
				$v  = $r->getAttribute('version')   ?: null;
				$op = $r->getAttribute('operation') ?: null;
				
				$this->_requires[] = new ComponentRequirement($t, $n, $v, $op);
			}
		}

		return $this->_requires;
	}
	
	/**
	 * Helper function for external classes and scripts to get this component XML DOM
	 *
	 * @return \DOMNode
	 */
	public function getRootDOM() : \DOMNode {
		return $this->_xmlLoader->getRootDOM();
	}
	
	/**
	 * Get an array of this component's registered user auth drivers.
	 *
	 * @return array
	 */
	public function getUserAuthDrivers() : array {
		if($this->_userAuthDrivers === null){
			$this->_userAuthDrivers = [];
			
			$nodes = $this->_xmlLoader->getElements('/users/userauth');
			foreach($nodes as $n){
				/** @var DOMElement $n */
				$name = $n->getAttribute('name');
				$class = $n->getAttribute('class');

				$this->_userAuthDrivers[ $name ] = $class;
			}
		}
		
		return $this->_userAuthDrivers;
	}
	
	/**
	 * Get this component's version according to the filesystem
	 *
	 * @return string
	 */
	public function getVersion() : string {
		return $this->_version;
	}
	
	/**
	 * Get the version of this Component as per the database or NULL if not installed yet
	 * 
	 * Useful to know if something needs updated.
	 * 
	 * @return string|null
	 */
	public function getVersionInstalled() {
		return $this->_versionDB;
	}
	
	/**
	 * Get an array of Widget Creates that are defined in the XML for this component.
	 * 
	 * @return array
	 */
	public function getWidgetCreatesDefined() : array {
		$widgets = [];

		// I need to get the schema definitions first.
		$node = $this->_xmlLoader->getElement('widgets');

		// Now, get every table under this node.
		foreach ($node->getElementsByTagName('widgetcreate') as $subnode) {
			/** @var DOMElement $subnode */

			$title       = $subnode->getAttribute('title');
			$description = $subnode->getAttribute('description');
			$image       = $subnode->getAttribute('image');
			
			if($subnode->getAttribute('baseurl')){
				$nodebaseurl = $subnode->getAttribute('baseurl');
			}
			elseif(($class = $subnode->getAttribute('class'))){
				$nodebaseurl = '/widget/create?class=' . $class;
			}
			else{
				\Core\log_debug('Invalid "widgetcreate" found in ' . $this->getName() . ', ' . $subnode->getAttribute('title') . '; at least baseurl or class are required.');
				continue;
			}

			// Toss this widget onto the stack
			$widgets[] = [
				'title'       => $title,
				'baseurl'     => $nodebaseurl,
				'description' => $description,
				'preview'     => $image,
			];
		}

		return $widgets;
	}
	
	/**
	 * Get the widgets defined in this component.
	 *
	 * @return array <<ComponentWidgetDefinition>>
	 */
	public function getWidgetsDefined() : array{
		if($this->_widgets === null){
			$this->_widgets = [];
			
			// I need to get the schema definitions first.
			$node = $this->_xmlLoader->getElement('widgets');
			
			// Now, get every table under this node.
			foreach ($node->getElementsByTagName('widget') as $subnode) {
				/** @var DOMElement $subnode */
				$this->_widgets[] = new ComponentWidgetDefinition($subnode, $this->_keyName);
			}
		}
		
		return $this->_widgets;
	}
	
	/**
	 * Get an array of widget names provided in this component.
	 *
	 * @return array <<string,string>>
	 */
	public function getWidgetList() : array {
		if($this->_widgetList === null){
			$this->_loadFilesDataFromXML();
		}
		
		return $this->_widgetList;
	}
	
	/**
	 * Get the XML Loader backend of this component.
	 *
	 * Useful for manipulating the XML structure.
	 *
	 * @return XMLLoader
	 */
	public function getXMLLoader() : XMLLoader {
		return $this->_xmlLoader;
	}
	
	// </editor-fold>
	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC is* Methods">
	
	/**
	 * Check if this component is marked as installed
	 * 
	 * @return bool
	 */
	public function isInstalled(): bool {
		return $this->_versionDB !== false;
	}
	
	/**
	 * Check if this component is marked as enabled
	 * 
	 * Will always return false if called before Core has initialized the database.
	 * 
	 * @return bool
	 */
	public function isEnabled(): bool {
		return $this->_enabled === true;
	}
	
	/**
	 * Get if this Component is ready based on if there were any errors during registration.
	 * 
	 * @return bool
	 */
	public function isReady(): bool {
		return $this->error === null;
	}
	
	/**
	 * Check if this component has an update pending
	 * 
	 * This means there's a version discrepancy between the filesystem and the database.
	 * 
	 * @return bool
	 */
	public function isPendingUpdate() : bool {
		return ($this->_version !== $this->_versionDB);
	}
	
	// </editor-fold>
	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC set* and save Methods">
	
	/**
	 * Rescan the various files contained in the XML and update the metadata so it can be saved.
	 * 
	 * Only really useful from within the packager.
	 */
	public function rescanFiles() {
		$viewdir    = $this->getBaseDir() . 'templates/';
		$assetdir   = $this->getBaseDir() . 'assets/';
		$skindir    = ($this->_type === 'theme') ? $this->getBaseDir() . 'skins/' : null;
		$basestrlen = strlen($this->getBaseDir());
		$assetfiles = [];
		$viewfiles  = [];
		$otherfiles = [];
		$skinfiles  = []; // Only themes have skins, but this is the new unified save handler, so it needs to be present.
		$autoloads = $this->getAutoLoads();
		
		$iterator = new DirectoryIterator(FileFactory::Directory($this->getBaseDir()));
		$baseDir = $this->getBaseDir();

		$iterator->findDirectories = false;
		$iterator->recursive = true;
		$iterator->ignores = [
			'component.xml',
		];

		foreach($iterator as $file){
			/** @var FileLocal $file */

			// If this ends with a "~", skip it!
			// These are backup files used by gedit.
			if(substr($file->getFilename(), -1) == '~'){
				continue;
			}

			// And then, scan this file for code, ie: classes, controllers, etc.
			$fname = substr($file->getFilename(), $basestrlen);
			
			if($viewdir && $file->inDirectory($viewdir)){
				// It's a template! (view)
				$viewfiles[] = ['file' => $fname, 'md5' => $file->getHash()];
			}
			elseif($assetdir && $file->inDirectory($assetdir) && $file->getExtension() == 'scss'){
				// SKIP scss files, they don't get deployed anyway.
				continue;
			}
			elseif($assetdir && $file->inDirectory($assetdir)){
				// It's an asset!
				$assetfiles[] = ['file' => $fname, 'md5' => $file->getHash()];
			}
			elseif($skindir && $file->inDirectory($skindir)){
				// It's a skin!
				$skinfiles[] = ['file' => $fname, 'md5' => $file->getHash()];
			}
			else{
				// It's a something..... it goes in the "files" array!
				// This will be slightly different though, as it needs to also check for classes.
				$filedat = [
					'file'               => $fname,
					'md5'                => $file->getHash(),
					'controllers'        => [],
					'models'             => [],
					'widgets'            => [],
					'modelsupplementals' => [],
					'classes'            => [],
					'interfaces'         => [],
					'traits'             => [],
					'autoload'           => in_array($file->getFilename(), $autoloads),
				];
				
				// PHP files get checked.
				if(preg_match('/\.(php|inc)$/i', $fname)){
					$contents = $file->getContentsObject();
					$classes = $contents->getClasses();
					$interfaces = $contents->getInterfaces();
					$traits = $contents->getTraits();
					
					foreach($classes as $classDat) {
						if($classDat['extends'] === 'Core\\Model'){
							$filedat['models'][] = $classDat['name'];
						}
						elseif($classDat['extends'] === 'Core\\Controller'){
							$filedat['controllers'][] = $classDat['name'];
						}
						elseif($classDat['extends'] === 'Core\\Widget'){
							$filedat['widgets'][] = $classDat['name'];
						}
						elseif( in_array('Core\\ModelSupplemental', $classDat['implements']) ){
							$filedat['modelsupplementals'][] = $classDat['name'];
						}
						else{
							$filedat['classes'][] = $classDat['name'];
						}
					}
					foreach($interfaces as $classDat){
						$filedat['interfaces'][] = $classDat['name'];
					}
					foreach($traits as $classDat){
						$filedat['traits'][] = $classDat['name'];
					}
				}
				
				$otherfiles[] = $filedat;
			}
		}
		
		$this->setFiles($otherfiles);
		$this->setViewFiles($viewfiles);
		$this->setAssetFiles($assetfiles);
		if($this->_type === 'theme'){
			$this->setSkinFiles($skinfiles);
		}
	}
	
	/**
	 * Save this component metadata back to its XML file.
	 * Useful in packager scripts.
	 */
	public function save() {
		// Set the schema version to the newest API version.
		$this->_xmlLoader->setSchema('https://corepl.us/api/6_3/component.dtd');
		// Ensure there's a required namespace on the root node.
		$this->_xmlLoader->getRootDOM()->setAttribute('xmlns:xsi', "http://www.w3.org/2001/XMLSchema-instance");

		// This needs to be the final step... write the XML doc back to the file.
		file_put_contents($this->_xmlFilename, $this->_xmlLoader->asPrettyXML());
	}
	
	/**
	 * Set all asset files in this component.  Only really usable in the installer.
	 *
	 * @param $files array Array of files to set.
	 */
	public function setAssetFiles($files) {
		// Clear out the array first.
		$this->_xmlLoader->removeElements('/assets/file');

		// It would be nice to have them alphabetical.
		usort($files, function($a, $b){
			return $a['file'] <=> $b['file'];
		});

		// And recreate them all.
		foreach ($files as $f) {
			$el = $this->_xmlLoader->createElement('/assets/file[@filename="' . $f['file'] . '"][@md5="' . $f['md5'] . '"]');
		}
	}
	
	/**
	 * Set and override the list of authors for this component.
	 *
	 * @param $authors array Array of authors to set
	 */
	public function setAuthors($authors) {
		// First, remove any authors currently in the XML.
		$this->_xmlLoader->removeElements('/authors');

		// Now I can add the ones in the authors array.
		foreach ($authors as $a) {
			if (isset($a['email']) && $a['email']) {
				$this->_xmlLoader->getElement('/authors/author[@name="' . $a['name'] . '"][@email="' . $a['email'] . '"]');
			}
			else {
				$this->_xmlLoader->getElement('/authors/author[@name="' . $a['name'] . '"]');
			}
		}
	}
	
	/**
	 * Set the description for this component
	 * @param $desc string
	 */
	public function setDescription($desc) {
		// Set the cache first.
		$this->_description = $desc;
		// And set the data in the original DOM.
		$this->_xmlLoader->getElement('/description')->nodeValue = $desc;
	}
	
	/**
	 * Set all files in this component.  Only really usable in the installer.
	 *
	 * @param $files array Array of files to set.
	 */
	public function setFiles($files) {
		// Clear out the array first.
		$this->_xmlLoader->removeElements('/files/file');

		// It would be nice to have them alphabetical.
		usort($files, function($a, $b){
			return $a['file'] <=> $b['file'];
		});

		// And recreate them all.
		foreach ($files as $f) {
			$el = $this->_xmlLoader->createElement('/files/file[@filename="' . $f['file'] . '"][@md5="' . $f['md5'] . '"]');
			if($f['autoload']){
				$el->setAttribute('autoload', '1');
			}

			if (isset($f['controllers']) && count($f['controllers'])) {
				foreach ($f['controllers'] as $c) {
					$this->_xmlLoader->createElement('class[@type="controller"][@name="' . $c . '"]', $el);
				}
			}
			if (isset($f['models']) && count($f['models'])) {
				foreach ($f['models'] as $c) {
					$this->_xmlLoader->createElement('class[@type="model"][@name="' . $c . '"]', $el);
				}
			}
			if (isset($f['widgets']) && count($f['widgets'])) {
				foreach ($f['widgets'] as $c) {
					$this->_xmlLoader->createElement('class[@type="widget"][@name="' . $c . '"]', $el);
				}
			}
			if (isset($f['modelsupplementals']) && count($f['modelsupplementals'])) {
				foreach ($f['modelsupplementals'] as $c) {
					$this->_xmlLoader->createElement('class[@type="modelsupplemental"][@name="' . $c . '"]', $el);
				}
			}
			if (isset($f['classes']) && count($f['classes'])) {
				foreach ($f['classes'] as $c) {
					$this->_xmlLoader->createElement('class[@name="' . $c . '"]', $el);
				}
			}
			if (isset($f['interfaces']) && count($f['interfaces'])) {
				foreach ($f['interfaces'] as $i) {
					$this->_xmlLoader->createElement('interface[@name="' . $i . '"]', $el);
				}
			}
			if (isset($f['traits']) && count($f['traits'])) {
				foreach ($f['traits'] as $i) {
					$this->_xmlLoader->createElement('trait[@name="' . $i . '"]', $el);
				}
			}
		}
	}

	/**
	 * Set and override the list of licenses for this component.
	 *
	 * @param $licenses array Array of licenses to set
	 */
	public function setLicenses($licenses) {
		// First, remove any licenses currently in the XML.
		$this->_xmlLoader->removeElements('/licenses');

		// Now I can add the ones in the licenses array.
		$path = '/licenses/';
		foreach ($licenses as $lic) {
			$el = 'license' . ((isset($lic['url']) && $lic['url']) ? '[@url="' . $lic['url'] . '"]' : '');
			$l  = $this->_xmlLoader->createElement($path . $el, false, 1);
			if ($lic['title']) $l->nodeValue = $lic['title'];
		}
	}
	
	/**
	 * Set the logo for this component from an absolutely-resolved filename.
	 * 
	 * @param string $filename
	 */
	public function setLogo($filename){
		$nf = FileFactory::File($filename);
		
		$current = $this->getLogo();
		if($current){
			$f = FileFactory::File($this->_baseDirectory . $current);
			if($nf->identicalTo($f)){
				// Files are the same, no change.
				return;
			}
		}
		
		$f = $nf->copyTo($this->_baseDirectory . 'assets/images/logos/');
		$file = substr($f->getFilename(), strlen($this->_baseDirectory));
		$this->_logo = $file;
		$this->_xmlLoader->getRootDOM()->setAttribute('logo', $file);
	}
	
	/**
	 * Set a require in the XML
	 * 
	 * This is used by the packager.
	 * 
	 * @param string      $name
	 * @param string      $type
	 * @param null|string $version
	 * @param null|string $op
	 */
	public function setRequires($name, $type, $version = null, $op = null){
		// Get the node, (and auto-create if it doesn't exist).
		$node = $this->_xmlLoader->getElement('/requires/require[@name="' . $name . '"][@type="' . $type . '"]');

		// Set the new version and operation for it.
		if($version){
			$node->setAttribute('version', $version);
			if($op){
				$node->setAttribute('operation', $op);
			}
			else{
				$node->removeAttribute('operation');
			}
		}
		else{
			$node->removeAttribute('version');
			$node->removeAttribute('operation');
		}
	}
	
	/**
	 * Set the version of this component
	 *
	 * This affects the xml metafile of the package.
	 *
	 * @param $vers string New version to set
	 */
	public function setVersion($vers) {
		if ($vers == $this->_version) return;

		// Switch over any unversioned upgrade directives to this version.
		// First, check just a plain <upgrade> directive.
		if (($upg = $this->_xmlLoader->getElement('/upgrades/upgrade[@from=""][@to=""]', false))) {
			// Add the current and dest. attribute to it.
			$upg->setAttribute('from', $this->_version);
			$upg->setAttribute('to', $vers);
		}
		elseif (($upg = $this->_xmlLoader->getElement('/upgrades/upgrade[@from="' . $this->_version . '"][@to=""]', false))) {
			$upg->setAttribute('to', $vers);
		}
		elseif (($upg = $this->_xmlLoader->getElement('/upgrades/upgrade[@from="next"][@to="next"]', false))) {
			// Support the <upgrade from="next" to="next"/> format.
			$upg->setAttribute('from', $this->_version);
			$upg->setAttribute('to', $vers);
		}
		else {
			// No node found... just create a new one.
			$this->_xmlLoader->getElement('/upgrades/upgrade[@from="' . $this->_version . '"][@to="' . $vers . '"]');
		}

		$this->_version = $vers;
		$this->_xmlLoader->getRootDOM()->setAttribute('version', $vers);
	}
	
	/**
	 * Set all template files in this component.  Only really usable in the installer.
	 *
	 * @param $files array Array of files to set.
	 */
	public function setViewFiles($files) {
		// Clear out the array first.
		$this->_xmlLoader->removeElements('/view/file');

		// It would be nice to have them alphabetical.
		usort($files, function($a, $b){
			return $a['file'] <=> $b['file'];
		});

		// And recreate them all.
		foreach ($files as $f) {
			$el = $this->_xmlLoader->createElement('/view/file[@filename="' . $f['file'] . '"][@md5="' . $f['md5'] . '"]');
		}
	}
	
	// </editor-fold>
	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC Install-Related Methods">
	
	/**
	 * Disable this component but keep its data present.
	 */
	public function disable() {
		\Core\log_trace('Disabling component ' . $this->_name);
		
		// Update the registration data for this component
		$model = new ComponentModel($this->_keyName);
		$model->set('enabled', 0);
		$model->save();
	}
	
	/**
	 * Enable this component.
	 */
	public function enable() {
		\Core\log_trace('Enabling component ' . $this->_name);
		
		// Update the registration data for this component
		$model = new ComponentModel($this->_keyName);
		$model->set('enabled', 1);
		$model->save();
	}
	
	/**
	 * Install this component
	 *
	 * Returns an array containing a list of changes performed
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function install() {
		$results = [
			'changes' => [],
			'errors'  => [],
		];
			
		if($this->isInstalled()){
			return $results;
		}
		
		\Core\log_trace('Installing component ' . $this->_name);
		
		$results = array_merge_recursive(
			$this->installDBSchema(),
			$this->installConfigs(),
			$this->installPages(),
			$this->installWidgets(),
			$this->installAssets(),
			$this->processUpgradeNode($this->_xmlLoader->getElement('/component/install')),
			$this->installUpdater()
		);
		
		
		// Set the version for this component in the database.
		$model = new ComponentModel($this->_keyName);
		$model->set('version', $this->_version);
		$model->set('enabled', 1);
		$model->save();
		
		return $results;
	}
	
	/**
	 * Copy in all the assets for this component into the assets location.
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function installAssets(){
		// Will contain the report for any changes performed herein.
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		\Core\log_trace('Installing assets for Component ' . $this->getName());
		
		$assets = $this->getAssets();
		foreach($assets as $asset){
			$action = $asset->exists() ? 'Replaced' : 'Installed';
			
			try{
				if($asset->install()) {
					$results['changes'][] = $action . ' ' . $asset->filename;
					\Core\log_debug($action . ' ' . $asset->filename);
				}
				else{
					\Core\log_trace('Skipped ' . $asset->filename);
				}
			}
			catch (Exception $ex) {
				\Core\log_notice($ex->getMessage());
				$results['errors'][] = $ex->getMessage();
			}
		}
		
		if(sizeof($results['changes']) === 0 && sizeof($results['errors']) === 0){
			\Core\log_debug('No changes detected');
		}

		return $results;
	}
	
	/**
	 * Install all configuration options for this component
	 * 
	 * This is used for installations and upgrades.
	 * 
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function installConfigs(){
		// Keep track of if this changed anything.
		$results = [
			'changes' => [],
			'errors'  => [],
		];

		\Core\log_trace('Installing configs for Component ' . $this->getName());
		
		$configs = $this->getConfigs();
		
		foreach($configs as $config){
			if($config->install()){
				$results['changes'][] = 'Installed/Updated configuration ' . $config->key;
			}
			else{
				\Core\log_trace('Skipped, no changes detected.');
			}
		}

		return $results;
	}
	
	/**
	 * Parse and handle the various Models and their DB Schemas
	 * 
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function installDBSchema() : array {
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		\Core\log_trace('Installing database schema for Component ' . $this->getName());
		
		/** @var \Core\Datamodel\BackendInterface $db */
		$db     = \Core\db();
		
		// Get the table structure as it exists in the database first, this will be the comparison point.
		$classes = $this->getModelList();

		// Do the actual processing of every Model.
		foreach ($classes as $m => $file) {
			// Ensure the Model exists, (it may not be loaded yet)
			require_once $file;
			
			$schema = ModelFactory::GetSchema($m);
			$tablename = $m::GetTableName();
			
			\Core\log_trace('Processing database table ' . $tablename);

			if ($db->tableExists($tablename)) {
				try{
					if(($res = $db->modifyTable($tablename, $schema)) !== false){
						// Changes detected!
						$results['changes'][] = 'Modified table ' . $tablename;
						\Core\log_trace(implode("\n", $res));
					}
					else{
						// No changes detected.
						\Core\log_trace('No changes to table detected');
					}
				}
				catch(DMIQueryException $ex){
					$results['errors'][] = 'Failed to modify table ' . $tablename .
						' due to ' . $ex->getMessage();
					\Core\log_debug(
						'Failed to modify table ' . $tablename . ' due to error in query ' .
						$ex->query
					);
				}
			}
			else {
				try{
					$db->createTable($tablename, $schema);
					$results['changes'][] = 'Created table ' . $tablename;
					\Core\log_trace('Created table ' . $tablename);
				}
				catch(DMIQueryException $ex){
					$results['errors'][] = 'Failed to create table ' . $tablename .
						' due to ' . $ex->getMessage();
					\Core\log_debug(
						'Failed to create table ' . $tablename . ' due to error in query ' .
						$ex->query
					);
				}
			}
		}
		
		return $results;
	}
	
	/**
	 * Install any page defined in this Component
	 * 
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function installPages() : array {
		// Keep track of if this changed anything.
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		\Core\log_trace('Installing pages for Component ' . $this->_name);
		
		// Retrieve a list of current pages attached to this Component.
		// This is used as a cache of what pages were deleted in the schema, (if any).
		// We include deleteurl to only delete pages that cannot be otherwise deleted.
		// This is to allow components to dynamically create admin-level pages,
		// as long as they can be deleted by the admin.
		$previousPages = Dataset::Init()
			->select('baseurl')
			->table('page')
			->whereEquals('admin', '1')
			->whereEquals('component', $this->_keyName)
			->whereEquals('deleteurl', '')
			->executeAndGet();


		// Load all the current pages as per the spec.
		$pages = $this->getPagesDefined();


		// Uncache any previously created page that exists in the new spec, this will just check for pages that have been deleted.
		foreach($pages as $page) {
			/** @var $page ComponentPageDefinition */
			// Drop this page from the cache of current pages, if set.
			if(($key = array_search($page->baseurl, $previousPages)) !== false){
				unset($previousPages[ $key ]);
			}
		}
		
		// Now $previousPages contains any admin page that was deleted since the last import.
		foreach($previousPages as $baseurl){
			Dataset::Init()
				->delete()
				->table('page')
				->whereEquals('admin', '1')
				->whereEquals('component', $this->_keyName)
				->whereEquals('baseurl', $baseurl)
				->execute();
			$results['changes'][] = 'Deleted page ' . $baseurl;
		}

		// Create any new page.
		foreach($pages as $page){
			/** @var $page ComponentPageDefinition */
			try {
				$isNew = $page->getModel()->checkIsNew();

				if($page->install()) {
					$results['changes'][] = ($isNew ? 'Created' : 'Updated') . ' page ' . $page->baseurl;
				}
			}
			catch(Exception $ex) {
				$results['errors'][] = 'Error when installing page ' . $page->baseurl . ': ' . $ex->getMessage();
			}
		}
		
		return $results;
	}
	
	/**
	 * Install any widgets defined in this Component
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function installWidgets() {
		// Keep track of if this changed anything.
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		\Core\log_trace('Installing Widgets for ' . $this->getName());
		
		$widgets = $this->getWidgetsDefined();
		
		foreach($widgets as $widget){
			/** @var $widget ComponentWidgetDefinition */
			if($widget->install()){
				$results['changes'][] = 'Created/Updated widget ' . $widget->baseurl;
			}
		}
		
		return $results;
	}

	/**
	 * Install the updater for this component
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function installUpdater(){
		// Will contain the report for any changes performed herein.
		$results = [
			'changes' => [],
			'errors'  => [],
		];

		\Core\log_trace('Installing updater for Component ' . $this->getName());

		if(file_exists($this->getBaseDir() . 'LICENSER.gpg')) {
			try {
				$gpg = new GPG();
				$loads = $gpg->importKeyFile($this->getBaseDir() . 'LICENSER.gpg');

				if($loads->changes > 0) {
					foreach($loads->keys as $key) {
						if($key) {
							$results['changes'][] = 'Imported GPG key ' . $key->getLabel();
						}
						else {
							$results['changes'][] = 'Imported GPG key';
						}
					}
				}
			}
			catch(Exception $ex) {
				$results['errors'][] = $ex->getMessage();
			}
		}

		if(file_exists($this->getBaseDir() . 'LICENSER.php')) {
			try {
				$settings = include($this->getBaseDir() . 'LICENSER.php');

				if(isset($settings['url'])) {
					if(!\UpdateSiteModel::Count(['url = ' . $settings['url']])) {
						$site = new \UpdateSiteModel();
						$site->url = $settings['url'];
						$site->save();

						$results['changes'][] = 'Installed update URL ' . $settings['url'];
					}
				}
			}
			catch(Exception $ex) {
				$results['errors'][] = $ex->getMessage();
			}
		}

		if(sizeof($results['changes']) === 0 && sizeof($results['errors']) === 0){
			\Core\log_debug('No changes detected');
		}

		return $results;
	}
	
	/**
	 * Process an upgrade or install command node.
	 * 
	 * 
	 * @param DOMNode $node
	 * 
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function processUpgradeNode(DOMNode $node) : array {
		$results = [
			'changes' => [],
			'errors' => [],
		];
		
		foreach($node->childNodes as $child){
			/** @var $child DOMNode */
			switch($child->nodeName){
				case 'dataset':
					try{
						if($this->_parseDatasetNode($child)){
							$results['changes'][] = 'Executed dataset ' . $child->getAttribute('action') . ' node on table ' . $child->getAttribute('table');
						}
					}
					catch (Exception $ex) {
						$results['errors'][] = $ex->getMessage();
					}
					break;
				case 'php':
					include $this->getBaseDir() . $child->getAttribute('file');
					$results['changes'][] = 'Included custom php file ' . $child->getAttribute('file');
					break;
				/*
				case 'sql':
					$contents = file_get_contents($this->getBaseDir() . $child->getAttribute('file'));
					$execs = 0;
					$parser = new SQL_Parser_Dataset($contents, SQL_Parser::DIALECT_MYSQL);
					$datasets = $parser->parse();
					foreach($datasets as $ds){
						$ds->execute();
						$execs++;
					}
					$changes[] = 'Executed custom sql file ' . $file . ' and ran ' . $execs . ($execs == 1 ? ' query' : 'queries');
					break;
				*/
				case '#text':
					// Text entries can silently be ignored.
					break;
				default:
					$results['errors'][] = 'Ignoring unsupported install directive: [' . $child->nodeName . ']';
			}
		}
		
		return $results;
	}
	
	/**
	 * Re-install this component
	 *
	 * Returns an array containing a list of changes performed
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function reinstall() {
		$results = [
			'changes' => [],
			'errors'  => [],
		];
			
		if(!$this->isInstalled()){
			// Only supported if this component is already installed.
			return $results;
		}
		
		\Core\log_trace('Reinstalling component ' . $this->_name);
		
		$results = array_merge_recursive(
			$this->installDBSchema(),
			$this->installConfigs(),
			$this->installPages(),
			$this->installWidgets(),
			$this->installAssets(),
			$this->installUpdater()
		);
		
		return $results;
	}
	
	/**
	 * UnInstall this component
	 *
	 * Returns an array containing a list of changes performed
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function uninstall() {
		$results = [
			'changes' => [],
			'errors'  => [],
		];
			
		if(!$this->isInstalled()){
			return $results;
		}
		
		\Core\log_trace('Uninstalling component ' . $this->_name);
		
		$results = array_merge_recursive(
			$this->uninstallDBSchema(),
			$this->uninstallConfigs(),
			$this->uninstallPages(),
			$this->uninstallWidgets(),
			$this->uninstallAssets()
		);
		
		
		// Set the version for this component in the database.
		$model = new ComponentModel($this->_keyName);
		$model->delete();
		
		return $results;
	}
	
	/**
	 * Remove all assets for this Component
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function uninstallAssets(){
		// Will contain the report for any changes performed herein.
		$results = [
			'changes' => [],
			'errors' => [],
		];
		
		\Core\log_trace('Uninstalling assets for Component ' . $this->getName());
		
		$assets = $this->getAssets();
		
		foreach($assets as $asset){
			if($asset->uninstall()) {
				$results['changes'][] = 'Removed ' . $asset->filename;
			}
			else{
				\Core\log_trace('Skipped removal of ' . $asset->filename);
			}
		}
		
		if(sizeof($results['changes']) === 0 && sizeof($results['errors']) === 0){
			\Core\log_debug('No changes detected');
		}

		return $results;
	}
	
	/**
	 * Uninstall all configuration options for this component
	 * 
	 * This is used for installations and upgrades.
	 * 
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function uninstallConfigs(){
		// Keep track of if this changed anything.
		$results = [
			'changes' => [],
			'errors'  => [],
		];

		\Core\log_trace('Uninstalling configs for Component ' . $this->getName());
		
		$configs = $this->getConfigs();
		
		foreach($configs as $config){
			if($config->uninstall()){
				$results['changes'][] = 'Removed configuration ' . $config->key;
			}
			else{
				\Core\log_trace('Skipped, no changes detected.');
			}
		}

		return $results;
	}
	
	/**
	 * Parse and handle the various Models and their DB Schemas
	 * 
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function uninstallDBSchema() : array {
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		\Core\log_trace('Uninstalling database schema for Component ' . $this->getName());
		
		/** @var \Core\Datamodel\BackendInterface $db */
		$db     = \Core\db();
		
		// Get the table structure as it exists in the database first, this will be the comparison point.
		$classes = $this->getModelList();

		// Do the actual processing of every Model.
		foreach ($classes as $m => $file) {
			// Ensure the Model exists, (it may not be loaded yet)
			require_once $file;
			
			$schema = ModelFactory::GetSchema($m);
			$tablename = $m::GetTableName();
			
			\Core\log_trace('Removing database table ' . $tablename);
			
			if($db->tableExists($tablename)){
				$db->dropTable($tablename);
				$results['changes'][] = 'Removed table ' . $tablename;
			}
		}
		
		return $results;
	}
	
	/**
	 * Uninstall any page defined in this Component
	 * 
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function uninstallPages() : array {
		// Keep track of if this changed anything.
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		\Core\log_trace('Uninstalling pages for Component ' . $this->_name);
		
		// Retrieve a list of current pages attached to this Component.
		// This is used as a cache of what pages were deleted in the schema, (if any).
		$pages = Dataset::Init()
			->select('baseurl')
			->table('page')
			->where('component = ' . $this->_keyName)
			->executeAndGet();
		
		foreach($pages as $baseurl){
			Dataset::Init()
				->delete()
				->table('page')
				->where('component = ' . $this->_keyName)
				->where('baseurl = ' . $baseurl)
				->execute();
			$results['changes'][] = 'Deleted page ' . $baseurl;
		}
		
		return $results;
	}
	
	/**
	 * Uninstall any widgets defined in this Component
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function uninstallWidgets() {
		// Keep track of if this changed anything.
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		\Core\log_trace('Uninstalling Widgets for ' . $this->getName());
		
		$widgets = $this->getWidgetsDefined();
		
		foreach($widgets as $widget){
			/** @var $page ComponentWidgetDefinition */
			if($widget->uninstall()){
				$results['changes'][] = 'Removed widget ' . $widget->baseurl;
			}
		}
		
		return $results;
	}
	
	/**
	 * Upgrade this component to the newer version, if possible.
	 *
	 * Returns an array containing a list of changes performed
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function upgrade() {
		$results = [
			'changes' => [],
			'errors'  => [],
		];
			
		if(!$this->isInstalled()){
			return $results;
		}
		
		\Core\log_trace('Upgrading component ' . $this->_name . ' from ' . $this->_versionDB . ' to ' . $this->_version);

		foreach ($this->_xmlLoader->getRootDOM()->getElementsByTagName('upgrade') as $u) {
			/** @var $u DOMElement */

			$from = $u->getAttribute('from');
			$to   = $u->getAttribute('to');

			\Core\log_trace('Incremental upgrade step from ' . $from . ' to ' . $to);

			// look for a valid upgrade path.
			if ($this->_versionDB == $from) {
				$results = array_merge_recursive(
					$results, 
					$this->processUpgradeNode($u)
				);
				$this->_versionDB = $to;
			}
		}


		// Only after all the upgrade directives have run, should we sync the codebase with the various destinations.
		// This is required because an upgrade is expecting a state from the *previous* version, so we shouldn't
		// actually perform the sync until after those have completed.
		$results = array_merge_recursive(
			$this->installDBSchema(),
			$this->installConfigs(),
			$this->installPages(),
			$this->installWidgets(),
			$this->installAssets()
		);
		
		// Set the version for this component in the database.
		$model = new ComponentModel($this->_keyName);
		$model->set('version', $this->_version);
		$model->save();
		
		return $results;
	}
	
	/**
	 * Upgrade this component to the newest "next" version, when present.
	 *
	 * Returns an array containing a list of changes performed
	 *
	 * @return array Contains changes <<string>> and errors <<string>>
	 */
	public function upgradeNext() {
		$results = [
			'changes' => [],
			'errors'  => [],
		];
		
		if(!$this->isInstalled()){
			return $results;
		}
		
		\Core\log_trace('Upgrading component ' . $this->_name . ' from ' . $this->_versionDB . ' to *NEXT*');
		
		foreach ($this->_xmlLoader->getRootDOM()->getElementsByTagName('upgrade') as $u) {
			/** @var $u DOMElement */

			$from = $u->getAttribute('from');
			$to   = $u->getAttribute('to');

			// look for a valid upgrade path.
			if ($this->_versionDB == $from && $to === 'next') {
				$results = array_merge_recursive(
					$results, 
					$this->processUpgradeNode($u)
				);
			}
		}
		
		return $results;
	}
	
	// </editor-fold>
	
	// <editor-fold defaultstate="collapsed" desc="PUBLIC Utility Methods">
	
	/**
	 * Query the registered licenser URL for this Component.
	 * 
	 * @throws Exception
	 * 
	 * @return array 'status' and 'message' are keys returned from this call.
	 */
	public function queryLicenser(){ 
		$data = $this->getLicenseData();

		// Every query voids the cache!
		$cacheKey = md5('LICENSER:' . SERVER_ID . $this->getKeyName());
		Cache::Delete($cacheKey);
		$this->_licenserFileData = null;

		// No data?  Simple!
		if(!sizeof($data)){
			return null;
		}       

		$url = $data['url'];

		$r = new \Core\Filestore\Backends\FileRemote();
		// I need to use POST here with a payload because Apache 2.4 is blocking custom X- headers!
		$r->setMethod('POST');
		$r->setPayload(['serverid' => SERVER_ID]);
		// Send the component and version as GET parameters so the retrieving server can gather analytics on who is using what version.
		$r->setFilename($url . '/licenser?component=' . $this->getKeyName() . '&version=' . $this->getVersion());

		$contents = $r->getContents();

		if(strpos($contents, '-----BEGIN PGP MESSAGE-----') === false){
			return [
				'status' => false,
				'message' => 'Unexpected return from the server! ' . htmlentities($contents),
			];
		}

		// Ensure that this content matches out with the recorded key.
		//$sig = \Core\Filestore\Factory::File('tmp/gpg-verify-' . \Core::RandomHex(6) . '.asc');
		//$sig->putContents($contents);
		$gpg = new GPG();
		try {
			$gpg->verifySignedData($contents);
		}
		catch(Exception $ex) {
			// Try to load the public key if it's set.
			try {
				if(file_exists($this->getBaseDir() . 'LICENSER.gpg')) {
					$gpg->importKeyFile($this->getBaseDir() . 'LICENSER.gpg');

					// and try it again
					$gpg->verifySignedData($contents);
				}
				else {
					return [
						'status' => false,
						'message' => 'No public key available!',
					];
				}
			}
			catch(Exception $ex) {
				return [
					'status' => false,
					'message' => 'Invalid GPG signed content from server!  Do you have the correct keys installed?',
				];
			}
		}


		// Verification was successful!  Record this onto the component.
		$c = ComponentModel::Construct($this->_name);
		$c->set('license', $contents);
		$c->save();
		$this->_licenseDBData = $contents;

		return [
			'status' => true,
			'message' => 'Retrieved license successfully!',
		];
	}

	/**
	 * Package this component to a tarball in the given location
	 *
	 * @param string $destination
	 * @param bool $compressed
	 *
	 * @throws Exception
	 */
	public function package(string $destination, bool $compressed = true) {
		$command = new Command('tar');

		// We're performing a create, so set the necessary commands to start.
		if($compressed) {
			// Compressed archives only support creating due to limitations in tar.
			$command->addArgKey('--gzip');
			$command->addArgKey('--create');
		}
		else {
			// Standard tar archives support appending content
			if(file_exists($destination)) {
				$command->addArgKey('--append');
			}
			else {
				$command->addArgKey('--create');
			}
		}

		// We want to dereference symbolic links, because most development environments will have the various packages linked.
		$command->addArgKey('--dereference');

		// Exclude some common files
		$command->addArgKey('--exclude-vcs');
		$command->addArgKey('--exclude=*~');
		$command->addArgKey('--exclude=.*');
		$command->addArgKey('--exclude=*.pyc');


		// Base directory of everything
		$command->addArgKeyValue('--file', $destination);
		$command->addArgKeyValue('--directory', ROOT_PDIR);


		$files = [
			substr($this->getBaseDir(), strlen(ROOT_PDIR))
		];

		if($this->getKeyName() === 'core') {
			$files[] = 'index.php';
			$files[] = 'cli';
			$files[] = 'robots.txt';

			// Include some default system directories for convenience.
			// This will include the DIRECTORIES ONLY!
			$files[] = 'themes/custom/';
			$command->addArgKey('--exclude=themes/custom/*');

			$files[] = 'files/public/';
			$command->addArgKey('--exclude=files/public/*');

			$files[] = 'files/assets/';
			$command->addArgKey('--exclude=files/assets/*');

			$files[] = 'files/private/';
			$command->addArgKey('--exclude=files/private/*');

			$files[] = 'gnupg/';
			$command->addArgKey('--exclude=gnupg/*');

			$files[] = 'logs/';
			$command->addArgKey('--exclude=logs/*');

			$files[] = 'config/';
			$command->addArgKey('--exclude=config/*');
		}

		// Add all the files to be archived.
		foreach($files as $file) {
			$command->addArgValue($file);
		}

		$command->execute();

		if($command->getError()) {
			throw new Exception($command->getError());
		}
	}
	
	// </editor-fold>
	
	// <editor-fold defaultstate="collapsed" desc="PRIVATE Load Methods">

	/**
	 * @throws Exception
	 */
	private function _init() {
		if (!$this->_xmlLoader->load()) {
			throw new \Exception('Invalid component.xml file provided, unable to create Component!');
		}

		$systemID = $this->_xmlLoader->getDOM()->doctype->systemId;
		switch($systemID){
			case 'https://corepl.us/api/6_3/component.dtd':
				// 6.3 was released Oct 2017
				$this->_apiVersion = 63;
				break;
			case 'http://corepl.us/api/2_4/component.dtd':
			case 'https://corepl.us/api/2_4/component.dtd':
				// 2.4 was released around Jan 2013
				$this->_apiVersion = 24;
				break;
			case 'http://corepl.us/api/2_1/component.dtd':
				// 2.1 was released early 2012
				$this->_apiVersion = 21;
				break;
			default:
				throw new \Exception('Unsupported component.xml version detected!');
		}

		$this->_type = 'component';
		$this->_name    = $this->_xmlLoader->getRootDOM()->getAttribute('name');
		$this->_version = $this->_xmlLoader->getRootDOM()->getAttribute("version");

		$this->_keyName = strtolower(str_replace(' ', '-', $this->_name));
	}
	
	/**
	 * Load all files from the XML and populate them into the necessary locations.
	 */
	private function _loadFilesDataFromXML() {
		$dir = $this->getBaseDir();
		
		$this->_autoLoads             = [];
		$this->_classList             = [];
		$this->_controllerList        = [];
		$this->_modelList             = [];
		$this->_supplementalModelList = [];
		$this->_widgetList            = [];
		
		
		foreach ($this->_xmlLoader->getElements('/files/file') as $f) {
			$filename = $dir . $f->getAttribute('filename');
			
			if($f->getAttribute('autoload') === '1'){
				$this->_autoLoads[] = $filename;
			}
			
			// Each <file> entry probably has one or more classes embedded inside.
			foreach ($f->getElementsByTagName('class') as $p) {
				$name = $p->getAttribute('name');
				$type = $p->getAttribute('type') ?: null;
				
				// ClassList gets everything!
				$this->_classList[$name] = $filename;
				
				if($type === 'controller'){
					$this->_controllerList[$name] = $filename;
				}
				elseif($type === 'model'){
					$this->_modelList[$name] = $filename;
				}
				elseif($type === 'widget'){
					$this->_widgetList[$name] = $filename;
				}
				elseif($type === 'modelsupplemental'){
					$this->_supplementalModelList[$name] = $filename;
				}
			}

			foreach ($f->getElementsByTagName('interface') as $p) {
				$n = $p->getAttribute('name');
				$this->_classList[$n] = $filename;
			}

			foreach ($f->getElementsByTagName('trait') as $p) {
				$n = $p->getAttribute('name');
				$this->_classList[$n] = $filename;
			}
		}
	}
	
	/**
	 * Load the various <provides> from the XML into memory
	 */
	private function _loadProvidesFromXML(){
		$this->_provides = [];
		
		// Every component is its own library
		$this->_provides[] = new ComponentProvide('library', $this->_keyName, $this->_version);
		$this->_provides[] = new ComponentProvide('component', $this->_keyName, $this->_version);

		foreach ($this->_xmlLoader->getElements('/provides/provide') as $p) {
			$type = $p->getAttribute('type');
			$name = $p->getAttribute('name');
			$v = $p->getAttribute('version') ?: $this->_version;
			$this->_provides[] = new ComponentProvide($type, $name, $v);
		}
	}
	
	// </editor-fold>
	
	// <editor-fold defaultstate="collapsed" desc="PRIVATE Install-Related Methods">
	
	/**
	 * Internal function to parse and handle the dataset in the <upgrade> and <install> tasks.
	 * This is used for installations and upgrades.
	 *
	 * @param $node DOMElement
	 * @param $verbose bool
	 *
	 * @throws Exception
	 */
	private function _parseDatasetNode(DOMElement $node){
		$action   = $node->getAttribute('action');
		$table    = $node->getAttribute('table');
		$haswhere = false;
		$sets     = array();
		$renames  = array();
		$ds       = new Dataset();


		$ds->table($table);

		foreach($node->getElementsByTagName('datasetset') as $el){
			$sets[$el->getAttribute('key')] = $el->nodeValue;
		}

		foreach($node->getElementsByTagName('datasetrenamecolumn') as $el){
			// <datasetrenamecolumn oldname="ID" newname="id"/>
			$renames[$el->getAttribute('oldname')] = $el->getAttribute('newname');
		}

		foreach($node->getElementsByTagName('datasetwhere') as $el){
			$haswhere = true;
			$ds->where(trim($el->nodeValue));
		}

		switch($action){
			case 'alter':
				if(sizeof($sets)){
					throw new Exception('Invalid mix of arguments on ' . $action . ' dataset request, datasetset is not supported!');
				}
				if($haswhere){
					throw new Exception('Invalid mix of arguments on ' . $action . ' dataset request, datasetwhere is not supported!');
				}

				foreach($renames as $k => $v){
					// ALTER TABLE `controllers` CHANGE `ID` `id` INT( 11 ) NOT NULL AUTO_INCREMENT
					$ds->renameColumn($k, $v);
				}
				break;
			case 'delete':
				if(sizeof($sets)){
					throw new Exception('Invalid mix of arguments on ' . $action . ' dataset request');
				}
				if(!$haswhere){
					throw new Exception('Cowardly refusing to delete with no where statement');
				}
				$ds->delete();
				break;
			case 'empty':
				if(sizeof($sets)){
					throw new Exception('Invalid mix of arguments on ' . $action . ' dataset request');
				}
				if($haswhere){
					throw new Exception('Unable to empty table when WHERE arguments are set!  Use delete instead.');
				}
				$ds->delete();
				break;
			case 'insert':
				foreach($sets as $k => $v){
					$ds->insert($k, $v);
				}
				break;
			case 'update':
				foreach($sets as $k => $v){
					$ds->update($k, $v);
				}
				break;
			default:
				throw new Exception('Invalid action type, '. $action);
		}

		$ds->execute();
		return ($ds->num_rows > 0);
	}
	
	// </editor-fold>
}
