<?php
/**
 * Core class of this entire system.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Core\Cache;
use Error;
use Exception;
use Core\Session;
use Core\Datamodel\DMI;
use Core\Datamodel\DatabaseLogSink as DatabaseLogger;
use Core\Logger\Logger;
use Core\CLI\LogSink as CLILogger;
use Core\Filestore\FileLogger;
use Core\Utilities\Profiler\ProfilerLogSink as ProfierLogger;
use Core\Utilities\SyslogLogger;
use Core\Utilities\Profiler\ProfilerEvent;


class Core implements SingletonInterface {
	
	// <editor-fold defaultstate="collapsed" desc="PRIVATE Variables">
	
	/**
	 * @var array <<string,string>> Hash of every registered Class to its full file path
	 */
	private $_classes;
	
	/**
	 * An array of every enabled Component in the system.
	 *
	 * @var array <<Component>>  Array of components keyed with the component keyname.
	 */
	private $_components;
	
	/**
	 * Array of the disabled components on the system.
	 *
	 * This is useful for the updater where the admin can enable/disable packages.
	 *
	 * @var array <<Component>>
	 */
	private $_componentsDisabled;
	
	/*
	 * @var array List of the order that components get registered in, used during initialization.
	 */
	private $_componentRegistrationOrder;
	
	/**
	 * @var array <<string,string>> Hash of registered lowercased controller names to their class
	 */
	private $_controllers;

	/**
	 * @var array Array of licensed feature codes and the URL that they point to.
	 */
	private $_features = [];
	
	/**
	 * @var bool Set to true when the framework has been started, (prevents Start() from running twice)
	 */
	private $_isStarted = false;
	
	/**
	 * @var bool Set to true when the framework detects that it has been installed
	 */
	private $_isInstalled = false;
	
	/**
	 * An array of every library in the system.
	 * This is useful because some components register additional libraries.
	 *
	 * @var array <<string,string>>
	 */
	private $_libraries;

	/**
	 * All permissions that are registered from components.
	 *
	 * @var array
	 */
	private $_permissions = [];

	/**
	 * List of every available jslibrary and its call..
	 * @var array <<String>>
	 */
	private $_scriptlibraries;

	/**
	 * List of every installed view class and its location on the system.
	 * @var array <<String>>
	 */
	private $_viewClasses;

	/**
	 * List of widgets available on the system.
	 *
	 * @var array
	 */
	private $_widgets;
	
	/**
	 * The singleton instance of the Core object.
	 * 
	 * Populated from the Singleton static method
	 *
	 * @var Core
	 */
	private static $_Instance;
	
	
	// </editor-fold>


	// <editor-fold defaultstate="collapsed" desc="PRIVATE Init and Load Methods">
	
	/**
	 * Attempt to auto-upgrade any components that are flagged as out of date.
	 */
	private function _autoUpgradeComponents(){
		// Scan the set and if no component is flagged as requiring an update, just exit out.
		$updateRequired = false;
		foreach($this->_components as $c){
			/** @var $c Component */
			if($c->isPendingUpdate()){
				$updateRequired = true;
				break;
			}
		}
		
		if(!$updateRequired){
			return;
		}
		
		// Lock the site first!
		// This is because some upgrade procedures take a long time to upgrade.
		file_put_contents(TMP_DIR . 'lock.message', 'auto-upgrade running');
		
		
		// These get checked and processed in the order that they were registered
		// to ensure that any Component dependency is met already.
		foreach($this->_componentRegistrationOrder as $key){
			/** @var $c Component */
			$c = $this->_components[ $key ];
			
			$c->upgrade();
		}
		
		// Ensure the caches are updated for everything.
		Cache::Delete('asset-resolveurl');
		
		// Unlock the site now that upgrades are done.
		unlink(TMP_DIR . 'lock.message');
	}
	
	/**
	 * Check and process any low-level redirects that need to run.
	 * 
	 * These ONLY include SSL checks, configuration enforcing, and maintenance checks.
	 */
	private function _checkMandatoryRedirect(){
		// Is this site configured to require SSL mode?  If so, might as well require that here.
		if(SSL_MODE == SSL_MODE_REQUIRED && !SSL){
			if(!DEVELOPMENT_MODE){
				// Skip the 301 when not in production because it makes things rather annoying to debug at times...
				// ie: the browser remembers it was a permanent redirect and doesn't even try the nonSSL version.
				header("HTTP/1.1 301 Moved Permanently");
			}
			header('Location: ' . ROOT_URL_SSL . substr(REL_REQUEST_PATH, 1));
			die('This site requires SSL, if it does not redirect you automatically, please <a href="' . ROOT_URL_SSL . substr(REL_REQUEST_PATH, 1) . '">Click Here</a>.');
		}
		elseif(SSL_MODE == SSL_MODE_DISABLED && SSL){
			if(!DEVELOPMENT_MODE){
				// Skip the 301 when not in production because it makes things rather annoying to debug at times...
				// ie: the browser remembers it was a permanent redirect and doesn't even try the nonSSL version.
				header("HTTP/1.1 301 Moved Permanently");
			}
			header('Location: ' . ROOT_URL_NOSSL . substr(REL_REQUEST_PATH, 1));
			die('This site has SSL disabled, if it does not redirect you automatically, please <a href="' . ROOT_URL_NOSSL . substr(REL_REQUEST_PATH, 1) . '">Click Here</a>.');
		}
		
		// If the admin defined an explicit hostname in the configuration, there may be a mismatch
		// between the calculated SERVERNAME and HOST.  Use this to detect if we should redirect
		// to that defined hostname.
		if(SERVERNAME_CURRENT !== SERVERNAME){
			if(!DEVELOPMENT_MODE){
				// Skip the 301 when not in production because it makes things rather annoying to debug at times...
				// ie: the browser remembers it was a permanent redirect and doesn't even try the nonSSL version.
				header("HTTP/1.1 301 Moved Permanently");
			}
			header('Location: ' . ROOT_URL . substr(REL_REQUEST_PATH, 1));
			die('Site is configured to use a different URL, please <a href="' . ROOT_URL . substr(REL_REQUEST_PATH, 1) . '">Click Here</a>.');
		}
		
		// Another "redirect" of sorts is the maintenance message triggered by lock.message in /tmp.
		if(file_exists(TMP_DIR . 'lock.message')){
			die(file_get_contents(ROOT_PDIR . 'core/resources/maintenance.html'));
		}
	}
	
	/**
	 * Initialize the bare minimum for Core to function
	 * 
	 * This includes the class list for Core and any registered autoload.
	 * 
	 * @throws \Exception
	 */
	private function _init(){
		// Core is first, (obviously)
		// Load the classes for Core so that the base system is available for whatever
		// may be needed.
		if(!file_exists(ROOT_PDIR . 'core/component.xml')){
			throw new \Exception('Core component.xml not located, unable to load the framework.');
		}

		if(!class_exists('DOMDocument')) {
			throw new \Exception('Core requires XML support before starting.');
		}
		
		$this->_classes                    = [];
		$this->_components                 = [];
		$this->_componentsDisabled         = [];
		$this->_componentRegistrationOrder = [];
		$this->_controllers                = [];
		$this->_libraries                  = [];
		$this->_permissions                = [];
		$this->_scriptlibraries            = [];
		$this->_widgets                    = [];
		$this->_features                   = [];
		
		// Now that the Core class list has been assembled, the native autoloader can be hooked into
		// to provide the bulk of the functionality of Class lookups and auto-includes.
		// (at least as soon as the component is registered via _registerComponent!)
		// This needs to be here before because the autoloads are handled in that register method.
		spl_autoload_register('\Core\Core::CheckClass');
		
		$coreComponent = new Component(ROOT_PDIR . 'core/component.xml');
		$this->_registerComponent($coreComponent);
	}
	
	/**
	 * Load all the components in the Core registry
	 * @throws CoreException
	 */
	private function _loadComponents() {
		\Core\log_trace('Starting loading of component metadata');
		
		// Now that the base system is available, we can also load Core's data from the database.
		// This will be done for the rest of the components as well within this method.
		$this->_components['core']->loadAllFromDB();
		
		// Will hold all components dynamically located herein
		$components = false;
		
		// If the site is in DEVELOPMENT mode, component caching would probably be a bad idea; ie: the developer probably wants
		// those component files loaded everytime.
		$enableCache = !DEVELOPMENT_MODE;

		// Is there a cache of elements available?  This is a primary system cache that greatly increases performance,
		// since it will no longer have to run through each component.xml file to register each one.
		if($enableCache){
			\Core\log_trace('Checking core-components cache');
			// Try to load up the cached components and check them first.
			$components = Cache::Get('core-components', (3600 * 24));
		}

		if(!$enableCache || $components === false || $components === null){
			\Core\log_trace('Scanning for component.xml files manually');
			$components = [];

			// First, build my cache of components, regardless if the component is installed or not.
			if(($dh = opendir(ROOT_PDIR . 'components'))){
				while (($file = readdir($dh)) !== false) {
					if($file[0] === '.'){
						// Skip hidden directories/files with no debug warning.
						continue;
					}
					
					if(
						// Enforce a very small subset of characters to be used for the directory name.
						!preg_match('/^[a-z0-9\-]*$/', $file) ||
						
						// Skip files, (we only want directories here!)
						!is_dir(ROOT_PDIR . 'components/' . $file) ||
							
						// They also need a component.xml
						!is_readable(ROOT_PDIR . 'components/' . $file . '/component.xml')
						
					){
						\Core\log_debug('Skipping component file/directory ' . $file . ' due to incorrect naming or file schema.');
						continue;
					}
					
					\Core\log_trace('Opened component ' . $file);
					$comp = new Component(ROOT_PDIR . 'components/' . $file . '/component.xml');
					
					if($enableCache){
						// If the cache is enabled, request the component to load the XML schema into memory
						// so that when the object gets saved, all that metadata gets saved as well
						// and the next page request doesn't need to perform the XML iteration.
						$comp->loadAllFromXML();
					}
					
					$components[ $comp->getKeyName() ] = $comp;
				}
				
				closedir($dh);
				\Core\log_debug('Component XML file scanning complete');
			}
			else{
				\Core\log_debug('Component directory does not exist or could not be read, no components available!');
			}
			

			// Cache this list!
			if($enableCache){
				\Core\log_trace(' * Caching core-components for next pass');
				Cache::Set('core-components', $components, (3600 * 24));
			}
		}
		
		\Core\log_trace(' * Component metadata loaded, starting registration');

		if($components === false || $components === null) {
			// No caching and no components either!
			$components = [];
		}
		
		foreach($components as $n => $c){
			/** @var $c Component */
			
			// Ensure it's loaded from the database.
			// This will have the data about whether the user disabled it.
			$c->loadAllFromDB();

			
			if($c->isInstalled() && !$c->isEnabled()){
				// Explicitly disabled, (but otherwise installed), components get moved to a disabled list
				\Core\log_trace('Component ' . $c->getName() . ' administratively disabled.');
				$this->_componentsDisabled[$n] = $c;
				unset($components[$n]);
			}
			
			if(!$c->isInstalled()){
				// Component that has not been installed yet, all new installs must be approved!
				\Core\log_trace('Component ' . $c->getName() . ' not installed yet.');
				$this->_componentsDisabled[$n] = $c;
				unset($components[$n]);
			}
		}
		
		// Run the requirement checks for each component and register them if successful
		// This runs on a loop because some components registered 
		do {
			$size = sizeof($components);
			foreach($components as $n => $c) {
				/** @var $c Component */
				try{
					// If registration works on this component, then we're all honky!
					$this->_registerComponent($c);
					
					// Clear out any errors previously encountered.
					$c->error = null;
					
					$order[] = $n;
					unset($components[ $n ]);
				}
				catch (Exception $ex) {
					// Record this issue on the Component in case something wants to know what went wrong, eg: /updater.
					$c->error = $ex->getMessage();
					\Core\log_trace('Component ' . $c->getName() . ' registration failed due to ' . $ex->getMessage() . ', will probably retry');
				}
			}
		} while ($size > 0 && ($size != sizeof($components)));
		
		
		// If there are still components in the original array,
		// that means that their registrations failed even after iterating through the rest of the apps.
		// DISABLED!
		foreach($components as $n => $c){
			\Core\log_notice('Component ' . $c->getName() . ' could not be registered due to unmet dependencies.');
			$this->_componentsDisabled[$n] = $c;
			unset($components[$n]);
		}
	}

	/**
	 * Load any licensing information for any component installed.
	 *
	 * This requires use of the LICENSER.php settings file.
	 */
	private function _loadLicenses() {
		foreach($this->_components as $component) {
			$licenser = $component->getLicenseData();
			if(sizeof($licenser) && isset($licenser['features'])){
				$this->_features += $licenser['features'];
			}
		}
	}
	
	/**
	 * Try to load the configuration system for Core
	 * 
	 * If succeeds, the Core configuration will be returned for further processing.
	 * 
	 * @throws \Exception
	 * 
	 * @return array
	 */
	private function _loadConfiguration() : array {
		ConfigHandler::Singleton();

		// Give me core settings!
		// This will do the defines for the site, and provide any core variables to get started.
		$coreSettings = ConfigHandler::LoadConfigFile("configuration");
		
		if(!$coreSettings){
			throw new \Exception('No configuration XML created yet');
		}
		
		return $coreSettings;
	}
	
	/**
	 * Load and install any ModelSupplemental onto their corresponding Models
	 * 
	 * Any calls prior to this will not have any supplemental Model data.
	 */
	private function _loadModelSupplementals(){
		// Will contain a list of all supplemental models on the system, with the base class as the key.
		// This is so that when the model is actually loaded, the corresponding supplemental data can be loaded along with it. 
		$modelSupplementals = [];
		
		// Iterate over every enabled component and retrieve the supplemental data.
		foreach($this->_components as $c){
			/** @var $c Component */
			$supplementals = $c->getSupplementalModelList();
			foreach($supplementals as $class => $file){
				$base = substr($class, strpos($class, '_') + 1, -12);
				
				if(!isset($modelSupplementals[$base])){
					$modelSupplementals[$base] = [];
				}

				$modelSupplementals[$base][] = $class;
			}
		}
		
		// Sort them, (to at least give them a reliable order).
		// Otherwise, they're pulled in by whatever order the filesystem gives them.
		// this could cause issues with the database because the column order would differ.
		foreach($modelSupplementals as $k => $dat){
			sort($modelSupplementals[$k]);
		}
		
		// Lastly iterate over the list again now that the data is available and patch that original data.
		foreach($this->_components as $c){
			/** @var $c Component */
			$modelList = $c->getModelList();
			
			// Merge any models found with the list of supplemental models by other components.
			foreach($modelList as $k => $file) {
				if(isset($modelSupplementals[ $k ])) {
					// This model has supplemental functionality!
					foreach($modelSupplementals[ $k ] as $e) {
						Model::AddSupplemental($k, $e);
					}
				}
			}
		}		
	}
	
	/**
	 * Perform a simple registration of a Component and its classes
	 * 
	 * If the component cannot be loaded due to a missing requirement or conflict,
	 * an exception is thrown and the component is NOT registered.
	 * 
	 * @param \Core\Component $component
	 * 
	 * @throws \Exception
	 */
	private function _registerComponent(Component $component) {
		$keyName = $component->getKeyName();

		if($this->_isStarted) {
			// Install/Enable status is only known if Core is already started,
			// and this logic is performed in _loadComponents.
			if(!$component->isInstalled()) {
				throw new Exception($component->getName() . ' is not installed yet, unable to register');
			}

			if(!$component->isEnabled()) {
				throw new Exception($component->getName() . ' has been disabled by an administrator, unable to register');
			}
		}

		foreach($component->getRequires() as $requirement){
			/** @var $requirement ComponentRequirement */
			$requirement->check();
		}

		if($this->_isStarted) {
			// Core has already started, so ensure the next load will include this updated registration.
			Cache::Delete('core-components');
		}
		
		foreach($component->getClassList() as $name => $path){
			if($this->_isStarted) {
				// Allow overrides to work if we're reregistering something
				$this->_classes[$name] = $path;
			}
			else {

				if(isset($this->_classes[ $name ])){
					throw new \Exception('Class ' . $name . ' provided in another component!');
				}

				$this->_classes[$name] = $path;
			}
		}

		foreach($component->getLibraryList() as $name => $version){
			if($this->_isStarted) {
				// Allow overrides to work if we're reregistering something
				$this->_libraries[$name] = $version;
			}
			else {

				if(isset($this->_libraries[ $name ])){
					throw new \Exception('Library ' . $name . ' provided in another component!');
				}

				$this->_libraries[$name] = $version;
			}
		}

		foreach($component->getScriptLibraryList() as $name => $call){
			$this->_scriptlibraries[$name] = $call;
		}

		// Ensure this new component register is available globally
		if(isset($this->_components[$keyName])) {
			$this->_components[ $keyName ] = $component;
		}
		else {
			$this->_components[ $keyName ] = $component;
			$this->_componentRegistrationOrder[] = $keyName;
		}

		// Remove it from the disabled components if set.
		if(isset($this->_componentsDisabled[$keyName])) {
			unset($this->_componentsDisabled[$keyName]);
		}
		
		// Register all controllers to their callable name
		foreach($component->getControllerList() as $name => $path ){
			$uri = strtolower( substr($name, 0, -10) );
			$this->_controllers[ $uri ] = $name;
		}
		
		// Much like controllers, widgets can be called from their "baseurl" path.
		foreach($component->getWidgetList() as $name => $path ){
			$uri = strtolower( substr($name, 0, -6) );
			$this->_widgets[ $uri ] = $name;
		}
		
		foreach($component->getAutoLoads() as $autoload){
			include $autoload;
		}
		
		$component->registerHooks();

		foreach($component->getPermissions() as $permission) {
			if(!in_array($permission, $this->_permissions)) {
				$this->_permissions[] = $permission;
			}
		}
		
		foreach($component->getFormElements() as $tag => $class) {
			Forms\Form::$Mappings[ $tag ] = $class;
		}
		
		// Register this component's user authdrivers, if any.
		$auths = $component->getUserAuthDrivers();
		foreach($auths as $name => $class){
			\Core\User\Helper::$AuthDrivers[$name] = $class;
		}
	}
	
	/**
	 * Generate all the necessary DEFINE directives used by the Core framework
	 * 
	 * @param array $settings Set of data from the Core configuration or null if not available yet
	 */
	private function _setupDefines(array $settings = null){
		
		/**
		 * @var bool Shortcut for if settings are null
		 */
		$set = ($settings !== null);
		
		/**
		 * If the execution mode is as a script, most web-based constants are simply null.
		 * This section sets up the following constants:
		 * SERVERNAME
		 * SERVERNAME_NOSSL
		 * SERVERNAME_SSL
		 * ROOT_URL
		 * ROOT_URL_NOSSL
		 * ROOT_URL_SSL
		 * CUR_CALL
		 */
		if (EXEC_MODE === 'CLI') {
			/**
			 * Hostname + Full URL of server.
			 * ie: http://www.example.com, https://127.0.0.1:8443, or NULL
			 */
			if($set && $settings['site_url'] != ''){
				$host = $settings['site_url'];
				$servername = 'http://' . $host;
			}
			elseif(isset($_SERVER['HTTP_HOST'])) {
				$host = $_SERVER['HTTP_HOST'];
				$servername = 'http://' . $host;
			}
			else{
				$host = 'localhost';
				$servername = null;
			}
			$servernameSSL       = null;
			$servernameNOSSL     = null;
	
			if($servername !== null){
				$rooturl = $servername . (defined('ROOT_WDIR') ? ROOT_WDIR : '');
			}
			else{
				$rooturl = null;
			}
			$servernameCurrent   = $servername;
			$hostConfigured      = $host;
			$rooturlNOSSL        = $rooturl;
			$rooturlSSL          = $rooturl;
			$curcall             = null;
			$relativerequestpath = null;
			$ssl                 = false;
			$sslmode             = 'disabled';
			$tmpdir              = $set ? $settings['tmp_dir_cli'] : '/tmp/core-cli/';
			
			// Check if this user has a .gnupg directory in the home directory.
			// This is because when the user runs a script, (ie: packager or create_repo),
			// it should use his/her private key, (which is not accesable from the website).
			if (isset($_SERVER['HOME']) && is_dir($_SERVER['HOME'] . '/.gnupg')){
				$gnupgdir = $_SERVER['HOME'] . '/.gnupg/';
			}
			else{
				$gnupgdir = false;
			}

			// CLI mode shouldn't have HTML error reporting.
			// Having this enabled just produces annoying HTML tags on stdout!
			ini_set('html_errors', 0);
		}
		else {
			/**
			 * @var string Current base hostname eg: "localhost" or "domain.tld"
			 */
			$host = $_SERVER['SERVER_NAME'];
			
			/**
			 * @var bool Determined if SSL is currently being used based on a bunch of different criteria.
			 */
			$ssl = (
				// Standard header provided 99% of the time.
				// If set, it's probably set to "ON", otherwise it's unset, so we'll set that to "off".
				strtolower($_SERVER['HTTPS'] ?? 'off') === 'on' ||

				// a de facto standard for identifying the originating protocol of an HTTP request,
				// since a reverse proxy (load balancer) may communicate with a web server using HTTP even if the request to the reverse proxy is HTTPS.
				strtolower($_SERVER['HTTP_X_FORWARDED_PROTO'] ?? '') === 'https' ||

				// Non-standard header field used by Microsoft applications and load-balancers
				strtolower($_SERVER['FRONT_END_HTTPS'] ?? 'off') === 'on' ||
				strtolower($_SERVER['HTTP_HTTPS'] ?? 'off') === 'on'
			);
			
			/**
			 * @var string Current prefix to use based on if SSL is detected
			 */
			$prefix = $ssl ? 'https://' : 'http://';
			
			/**
			 * @var string Configured base hostname, or HOST if not defined otherwise.
			 */
			$hostConfigured = ($set && $settings['site_url'] !== '') ? $settings['site_url'] : $host;
			
			$currentPort = (int)$_SERVER['SERVER_PORT'];
			$portSSL = defined('PORT_NUMBER_SSL') ? PORT_NUMBER_SSL : ($ssl ? $currentPort : 443);
			$portNOSSL = defined('PORT_NUMBER') ? PORT_NUMBER : (!$ssl ? $currentPort : 80);
			
			/**
			 * Full URL of server as configured
			 * ie: http://www.example.com or https://127.0.0.1:8443
			 */
			$servername = $prefix . $hostConfigured;
			if($ssl && $portSSL !== 443){
				// Tack on non-default port if necessary
				$servername .= ':' . $portSSL;
			}
			elseif(!$ssl && $portNOSSL !== 80){
				// Tack on non-default port if necessary
				$servername .= ':' . $portNOSSL;
			}
			
			/**
			 * Full URL of server as currently loaded
			 * ie: http://www.example.com or https://127.0.0.1:8443
			 */
			$servernameCurrent = $prefix . $host;
			if($ssl && $portSSL !== 443){
				// Tack on non-default port if necessary
				$servernameCurrent .= ':' . $portSSL;
			}
			elseif(!$ssl && $portNOSSL !== 80){
				// Tack on non-default port if necessary
				$servernameCurrent .= ':' . $portNOSSL;
			}
			
			/**
			 * Full URL of server in non-SSL mode.
			 * ie: http://www.example.com:80 or http://localhost:880
			 */
			$servernameNOSSL = 'http://' . $hostConfigured;
			if($portNOSSL !== 80){
				// Tack on non-default port if necessary
				$servernameNOSSL .= ':' . $portNOSSL;
			}

			/**
			 * Determine how to handle the SSL settings.
			 */
			if(defined('SSL_MODE')){
				define('ENABLE_SSL', (SSL_MODE !== 'disabled'));
			}
			else{
				define('SSL_MODE', 'disabled');
				define('ENABLE_SSL', false);
			}


			/**
			 * Full URL of server in SSL mode.
			 * ie: https://www.example.com:443 or https://localhost:8443
			 *
			 * (defaults back to SERVERNAME_NOSSL if ENABLE_SSL is disabled).
			 */
			if (ENABLE_SSL) {
				// Create the server name for SSL connections.  This should override any previous port number.
				$servernameSSL = 'https://' . $hostConfigured;
				if($portSSL !== 443){
					// Tack on non-default port if necessary
					$servernameSSL .= ':' . $portSSL;
				}
			}
			else {
				// SSL is disabled in the configuration, reset the SSL option back to NOSSL just in case.
				$servernameSSL = $servernameNOSSL;
			}

			$rooturl             = $servername . ROOT_WDIR;
			$rooturlNOSSL        = $servernameNOSSL . ROOT_WDIR;
			$rooturlSSL          = $servernameSSL . ROOT_WDIR;
			$requestURI          = $_SERVER['REQUEST_URI'];
			$curcall             = '/' . substr($requestURI, strlen(ROOT_WDIR));
			$relativerequestpath = strtolower('/' . substr($requestURI, strlen(ROOT_WDIR)));
			if (strpos($relativerequestpath, '?') !== false){
				$relativerequestpath = substr($relativerequestpath, 0, strpos($relativerequestpath, '?'));
			}


			/*

			X-Forwarded-For
			a de facto standard for identifying the originating IP address of a client connecting to a web server through an HTTP proxy or load balancer

			X-Forwarded-For: client1, proxy1, proxy2
			X-Forwarded-For: 129.78.138.66, 129.78.64.103

			X-Forwarded-Host
			a de facto standard for identifying the original host requested by the client in the Host HTTP request header,
			since the host name and/or port of the reverse proxy (load balancer) may differ from the origin server handling the request.

			X-Forwarded-Host: en.wikipedia.org:80
			X-Forwarded-Host: en.wikipedia.org
			*/

			// @todo Implement support for trusted proxy IP addresses!
			

			$tmpdir = $set ? $settings['tmp_dir_web'] : '/tmp/core-web/';

			$gnupgdir = false;
			
			// Web uses HTML output, (obviously)
			ini_set('html_errors', 1);
		}
		
		if($set && CDN_TYPE === 'local'){
			$assetdir = (strpos(CDN_LOCAL_ASSETDIR, '/') === 0) ? CDN_LOCAL_ASSETDIR : (ROOT_PDIR . CDN_LOCAL_ASSETDIR);
			$publicdir = (strpos(CDN_LOCAL_PUBLICDIR, '/') === 0) ? CDN_LOCAL_PUBLICDIR : (ROOT_PDIR . CDN_LOCAL_PUBLICDIR);
			$privatedir = (strpos(CDN_LOCAL_PRIVATEDIR, '/') === 0) ? CDN_LOCAL_PRIVATEDIR : (ROOT_PDIR . CDN_LOCAL_PRIVATEDIR);
		}
		elseif($set && CDN_TYPE === 'ftp'){
			$assetdir = (strpos(CDN_FTP_ASSETDIR, '/') === 0) ? CDN_FTP_ASSETDIR : (CDN_FTP_PATH . CDN_FTP_ASSETDIR);
			$publicdir = (strpos(CDN_FTP_PUBLICDIR, '/') === 0) ? CDN_FTP_PUBLICDIR : (CDN_FTP_PATH . CDN_FTP_PUBLICDIR);
			$privatedir = (strpos(CDN_FTP_PRIVATEDIR, '/') === 0) ? CDN_FTP_PRIVATEDIR : (CDN_FTP_PATH . CDN_FTP_PRIVATEDIR);
		}
		else{
			$assetdir = null;
			$publicdir = null;
			$privatedir = null;
		}
		
		if($assetdir !== null && strrpos($assetdir, '/') !== strlen($assetdir) - 1){
			// Ensures it ends in a trailing slash
			$assetdir .= '/';
		}
		if($publicdir !== null && strrpos($publicdir, '/') !== strlen($publicdir) - 1){
			// Ensures it ends in a trailing slash
			$publicdir .= '/';
		}
		if($privatedir !== null && strrpos($privatedir, '/') !== strlen($privatedir) - 1){
			// Ensures it ends in a trailing slash
			$privatedir .= '/';
		}
		if($tmpdir !== null && strrpos($tmpdir, '/') !== strlen($tmpdir) - 1){
			// Ensures it ends in a trailing slash
			$tmpdir .= '/';
		}
		

		/**
		 * Full URL of server.
		 * eg: http://www.example.com or https://127.0.0.1:8443
		 */
		define('SERVERNAME', $servername);
		/**
		 * Full URL of the server forced non-ssl mode.
		 * eg: http://www.example.com
		 */
		define('SERVERNAME_NOSSL', $servernameNOSSL);
		/**
		 * Full URL of the server forced SSL mode.
		 * eg: https://www.example.com or https://127.0.0.1:8443
		 */
		define('SERVERNAME_SSL', $servernameSSL);
		/**
		 * Full URL of the server as currently loaded
		 * eg: https://www.example.com or https://127.0.0.1:8443
		 */
		define('SERVERNAME_CURRENT', $servernameCurrent);
		/**
		 * URL of web root.
		 * eg: http://www.example.com/foo/man/choo/
		 */
		define('ROOT_URL', $rooturl);
		/**
		 * URL of web root.
		 * eg: http://www.example.com/foo/man/choo/
		 */
		define('ROOT_URL_NOSSL', $rooturlNOSSL);
		/**
		 * URL of web root.
		 * eg: https://www.example.com/foo/man/choo/
		 */
		define('ROOT_URL_SSL', $rooturlSSL);
		/**
		 * Current call/request.
		 * eg: /user/login?somevariable=true&somethingelse=false
		 */
		define('CUR_CALL', $curcall);

		/**
		 * Relative requested path.
		 * eg: /user/login or '/' for the index.
		 */
		define('REL_REQUEST_PATH', $relativerequestpath);

		/**
		 * Simple true/false if current page call is via SSL.
		 */
		define('SSL', $ssl);

		/**
		 * SSL Mode for SSL being disabled completely
		 */
		define('SSL_MODE_DISABLED', 'disabled');
		/**
		 * SSL is allowed on pages that require it only, (standard pages redirect to non-ssl)
		 */
		define('SSL_MODE_ONDEMAND', 'ondemand');
		/**
		 * SSL is allowed on any page throughout the site
		 */
		define('SSL_MODE_ALLOWED',  'allowed');
		/**
		 * SSL is always required for all pages
		 */
		define('SSL_MODE_REQUIRED', 'required');


		if(!defined('TMP_DIR')) {
			/**
			 * Temporary directory
			 */
			define('TMP_DIR', $tmpdir);
		}

		/**
		 * Temporary directory for web only
		 * (useful in the packager)
		 */
		define('TMP_DIR_WEB', $set ? $settings['tmp_dir_web'] : '/tmp/core-web/');

		/**
		 * Temporary directory for cli only
		 * (useful in the packager)
		 */
		define('TMP_DIR_CLI', $set ? $settings['tmp_dir_cli'] : '/tmp/core-cli/');
		
		/**
		 * @var string Fully resolved files/assets path
		 */
		define('ASSET_PDIR', $assetdir);
		
		/**
		 * @var string Fully resolved files/public path
		 */
		define('PUBLIC_PDIR', $publicdir);
		
		/**
		 * @var string Fully resolved files/private path
		 */
		define('PRIVATE_PDIR', $privatedir);
		
		/**
		 * Host is a more human-friendly version of SERVERNAME.
		 * It does not include port number or protocol, but just the hostname itself.
		 * eg: domain.tld
		 */
		define('HOST', $hostConfigured);

		// The TMP_DIR needs to be writable!
		if (!is_dir(TMP_DIR)) {
			mkdir(TMP_DIR, 0777, true);
		}
		
		/**
		 * The GnuPG home directory to store keys in.
		 */
		if (!defined('GPG_HOMEDIR')) {
		   define('GPG_HOMEDIR', ($gnupgdir) ? $gnupgdir : ROOT_PDIR . 'gnupg/');
		}
		// PECL expects this variable to be set, so set it!
		putenv('GNUPGHOME=' . GPG_HOMEDIR);

		if(!defined('XHPROF')){
		   define('XHPROF', 0);
		}

		$profilingEnabled = (XHPROF == 100 || (XHPROF > rand(1,100)));

		if(function_exists('xhprof_enable')){
		   define('ENABLE_XHPROF', $profilingEnabled);
		   if($profilingEnabled){
			   xhprof_enable(XHPROF_FLAGS_CPU | XHPROF_FLAGS_MEMORY);
		   }
		}
		elseif(function_exists('xdebug_start_trace')){
		   // Use xdebug instead!
		   ini_set('xdebug.profiler_enable_trigger', $profilingEnabled ? 1 : 0);
		}
		
		if(defined('TIME_DEFAULT_TIMEZONE')){
			// Let the core override the server's timezone as well.
			// In the case that the timezone isn't set, this will prevent the "strftime unsafe" error.
			date_default_timezone_set(TIME_DEFAULT_TIMEZONE);
		}
		
		if(!defined('DEVELOPMENT_MODE')){
			define('DEVELOPMENT_MODE', false);
		}
		
		if(!defined('FULL_TRACE')){
			define('FULL_TRACE', false);
		}
		
		/**
		 * The site has already been configured to hide errors in the bootstrap by default.
		 * 
		 * This will re-enable them if the user selected DEVELOPMENT_MODE and/or trace output.
		 */
		if( DEVELOPMENT_MODE || FULL_TRACE ) {
			error_reporting(E_ALL | E_STRICT);
			ini_set('display_errors', 1);
		}
		else{
			// Make sure that errors are set to be displayed to the fullest extent.
			ini_set('display_errors', 0);
		}
	}
	
	/**
	 * Setup all the default Loggers and necessary error handling
	 */
	private function _setupLoggersAndErrorChecking(){
		// Handle errors generated internally.
		// These should be fewer and further between as PHP 7 has sifted a great portion of
		// bubbling to Exceptions, but some things still register errors.
		set_error_handler('Core\\ErrorManagement\\error_handler', error_reporting());
		
		
		// Register the HookHandler to dispatch the hook "/core/shutdown" automatically at the end of execution.
		register_shutdown_function('Core\\HookHandler::DispatchHook', '/core/shutdown');
		
		
		// Also register the hook for check for any fatal error.
		register_shutdown_function('Core\\ErrorManagement\\check_for_fatal');


		// Configure and egister all the default Core log sinks.
		if(FULL_TRACE && EXEC_MODE === 'CLI'){
			// Standard out is the only reliable output mechanism for CLI and full TRACES are requested.
			Logger::SetSinkLevel('stdout', LOG_LEVEL_TRACE);
		}
		elseif(FULL_TRACE && EXEC_MODE === 'WEB'){
			// The profiler better handles logging, but standard out should be used for errors.
			Logger::SetSinkLevel('profiler', LOG_LEVEL_TRACE);
			Logger::SetSinkLevel('stdout', LOG_LEVEL_HIGH);
		}
		elseif(DEVELOPMENT_MODE && EXEC_MODE === 'CLI'){
			// Standard out is the only reliable output mechanism for CLI and DEBUG info is requested.
			Logger::SetSinkLevel('stdout', LOG_LEVEL_INFO);
		}
		elseif(DEVELOPMENT_MODE && EXEC_MODE === 'WEB'){
			// The profiler better handles logging, but standard out should be used for errors.
			Logger::SetSinkLevel('profiler', LOG_LEVEL_DEBUG);
			Logger::SetSinkLevel('stdout', LOG_LEVEL_HIGH);
		}
		
		
		Logger::RegisterSink('profiler', new ProfierLogger());
		Logger::RegisterSink('stdout', new CLILogger());
		Logger::RegisterSink('file', new FileLogger());
		Logger::RegisterSink('database', new DatabaseLogger());
		Logger::RegisterSink('syslog', new SyslogLogger());
		\Core\log_trace('Core framework and base dependencies ready');
	}
	
	// </editor-fold>

	/**
	 * Internally used method to notify the rest of the system that a given
	 *    component has been loaded and is available.
	 *
	 * Expects all checks to be done already.
	 */
	public function _registerComponentOLD(Component $c) {
		$name = str_replace(' ', '-', strtolower($c->getName()));

		if ($c->hasLibrary()) {

			$liblist = $c->getLibraryList();

			$this->_libraries = array_merge($this->_libraries, $liblist);

			// Register the include paths if set.
			foreach ($c->getIncludePaths() as $path) {
				set_include_path(get_include_path() . PATH_SEPARATOR . $path);
			}
		}

		

		if ($c->hasModule()) $this->_modules[$name] = $c->getVersionInstalled();

		$this->_classes           = array_merge($this->_classes, $c->getClassList());
		$this->_viewClasses       = array_merge($this->_viewClasses, $c->getViewClassList());
		$this->_widgets           = array_merge($this->_widgets, $c->getWidgetList());
		$this->_components[$name] = $c;

		// All models get a control link registered automatically :)
		$models = $c->getModelList();
		foreach($models as $class => $file){
			if(!HookHandler::GetHook('/core/controllinks/' . $class)){
				$h = new Hook('/core/controllinks/' . $class);
				$h->returnType = Hook::RETURN_TYPE_ARRAY;
				$h->description = 'Automatic hook for control links on the ' . $class . ' object.  Attach onto this hook if you want to add a custom link anytime this object\'s control is displayed.';
			}
		}
		
		$licenser = $c->getLicenseData();
		if(sizeof($licenser) && isset($licenser['features'])){
			$this->_features += $licenser['features'];
		}

		// Lastly, mark this component as available!
		$c->_setReady(true);
	}


	// <editor-fold defaultstate="collapsed" desc="PUBLIC Static Utility Methods">
	
	/**
	 * Simple autoload register function to lookup a classname and resolve it.
	 *
	 * @throws Exception
	 *
	 * @param string $classname The class to lookup and auto-include
	 */
	public static function CheckClass($classname) {
		if (class_exists($classname)){
			// If the class already exists, nothing needs to be done
			return true;
		}
		
		// Grab the instance of Core that will contain the full app registery
		$core = self::Singleton();
		
		if(!isset($core->_classes[ $classname ])){
			return false;
		}
		
		if(!file_exists($core->_classes[ $classname ])){
			throw new \Exception('Class ' . $classname . ' defined but source file ' . $core->_classes[ $classname ] . ' not present!');
		}
		
		require_once $core->_classes[ $classname ];
	}
	
	/**
	 * Get all registered controllers mapped with their URL-version to the resolved class
	 * @return array <<string,string>>
	 */
	public static function GetControllers() : array {
		return self::Singleton()->_controllers;
	}
	
	/**
	 * Get all registered components
	 * @return Component[]
	 */
	public static function GetComponents() : array {
		return self::Singleton()->_components;
	}

	/**
	 * Get all disabled components
	 * @return array <<Component>>
	 */
	public static function GetDisabledComponents() : array {
		return self::Singleton()->_componentsDisabled;
	}
	
	/**
	 * Get all registered widgets mapped with their URL-version to the resolved class
	 * @return array <<string,string>>
	 */
	public static function GetWidgets() : array {
		return self::Singleton()->_widgets;
	}
	
	/**
	 * Start everything in Core!
	 * 
	 * None of the script can be called before running this!
	 */
	public static function Start() {
		$core = self::Singleton();
		if($core->_isStarted){
			// Framework already started
			return;
		}
		
		$profiler = new ProfilerEvent('Core Framework Initialization');
		$profiler->start();
		
		// Initialize the base framework so automatic file includes work at a minimum!
		$core->_init();
		
		// Try to load the base configuration  
		// If this succeeds we can safely set the framework into an "installed" state.
		try{
			$coreSettings = $core->_loadConfiguration();
		}
		catch(\Exception $ex){
			$coreSettings = null;
		}
		
		// Load those settings into the necessary defines.
		$core->_setupDefines($coreSettings);
		
		
		// If the request is a WEB request, check if we need to redirect the user now!
		// This is done as soon as possible to boost performance.
		// It doesn't make much sense to load all the components if all we're doing is redirecting the user!
		if(EXEC_MODE === 'WEB'){
			$core->_checkMandatoryRedirect();
		}
		
		// Register all of the default loggers and default error management for Core
		$core->_setupLoggersAndErrorChecking();
		
		
		if($coreSettings === null){
			// Core is not installed yet, (no configuration file).
			// nothing left to do here.
			$profiler->stop();
			return;
		}
		
		
		// Try to initialize the database next!
		try{
			DMI::GetSystemDMI();
		}
		catch (Exception $ex) {
			// Guess it's not installed yet; no database setup or otherwise connection error.
			$core->_isInstalled = false;
			$profiler->stop();
			return;
		}
		
		
		// DB is ready and working, next try the dynamic configuration retrieval.
		try{
			ConfigHandler::LoadFromDatabase();
		}
		catch (Exception $ex) {
			// Guess it's not installed yet; no database setup or otherwise connection error.
			$core->_isInstalled = false;
			$profiler->stop();
			return;
		}
		catch(Error $ex) {
			// An error was thrown, also not installed.
			$core->_isInstalled = false;
			$profiler->stop();
			return;
		}
		
		
		$core->_loadComponents();
		$core->_loadModelSupplementals();
		$core->_autoUpgradeComponents();
		
		
		Logger::SetSinkLevel('database', ConfigHandler::Get('/core/logs/db/level'));
		Logger::SetSinkLevel('file', ConfigHandler::Get('/core/logs/level'));
		Logger::SetSinkLevel('syslog', ConfigHandler::Get('/core/logs/syslog/level'));


		// The base framework and all components are ready, go ahead and poll for the licensing information for any component.
		$core->_loadLicenses();
		
		$core->_isStarted = true;
		$core->_isInstalled = true;
		
		// Dispatch the hook to notify the components that Core is ready!
		HookHandler::DispatchHook('/core/started');
		$profiler->stop();
	}
	
	/**
	 * Register a non-standard Controller into the system.
	 * 
	 * Primary use case is in the testing framework for controllers that are not real controllers.
	 * 
	 * @param string $class Class name to register
	 */
	public static function RegisterController(string $class) {
		$uri = strtolower( substr($class, 0, -10) );
		self::$_Instance->_controllers[ $uri ] = $class;
	}

	/**
	 * Register a component after the system has already loaded, for example during an update
	 *
	 * @param Component $component
	 *
	 * @throws Exception
	 */
	public static function RegisterComponent(Component $component) {
		self::$_Instance->_registerComponent($component);
	}
	
	// </editor-fold>


	
	/**
	 * Translate a dimension, (or dimensions), to a "preview size"
	 * of sm, med, lg or xl.
	 *
	 * @param string $dimensions Dimensions to translate
	 * @param [optional] int $width If second parameter is sent, assume width, height.
	 *
	 * @return string
	 */
	public static function TranslateDimensionToPreviewSize($dimensions) {
		// Load in the theme sizes for reference.
		$themesizes = array(
			'sm'  => ConfigHandler::Get('/theme/filestore/preview-size-sm'),
			'med' => ConfigHandler::Get('/theme/filestore/preview-size-med'),
			'lg'  => ConfigHandler::Get('/theme/filestore/preview-size-lg'),
			'xl'  => ConfigHandler::Get('/theme/filestore/preview-size-xl'),
		);

		if (sizeof(func_get_args()) == 2) {
			// Assume $width, $height.
			$width  = (int)func_get_arg(0);
			$height = (int)func_get_arg(1);
		}
		elseif (is_numeric($dimensions)) {
			// It's a straight single number, use that for both dimensions.
			$width  = $dimensions;
			$height = $dimensions;
		}
		elseif (stripos($dimensions, 'x') !== false) {
			// It's a string joining both dimensions.
			$ds     = explode('x', strtolower($dimensions));
			$width  = trim($ds[0]);
			$height = trim($ds[1]);
		}
		else {
			// Invalid size given.
			return null;
		}

		$smaller = min($width, $height);

		if ($smaller >= $themesizes['xl']) return 'xl';
		elseif ($smaller >= $themesizes['lg']) return 'lg';
		elseif ($smaller >= $themesizes['med']) return 'med';
		else return 'sm';
	}

	/**
	 * Get all registered permissions for all loaded components.
	 *
	 * @static
	 * @return array
	 */
	public static function GetPermissions(){
		return self::Singleton()->_permissions;
	}

	/**
	 * Get the component object by its name.
	 *
	 * @param string $name Name of the requested component
	 *
	 * @return Component
	 */
	public static function GetComponent($name = 'core') {
		$s = self::Singleton();
		if(isset($s->_components[$name])) return $s->_components[$name];

		// maybe it's a disabled component.  Those should be returned too.
		if(isset($s->_componentsDisabled[$name])) return $s->_componentsDisabled[$name];

		// Not that either?
		return null;
	}
	

	/**
	 * Lookup a component by a controller.
	 * Useful for figuring out what API version a given controller needs to be handled as.
	 *
	 * @param string $controller
	 *
	 * @return Component|null
	 */
	public static function GetComponentByController($controller) {
		$controller = strtolower($controller);

		$self = self::Singleton();
		foreach ($self->_components as $c) {
			$controllers = $c->getControllerList();
			if (isset($controllers[$controller])) return $c;
		}

		// No?
		return null;
	}
	
	

	/**
	 * Get the core singleton object
	 *
	 * @return Core
	 */
	public static function Singleton() {
		if(self::$_Instance === null){
			self::$_Instance = new self();
		}

		return self::$_Instance;
	}

	/**
	 * Check if a given class is available in the system as-is.
	 *
	 * @param string $classname
	 *
	 * @return boolean
	 */
	public static function IsClassAvailable($classname) {
		// let's see if I can't speed this function up a little...
		if(self::$instance == null){
			self::Singleton();
		}

		if(isset(self::$instance->_classes[$classname])){
			return true;
		}
		else{
			return false;
		}
	}

	public static function IsLibraryAvailable($name, $version = false, $operation = 'ge') {
		$ch   = self::Singleton();
		$name = strtolower($name);
		//var_dump($ch->_libraries[$name], version_compare(str_replace('~', '-', $ch->_libraries[$name]), $version, $operation));
		//if($name == 'DB') return true;
		//echo "Checking library name[$name] v[$version] op[$operation]<br>";
		if (!isset($ch->_libraries[$name])) {
			//echo "Library " . $name . " is not available!"; // DEBUG //
			return false;
		}
		// There's a bit of an issue with the debian-style versions... PHP considers 1.2.3~1 < 1.2.3...
		elseif ($version !== false) {
			//var_dump($ch->_libraries[$name], $operation, $version, version_compare($ch->_libraries[$name], $version, $operation));
			//var_dump(Core::VersionCompare($ch->_libraries[$name], $version, $operation));
			//return version_compare(str_replace('~', '-', $ch->_libraries[$name]), $version, $operation);

			// Core provides more accurate comparison for Debian-style versions.
			return Core::VersionCompare($ch->_libraries[$name], $version, $operation);
		}
		else return true;
	}

	public static function IsJSLibraryAvailable($name, $version = false, $operation = 'ge') {
		$ch   = self::Singleton();
		$name = strtolower($name);
		//if($name == 'DB') return true;
		//echo "Checking jslibrary name[$name] v[$version] op[$operation]<br>";
		if (!isset($ch->_scriptlibraries[$name])) return false;
		// There's a bit of an issue with the debian-style versions... PHP considers 1.2.3~1 < 1.2.3...
		elseif ($version) return version_compare(str_replace('~', '-', $ch->_scriptlibraries[$name]->version), $version, $operation);
		else return true;
	}

	/**
	 * Get the version for a loaded library or NULL if not loaded.
	 * 
	 * @param string $library
	 *
	 * @return string|null
	 */
	public static function GetLibraryVersion($library){
		$s = self::Singleton();
		return isset($s->_libraries[$library]) ? $s->_libraries[$library] : null;
	}

	/**
	 * @param string $library
	 *
	 * @return null|array
	 */
	public static function GetJSLibrary($library) {
		$library = strtolower($library);
		return self::Singleton()->_scriptlibraries[$library];
	}

	/**
	 * Get a flat list of javascript libraries currently available.
	 *
	 * @return array
	 */
	public static function GetJSLibraries() {
		return self::Singleton()->_scriptlibraries;
	}

	/**
	 * Get all the classes that are currently available and loaded.
	 *
	 * @return array
	 */
	public static function GetClasses(){
		return self::Singleton()->_classes;
	}

	/**
	 * Load a Script library from the available resources,
	 * will call the underlying handler which should take care of all dependencies, styles,
	 * and related files necessary for the requested library to operate.
	 *
	 * If the library could not be located, an Exception is thrown with possible fixes.
	 *
	 * @param string $library
	 *
	 * @return boolean
	 *
	 * @throws Exception
	 */
	public static function LoadScriptLibrary(string $library) : bool {
		$library = strtolower($library);
		$obj     = self::Singleton();

		if (isset($obj->_scriptlibraries[$library])) {
			return call_user_func($obj->_scriptlibraries[$library]);
		}

		// Otherwise, it wasn't found, try to provide some useful hinting to the user
		$libs = array_keys(self::GetJSLibraries());

		$possibilities = [];
		foreach($libs as $l) {
			$perc = 0;
			similar_text($library, $l, $perc);
			if($perc > 50) {
				$possibilities[] = $l;
			}
		}

		if(count($possibilities) === 1) {
			$msg = 'Did you mean ' . $possibilities[0] . '?';
		}
		elseif(count($possibilities) > 1) {
			$msg = 'Did you mean one of the following: ' . implode(', ', $possibilities) . '?';
		}
		else {
			$msg = 'Library ' . $library . ' not available in the current system!';
		}

		throw new Exception($msg);
	}


	public static function IsComponentAvailable($name, $version = false, $operation = 'ge') {
		$self = self::Singleton();

		$name = strtolower($name);

		//echo "Checking component name[$name] v[$version] op[$operation]<br>";
		if (!isset($self->_components[$name])){
			return false;
		}

		// Only included enabled components.
		elseif (!$self->_components[$name]->isEnabled()){
			return false;
		}

		elseif ($version){
			return Core::VersionCompare($self->_components[$name]->getVersionInstalled(), $version, $operation);
		}

		else{
			return true;
		}
	}

	/**
	 * Utility check to see if a requested component is ready to use.
	 *
	 * This is different from isAvailable, as it will query the component to see if it's configured and ready for use!
	 *
	 * If it's not ready, either an error message or URL is to be returned.
	 * Otherwise, TRUE is returned to indicate that it's ready.
	 *
	 * @param $name
	 *
	 * @return string|true
	 */
	public static function IsComponentReady($name){
		if(!self::IsComponentAvailable($name)){
			return 'Component ' . $name . ' is not available!';
		}

		$self = self::Singleton();

		$name = strtolower($name);
		/** @var Component $c */
		$c = $self->_components[$name];

		$attr = $c->getRootDOM()->getAttribute('isready');

		if($attr === null || $attr === ''){
			// No check installed, OK!
			return true;
		}

		// It should be a method, call that method and return the result.
		return call_user_func($attr);
	}


	/**
	 * Check and see if the framework is installed
	 * 
	 * @return bool
	 */
	public static function IsInstalled() : bool {
		return Core::Singleton()->_isInstalled;
	}

	public static function NeedsUpdated() {
		return Core::Singleton()->_needsUpdated();
	}

	public static function GetVersion() {
		return Core::GetComponent()->getVersionInstalled();
	}

	/**
	 * Resolve an asset to a fully-resolved URL.
	 *
	 * @deprecated
	 * @see \Core\resolve_asset
	 *
	 * @param string $asset
	 *
	 * @return string The full url of the asset, including the http://...
	 */
	public static function ResolveAsset($asset) {
		trigger_error('Core::ResolveAsset is deprecated, please use \\Core\\resolve_asset() instead.', E_USER_DEPRECATED);
		return \Core\resolve_asset($asset);
	}

	/**
	 * Resolve a url or application path to a fully-resolved URL.
	 *
	 * This can also be an already-resolved link.  If so, no action is taken
	 * and the original URL is returned unchanged.
	 *
	 * @deprecated
	 * @see \Core\resolve_link
	 *
	 * @param string $url
	 *
	 * @return string The full url of the link, including the http://...
	 */
	public static function ResolveLink($url) {
		trigger_error('Core::ResolveLink is deprecated, please use \\Core\\resolve_link() instead.', E_USER_DEPRECATED);
		return \Core\resolve_link($url);
	}

	/**
	 * Redirect the user to another page via sending the Location header.
	 *    Prevents any POST data from being reloaded.
	 *
	 * @deprecated 2013.06.11 Please use the namespaced versions.
	 *
	 * @param  string $page The page URL to redirect to
	 * @param  int    $code  The HTTP status code to send to the browser, MUST be 301 or 302.
	 *
	 * @throws \Exception
	 *
	 * @return bool|null False on failure, success will halt the script.
	 */
	static public function Redirect($page, $code = 302) {
		trigger_error('Core::Redirect is deprecated, please use \\Core\\redirect() instead.', E_USER_DEPRECATED);
		\Core\redirect($page, $code);
	}

	/**
	 * @deprecated 2013.06.11 Please use the namespaced versions.
	 */
	static public function Reload() {
		trigger_error('Core::Reload is deprecated, please use \\Core\\reload() instead.', E_USER_DEPRECATED);
		\Core\reload();
	}

	/**
	 * Add a message to the user's stack.
	 *    It will be displayed the next time the user (or session) renders the page.
	 *
	 * @param string $messageText The message to send to the user
	 * @param string $messageType The type of message, "success", "info", or "error"
	 *
	 * @return void
	 */
	static public function SetMessage($messageText, $messageType = 'info') {
		if(trim($messageText) == '') return;

		$messageType = strtolower($messageType);

		// CLI doesn't use sessions.
		if(EXEC_MODE == 'CLI'){
			$messageText = preg_replace('/<br[^>]*>/i', "\n", $messageText);
			echo "[" . $messageType . "] - " . $messageText . "\n";
		}
		else{
			$stack = Session::Get('message_stack', []);

			$stack[] = array(
				'mtext' => $messageText,
				'mtype' => $messageType,
			);
			Session::Set('message_stack', $stack);
		}
	}

	/**
	 * Retrieve the messages and optionally clear the message stack.
	 *
	 * @param bool $returnSorted
	 * @param bool $clearStack
	 *
	 * @return array
	 */
	static public function GetMessages($returnSorted = false, $clearStack = true) {
		$stack = Session::Get('message_stack', []);

		if($returnSorted){
			usort($stack, function($a, $b){
				return $a['mtype'] <=> $b['mtype'];
			});
		}

		if($clearStack){
			Session::UnsetKey('message_stack');
		}
		return $stack;
	}

	static public function SortByKey($named_recs, $order_by, $rev = false, $flags = 0) {
		// Create 1-dimensional named array with just
		// sortfield (in stead of record) values
		$named_hash = array();
		foreach ($named_recs as $key=> $fields) $named_hash["$key"] = $fields[$order_by];

		// Order 1-dimensional array,
		// maintaining key-value relations
		if ($rev) arsort($named_hash, $flags);
		else asort($named_hash, $flags);

		// Create copy of named records array
		// in order of sortarray
		$sorted_records = array();
		foreach ($named_hash as $key=> $val) $sorted_records["$key"] = $named_recs[$key];

		return $sorted_records;
	}


	/**
	 * Return a string of the keys of the given array glued together.
	 *
	 * @param $glue string
	 * @param $array array
	 *
	 * @return string
	 *
	 * @version 2008.06.05
	 * @author Charlie Powell <charlie@evalagency.com>
	 */
	static public function ImplodeKey($glue, &$array) {
		$arrayKeys = array();
		foreach ($array as $key => $value) {
			$arrayKeys[] = $key;
		}
		return implode($glue, $arrayKeys);
	}


	/**
	 * Generate a random hex-deciman value of a given length.
	 *
	 * @param int     $length
	 * @param boolean $casesensitive [false] Set to true to return a case-sensitive string.
	 *                              Otherwise the resulting string will simply be all uppercase.
	 *
	 * @return string
	 */
	static public function RandomHex($length = 1, $casesensitive = false) {
		return \Core\random_hex($length, $casesensitive);
	}


	/**
	 * Utility function to translate a filesize in bytes into a human-readable version.
	 *
	 * @deprecated 2013.05.31
	 *
	 * @param int $filesize Filesize in bytes
	 * @param int $round Precision to round to
	 *
	 * @return string
	 */
	public static function FormatSize($filesize, $round = 2) {
		return \Core\Filestore\format_size($filesize, $round);
	}

	public static function GetExtensionFromString($str) {
		// File doesn't have any extension... easy enough!
		if (strpos($str, '.') === false) return '';

		return substr($str, strrpos($str, '.') + 1);
	}

	/**
	 * @deprecated 2013.05.31
	 * @return float
	 */
	public static function GetProfileTimeTotal() {
		error_log(__FUNCTION__ . ' is deprecated, please use \Core\Utilities\Profiler\Profiler::GetDefaultProfiler()->getTime() instead', E_USER_DEPRECATED);
		return \Core\Utilities\Profiler\Profiler::GetDefaultProfiler()->getTime();
	}

	/**
	 * Validate an email address.
	 * Provide email address (raw input)
	 * Returns true if the email address has the email
	 * address format and the domain exists.
	 *
	 * Copied (almost) verbatim from http://www.linuxjournal.com/article/9585?page=0,3
	 * @author Douglas Lovell @ Linux Journal
	 *
	 * @param string $email The email to validate
	 * @return boolean|string
	 */
	public static function CheckEmailValidity($email) {
		return \Core\CheckEmailValidity($email);
	}

	/**
	 * Simple function to check if a number is an int and greater than 0.
	 * This is useful as a default validation option for model properties.
	 *
	 * @param $val
	 * @return bool
	 */
	public static function CheckIntGT0Validity($val){
		if(!(is_int($val) || ctype_digit($val))){
			return false;
		}
		if($val <= 0){
			return false;
		}

		return true;
	}
	
	/**
	 * Retrieve the dynamic legal footer based on licensed options.
	 * 
	 * @return string
	 */
	public static function GetLegalFooterContent() : string {
		
		$lic = self::GetLicensedFeature('/core/license');
		$licurl = self::GetLicensedFeature('/core/license_url');
		$licto = self::GetLicensedFeature('/core/licensed_to');
		$hide = self::GetLicensedFeature('/core/hide_legal_notice');

		if(!$lic) {
			$lic = 'AGPLv3';
			$licurl = 'https://www.gnu.org/licenses/agpl';
			$hide = 0;
		}
		
		if($hide == '1'){
			return '';
		}
		
		$licclass = strtolower($lic) . '-tag';
		
		if($licto){
			$lictext = t('STRING_LICENSED_TO_S', $licto);
			if($lic && $licurl){
				$lictext .= ' (<a href="' . $licurl . '" target="_blank" class="' . $licclass . '" rel="noopener">' . $lic . '</a>)';
			}
		}
		elseif($lic && $licurl){
			$lictext = t('STRING_LICENSED_UNDER');
			$lictext .= ' <a href="' . $licurl . '" target="_blank" class="' . $licclass . '" rel="noopener">' . $lic . '</a>';
		}
		elseif($lic){
			$lictext = t('STRING_LICENSED_UNDER') . ' ' . $lic;
		}
		
		$poweredBy = t('STRING_POWERED_BY') . ' <a href="https://corepl.us" target="_blank" rel="noopener">Secure PHP Framework and CMS, Core Plus</a>';
		
		return '<p class="legal-notice">' . $lictext . '&nbsp;&nbsp;' . $poweredBy . '</p>';
	}


	/**
	 * Clone of the php version_compare function, with the exception that it treats
	 * version numbers the same that Debian treats them.
	 *
	 * @param string $version1 Version to compare
	 * @param string $version2 Version to compare against
	 * @param string $operation Operation to use or null
	 *
	 * @return bool | int Boolean if $operation is provided, int if omited.
	 */
	public static function VersionCompare($version1, $version2, $operation = null) {
		return \Core\version_compare($version1, $version2, $operation);
	}

	/**
	 * Break a version string into the corresponding parts.
	 *
	 * Major Version
	 * Minor Version
	 * Point Release
	 * Core Version
	 * Developer-Specific Version
	 * Development Status
	 *
	 * Optimized 2013.08.17
	 *
	 * @param string $version
	 *
	 * @return \Core\VersionString
	 */
	public static function VersionSplit($version) {
		return new \Core\VersionString($version);
	}

	/**
	 * Simple method to compare two values with each other in a more restrictive manner than == but not quite fully typecasted.
	 *
	 * This is useful for the scenarios that involve needing to check that "3" == 3, but "" != 0.
	 *
	 * @param $val1
	 * @param $val2
	 *
	 * @deprecated 2013.09 Please use the namespaced function instead.
	 *
	 * @return boolean
	 */
	public static function CompareValues($val1, $val2){
		return \Core\compare_values($val1, $val2);
	}

	/**
	 * Compare two values as strings explictly.
	 * This is useful for numbers that need to behave like strings, ie: postal codes with their leading zeros.
	 *
	 * @param $val1
	 * @param $val2
	 *
	 * @deprecated 2013.09 Please use the namespaced function instead.
	 *
	 * @return boolean
	 */
	public static function CompareStrings($val1, $val2) {
		return \Core\compare_strings($val1, $val2);
	}

	/**
	 * Generate a globally unique identifier that can be used as a replacement for an autoinc or similar.
	 *
	 * This method IS compatible with multiple servers on a single codebase!
	 *
	 * An example of a UUID returned by this function would be: "1-c5dbcaaf9db-8d77"
	 *
	 * @since 2.4.2
	 *
	 * @return string
	 */
	public static function GenerateUUID(){
		return \Core\generate_uuid();
	}
	
	public static function GetSupplementalModels($modelname){
		$ret              = [];
		$supplementalName = strtolower($modelname . 'supplemental');
		$supplementalLen  = strlen($supplementalName);
		foreach(Core::GetComponents() as $c){
			/** @var Component $c */
			$list = $c->getSupplementalModelList();
			foreach($list as $key => $location){
				// Disregard $location, I just need the keys.
				if(strpos($key, $supplementalName) == strlen($key) - $supplementalLen){
					// A match was found!
					// This means that an external component is extending this Model.
					$ret[] = $key;
				}
			}
		}
		
		return $ret;
	}

	/**
	 * Get the requested licensed feature code from the licenser, or false if invalid/not set.
	 * 
	 * @param string $featureCode
	 *
	 * @return bool|string
	 */
	public static function GetLicensedFeature($featureCode) {
		$s = self::Singleton();
		
		return isset($s->_features[$featureCode]) ? $s->_features[$featureCode] : false;
	}

	/**
	 * Get a dump of all licensed features on the site, along with their current value and expiration.
	 * 
	 * Useful for the system health page and developer debug output.
	 * 
	 * @return array
	 */
	public static function GetLicensedDump() {
		$components = self::GetComponents();
		$licenses = [];
		
		foreach($components as $c){
			/** @var Component $c */
			$dat = $c->getLicenseData();
			if(sizeof($dat)){
				$licenses[] = $dat;
			}
		}
		
		return $licenses;
	}
}



