<?php
/**
 * @author Charlie Powell <charlie@evalagency.com>
 * @since 2017.09.28
 * @package Core\Forms
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Core\DictItem;
use ArrayAccess;
use Countable;
use Iterator;

/**
 * This addresses the issue of PHP arrays not acting like real hash tables when I need them to.
 * 
 * Example for a set of form options:
 * $options = [
 *   0 => 'First Option',
 *   1 => 'Second Option',
 *   2 => 'Third Option',
 * ]
 * 
 * The system is unable to know if the developer meant the option values to be 0 or "First Option".
 * 
 * With this object, assigning the options from a source array of ['First Option', 'Second Option', 'Third Option']
 * via the standard code of:
 * ```
 * $options = new Dict();
 * foreach($source as $item){
 *   $options[] = $item;
 * }
 * ```
 * 
 * will result in the dict containing
 * ```
 * [
 * { key: 'First Option', val:  'First Option'},
 * { key: 'Second Option', val: 'Second Option'},
 * { key: 'Third Option', val:  'Third Option'},
 * ]
 * ```
 * 
 * AND supports explicit numeric key setting: [0 => 'Option Zero', 1 => 'Option One', 2 => 'Option Two']
 * ```
 * $options = new Dict();
 * foreach($source as $item => $title){
 *   $options[$item] = $title;
 * }
 * ```
 * 
 * will result in the dict containing
 * ```
 * [
 * { key: 0, val: 'Option Zero'},
 * { key: 1, val: 'Option One'},
 * { key: 2, val: 'Option Two'},
 * ]
 * ```

 *
 * @author charlie
 */
class Dict implements Iterator, ArrayAccess, Countable {
	
	/**
	 * Set this to true to auto-convert all applicable values returned to their translated version.
	 * 
	 * This only applies to items that are scalar and start with "t:".
	 * 
	 * @var bool
	 */
	public $autoI18N = false;
	
	/**
	 * Internal storage of items
	 * 
	 * @var DictItem[]
	 */
	private $_items = [];
	
	/**
	 * Internal tracker of item key to its index in the items
	 * 
	 * @var string[]
	 */
	private $_itemMaps = [];
	
	/**
	 * Positional tracker of the array for use in the iterator.
	 * 
	 * @var int
	 */
	private $_pos = 0;
	
	/**
	 * Helping tracker for sizeof/count.
	 * 
	 * @var int
	 */
	private $_size = 0;
	
	/**
	 * Add an item to this Dictionary by its key and value.
	 * 
	 * @param string|int $key The key to add to this Dict, must be scalar.
	 * @param mixed $value    The value for the key, can be whatever PHP can save!
	 */
	public function addItem($key, $value) {
		// Before an item is blindly added, search through the array to see if it exists already.
		$idx = isset($this->_itemMaps[ $key ]) ? $this->_itemMaps[ $key ] : null;
		
		if($idx === null){
			$this->_items[] = new DictItem($key, $value);
			end($this->_items);
			$this->_itemMaps[ $key ] = key($this->_items);
			++ $this->_size;
		}
		else{
			$this->_items[ $idx ]->setValue($value);
		}
	}
	
	/**
	 * Add multiple items to this Dictionary, à la array_merge().
	 * 
	 * @param array $items List of items to add (append)
	 */
	public function addItems(array $items) {
		foreach($items as $key => $value){
			$this->addItem($key, $value);
		}
	}

	/**
	 * Force a reindex of this Dictionary, useful if called as an array.
	 */
	public function reindex() {

		// This will reset the index of any items, gaps can appear if unset is called.
		$this->_items = array_values($this->_items);
		$this->_itemMaps = [];
		$this->_size = count($this->_items);
		$c = 0;

		foreach($this->_items as $idx => $i) {
			$this->_itemMaps[ $i->key ] = $c++;
		}
	}
	
	/**
	 * Check if this Dictionary has a given key
	 * 
	 * @param string|int $key
	 * @return bool
	 */
	public function hasKey($key) : bool {
		return isset($this->_itemMaps[ $key ]);
	}

	/**
	 * Searches the Dict for a given value and returns the first corresponding key if successful
	 *
	 * If the value is not located, then NULL is returned.
	 * This is important to note because if the first key contains the value and that key is
	 * 0-indexed, then (int)0 will be returned.
	 *
	 * Will search both raw values AND translated values!
	 * 
	 * @see https://www.php.net/manual/en/function.array-search.php
	 * @param mixed $needle
	 * @return mixed|null
	 */
	public function search($needle) {
		foreach($this->_items as $item) {
			if($item->value === $needle || $item->getTranslated() === $needle) {
				return $item->key;
			}
		}

		return null;
	}
	
	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Return the current element
	 * @link http://php.net/manual/en/iterator.current.php
	 * @return mixed Can return any type.
	 */
	public function current() {
		if(isset($this->_items[$this->_pos])){
			return $this->autoI18N ? 
				$this->_items[$this->_pos]->getTranslated() : 
				$this->_items[$this->_pos]->value;
		}
		else{
			return null;
		}
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Move forward to next element
	 * @link http://php.net/manual/en/iterator.next.php
	 * @return void Any returned value is ignored.
	 */
	public function next() {
		++$this->_pos;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Return the key of the current element
	 * @link http://php.net/manual/en/iterator.key.php
	 * @return mixed scalar on success, or null on failure.
	 */
	public function key() {
		return isset($this->_items[$this->_pos]) ? $this->_items[$this->_pos]->key : null;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Checks if current position is valid
	 * @link http://php.net/manual/en/iterator.valid.php
	 * @return boolean The return value will be casted to boolean and then evaluated.
	 * Returns true on success or false on failure.
	 */
	public function valid() {
		return isset($this->_items[$this->_pos]);
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Rewind the Iterator to the first element
	 * @link http://php.net/manual/en/iterator.rewind.php
	 * @return void Any returned value is ignored.
	 */
	public function rewind() {
		$this->_pos = 0;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Whether a offset exists
	 * @link http://php.net/manual/en/arrayaccess.offsetexists.php
	 *
	 * @param mixed $offset <p>
	 * An offset to check for.
	 * </p>
	 *
	 * @return boolean true on success or false on failure.
	 * </p>
	 * <p>
	 * The return value will be casted to boolean if non-boolean was returned.
	 */
	public function offsetExists($offset) {
		foreach($this->_items as $i){
			if($i->key === $offset){
				return true;
			}
		}
		return false;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to retrieve
	 * @link http://php.net/manual/en/arrayaccess.offsetget.php
	 * @param mixed $offset <p>
	 * The offset to retrieve.
	 * </p>
	 * @return mixed Can return all value types.
	 */
	public function offsetGet($offset) {
		foreach($this->_items as $i){
			if($i->key === $offset){
				return $this->autoI18N ? 
					$i->getTranslated() : 
					$i->value;
			}
		}
		return null;
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to set
	 * @link http://php.net/manual/en/arrayaccess.offsetset.php
	 * @param mixed $offset <p>
	 * The offset to assign the value to.
	 * </p>
	 * @param mixed $value <p>
	 * The value to set.
	 * </p>
	 * @return void
	 *
	 * @throws Exception
	 */
	public function offsetSet($offset, $value) {
		if($offset === null){
			// If this happens, just map the user's value to the option key.
			$offset = $value;
		}
		
		// Add item will handle searching for the offset and adding or setting it as necessary.
		$this->addItem($offset, $value);
	}

	/**
	 * (PHP 5 &gt;= 5.0.0)<br/>
	 * Offset to unset
	 * @link http://php.net/manual/en/arrayaccess.offsetunset.php
	 * @param mixed $offset <p>
	 * The offset to unset.
	 * </p>
	 * @return void
	 */
	public function offsetUnset($offset) {
		foreach($this->_items as $idx => $i){
			if($i->key === $offset){
				unset($this->_items[$idx]);
				unset($this->_itemMaps[ $i->key ]);
				--$this->_size;
				return;
			}
		}
	}
	
	/**
	 * Get the number of items in this dictionary
	 * 
	 * @return int
	 */
	public function count() : int {
		return $this->_size;
	}
	
	public function __toString() : string {
		$hash = [];
		
		foreach($this->_items as $item){
			$hash[ $item->key ] = $this->autoI18N ? $item->getTranslated() : $item->value;
		}
		
		return json_encode($hash);
	}
	
	/**
	 * Create a new Dict from an associative array.
	 * 
	 * @param array $array
	 * 
	 * @return \Core\Dict
	 */
	public static function FromAssoc($array): Dict{
		$dict = new Dict();
		foreach($array as $k => $v){
			$dict->addItem($k, $v);
		}
		return $dict;
	}
	
	/**
	 * Create a new Dict from a flat array
	 * 
	 * @param array $array
	 * 
	 * @return \Core\Dict
	 */
	public static function FromList($array): Dict{
		$dict = new Dict();
		foreach($array as $v){
			$dict->addItem($v, $v);
		}
		return $dict;
	}
}
