<?php


namespace Core;


use Core\Filestore\Contents\ContentTARGZ;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Core\GPG\GPG;
use Core\GPG\Signature;
use Core\Utilities\Changelog\Section;
use Exception;

class Package {

	/** @var File */
	private $_file;
	/** @var File */
	private $_extractedFile;
	/** @var ContentTARGZ */
	private $_fileContents;
	/** @var Signature */
	private $_signature;
	/** @var string[] */
	private $_fileList;
	/** @var Component */
	private $_component;
	/** @var Section */
	private $_changelog;

	/**
	 * @throws Exception
	 */
	public function getSignature() : Signature {
		if(!$this->_signature) {
			if(!$this->_file) {
				throw new Exception('No file set, unable to verify signature');
			}

			if(!$this->_file->exists()) {
				throw new Exception($this->_file->getFilename() . ' does not exist, unable to verify signature');
			}
			$gpg = new GPG();
			$this->_signature = $gpg->verifyFileSignature($this->_file->getFilename());
		}

		return $this->_signature;
	}

	/**
	 * @throws Exception
	 */
	public function getExtractedPackage() : File {
		if(!$this->_extractedFile) {
			$tmp = FileFactory::Directory('tmp/package/');
			$tmp->mkdir();

			$dst = $this->_file->getBaseFilename(true);

			$gpg = new GPG();
			$this->_extractedFile = $gpg->decryptFile($this->_file->getFilename(), $tmp->getPath() . $dst);
		}

		return $this->_extractedFile;
	}

	/**
	 * @throws Exception
	 */
	public function getContentsObject() : ContentTARGZ {
		if(!$this->_fileContents) {
			$this->_fileContents = $this->getExtractedPackage()->getContentsObject();
		}

		return $this->_fileContents;
	}

	/**
	 * @throws Exception
	 */
	public function getFileList() : array {
		if(!$this->_fileList) {
			$this->_fileList = $this->getContentsObject()->listfiles();
		}

		return $this->_fileList;
	}

	/**
	 * @throws Exception
	 */
	public function getComponent() : Component {
		if($this->_component === null) {
			$list = $this->getFileList();
			$file = null;
			$type = null;

			foreach($list as $l) {
				if(strpos($l, 'component.xml')) {
					$file = $l;
					$type = 'component';
					break;
				}

				if(strpos($l, 'theme.xml')) {
					$file = $l;
					$type = 'theme';
					break;
				}
			}

			switch($type) {
				case 'component':
					$this->_component = new Component();
					break;
				case 'theme':
					$this->_component = new Theme();
					break;
				default:
					throw new Exception('Unsupported package type, unable to determine if Component or Theme.');
			}

			// Extract the package XML for all the necessary metadata.
			$xml = $this->getContentsObject()->extractSingleFile($file);

			$this->_component->initFromString($xml);
		}

		return $this->_component;
	}

	/**
	 * Get applicable source upgrades for this package
	 *
	 * @return array
	 */
	public function getUpgrades() : array {
		$component = $this->getComponent();

		$upgrades = [];
		$versions = [ $component->getVersion() ];
		foreach($component->getXMLLoader()->getElements('upgrades/upgrade') as $element) {
			$upgrades[] = [
				$element->getAttribute('from'),
				$element->getAttribute('to')
			];
		}

		// Upgrades are in standard order, so reverse it as we're working our way from the current back.
		$upgrades = array_reverse($upgrades);

		foreach($upgrades as $upgrade) {
			$from = $upgrade[0];
			$to = $upgrade[1];

			if(in_array($to, $versions)) {
				$versions[] = $from;
			}
		}

		return $versions;
	}

	/**
	 * Retrieve the requirements from the XML/Component.
	 *
	 * They will be Objects, so flatten them to standard arrays.
	 *
	 * @return array
	 */
	public function getRequires() : array {
		$ret = [];

		$component = $this->getComponent();
		foreach($component->getRequires() as $requirement) {
			$ret[] = [
				'type' => $requirement->getType(),
				'name' => $requirement->getName(),
				'version' => $requirement->getVersion(),
				'operation' => $requirement->getOperation(),
			];
		}

		return $ret;
	}

	/**
	 * Get the Changelog section for this package,
	 * contains the changes and release information.
	 *
	 * @throws Exception
	 *
	 * @return Section
	 */
	public function getChangelog() : Section {
		if(!$this->_changelog) {
			$component = $this->getComponent();
			$list = $this->getFileList();
			$file = null;
			$depth = 99;

			foreach($list as $l) {
				if(strpos($l, 'CHANGELOG')) {
					// This is weird because a package may contain multiple changelog files.
					// we want the top-most file, which should refer to the package as a whole.
					$c = substr_count($l, '/');
					if($c < $depth) {
						$file = $l;
						$depth = $c;
					}
				}
			}

			if(!$file) {
				throw new Exception('No CHANGELOG provided in package');
			}

			$parser = new Utilities\Changelog\Parser($component->getName());
			$parser->parseString($this->getContentsObject()->extractSingleFile($file));

			$this->_changelog = $parser->getSection($component->getVersion());
		}

		return $this->_changelog;
	}

	/**
	 * Get the logo filename, derived from the XML, but resolved to the packaged path
	 *
	 * @return string
	 */
	public function getLogo() : string {
		$logo = $this->getComponent()->getLogo();

		if(!$logo) {
			return '';
		}

		$files = $this->getFileList();
		foreach($files as $file) {
			if(strpos($file, $logo) !== false) {
				return $file;
			}
		}

		return '';
	}


	/**
	 * @throws Exception
	 */
	public function getType() : string {
		$list = $this->getFileList();

		foreach($list as $file) {

			if($file === 'core/component.xml') {
				// Core base package registers and is treated differently
				return 'core';
			}

			if(strpos($file, 'component.xml')) {
				// standard components
				return 'component';
			}

			if(strpos($file, 'theme.xml')) {
				// standard themes
				return 'theme';
			}
		}

		throw new Exception('Unsupported or unrecognized package type');
	}

	public function __destruct() {
		if($this->_extractedFile) {
			$this->_extractedFile->delete();
		}
	}

	// getVersion
	// getName
	// getSignatureKey
	// getPackager
	// getDescription
	// getRequires
	// getProvides
	// getUpgrades
	// getChangelog
	// getLogo
	// getScreenshots
	// getFeatures
	// getLicenserDetails

	public static function FromLocalFile(File $file) : self {
		$package = new self();
		$package->_file = $file;

		return $package;
	}
}