<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core;

use Exception;
use PageModel;
use ReflectionClass;
use ReflectionMethod;
use ReflectionParameter;
use Core\Logger\LogEntry;
use Core\Utilities\Profiler\ProfilerEvent;

/**
 * Core router for Core that handles accepting incoming requests
 * and mapping to all the necessary subsystems.
 *
 * @author charlie
 */
class Router {
	
	/**
	 * @var BaseURLMap The map that connects incoming URLs with the necessary components
	 */
	private $_baseurl;
	
	/**
	 * @var Request The processed Request object based from setURI
	 */
	private $_request;
	
	/**
	 * @var View View object that gets populated with the necessary front-end code from the Controller
	 */
	private $_view;
	
	/**
	 * @var PageModel Associated PageModel that corresponds to this request, pulled from the Request
	 */
	private $_model;
	
	/**
	 * @var string The keyname of the controller; acceptable for use in template mapping
	 */
	private $_controllerName;
	
	/**
	 * @var string The keyname of the method; acceptable for use in template mapping
	 */
	private $_methodName;
	
	/**
	 * @var View Default View to use as the target for {script} and {css} includes from sub templates.
	 */
	public static $DefaultView;
	
	/**
	 * @var Request Default Request that was used for the last router instance
	 */
	public static $DefaultRequest;
	
	/**
	 * @var BaseURLMap Default BaseURL for the incoming request
	 */
	public static $DefaultBaseURL;
	
	/**
	 * @var PageModel|null The page model behind this page, or NULL if not defined
	 */
	public static $DefaultPageModel;
	
	/**
	 * @var Router Default Core Router as set from the index and CLI.
	 */
	public static $DefaultRouter;

	public function __construct(){
		// Only one router can be used at a time!
		self::$DefaultRouter = $this;
	}
	
	/**
	 * Call the controller that is bound to this request and any pre-exec hooks that are defined.
	 */
	public function execute() {
		try{
			
			// Check if this page has a rewrite URL and tell the browser to jump there if necessary.
			if($this->_baseurl->getRewriteURL() !== $this->_request->uriBase){
				throw new PermanentFollowRedirectException($this->_baseurl->getRewriteURL());
			}
			
			// Anything that needs to fire off *before* the page is rendered.
			// This includes widgets, script addons, and anything else that needs the routing system to be available
			HookHandler::DispatchHook('/core/page/preexecute');
			
			// The primary application's caller logic!
			// This sets up and calls the custom controller for the application based on the bound Request.
			$profiler = new ProfilerEvent('Core Application Controller Load');
			$profiler->start();
			$this->_loadController();
			$profiler->stop();
		}
		catch (RedirectException $ex ){
			// Redirects should populate the Location field on the View and will not use any master templates.
			$this->_view->addHeader('Location', $ex->location);
			$this->_view->status = $ex->getCode();
		}
		catch( ViewStatusExceptionInterface $ex ){
			// Make a note of this Exception in case the developer wants to know.
			LogEntry::Init()
				->exception($ex)
				->level(LOG_LEVEL_INFO)
				->icon('globe')
				->send();
			
			// All other view-friendly exceptions trigger the View's status code and carry on.
			$this->_handleViewDataForExceptions($ex);
			
			HookHandler::DispatchHook('/core/page/error-' . $ex->getCode(), $this->_view);
		}
		catch (Exception $ex) {
			LogEntry::Init()
				->exception($ex)
				->level(LOG_LEVEL_HIGH)
				->icon('globe')
				->send();

			// Let this logic handle displaying a pretty error to the user,
			// it'll strip out sensitive information when necessary.
			$this->_handleViewDataForExceptions($ex);
		}

		try {
			// And allow scripts to tap into the routing system after the controller executes.
			// Just in case that's something they need.
			Hookhandler::DispatchHook('/core/page/postexecute');
		}
		catch(Exception $ex) {
			LogEntry::Init()
				->exception($ex)
				->send();
		}

		
		$this->_postProcessView();
	}
	
	/**
	 * Set the URI for this router instance, will create the Request object based off the URI.
	 *
	 * This is only meant to be called from the bootstrap.
	 * 
	 * @param string $uri
	 */
	public function setURI(string $uri = null) {
		$profiler = new ProfilerEvent('Core Router Load URL and Request');
		$profiler->start();
		
		if($uri === null && EXEC_MODE === 'CLI') {
			// CLI requests pull the URL from the global arguments passed in.
			// Also allow arguments to be passed in with --key[= ]value syntax common with CLI scripts.
			global $argv;
			$uri = CLI\CLI::ArgumentsToURI($argv);
		}
		elseif($uri === null) {
			// NULL requests should default to the current request.
			$uri = CUR_CALL;
		}
		elseif($uri === '') {
			// Empty requests should resolve to the root site however!
			$uri = '/';
		}
		else {
			// Sanitize the input via PHP's filter function for sanity.
			$uri = filter_var($uri, FILTER_SANITIZE_URL);
		}

		if(ROOT_WDIR !== null && strpos($uri, ROOT_WDIR) === 0){
			// If the uri includes the root path of the site, (which it probably does),
			// Trim that off as we already know where we're at.
			$uri = '/' . substr($uri, strlen(ROOT_WDIR));
		}
		
		
		// Now that the request is sanitized and filtered against the Core root,
		// create the request object to use
		$this->_request = new Request($uri);
		self::$DefaultRequest = $this->_request;
		
		
		// The BaseURL object can be built off the request now that the additional data
		// such as query strings and extensions have been stripped off.
		$this->_baseurl = new BaseURLMap($this->_request->uriBase);
		self::$DefaultBaseURL = $this->_baseurl;

		// The model is pulled from the URL Map, but it may be NULL!
		$this->_model = $this->_baseurl->getPageModel();
		self::$DefaultPageModel = $this->_model;
		
		// Positional arguments cannot be detected from the URL alone, but the
		// BaseURLMap knows about Controllers and their Methods in Core, so they're able
		// to be derived from that class.
		foreach($this->_baseurl->getPositionalParameters() as $k => $v){
			$this->_request->setParameter($k, $v);
		}
		
		
		// Populate the View object or update an already-set View
		if($this->_view === null){
			$this->_view = new View();
			self::$DefaultView = $this->_view;
		}
		
		
		// Set the defaults for Page-type Views, these can be overridden in the Controller if necessary.
		$this->_view->mastertemplate = 'default';
		$this->_view->templatename   = $this->_baseurl->getDefaultTemplateName();
		$this->_view->contenttype    = EXEC_MODE === 'WEB' ? View::CTYPE_HTML : View::CTYPE_PLAIN;
		$this->_view->status         = $this->_baseurl->getStatus();
		$this->_view->meta->setMeta('canonical', $this->_baseurl->getRewriteURL());
		$this->_view->meta->setMeta('charset', 'UTF-8');
		$this->_view->meta->setMeta('viewport', null);
		$this->_view->meta->setMeta('generator', null);

		if($this->_model !== null) {
			// Assign all the default PageModel data onto the view
			// this includes the title, metas, and breadcrumbs.
			// This only applies for linked pages, eg: content, gallery, blogs, etc.

			if($this->_model->admin) {
				// Admin pages get remapped to use the admin skin by default.
				$this->_view->mastertemplate = 'admin';
			}
			elseif($this->_model->theme_template !== '') {
				$this->_view->mastertemplate = 'skins/' . $this->_model->theme_template;
			}

			if($this->_model->page_template !== '') {
				// blank indicates default, so anything other is an assigned page template.
				// in this case, it's a child template of the original template.
				$this->_view->templatename .= '/' . $this->_model->page_template;
			}

			$this->_view->title = $this->_model->title;
			$this->_view->access = $this->_model->access;
			$this->_view->updated = $this->_model->updated;

			foreach($this->_model->getParentTree() as $page){
				$this->_view->addBreadcrumb($page->getTitle(), $page->get('baseurl'));
			}

			if(!$this->_model->indexable) {
				$this->_view->meta->setMeta('robots', 'noindex');
			}

			$this->_view->meta->setMeta(
				'canonical',
				$this->_model->get('baseurl')
			);

			$metas = $this->_model->getLink('PageMeta');
			foreach($metas as $pageMeta){
				$this->_view->meta->setMeta(
					$pageMeta->get('meta_key'),
					$pageMeta->get('meta_value_title'),
					$pageMeta->get('meta_value')
				);
			}
		}

		$profiler->stop();
	}
	
	/**
	 * Render the contained View to the browser
	 */
	public function render(){
		try{
			if($this->_view->mode !== View::MODE_STREAM) {
				// Streaming output should already be rendered, so don't try to re-render the data again
				$this->_view->render();
			}

			// Issue the hook to allow external scripts to hook into the process after everything
			// has been written to the DOM and all headers sent.
			HookHandler::DispatchHook('/core/page/postrender');
		}
		catch (Exception $ex) {
			// The user expects something, so give them something.
			$this->_view->status = View::STATUS_SERVERERROR;
			$this->_view->sendHeaders();
			
			$file = $this->_view->contenttype === View::CTYPE_HTML ?
				ROOT_PDIR . 'core/resources/servererror.html' :
				ROOT_PDIR . 'core/resources/servererror.txt';
			echo file_get_contents($file);

			LogEntry::Init()->exception($ex)->level(LOG_LEVEL_CRITICAL)->send();
		}
	}
	
	public function getView() : View {
		return $this->_view;
	}
	
	private function _loadController(){
		// The request will have the rendered baseurl that should map to a controller.
		if($this->_request === null){
			throw new Exception('No request set');
		}
		
		
		// If the baseurl has a non-200 status code, throw the corresponding exception here.
		// These are pre-controller exceptions caused by the class or method not being present.
		switch($this->_baseurl->getStatus()){
			case View::STATUS_BADREQUEST:
				throw new BadRequestException();
			case View::STATUS_URITOOLARGE:
				throw new URITooLargeException();
			case View::STATUS_NOTFOUND:
				throw new NotFoundException();
		}
		
		$class  = $this->_baseurl->getController();
		$method = $this->_baseurl->getMethod();
		
		
		// Use the CM Describer to get the directives and metadata for this method.
		$describer = new ControllerMethodDescriber($method);
		$describer->checkCanAccess();

		// Parameter checks for the GET/POST data
		$callArguments = [];
		foreach($method->getParameters() as $p){
			$callArguments[] = $this->_parseMethodParameter($p);
		}
		
		
		// Start loading the developer's Controller and populating with the required parameters.
		/** @var $controller Controller */
		$controller = $class->newInstance();
		
		
		// Make sure that the controller can access all expected objects
		$controller->request = $this->_request;
		$controller->view    = $this->_view;
		$controller->model   = $this->_model;
		
		
		// Controllers can have an access string defined to set access to anything in that controller.
		// Perform that check here.
		if (
			$controller->accessstring !== null &&
			!\Core\user()->checkAccess($controller->accessstring)
		) {
			throw new AccessDeniedException();
		}
		
		
		// In addition to the static checks, Controllers as of Core 6.3 also support a dynamic
		// __common method that can be called to run any custom check across all contained methods.
		// By default this just does nothing.
		$controller->__common();
		
		
		// Lastly, call the controller method and process its return.
		// A controller method can return a different View, exit codes, or nothing at all.
		\Core\log_trace('Calling controller logic contained in ' . $class->name . '->' . $method->name . '()');
		$result = $method->invoke($controller, ...$callArguments);
		
		if (is_int($result)) {
			// A generic error code was returned.  Create a View with that code and return that instead.
			$this->_view->status = $result;
		}
		elseif($result instanceof View && $result != $this->_view){
			// The controller method changed the view, (which is allowed),
			// but this needs to be remapped to this object so render knows about it.
			$this->_view = $result;
		}
	}
	
	private function _parseMethodParameter(ReflectionParameter $parameter) {
		$type     = $parameter->hasType() ? $parameter->getType()->getName() : null;
		$required = !$parameter->isOptional();
		$method   = $this->_request->method;
		
		// If a Form was requested as this parameter, retrieve the saved Form from the session!
		if($type === 'Core\\Forms\\Form'){
			return Forms\Form::GetSubmittedForm($method);
		}
		
		if(
			$method === Request::METHOD_POST &&
			filter_has_var (INPUT_POST, $parameter->name)
		){
			switch($type){
				case 'string':
					return filter_input(INPUT_POST, $parameter->name, FILTER_SANITIZE_STRING);
				case 'int':
					return (int)filter_input(INPUT_POST, $parameter->name, FILTER_SANITIZE_NUMBER_INT);
				default:
					return $_POST[$parameter->name] ?? null;
			}
		}
		elseif(
			// Query string parameters are supported for all METHOD types, not just GET.
			isset($this->_request->parameters[ $parameter->name ])
		){
			// GET parameters can be set manually in the URL, so accessing the GET directly
			// is avoided here to keep support for that.
			switch($type){
				case 'string':
					return filter_var($this->_request->parameters[ $parameter->name ], FILTER_SANITIZE_STRING);
				case 'int':
					return (int)filter_var($this->_request->parameters[ $parameter->name ], FILTER_SANITIZE_NUMBER_INT);
				default:
					// @todo!
					return $this->_request->parameters[ $parameter->name ];
			}
		}
		elseif($required){
			$this->_view->status = View::STATUS_BADREQUEST;
			throw new Exception($parameter->name . ' is required!');
		}
		else{
			// This is an optional variable and is not set, so set to null!
			return null;
		}
	}

	/**
	 * Gracefully handle exceptions thrown from the application based on various criteria
	 *
	 * @param ViewStatusExceptionInterface $ex
	 */
	private function _handleViewDataForExceptions(Exception $ex) {

		if($ex instanceof ViewStatusExceptionInterface) {
			// View Status exceptions have error codes that map to HTTP status codes
			$code = $ex->getCode();
		}
		else {
			// All other exceptions should have a 500 status code, as they're unexpected or unknown.
			$code = 500;
		}

		$message = $ex->getMessage();


		// Assign the template based on the code, if it's used.
		if($code >= 400 && $code <= 417) {
			// This is a 4xx series error that has a corresponding template
			$errorTemplate = 'pages/error/error' . (string)$code;
		}
		elseif($code === 500) {
			// This is a 5xx series error that has a corresponding template
			// (only 500 is supported at the moment)
			$errorTemplate = 'pages/error/error' . (string)$code;
		}
		else {
			// Generic or unknown error.
			$errorTemplate = 'pages/error/error';
		}


		if(EXEC_MODE === 'CLI') {
			// Command line executions use standard templates
			$this->_view->templatename = 'pages/error/error.txt';
			$this->_view->assign('error_message', $message);
		}
		elseif($this->_view->contenttype === View::CTYPE_JSON) {
			// JSON data, no templates are used but the data needs to be populated.
			if($this->_view->jsondata === null) {
				$this->_view->jsondata = [];
			}

			$this->_view->jsondata['status'] = $code;
			$this->_view->jsondata['message'] = $message;
		}
		else {
			$this->_view->templatename = $errorTemplate;
			$this->_view->assign('error_message', $message);
			$this->_view->assign('exception', $ex);
		}

		// @todo support images and other content types
	}

	/**
	 * Resolve template filenames for the View based on common parameters.
	 *
	 * @return string|null
	 */
	private function _resolveTemplateName() : ?string {
		if($this->_view->mode === View::MODE_NOOUTPUT) {
			// Pages that are set to have no output will *ALWAYS* have no template.
			// This is used by files that are sent to the useragent without templates.
			return null;
		}

		if($this->_view->contenttype === View::CTYPE_JSON && $this->_view->jsondata !== null) {
			// JSON pages that have JSON data set will not have any templates; the data is already set.
			return null;
		}

		if($this->_view->templatename === false) {
			// False is a special case from the script where no template is specifically requested.
			return null;
		}

		if($this->_view->templatename !== null) {
			// The template was explictly set, just ensure it's in the correct format and return that.

			if (strpos($this->_view->templatename, '/') === false) {
				// The template was set, but only the base version.
				// This should check for an alt version contained in the directory with the same name.
				return $this->_baseurl->getDefaultTemplateName() . '/' . $this->_view->templatename;
			}
			elseif (strpos($this->_view->templatename, '/pages/') === 0) {
				// Page starts with /pages/[foo]... we can safely trim off the beginning slash.
				return substr($this->_view->templatename, 1);
			}
			elseif (strpos($this->_view->templatename, 'pages/') !== 0) {
				// Allow the user to lazy-set a template name without defining "pages/" first.
				return 'pages/' . $this->_view->templatename;
			}
			else {
				// Templatename appears to not have any common issues, allow this to go through unmodified.
				return $this->_view->templatename;
			}
		}


		// At this stage, the mode is NOT no output, is NOT JSON data, and does NOT have a template set.
		// This is the full auto mode!


		if($this->_view->status !== View::STATUS_OK) {
			// There was an error on this page of some sort and the template was not overwritten
			// from a hooked call.  Resolve this to Core's built-in error templates.
			switch($this->_view->contenttype) {
				case View::CTYPE_JSON:
					return null;
				case View::CTYPE_PLAIN:
					return 'pages/error/error.txt';
				case View::CTYPE_HTML:
				default:
					return 'pages/error/error';
			}
		}

		return $this->_baseurl->getDefaultTemplateName();
	}
	
	private function _resolveSkinName() : ?string {
		if($this->_view->mode === View::MODE_NOOUTPUT) {
			// Pages that are set to have no output will *ALWAYS* have no template.
			// This is used by files that are sent to the useragent without templates.
			return null;
		}

		if($this->_view->mastertemplate === false) {
			// False is a special case from the script where no template is specifically requested.
			return null;
		}

		if($this->_view->contenttype === View::CTYPE_JSON && $this->_view->jsondata !== null) {
			// JSON pages that have JSON data set will not have any templates; the data is already set.
			return null;
		}


		// Block to handle default skins
		if($this->_view->mastertemplate === 'admin'){
			// Allows the page to declare that it's to use the configured "admin" skin
			return 'skins/' . ConfigHandler::Get('/theme/default_admin_template');
		}
		elseif($this->_view->mastertemplate === 'default'){
			// Default skin; (configured default skin)
			return 'skins/' . ConfigHandler::Get('/theme/default_template');
		}
		elseif($this->_view->mastertemplate === 'blank') {
			return 'skins/_blank';
		}
		elseif(
			$this->_view->mode === View::MODE_EMAILORPRINT &&
			$this->_view->contenttype === View::CTYPE_HTML &&
			$this->_view->mastertemplate === null
		){
			// Emails get the email skin
			return 'skins/' . ConfigHandler::Get('/theme/default_email_template');
		}
		elseif(strpos($this->_view->mastertemplate, 'skin/') === false) {
			// Skin isn't resolved, try to map it to a common alias
			$c = ConfigHandler::Get('/theme/siteskin/' . $this->_view->mastertemplate);
			if(!$c) {
				// Simply remap invalid aliases to the default skin.
				return 'skins/' . ConfigHandler::Get('/theme/default_template');
			}
			else {
				return 'skins/' . $c;
			}
		}
		else {
			return $this->_view->mastertemplate;
		}
	}
	
	private function _postProcessView() {
		
		// High-level operations based on the view mode generally
		if($this->_view->mode === View::MODE_PAGEORAJAX && !$this->_request->isAjax()) {
			// Page or Ajax requests by default get mapped to HTML as a page if not set otherwise.
			$this->_view->mode = View::MODE_PAGE;
		}
		elseif($this->_view->mode === View::MODE_PAGEORAJAX && $this->_request->isAjax()) {
			// Page or Ajax request, and this request WAS an ajax request.
			$this->_view->mode = View::MODE_AJAX;
		}
		elseif($this->_view->contenttype === View::CTYPE_HTML && $this->_request->isAjax()) {
			// Any HTML pages requested via AJAX should use JSON formatting!
			$this->_view->mode = View::MODE_AJAX;
			$this->_view->contenttype = View::CTYPE_JSON;
		}

		//$this->_view->templatename = $this->_resolveTemplateName();
		//$this->_view->mastertemplate = $this->_resolveSkinName();

		
		

		

		if($this->_view->contenttype === View::CTYPE_HTML) {
			// Block to handle assigning dynamic body classes based on the Request and any error codes
			$url = strtolower(trim(preg_replace('/[^a-z0-9\-]*/i', '', str_replace('/', '-', $this->_baseurl->getBaseURL())), '-'));
			while($url != ''){
				$this->_view->bodyclasses[] = 'page-' . $url;
				$url = substr($url, 0, strrpos($url, '-'));
			}

			$this->_view->bodyclasses[] = 'page-rewriteurl-' . \Core\str_to_url($this->_baseurl->getRewriteURL());

			if($this->_view->status !== View::STATUS_OK){
				$this->_view->bodyclasses[] = 'page-error';
				$this->_view->bodyclasses[] = 'page-error-' . $this->_view->status;
			}
		}
	}

	/**
	 * Change the base URL of the current request to a new path.
	 *
	 * This is useful inside /core/page/preexecute where the Controller needs to be remapped to another controller.
	 *
	 * @param string $newurl
	 */
	public static function ChangeBaseURL(string $newurl) {
		$instance = self::$DefaultRouter;

		$instance->_baseurl = new BaseURLMap($newurl);
		self::$DefaultBaseURL = $instance->_baseurl;

		// The model is pulled from the URL Map, but it may be NULL!
		$instance->_model = $instance->_baseurl->getPageModel();
		self::$DefaultPageModel = $instance->_model;
	}
}
