<?php
/**
 * Session system, responsible for saving and retrieving all data to and from the database.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

//This is a hack to allow classes to still be available after the page has been rendered.
//register_shutdown_function("session_write_close");

namespace Core;
use Exception;
use SessionModel;
use Core\Datamodel\Dataset;
use Core\Date\DateTime;

class Session implements \SessionHandlerInterface {
	
	private $_privateKey;

	/**
	 * @var Session
	 */
	public static $Instance;

	/** @var array Any externally set information on this session model. */
	public static $Externals = [];

	private static $_IsReady = false;

	public function __construct(){
		/*if(self::$Instance === null){
			self::$Instance = $this;

			//$session = new Session();
			session_set_save_handler(self::$Instance, true);
			session_start();
		}*/
	}

	/**
	 * PHP >= 5.4.0<br/>
	 * Close the session
	 * @link http://php.net/manual/en/sessionhandlerinterafce.close.php
	 * @return bool <p>
	 * The return value (usually TRUE on success, FALSE on failure).
	 * Note this value is returned internally to PHP for processing.
	 * </p>
	 */
	public function close() {
		// Sessions are persistent!
		return true;
	}

	/**
	 * PHP >= 5.4.0<br/>
	 * Initialize session
	 * @link http://php.net/manual/en/sessionhandlerinterafce.open.php
	 *
	 * @param string $save_path The path where to store/retrieve the session.
	 * @param string $session_id The session id.
	 *
	 * @return bool <p>
	 * The return value (usually TRUE on success, FALSE on failure).
	 * Note this value is returned internally to PHP for processing.
	 * </p>
	 */
	public function open($save_path, $session_id) {
		// Anything waiting for the session?
		//\HookHandler::DispatchHook('/core/session/ready');

		self::$_IsReady = true;

		// Open really doesn't need to do anything, everything that will be controlling the session will be built into Core Plus.
		return true;
	}

	/**
	 * PHP >= 5.4.0<br/>
	 * Destroy a session
	 * @link http://php.net/manual/en/sessionhandlerinterafce.destroy.php
	 *
	 * @param int $session_id The session ID being destroyed.
	 *
	 * @return bool <p>
	 * The return value (usually TRUE on success, FALSE on failure).
	 * Note this value is returned internally to PHP for processing.
	 * </p>
	 */
	public function destroy($session_id) {
		// Low-level datasets are used here because they have less overhead than
		// the full-blown model system.
		$dataset = new Dataset();
		$dataset->table('session');
		$dataset->where('session_id = ' . $session_id);
		$dataset->where('ip_addr = ' . REMOTE_IP);

		$dataset->delete();
		$dataset->execute();

		// Blow away the current session data too!
		$_SESSION = null;
		self::$_IsReady = false;

		return true;
	}

	/**
	 * PHP >= 5.4.0<br/>
	 * Read session data
	 * @link http://php.net/manual/en/sessionhandlerinterafce.read.php
	 * @param string $session_id The session id to read data for.
	 * @return string <p>
	 * Returns an encoded string of the read data.
	 * If nothing was read, it must return an empty string.
	 * Note this value is returned internally to PHP for processing.
	 * </p>
	 */
	public function read($session_id) {
		$model = self::_GetModel($session_id);

		$this->_privateKey = $model->get('session_private_key');
		self::$Externals = $model->getExternalData();
		#return $model->getData();
		
		# SessionModel's data *can* be null, as that is the model schema default. This is no good on PHP 7.0+
        # read() must return an empty string or string data...not null or false!
        return $model->getData() ? $model->getData() : '';

	}

	/**
	 * PHP >= 5.4.0<br/>
	 * Write session data
	 * @link http://php.net/manual/en/sessionhandlerinterafce.write.php
	 * @param string $session_id The session id.
	 * @param string $session_data <p>
	 * The encoded session data. This data is the
	 * result of the PHP internally encoding
	 * the $_SESSION superglobal to a serialized
	 * string and passing it as this parameter.
	 * Please note sessions use an alternative serialization method.
	 * </p>
	 * @return bool <p>
	 * The return value (usually TRUE on success, FALSE on failure).
	 * Note this value is returned internally to PHP for processing.
	 * </p>
	 */
	public function write($session_id, $session_data) {
		$model = self::_GetModel($session_id);
		$model->setData($session_data);
		$model->setExternalData(self::$Externals);
		
		if(!$model->get('session_private_key')){
			$model->set('session_private_key', self::GetPrivateKey());
		}
		
		$model->save();
		
		return true; // PHP requires this to be true otherwise it assumes an error occurred.
	}

	/**
	 * PHP >= 5.4.0<br/>
	 * Cleanup old sessions
	 * @link http://php.net/manual/en/sessionhandlerinterafce.gc.php
	 *
	 * @param int $maxlifetime <p>
	 * Sessions that have not updated for
	 * the last maxlifetime seconds will be removed.
	 * </p>
	 *
	 * @return bool <p>
	 * The return value (usually TRUE on success, FALSE on failure).
	 * Note this value is returned internally to PHP for processing.
	 * </p>
	 */
	public function gc($maxlifetime) {
		return self::CleanupExpired();
	}

	/**
	 * Set the current session to be owned by the given user,
	 * effectively logging the user in.
	 *
	 * Drop support for the User class in favour of UserModel after pre-2.8.x is no longer supported.
	 *
	 * @param \UserModel $u
	 */
	public static function SetUser($u) {
		$model = self::_GetModel(session_id());

		$model->set('user_id', $u->get('id'));
		$model->save();

		// If this user is currently SUDO, then set the SUDO'd user and NOT the actual user.
		if(isset($_SESSION['user_sudo'])){
			Session::Set('user_sudo', $u);
		}
		else{
			Session::Set('user', $u);
		}
	}

	/**
	 * Shortcut static function to call that will destroy the current session and logout the user.
	 *
	 */
	public static function DestroySession(){
		if(self::$Instance !== null){
			self::$Instance->destroy(session_id());
		}
	}

	/**
	 * Force the saving of the contents of $_SESSION back to the database.
	 */
	public static function ForceSave(){
		/*$session = self::$Instance;
		if($session){
			$session->write(session_id(), serialize($_SESSION));
		}*/
	}
	
	/**
	 * Handler to be called after a page has been loaded.
	 * 
	 * This ensures that the user's session doesn't terminate while they're browsing around.
	 * (otherwise, there's nothing on the session data that changes, so the updated timestamp never changes).
	 */
	public static function ClosePage(){
		if(
			self::$Instance !== null && 
			self::$Instance !== false
		){
			// Perform this update at a lower level because the Model system will refuse to proceed
			// if there were no actual data changes.
			Datamodel\Dataset::Init()
				->update('updated', Date\DateTime::NowGMT())
				->table('session')
				->where('session_id = ' . session_id())
				->execute();
		}
	}

	/**
	 * Cleanup any expired sessions from the database.
	 *
	 * @return bool Always returns true :)
	 */
	public static function CleanupExpired(){
		static $lastexecuted = 0;

		/**
		 * Delete ANY session that has expired.
		 */
		$ttl = ConfigHandler::Get('/core/session/ttl');
		$datetime = (\Time::GetCurrentGMT() - $ttl);

		if($lastexecuted == $datetime){
			// This operation was already called this second.  No need to do it again.
			// This is used because if a LOT of user operations occur on a given page,
			// this method may be called many many many times.
			return true;
		}

		// Low-level datasets are used here because they have less overhead than
		// the full-blown model system.
		$dataset = new Datamodel\Dataset();
		$dataset->table('session');
		$dataset->where('updated < ' . $datetime);
		$dataset->delete()->execute();

		$lastexecuted = $datetime;

		// Always return TRUE
		return true;
	}

	/**
	 * Get the value of a key from the session or $default if that key is explicitly not set.
	 *
	 * If the key "example/*" is provided, then all subkeys under the example array are returned.
	 * If no keys exist, an empty array is returned.
	 *
	 * @param string     $key
	 * @param null|mixed $default
	 *
	 * @return mixed
	 */
	public static function Get($key, $default = null){
		if(strpos($key, '/*') !== false){
			// If a group of keys are selected, the default is forced to be an empty array.
			// This ensures that an array is returned even if there are no COOKIEs.
			$default = [];
		}

		if(sizeof($_COOKIE) == 0){
			// If no cookie is set, then don't try to look anything up!
			// The act of looking up a variable will create the session, which sets the cookie value!
			// This doesn't help caching servers very much.
			return $default;
		}

		self::_GetInstance();

		if($key[0] === '/') {
			// Keys should not start with a '/', root is already assumed.
			$key = ltrim($key, '/');
		}
		
		if(strpos($key, '/*') !== false){
			// User requested to retrieve a set of keys, retrieve anything contained in this subarray
			$sub = substr($key, 0, strpos($key, '/*'));
			foreach($_SESSION as $k => $v){
				if($k == $sub){
					return $_SESSION[$k];
				}
			}
		}
		elseif(strpos($key, '/') !== false){
			// User requested a subkey, look for it!
			$sub = substr($key, 0, strpos($key, '/'));
			$key = substr($key, strlen($sub)+1);
			if(!isset($_SESSION[$sub])){
				return $default;
			}
			return isset($_SESSION[$sub][$key]) ? $_SESSION[$sub][$key] : $default;
		}

		return isset($_SESSION[$key]) ? $_SESSION[$key] : $default;
	}
	
	public static function GetPrivateKey() : string {
		$instance = self::_GetInstance();
		
		// Ensure there's a private key for this session; used for server-side salt operations
		// where the salt needs to be reproducable for a given user session, (such as form submissions)
		if(!$instance->_privateKey){
			$instance->_privateKey = \Core\random_hex(16, true);
		}
		
		return $instance->_privateKey;
	}
	
	/**
	 * Quick lookup if the requested key is present in the session yet
	 * 
	 * @param string $key
	 * @return bool True/False if the key in question exists in the session
	 */
	public static function IsKeySet($key) : bool {
		if(sizeof($_COOKIE) == 0){
			// If no cookie is set, then don't try to look anything up!
			// The act of looking up a variable will create the session, which sets the cookie value!
			// This doesn't help caching servers very much.
			return false;
		}
		
		self::_GetInstance();

		if($key[0] === '/') {
			// Keys should not start with a '/', root is already assumed.
			$key = ltrim($key, '/');
		}

		if(strpos($key, '/') !== false){
			$sub = substr($key, 0, strpos($key, '/'));
			$spr = substr($key, strlen($sub) + 1);

			return isset($_SESSION[ $sub ]) && isset($_SESSION[ $sub ][ $spr ]);
		}
		else{
			return isset($_SESSION[$key]);
		}
	}

	/**
	 * Set the value of a key to some value.
	 *
	 * @param string $key     The key to save into the session, can contain a '/' to indicate an array.
	 * @param mixed  $value   The value to save on that key
	 * @param int    $expires Optionally set an expiration to auto-delete this data after N seconds
	 */
	public static function Set(string $key, $value, int $expires = null) {
		$session = self::_GetInstance();

		if($key[0] === '/') {
			// Keys should not start with a '/', root is already assumed.
			$key = ltrim($key, '/');
		}

		if(strpos($key, '/') !== false){
			// User requested to retrieve a set of keys, retrieve anything contained in this subarray
			$sub = substr($key, 0, strpos($key, '/'));
			$spr = substr($key, strlen($sub) + 1);
			
			if( !(isset($_SESSION[$sub]) && is_array($_SESSION[$sub])) ) {
				$_SESSION[$sub] = [];
			}
			
			$_SESSION[$sub][$spr] = $value;
		}
		else{
			$_SESSION[$key] = $value;
		}
		
		// Auto-expire requested?
		if($expires !== null){
			$expiration = DateTime::NowGMT('U') + $expires;
			
			if( !(isset($_SESSION['___autoexpires']) && is_array($_SESSION['___autoexpires'])) ) {
				$_SESSION['___autoexpires'] = [];
			}
			$_SESSION['___autoexpires'][ $key ] = $expiration;
		}

		// Immediately save the data to the database
		// This incurs a performance hit, but will yield more consistent states for the application.
		// Example, if a Form renders to the page, but an error after prevents the page from exiting cleanly,
		// the session data saved will be lost.
		//echo 'Saving session data for ' . session_id() . ' NOW' . "<br/>\n";
		if($session) {
			$session->write(session_id(), session_encode());
		}
	}

	/**
	 * Explictly unset a key from the session.
	 *
	 * @param $key string
	 */
	public static function UnsetKey(string $key) {
		if(sizeof($_COOKIE) == 0){
			// If no cookie is set, then don't try to look anything up!
			// The act of looking up a variable will create the session, which sets the cookie value!
			// This doesn't help caching servers very much.
			return;
		}

		self::_GetInstance();

		if($key[0] === '/') {
			// Keys should not start with a '/', root is already assumed.
			$key = ltrim($key, '/');
		}

		if($key === '*'){
			// If the user requested to purge all the session data.
			$_SESSION = [];
		}
		elseif(strpos($key, '/*') !== false){
			// User requested to remove a set of keys, remove anything starting with this.
			$sub = substr($key, 0, strpos($key, '/*'));
			foreach($_SESSION as $k => $v){
				if($k == $sub){
					unset($_SESSION[$k]);
				}
			}
		}
		elseif(strpos($key, '/') !== false){
			// User requested to retrieve a set of keys, retrieve anything contained in this subarray
			$sub = substr($key, 0, strpos($key, '/'));
			$spr = substr($key, strlen($sub) + 1);
			if(isset($_SESSION[$sub]) && isset($_SESSION[$sub][$spr])){
				unset($_SESSION[$sub][$spr]);
			}
		}
		elseif(isset($_SESSION[$key])){
			unset($_SESSION[$key]);
		}
	}
	
	/**
	 * Alias of UnsetKey
	 * 
	 * @param string $key
	 */
	public static function Delete(string $key) {
		self::UnsetKey($key);
	}

	/**
	 * @return Session
	 */
	private static function _GetInstance(){
		if(self::$Instance === null){
			ini_set('session.hash_bits_per_character', 5);
			ini_set('session.hash_function', 1);
			if(!defined('SESSION_COOKIE_NAME')){
				define('SESSION_COOKIE_NAME', 'CorePlusSession');
			}
			session_name(SESSION_COOKIE_NAME);

			// Allow a config-set cookie domain.  This is required for xsite sessions in multimode.
			if(defined('SESSION_COOKIE_DOMAIN') && SESSION_COOKIE_DOMAIN){
				// A valid session name is required for xsite sessions to work. (not sure why)
				session_set_cookie_params(0, '/', SESSION_COOKIE_DOMAIN);
			}
			
			try{
				// Try to load the system DMI to use database sessions.
				if(!\Core\Datamodel\DMI::GetSystemDMI()->connection()->tableExists('session')){
					// session table does not exist yet; use PHP's default.
					self::$Instance = false;
					session_start();
				}
				else{
					self::$Instance = new Session();

					session_set_save_handler(self::$Instance, true);
					session_start();
					
					self::_AutoExpireKeys();
				}
			} catch (Exception $ex) {
				// If that fails however, fallback to PHP's default session handler.
				self::$Instance = false;
				session_start();
			}
		}

		return self::$Instance;
	}


	/**
	 * Get the Model for this current session.
	 * This method will NOT cache the results of the model.  This is due to race conditions at some point...
	 *
	 * @param string $session_id The session id to read the model for.
	 * @return \SessionModel
	 */
	private static function _GetModel($session_id) {
		
		// Do some sanitizing just to prevent clutter in the database from attack attempts.
		$session_id = preg_replace('/[^a-zA-Z0-9]/', '', $session_id);
		
		$model = new SessionModel($session_id);
		
		// If the IP doesn't match, invalidate it.
		if($model->get('ip_addr') == '') {
			// Ensure the data is matched up.
			$model->set('ip_addr', REMOTE_IP);
		}
		elseif($model->get('ip_addr') !== REMOTE_IP) {
			$model->set('data', null);
			$model->set('ip_addr', REMOTE_IP);
		}
		
		return $model;
	}
	
	/**
	 * Purge any key that's set to auto-expire if there are any.
	 * 
	 * Useful for internal cleanup without purging the entire Session.
	 */
	private static function _AutoExpireKeys() {
		if( !(isset($_SESSION['___autoexpires']) && is_array($_SESSION['___autoexpires'])) ) {
			return;
		}
		
		$autoExpires =& $_SESSION['___autoexpires'];
		$now = DateTime::NowGMT('U');
		
		foreach($autoExpires as $key => $expires){
			if($expires <= $now){
				self::UnsetKey($key);
				unset($autoExpires[ $key ]);
			}
		}
	}
}

