<?php
/**
 * Systems for the control widgets used for page-level operations and inline operations.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @since 2.1.2
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use ArrayAccess;
use Iterator;
use Exception;

/**
 * Just a tiny class for handling control links in the main page view.
 *
 * These are usually tiny icons or snippets of text that provide a bit of inline administrative
 * functionality for pages.
 */
class ViewControl implements ArrayAccess {

	/**
	 * Link for this control
	 *
	 * @var string
	 */
	public $link = '#';

	/**
	 * Title for this control
	 *
	 * @var string
	 */
	public $title = '';

	/**
	 * CSS class name for this control
	 *
	 * @var string
	 */
	public $class = '';

	/**
	 * Icon class name for this control
	 *
	 * Set to blank to omit the icon
	 *
	 * @var string
	 */
	public $icon = '';

	/**
	 * Confirm text for this link, useful for setting them as POST links.
	 *
	 * @var string|null
	 */
	public $confirm = null;

	/**
	 * The group for this control
	 *
	 * Use a blank string to indicate a top-level control
	 *
	 * @var string
	 */
	public $group = '';

	/**
	 * Any other attributes for the a tag.
	 *
	 * @var array
	 */
	public $otherattributes = [
		'role' => 'menuitem'
	];

	public $liattributes = [
		'data-important' => 0,
	];

	public $weight = self::WEIGHT_UNDEFINED;

	
	const ICON_ADD      = 'add';
	const ICON_DELETE   = 'times';
	const ICON_DOWNLOAD = 'download';
	const ICON_EDIT     = 'edit';
	const ICON_IMPORT   = 'upload';
	const ICON_UPLOAD   = 'upload';
	const ICON_VIEW     = 'eye';

	const WEIGHT_VIEW      = 10;
	const WEIGHT_UTILITIES = 20;
	const WEIGHT_ADD       = 30;
	const WEIGHT_EDIT      = 40;
	const WEIGHT_FINALIZE  = 70;
	const WEIGHT_DELETE    = 90;
	const WEIGHT_UNDEFINED = 99;


	const GROUP_ADMIN  = 't:STRING_ADMIN';
	const GROUP_VIEW   = 't:STRING_VIEW';
	const GROUP_MANAGE = 't:STRING_MANAGE';
	const GROUP_CREATE = 't:STRING_CREATE';

	/**
	 * Get this control as HTML
	 *
	 * @return string (HTML)
	 */
	public function fetch(){
		if(strpos($this->title, 't:') === 0) {
			// Translate this title.
			$title = '<span>' . t(substr($this->title, 2)) . '</span>';
		}
		else {
			$title = '<span>' . $this->title . '</span>';
		}
		
		
		if($this->icon) {
			$icon = '<i class="icon icon-' . $this->icon . '"></i> ';
		}
		else {
			$icon = '<i class="icon">&nbsp;</i>';
		}

		return
			$this->_fetchLIStartFragment() .
			$this->_fetchAStartFragment() .
			$icon .
			$title .
			'</a>' .
			'</li>';
	}

	/**
	 * Get this control as HTML, without any surrounding LI.
	 *
	 * @return string (HTML)
	 */
	public function fetchSingle(){
		if(strpos($this->title, 't:') === 0) {
			// Translate this title.
			$title = '<span>' . t(substr($this->title, 2)) . '</span>';
		}
		else {
			$title = '<span>' . $this->title . '</span>';
		}


		if($this->icon) {
			$icon = '<i class="icon icon-' . $this->icon . '"></i> ';
		}
		else {
			$icon = '<i class="icon">&nbsp;</i>';
		}

		return
			$this->_fetchAStartFragment() .
			$icon .
			$title .
			'</a>';
	}
	

	/**
	 * Fetch the A tag for this element.
	 *
	 * This is broken out into its own function since it has a decent amount of logic contained herein.
	 *
	 * @return string (HTML fragment)
	 */
	private function _fetchAStartFragment() : string {
		
		$href = $this->link ? $this->link : '#';

		// Start by grabbing the "other" attributes, these will get overwrote with the specific elements.
		$dat = [];
		foreach($this->otherattributes as $k => $v) {
			if($k[0] !== '_') {
				$dat[ $k ] = htmlentities($v);
			}
		}

		if($this->confirm !== null){

			$confirm = $this->confirm;
			if(strpos($confirm, 't:') === 0){
				// Translate this title.
				$confirm = t(substr($confirm, 2));
			}

			$this->class .= ' confirm-link';
			//$dat['onclick'] = 'return Core.ConfirmEvent(this);';
			$dat['data-href'] = \Core\resolve_link($href);
			$dat['data-confirm'] = $confirm;
			$dat['href'] = '#false';
		}
		else{
			$dat['href'] = $href;
		}

		$title = $this->title;
		if(strpos($title, 't:') === 0){
			// Translate this title.
			$title = t(substr($title, 2));
		}

		$dat['title'] = $title;
		if($this->class) {
			$dat['class'] = trim($this->class);
		}

		$html = '<a ';
		foreach($dat as $k => $v){
			$html .= " $k=\"$v\"";
		}
		$html .= '>';
		return $html;
	}

	/**
	 * Fetch the LI tag for this element.
	 *
	 * This is broken out into its own function since it has a decent amount of logic contained herein.
	 *
	 * @return string (HTML fragment)
	 */
	private function _fetchLIStartFragment() : string {
		// Start by grabbing the "other" attributes, these will get overwrote with the specific elements.
		$dat = [];
		foreach($this->liattributes as $k => $v) {
			if($k[0] !== '_') {
				if(is_array($v)) {
					$v = implode(' ', array_unique($v));
				}

				$dat[ $k ] = htmlentities($v);
			}
		}


		$dat['data-control-weight'] = $this->weight;

		$html = '<li ';
		foreach($dat as $k => $v){
			$html .= " $k=\"$v\"";
		}
		$html .= '>';
		return $html;
	}


	public function set($key, $value) : ViewControl{
		switch($key){
			case 'class':
				$this->classname($value);
				break;
			case 'confirm':
				$this->confirm($value);
				break;
			case 'icon':
				$this->icon($value);
				break;
			case 'group':
				$this->group($value);
				break;
			case 'link':
			case 'url':
			case 'href': // Just for an alias of the link.
				$this->link($value);
				break;
			case 'title':
				$this->title($value);
				break;
			case 'weight':
				$this->weight( (int) $value );
				break;
			case 'important':
				$this->important((bool) $value);
				break;
			case 'liclass':
				$this->liClass($value);
				break;
			default:
				$this->otherattributes[$key] = $value;
				break;
		}
		
		// Allow chaining
		return $this;
	}
	
	public function classname(string $class) : ViewControl {
		$this->class = $class;
		
		// Allow chaining
		return $this;
	}
	
	public function addClass(string $class) : ViewControl {
		$this->class .= ' ' . $class;
		
		// Allow chaining
		return $this;
	}
	
	public function confirm(string $text) : ViewControl {
		$this->confirm = $text;
		
		// Allow chaining
		return $this;
	}

	public function data(string $key, $value) : ViewControl {
		$this->otherattributes['data-' . $key] = $value;

		// Allow chaining
		return $this;
	}

	public function group(string $group) : ViewControl {
		$this->group = $group;

		// Allow chaining
		return $this;
	}
	
	public function icon(string $icon) : ViewControl {
		$this->icon = $icon;
		
		// Allow chaining
		return $this;
	}

	public function important(bool $isImportant) : ViewControl {
		$this->liattributes['data-important'] = $isImportant ? '1' : '0';

		// Allow chaining
		return $this;
	}

	public function liAttribute(string $attribute, $value) : ViewControl {
		$this->liattributes[ $attribute ] = $value;

		// Allow chaining
		return $this;
	}

	public function liClass(string $classname) : ViewControl {
		if(!isset($this->liattributes['class'])) {
			$this->liattributes['class'] = [];
		}

		$this->liattributes['class'][] = $classname;

		// Allow chaining
		return $this;
	}
	
	public function link(string $link) : ViewControl {
		$this->link = \Core\resolve_link($link);
		
		// Allow chaining
		return $this;
	}
	
	public function title(string $title) : ViewControl {
		$this->title = $title;
		
		// Allow chaining
		return $this;
	}

	public function weight(int $weight) : ViewControl {
		$this->weight = $weight;

		// Allow chaining
		return $this;
	}

	//// A few array access functions \\\\

	/**
	 * Whether an offset exists
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetexists.php
	 *
	 * @param mixed $offset An offset to check for.
	 *
	 * @return boolean Returns true on success or false on failure.
	 */
	public function offsetExists($offset) {
		return(property_exists($this, $offset));
	}

	/**
	 * Offset to retrieve
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetget.php
	 *
	 * @param mixed $offset The offset to retrieve.
	 *
	 * @return mixed Can return all value types.
	 */
	public function offsetGet($offset) {
		$dat = get_object_vars($this);

		if(isset($dat[$offset])){
			return $dat[$offset];
		}
		elseif(isset($this->otherattributes[$offset])){
			return $this->otherattributes[$offset];
		}
		else{
			return null;
		}
	}

	/**
	 * Offset to set
	 *
	 * Alias of Model::set()
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetset.php
	 *
	 * @param mixed $offset The offset to assign the value to.
	 * @param mixed $value The value to set.
	 *
	 * @return void
	 */
	public function offsetSet($offset, $value) {
		$this->set($offset, $value);
	}

	/**
	 * Offset to unset
	 *
	 * This actually doesn't do anything.
	 *
	 * @link http://php.net/manual/en/arrayaccess.offsetunset.php
	 *
	 * @param mixed $offset The offset to unset.
	 *
	 * @return void
	 */
	public function offsetUnset($offset) {
		return void;
	}

	public static function Factory(array $parameters = null) : ViewControl {
		$control = new ViewControl();

		if($parameters !== null) {
			foreach($parameters as $k => $v) {
				$control->set($k, $v);
			}
		}

		return $control;
	}

	/**
	 * Factory constructor to create a ViewControl from its title, link, and icon.
	 *
	 * @param string $title
	 * @param string $link
	 * @param string $icon
	 *
	 * @return ViewControl
	 */
	public static function FactoryTLI(string $title, string $link, string $icon = null) : ViewControl {
		$control = new ViewControl();

		$control->title($title)
			->link($link);

		if($icon) {
			$control->icon($icon);
		}

		return $control;
	}

	/**
	 * Factory constructor to create a ViewControl from a set of array data
	 *
	 * @param array $parameters
	 * @return ViewControl
	 */
	public static function FactoryArray(array $parameters) : ViewCOntrol {
		// @todo Should this be removed???
		$control = new ViewControl();
		foreach($parameters as $k => $v) {
			$control->set($k, $v);
		}
		return $control;
	}
}

