<?php
/**
 * File: CacheInterface
 * Interface that all storage-specific adapters must adhere to.
 *
 * @version 6.3
 * @package Core\Cache
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Cache;


/**
 * Interface: CacheInterface
 * 	Defines the methods that all implementing classes MUST have. Covers CRUD 
 * (create, read, update, delete) methods, as well as others that are used in 
 * the base CacheCore class.
 */
interface CacheInterface {

	/**
	 * Construct a new cache backend, capable of storing and retrieving values
	 */
	public function __construct();

	/**
	 * Writes the requested key and data optionally with an expiration date.
	 * 
	 * @param string $key  Key to save
	 * @param mixed $data  The data to cache
	 * @param int $expires Expiration in number of seconds from now
	 * 
	 * @return boolean Whether the operation was successful.
	 */
	public function write(string $key, $data, int $expires = null) : bool;

	/**
	 * Read the given key from the cache backend and return the contents
	 * 
	 * @param string $key Key to read
	 * 
	 * @return mixed Either the content of the cache object, or _boolean_ false.
	 */
	public function read(string $key);

	/**
	 * Delete the cache entry matching the key
	 * 
	 * @param string $key Key to delete
	 * 
	 * @return boolean Whether the operation was successful.
	 */
	public function delete(string $key) : bool;
	
	/**
	 * Invalidate all items in the cache
	 * 
	 * @since 2011.07.28
	 * @return boolean Whether the operation was successful.
	 */
	public function flush() : bool;
	
	/**
	 * Return a list of all keys on this backend, not supported by all processors!
	 * 
	 * @since 2017.04
	 * @return array
	 */
	public function ls() : array ;
	
	/**
	 * Check if this interface is ready to store/receive values.
	 * 
	 * @return bool
	 */
	public function isReady() : bool;
}
