<?php
/**
 * This is a basic class to provide some basic CLI functionality.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\CLI;

use Core\Dict;
use Core\DictItem;

class CLI {

	/**
	 * @var int Global max default width (length) of command line output
	 */
	public static $MaxLineWidth = 90;
	
	/**
	 * Parse CLI arguments to a valid GET string that can be passed into the Request object.
	 * 
	 * @param array $args Array of arguments, usually straight from $argv
	 * 
	 * @return string
	 */
	public static function ArgumentsToURI(array $args) : string {
		$count = count($args);
		if($count <= 1){
			// If only the script name was passed in, resolve that to '/'.
			return '/';
		}
		
		$uri = $args[1];
		
		if(strpos($uri, '/') !== 0){
			// Ensure it's prefixed with a '/'.
			$uri = '/' . $uri;
		}
		
		$params = [];
		$positionalC = 0;
		
		for($i = 2; $i < $count; $i++){
			if(
				preg_match('/^--[a-z_]*=.*/', $args[$i])
			){
				// This is a named argument with a value attached
				$pos = strpos($args[$i], '=');
				$k   = substr($args[$i], 2, $pos - 2);
				$v   = substr($args[$i], $pos + 1);
			}
			elseif(
				preg_match('/^--[a-z_]*$/', $args[$i]) &&
				$i + 1 < $count &&
				strpos($args[$i + 1], '--') !== 0
			){
				// This is a named argument with a value detached, 
				// (separated by a space as opposed to an '=' sign)
				$k = substr($args[$i], 2);
				$v = $args[$i + 1];
				$i++; // Skip the next argument; it's the val of this one.
			}
			elseif(
				preg_match('/^--[a-z_]*$/', $args[$i])
			){
				// Named argument with no value after it, (or another named after)
				$k = substr($args[$i], 2);
				$v = 1;
			}
			else{
				// This is a positional argument
				$k = $positionalC++;
				$v = $args[$i];
			}
			
			if(isset($params[ $k ]) && is_array($params[ $k ])){
				$params[ $k ][] = $v;
			}
			elseif(isset($params[ $k ])){
				$params[ $k ] = [ $params[ $k ], $v];
			}
			else{
				$params[ $k ] = $v;
			}
		}
		
		$qstrs = [];
		foreach($params as $k => $v){
			if(is_array($v)){
				foreach($v as $subv){
					$qstrs[] = $k . '[]=' . htmlentities($subv);
				}
			}
			else{
				$qstrs[] = $k . '=' . htmlentities($v);
			}
		}
		
		return $uri . (count($qstrs) ? '?' . implode('&', $qstrs) : '');
	}
	
	/**
	 * Simple command to clear the screen
	 */
	public static function Clear() {
		system('tput reset');
	}
	
	public static function PromptFormElement(string $type, array $attributes){
		if(isset($attributes['description'])){
			CLI::PrintDebug($attributes['description']);
		}
		
		$default = isset($attributes['value']) ? $attributes['value'] : null;
		$title   = isset($attributes['title']) ? $attributes['title'] : 'Question';
		
		switch($type){
			case 'select':
				return self::PromptUser($title, $attributes['options'], $default);
				
			case 'checkbox':
				// Allow auto-conversion to and from strings.
				if($default === '0' || $default === '1'){
					$conv = [ '0', '1' ];
				}
				elseif($default === 'false' || $default === 'true'){
					$conv = [ 'false', 'true' ];
				}
				elseif($default === 'no' || $default === 'yes'){
					$conv = [ 'no', 'yes' ];
				}
				else{
					$conv = [ false, true ];
				}
				
				return(self::PromptUser($title, 'bool', (array_search($default, $conv) === 1))) ?
					$conv[1] :
					$conv[0];
				
			default:
				return self::PromptUser($title, 'text', $default);
		}
	}
	
	/**
	 * Prompt the user a question and return the result.
	 *
	 * @param string       $question The question to prompt to the user.
	 * @param array|string $answers  What answers to provide to the user.
	 *                               array           - Will prompt the user with the value of each pair, returning the key.
	 *                               "boolean"       - Will ask for a yes/no response and return true/false.
	 *                               "text"          - Open-ended text input, user can type in anything and that input is returned.
	 *                               "text-required" - Open-ended text input, user can type in anything (non-blank), and that value is returned.
	 * @param mixed  $default  string The default answer if the user simply presses "enter". [optional]
	 *
	 * @throws \Exception
	 * @return bool|string
	 */
	public static function PromptUser($question, $answers, $default = null, $description = null) {
		
		if (is_array($answers)) {
			return self::_DisplayOptions($question, Dict::FromList($answers), $default, $description);
		}
		elseif ($answers instanceof Dict) {
			return self::_DisplayOptions($question, $answers, $default, $description);
		}
		elseif($answers instanceof CLITable) {
			return self::_DisplayTableSelect($question, $answers, $default, $description);
		}
		elseif(
			$answers === 'bool' || 
			$answers === 'boolean' ||
			$answers === 'checkbox'
		){
			return self::_DisplayCheckbox($question, $default, $description);
		}
		elseif($answers === 'text'){
			return self::_DisplayText($question, $default, false, $description);
		}
		elseif($answers === 'text-required'){
			return self::_DisplayText($question, $default, true, $description);
		}
		elseif($answers === 'file'){
			return self::_DisplayFileSelect($question, $default, $description);
		}
		elseif($answers === 'textarea'){
			return self::_DisplayTextarea($question, $default, $description);
		}
		else{
			throw new \Exception('Unsupported answer choice [' . strtolower($answers) . '], please ensure it is either an array of options, "boolean", "text", "text-required", or "textarea"!');
		}
	}

	/**
	 * Set the 'EDITOR' variable to be set.
	 * This is a linux-specific thing that svn shares also.
	 *
	 * The user will usually set their preferred EDITOR, be it vi/vim, emacs or nano.
	 * If they didn't, ask the user for their choice.
	 */
	public static function RequireEditor() {
		global $previous_editor;

		// First, check the editor in the "session" file.
		\Core\CLI\CLI::LoadSettingsFile('editor');
		if (isset($previous_editor)) $_SERVER['EDITOR'] = $previous_editor;


		if (!isset($_SERVER['EDITOR']) || $_SERVER['EDITOR'] == '') {
			// I need to assemble a list of editors currently on the system.
			$opts    = array();
			$default = false;
			if (($loc = trim(`which vi`))) $opts[$loc] = $loc;
			if (($loc = trim(`which vim`))) $opts[$loc] = $loc;
			if (($loc = trim(`which emacs`))) $opts[$loc] = $loc;
			if (($loc = trim(`which nano`))) {
				$opts[$loc] = $loc;
				$default    = $loc;
			}

			$_SERVER['EDITOR'] = \Core\CLI\CLI::PromptUser(
				'Which editor do you want to use for editing text files?  If you are unsure, simply press enter if there is a default option.',
				$opts,
				$default
			);

			// And remember this option.
			$previous_editor = $_SERVER['EDITOR'];
			\Core\CLI\CLI::SaveSettingsFile('editor', array('previous_editor'));
		}
	}

	/**
	 * Print a stylized header to stdout
	 * 
	 * @param string $line   The header string to output
	 */
	public static function PrintHeader($line) {
		$nl = (EXEC_MODE == 'WEB') ? NL . '<br/>' : NL;
		
		echo COLOR_LINE;
		echo "| " . $nl;
		echo "+" . str_repeat('=', self::$MaxLineWidth - 1) . $nl;
		echo "| " . COLOR_RESET . COLOR_HEADER;

		// Make this text centered.
		if(strlen($line) < self::$MaxLineWidth){
			echo str_repeat(NBSP, (self::$MaxLineWidth - strlen($line)) / 2);
		}

		echo $line . $nl . COLOR_RESET . COLOR_LINE;
		echo "+" . str_repeat('=', self::$MaxLineWidth - 1);
		echo COLOR_RESET . $nl;

		if(EXEC_MODE == 'WEB'){
			echo '<!--CLI-DATA:HEADER=' . $line . '-->';
		}
		
		if(self::_FlushRequired()){
			ob_flush();
			if(phpversion() < '7.2.19') {
				flush();
			}
		}
	}

	/**
	 * Print a single line or multiple lines of text to the screen or console.
	 *
	 * @param string|array $line  Line (or array of lines) to output
	 * @param string       $color Colour to render the output with
	 */
	public static function PrintLine($line, $color = COLOR_NORMAL) {
		if(is_array($line)){
			foreach($line as $l){
				self::PrintLine($l, $color);
			}
		}
		elseif(strpos($line, "\n") !== false){
			// This is a multi-line string, re-run this script on each line for proper formatting.
			$lines = explode("\n", $line);
			foreach($lines as $l){
				self::PrintLine($l, $color);
			}
		}
		elseif(EXEC_MODE === 'WEB'){
			echo $color . htmlentities($line) . COLOR_RESET . NL . '<br/>';

			switch($color){
				case COLOR_DEBUG:
					echo '<!--CLI-DATA:LINE=' . $line . ';TYPE=debug-->';
					break;
				case COLOR_ERROR:
					echo '<!--CLI-DATA:LINE=' . $line . ';TYPE=error-->';
					break;
				case COLOR_WARNING:
					echo '<!--CLI-DATA:LINE=' . $line . ';TYPE=warning-->';
					break;
				case COLOR_SUCCESS:
					echo '<!--CLI-DATA:LINE=' . $line . ';TYPE=success-->';
					break;
				default:
					echo '<!--CLI-DATA:LINE=' . $line . '-->';
					break;
			}

			if(self::_FlushRequired()){
				ob_flush();
				if(phpversion() < '7.2.19') {
					flush();
				}
			}
		}
		else{
			echo $color . $line . COLOR_RESET . NL;
		}
	}

	/**
	 * Print an error to stdout
	 * 
	 * @param string $line
	 */
	public static function PrintError($line) {
		self::PrintLine($line, COLOR_ERROR);
	}

	/**
	 * Print a success message to stdout
	 * 
	 * @param string $line
	 */
	public static function PrintSuccess($line) {
		self::PrintLine($line, COLOR_SUCCESS);
	}

	/**
	 * Print a warning to stdout
	 * 
	 * @param string $line
	 */
	public static function PrintWarning($line) {
		self::PrintLine($line, COLOR_WARNING);
	}

	/**
	 * Print a debug message to stdout
	 * 
	 * @param string $line
	 */
	public static function PrintDebug($line) {
		self::PrintLine($line, COLOR_DEBUG);
	}

	/**
	 * Print an action start line
	 * 
	 * This is usually rendered in the format of 
	 * 
	 * ```
	 * Performing Some Action ...           [ OK ]
	 * ```
	 * 
	 * @param        $line
	 * @param string $suffix
	 */
	public static function PrintActionStart($line, string $suffix = null) {
		// Use NULL for defaults instead of explicit numbers, makes skipping easier.
		if($suffix === null) {
			$suffix = '...';
		}

		$flen = strlen($line) + strlen($suffix) + 8;
		echo $line .
			$suffix .
			str_repeat(NBSP, max(self::$MaxLineWidth - $flen, 1));
		
		if(EXEC_MODE == 'WEB'){
			echo '<!--CLI-DATA:LINE=' . $line . '-->';	
		}

		if(self::_FlushRequired()){
			ob_flush();
			if(phpversion() < '7.2.19') {
				flush();
			}
		}
	}

	/**
	 * Print the result of a previous "ActionStart" command.
	 *
	 * The rendered text and color depends on the parameter provided as such:
	 *
	 * * '[  OK  ]': TRUE, 1, "OK", "TRUE"
	 * * '[  !!  ]': FALSE, 0, "FALSE"
	 * * '[ FAIL ]': "FAIL"
	 * * '[ SKIP ]': "SKIP"
	 * * '[ WARN ]': "WARN"
	 * * '[  ??  ]': anything else
	 *
	 * @param string|int|bool $status
	 */
	public static function PrintActionStatus($status) {
		$nl = (EXEC_MODE == 'WEB') ? NL . '<br/>' : NL;

		if($status === true || $status === 1){
			$status = 'OK';
		}
		elseif($status === false || $status === 0){
			$status = 'FALSE';
		}
		else {
			$status = strtoupper($status);
		}

		switch($status) {
			case 'OK':
			case 'TRUE':
				echo COLOR_SUCCESS . '[  OK  ]' . COLOR_RESET;
				break;

			case 'SKIP':
				echo '[ SKIP ]';
				break;

			case 'WARN':
				echo COLOR_WARNING . '[ WARN ]' . COLOR_RESET;
				break;

			case 'FAIL':
				echo COLOR_ERROR . '[ FAIL ]' . COLOR_RESET;
				break;

			case 'FALSE':
				echo COLOR_ERROR . '[  !!  ]' . COLOR_RESET;
				break;

			default:
				echo '[  ??  ]';
		}

		echo $nl;

		if(self::_FlushRequired()){
			ob_flush();
			if(phpversion() < '7.2.19') {
				flush();
			}
		}
	}

	/**
	 * Print a progress bar and/or a status update on an existing progress bar.
	 * 
	 * If an absolute value is set, then the bar is jumped to that new value.
	 * Any absolute value less than the current will reset the progress bar to a new line.
	 * 
	 * A relative percentage can be set via the prefix '+'.
	 * For example, '+1' will bump the progress bar up by 1%.
	 * 
	 * @param $percent
	 */
	public static function PrintProgressBar($percent, string $title) {
		static $last = 0;
		
		// Allow "+1" to be sent, to indicate 1 percent more.
		if(strpos($percent, '+') === 0){
			$percent = substr($percent, 1);
			$percent += $last;
		}

		// Remember the current percent, to allow "+.." to be used for relative positions.
		$last = $percent;

		// This progress bar displays up to (VARIABLE) characters.
		// Always start with at least 1 point on the start position.
		$newPos = max(1, round($percent / 100 * self::$MaxLineWidth));
		
		self::Clear();
		self::PrintHeader($title);
		
		// The line is formatted in the fashion:
		// =====         [percentage]
		// ============= [percentage] ==========

		// Half length is halfway of the viewport width; where the [..] text starts.
		$halfLength = floor(self::$MaxLineWidth / 2);

		// Max length of the first part of the progress bar
		$progress1Max = $halfLength - 3;

		// Start position of the second part of the progress bar
		$progress2Start = $halfLength + 4;

		// Length of the first part of the progress bar, goes to a max of the text.
		$progressLength1  = min($newPos, $progress1Max);

		// Length of the second part of the progress bar, fills the rest of the screen after the text.
		$progressLength2  = ($newPos > $progress2Start) ? $newPos - $progress2Start : 0;

		// Position of the indicator
		$indicatorStart = $halfLength - ($percent < 10 ? 2 : 1);

		// Post padding of the indicator
		$indicatorPadding2 = ($percent > 99) ? 1 : 2;

		// Trim the extra precision from the precentage.
		$percent = round($percent,0);
		
		if(EXEC_MODE === 'WEB') {
			CLI::PrintLine(round($percent, 0) . '% Complete');
			echo '<!--CLI-DATA:PROGRESSBAR=' . (int)$percent . '-->';
		}
		else {
			// Render the bar with all the calculated values!
			echo COLOR_SUCCESS . str_repeat('=', $progressLength1) . COLOR_RESET .
				str_repeat(' ', $indicatorStart - $progressLength1) . $percent . '%' .
				str_repeat(' ', $indicatorPadding2);

			// str_repeat will complain if the length is 0, so check that before blindly rendering the second bar
			if($progress2Start > 0) {
				echo COLOR_SUCCESS . str_repeat('=', $progressLength2) . COLOR_RESET;
			}
		}
		
		if(self::_FlushRequired()){
			ob_flush();
			if(phpversion() < '7.2.19') {
				flush();
			}
		}
	}


	/**
	 * This can be used to load a saved session from the user's home directory.
	 * It's useful for saving common per-user data across different executions.
	 *
	 * Since CLI scripts are per-user and localhost only, these can, and should be, saved locally.
	 *
	 * Note, no error is generated if file doesn't exist, but false is returned instead of true.
	 *
	 * @param string $filebasename The basename of the settings file, the .php is added automatically.
	 *
	 * @return array
	 */
	public static function LoadSettingsFile(string $filebasename, array $defaultVariables) : array {
		// @todo Is there a better way of getting the home directory of the user?
		$home = $_SERVER['HOME'];
		$dir  = $home . '/.cae2/';
		$file = $dir . $filebasename . '.php';

		$default = [];
		foreach($defaultVariables as $key) {
			$default[$key] = null;
		}

		if (!is_dir($dir)) {
			return $default;
		}
		if (!is_readable($file)) {
			return $default;
		}

		$variables = include_once($file);
		if(!is_array($variables)) {
			$variables = [];
		}

		foreach($defaultVariables as $key) {
			if(!isset($variables[$key])) {
				$variables[$key] = null;
			}
		}

		return $variables;
	}

	/**
	 * Save the user settings back to the settings file.
	 * Any parameter given after the first is written to the settings file.
	 *
	 * @param string $filebasename The basename of the settings file, the .php is added automatically.
	 * @param ... Any variables to save.
	 */
	public static function SaveSettingsFile(string $filebasename, array $variables) {
		// @todo Is there a better way of getting the home directory of the user?
		$home = $_SERVER['HOME'];
		$dir  = $home . '/.cae2/';
		$file = $dir . $filebasename . '.php';

		$out = '<?php' . NL . 'return [' . NL;

		foreach ($variables as $name => $val) {
			$out .= TAB . "'" . $name . "' => ";

			if (is_null($val)) $out .= 'NULL';
			elseif (is_numeric($val)) $out .= $val;
			elseif (is_string($val)) $out .= '"' . str_replace('"', '\\"', $val) . '"';
			elseif ($val === true) $out .= 'TRUE';
			elseif ($val === false) $out .= 'FALSE';
			elseif (is_array($val) || is_object($val)) $out .= 'unserialize("' . str_replace('"', '\\"', serialize($val)) . '")';
			else $out .= '""';

			// And the common line-endings for all variable types.
			$out .= ',' . NL;
		}

		$out .= '];' . NL;

		// This output will be saved to the file.
		if (!is_dir($dir)) mkdir($dir);
		file_put_contents($file, $out);
	}

	/**
	 * Render a header with the Core logo and some information about Core.
	 */
	public static function PrintLogoHeader() {
		$logo = file_get_contents(ROOT_PDIR . 'core/resources/logo.txt');
		$version = 'Version ' . \Core\Core::GetComponent('core')->getVersion();

		$logoColor = COLOR_BLK2;
		$nameColor = COLOR_HEADER;
		$versionColor = COLOR_DEBUG;
		$minWidth = 60;

		if(self::$MaxLineWidth <= $minWidth) {
			// Basic rendering, this is easy.
			echo $logoColor . $logo . COLOR_RESET .
				NL . NL .
				$nameColor . 'Core Plus' . COLOR_RESET . NL .
				'Development Framework' . NL .
				$versionColor . $version . COLOR_RESET . NL . NL .
				'https://corepl.us' . NL;
		}
		else {
			// Complex rendering; they're side-by-side.
			$logoLines = explode("\n", $logo);
			foreach($logoLines as $lineNumber => $logoSnippet) {
				echo $logoColor . $logoSnippet . COLOR_RESET;

				if($lineNumber === 3) {
					$text = 'Core Plus';
					$color = $nameColor;
				}
				elseif($lineNumber === 4) {
					$text = 'Development Framework';
					$color = COLOR_NORMAL;
				}
				elseif($lineNumber === 6) {
					$text = $version;
					$color = $versionColor;
				}
				elseif($lineNumber === 8) {
					$text = 'https://corepl.us';
					$color = COLOR_NORMAL;
				}
				else {
					$text = null;
				}

				if($text) {
					// There's supplemental text, render that right aligned.
					echo str_repeat(
						' ', self::$MaxLineWidth - strlen($logoSnippet) - strlen($text)
					);
					echo $color . $text . COLOR_RESET;
				}

				echo NL;
			}
		}
	}

	/**
	 * Get if an obflush is required to send output to the end browser.
	 * 
	 * If another output buffer is active other than the primary one, do not flush anything!
	 */
	private static function _FlushRequired(){
		if(EXEC_MODE != 'WEB'){
			// ob_flush is only required when in web.
			return false;
		}
		
		$l = ob_get_level();
		if($l == 0){
			// No output buffers present!  YAY!
			return true;
		}
		
		if($l == 1 && ini_get('output_buffering')){
			// 1 is also allowed if output buffering is turned on.
			return true;
		}
		
		// Otherwise?
		return false;
	}
	
	/**
	 * Display a Dictionary as a form element for the user to select from.
	 * 
	 * The resulting key the user selected is returned.
	 *
	 * If the user cancels out, false will be returned.
	 * 
	 * @param Dict $list
	 * @param type $default
	 * @return string|false
	 */
	private static function _DisplayOptions(string $question, Dict $list, $default, string $description = null){
		// Build the list of options that will be used for rendering as columns if necessary.
		$counter    = 0;
		$options    = [];
		$answers    = [];
		$total      = count($list);
		$indentOnes = $total >= 100 ? '  ' : ($total >= 10 ? ' ' : '');
		$indentTens = $total >= 100 ? ' ' : '';
		$extras     = [];
		$optsUnfmtd = [];
		
		
		if($total === 0){
			self::PrintError('No options to select from :(');
			return '';
		}

		foreach($list as $key => $text){
			if(preg_match('/^_[A-Z]*_$/', $key) && $key === $text){
				// Special match for "extra" actions supported by the calling script.
				// This is useful if you need to include an option to quit, go back, etc.
				$text = strtolower(substr($text, 1, -1));
				
				$extras[]               = $text[0];
				$answers[ $text[0] ]    = $key;
				$options[]              = $indentOnes . $text[0] . ' - ' . $text;
				$optsUnfmtd[ $text[0] ] = $text;
			}
			elseif(preg_match('/_[A-Za-z]_/', $text)){
				// The text contains a single key letter.
				$hotkey = strtolower(preg_replace('/.*_([A-Za-z])_.*/', '\1', $text));
				$textUn = preg_replace('/_([A-Za-z])_/', '\1', $text);
				$text   = preg_replace('/_([A-Za-z])_/', COLOR_PRP2 . '\1' . COLOR_RESET, $text);
				
				$extras[]              = $hotkey;
				$answers[ $hotkey ]    = $key;
				$options[]             = $indentOnes . $hotkey . ' - ' . $text;
				$optsUnfmtd[ $hotkey ] = $textUn;
			}
			else{
				++ $counter;
				
				$answers[ $counter ] = $key;
				$options[] = ($counter < 10 ? $indentOnes : ( $counter < 100 ? $indentTens : '' )) .
					$counter .
					($default !== null && $default === $key ? '*' : ' ') .
					'- ' .
					$text .
					($default !== null && $default === $key ? ' (default)' : '');
				$optsUnfmtd[ $counter ] = $text;
			}
		}
		
		if(self::_ZenityAvailable()){
			$opts = '';
			foreach($optsUnfmtd as $title){
				$opts .= ' FALSE ' . escapeshellarg($title);
			}
			$opts = substr($opts, 1); // Trim the first " "
			$line = self::_DisplayZenity(
				'list',
				[
					'separator'   => ' ',
					'radiolist'   => false,
					'hide-header' => false,
					'title'       => $question,
					'text'        => $description,
				],
				'--column SEL --column TITLE ' . $opts
			);
			
			// Remap this to the value of the title.
			$line = array_search($line, $optsUnfmtd);
		}
		else{
			echo NL . $question . NL;
		
			// Get the dimension of the browser currently.
			$rows = trim(exec('tput lines')) - 6;
			$cols = trim(exec('tput cols'));
			$columnWidth = max(array_map('strlen', $options)) + 4;

			if($total <= $rows){
				// No formatting required, just output them.
				foreach($options as $line){
					echo '    ' . $line . NL;
				}
			}
			else{
				// Determine how many columns are required to fit it best.
				$maxColumns    = floor($cols / $columnWidth);
				$bestColumns   = ceil($total / $rows);
				$actualColumns = min($maxColumns, $bestColumns);

				$jump = ceil($total / $actualColumns);
				for($row = 0; $row < $jump; $row++){
					for($col = 0; $col < $actualColumns; $col++){
						echo '    ' . str_pad($options[ $col * $jump + $row ], $columnWidth, ' ');
					}
					echo NL;
				}
			}

			// Print the "enter a number 1-10..." text.
			if(count($extras)){
				sort($extras);

				if($counter > 1){
					$extras = array_merge(['1-' . $counter], $extras);
				}
				elseif($counter === 1){
					$extras = array_merge(['1'], $extras);
				}

				$last = array_pop($extras);

				echo NL . 'Enter ' . implode(', ', $extras) . ' or ' . $last;
			}
			else{
				echo NL . 'Enter 1-' . $counter;
			}

			if($default !== null){
				echo ' (or [ENTER] for ' . array_search($default, $answers) . ')';
			}
			
			echo ': ';
			
			// Read the response.
			//$line = readline();
			$line = strtolower(trim(fgets(STDIN)));
			echo NL;
		}


		if($line === false) {
			// Allow the user to cancel out, it's up to the calling script to understand this.
			return false;
		}
		elseif ($line === '' && $default !== null) {
			// Maybe there's a default.
			return $default;
		}
		elseif (!isset($answers[$line])) {
			// An answer was provided, but it was not a valid one.  Ask again!
			self::PrintError('Invalid Response!');
			sleep(1.5);
			
			// RECURSE!!!
			return self::_DisplayOptions($question, $list, $default, $description);
		}
		else {
			// Valid answer was provided.
			return $answers[$line];
		}
	}
	
	private static function _DisplayCheckbox(string $question, $default, string $description = null){
		
		echo NL . $question . NL;
		
		if($default === null){
			echo 'Enter y or n: ';
		}
		elseif($default === true){
			echo 'Enter Y or n ([ENTER] for yes): ';
		}
		elseif($default === false){
			echo 'Enter y or N ([ENTER] for no): ';
		}
		
		$line = strtolower(trim(fgets(STDIN)));
		echo NL;
		
		// Maybe there's a default.
		if ($line === '' && $default !== null) {
			return $default;
		}
		elseif ($line === 'y' || $line === 'yes') {
			return true;
		}
		elseif ($line === 'n' || $line === 'no') {
			return false;
		}
		else {
			self::PrintError('Invalid Response!');
			sleep(1.5);
			
			return self::_DisplayCheckbox($default);
		}
	}

	private static function _DisplayTableSelect(string $question, CLITable $answers, $default = null, $description = null) {
		if(self::_ZenityAvailable()) {
			// Build the Zenity command

			$keyMatches = [];
			$args = '';
			foreach($answers->getColumns() as $column) {
				$args .= ' --column=' . escapeshellarg($column['text']);
			}

			foreach($answers->getRecords() as $record) {
				// First key on each record is the index for selecting it.
				$key = $record[0]['text'];
				$selected = ($default !== null && $key == $default) ? 'TRUE' : 'FALSE';
				$keyMatch = [];

				foreach($record as $idx => $data) {
					if($idx === 0) {
						$args .= ' ' . $selected;
					}
					else {
						$args .= ' ' . escapeshellarg($data['text']);
						$keyMatch[] = $data['text'];
					}
				}

				// Keep track of the full text of the record to know which one the user selected.
				$keyMatches[ implode('|', $keyMatch) ] = $key;
			}

			$line = self::_DisplayZenity(
				'list',
				[
					'radiolist'    => false,
					'title'        => $question,
					'text'         => $description,
					'print-column' => 'ALL',
				],
				$args
			);

			// Reverse the selection to the key based on the string match.
			if(isset($keyMatches[$line])) {
				$line = $keyMatches[$line];
			}
			else {
				$line = null;
			}
		}
		else {
			$answers->render();

			$total = $answers->getTotal();
			echo NL. $question . '(1-' . $total . ')' . NL;
			if ($default !== null) {
				echo 'Press ENTER for [' . $default . ']' . NL;
			}

			$line = trim(fgets(STDIN));
			echo NL;

			if(!is_numeric($line) || $line < 1 || $line > $total) {
				// Invalid responses just get blanked out
				$line = null;
			}
		}

		if(
			($line === '' || $line === null) &&
			$default !== null
		){
			$line = $default;
		}

		return $line;
	}
	
	private static function _DisplayFileSelect(string $question, $default, string $description = null){
		if(self::_ZenityAvailable()){
			$line = self::_DisplayZenity(
				'file-selection',
				[
					'filename'   => $default,
					'title'      => $question,
					'text'       => $description,
				]
			);
		}
		else{
			echo NL . $question . NL;
		
			if($description){
				self::PrintDebug($description);
			}

			if ($default !== null) {
				echo 'Press ENTER for [' . $default . ']' . NL;
			}
			$line = trim(fgets(STDIN));
			echo NL;
		}
		
		if(
			($line === '' || $line === null) && 
			$default !== null
		){
			$line = $default;
		}
		
		if(!file_exists($line)){
			self::PrintError($line . ' does not appear to exist!');
			sleep(1.5);
			
			return self::_DisplayFileSelect($question, $default, $description);
		}
		
		return $line;
	}
	
	private static function _DisplayText(string $question, $default, bool $required = false, string $description = null){
		
		if(self::_ZenityAvailable()){
			$line = self::_DisplayZenity(
				'entry',
				[
					'title'      => $question,
					'text'       => $description,
					'entry-text' => $default,
				]
			);
		}
		else{
			echo NL . $question . NL;
		
			if($description){
				self::PrintDebug($description);
			}

			if ($default !== null) {
				echo "Press enter for [" . $default . "]" . NL;
			}
			$line = trim(fgets(STDIN));
			echo NL;
		}
		
		if(
			($line === '' || $line === null) && 
			$default !== null
		){
			$line = $default;
		}

		if($line === '' && $required){
			self::PrintError('Invalid Response!');
			sleep(1.5);
			return self::_DisplayText($question, $default, $required, $description);
		}
		else {
			return $line;
		}
	}
	
	private static function _DisplayTextarea(string $question, $default, string $description = null){
		
		$file = "/tmp/cae2-cli-textarea-" . \Core\random_hex(4) . '.tmp';
		file_put_contents($file, $default);
		
		if(self::_ZenityAvailable()){
			$data = self::_DisplayZenity(
				'text-info',
				[
					'editable'   => false,
					'title'      => $question,
					'text'       => $description,
					'filename'   => $file,
				]
			);

			// Zenity will set the content to NULL if the user clicks cancel, we want it to be FALSE instead.
			if($data === null) {
				$data = false;
			}
		}
		else{
			system($_SERVER['EDITOR'] . " " . $file . " > `tty`");
			
			// And read back in that file.
			$data = file_get_contents($file);
		}
		
		// Remove the file from the filesystem, no need for clutter.
		unlink($file);
		return $data;
	}

	/**
	 * Display a Zenity window with the requested type and options.
	 * @param string $type
	 * @param array $arguments
	 * @param string|null $append
	 * @return string|null
	 */
	private static function _DisplayZenity(string $type, array $arguments, string $append = null){
		$out = [];
		$ret = null;
		$cmd = 'zenity --' . $type;
		foreach($arguments as $key => $val){
			$cmd .= ' --' . $key . ($val === false ? '' : '=' . escapeshellarg($val));
		}
		if($append !== null){
			$cmd .= ' ' . $append;
		}
		$cmd .= ' 2>/dev/null';
		//CLI::PrintDebug($cmd);
		exec( $cmd, $out, $ret );
		
		//var_dump($out, $ret); die();
		
		if($ret !== 0){
			// User cancelled the operation, return default?
			return null;
		}
		else{
			return implode("\n", $out);
		}
	}
	
	/**
	 * Check to see if zenity is available and can be used
	 * 
	 * @return bool
	 */
	private static function _ZenityAvailable() : bool {
		return (
			trim(exec('which zenity 2>/dev/null')) !== '' &&
			isset($_SERVER['DISPLAY']) &&
			$_SERVER['DISPLAY'] !== ''
		);
	}
}