<?php
/**
 * File for class DatasetWhere definition in the coreplus project
 * 
 * @package Core\Datamodel
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20131022.1745
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Datamodel;


/**
 * An individual where item, consisting of a field key, an operation, and a value.
 * 
 * ### Example Usage:
 * 
 * ```{.syntax-php}
 * use \Core\Datamodel\DatasetWhere;
 * 
 * $item1 = new DatasetWhere('keyname = value');
 * 
 * $item2 = new DatasetWhere();
 * $item2->field = 'keyname';
 * $item2->op = '=';
 * $item2->value = 'value';
 * ```
 * 
 * @package Core\Datamodel
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class DatasetWhere{

	/**
	 * @var string Table for the base of this check
	 */
	public $table;

	/**
	 * @var string Field or alias to check
	 */
	public $field;

	/**
	 * @var string Operation of comparison
	 */
	public $op;

	/**
	 * @var string|int|float Value to search, parsed as unsafe data
	 */
	public $value;

	/**
	 * @var string Reference key to search, parsed as safe data
	 */
	public $valueKey;

	/**
	 * @var string Referenced table to search
	 */
	public $valueTable;

	const OP_EQ = '=';
	const OP_NOT_EQ = '!=';
	const OP_GT = '>';
	const OP_LT = '<';
	const OP_GE = '>=';
	const OP_LE = '<=';
	const OP_LIKE = 'LIKE';
	const OP_NOT_LIKE = 'NOT LIKE';
	const OP_IN = 'IN';
	const OP_NULL = 'NULL';
	const OP_NOT_NULL = 'NOT NULL';



	public function __construct($arguments = null){
		if($arguments){
			$this->parseWhere($arguments);
		}
	}

	private function _parseColumnName(string $name) : array {
		// Does the key include a tablename?
		if(($pos = strpos($name, '.')) !== false){
			$table = substr($name, 0, $pos);
			$field = substr($name, $pos + 1);
		}
		else{
			$table = null;
			$field = $name;
		}

		return [$table, $field];
	}

	public function setColumnName(string $name) : self {
		list($this->table, $this->field) = $this->_parseColumnName($name);
		return $this;
	}

	public function setValueColumnName(string $name) : self {
		list($this->valueTable, $this->valueKey) = $this->_parseColumnName($name);
		return $this;
	}

	public function op(string $operation) : self {
		$this->op = $operation;
		return $this;
	}

	/**
	 * Parse a single where statement for the key, operation, and value.
	 *
	 * @param string $statement The where statement to parse and evaluate
	 * @param bool $keyBasedValue Set to true if the value is a key reference
	 *
	 * @return void
	 *
	 * @throws DMIQueryException
	 */
	public function parseWhere(string $statement, bool $keyBasedValue = false) {
		// The user may have sent something like "blah = mep" or "datecreated < somedate"
		$operations = self::GetSupportedOperations();

		// First, extract out the key.  This is the simplest thing to look for.
		$k = preg_replace('/^([^ !=<>]*).*/s', '$1', $statement);
		$this->setColumnName($k);

		// and the rest of the query...
		$fragment = trim(substr($statement, strlen($k)));


		// Now I can sift through each operation and find the one that this query is.
		foreach($operations as $c) {
			// The match MUST be the first character.
			if(($pos = strpos($fragment, $c)) === 0) {
				$this->op = $c;
				$fragment = trim(substr($fragment, strlen($this->op)));

				if($this->op == 'IN') {
					// the IN statement has a bit of an extra functionality.
					// This expects a comma separated list of values.
					// If there are any spaces or parentheses, remove them first.
					$fragment = ltrim($fragment, " \t\n\r\0\x0B(");
					$fragment = rtrim($fragment, " \t\n\r\0\x0B)");

					$fragment = array_map('trim', explode(',', $fragment));
				}
				break;
			}
		}

		if($this->op === null) {
			throw new DMIQueryException('Unsupported operation for WHERE statement ' . $statement);
		}

		if($keyBasedValue) {
			$this->setValueColumnName($fragment);
		}
		else {
			$this->value = $fragment;
		}
	}

	public static function GetSupportedOperations() : array {
		return [
			self::OP_NOT_EQ,
			self::OP_LE,
			self::OP_GE,
			self::OP_EQ,
			self::OP_LT,
			self::OP_GT,
			self::OP_NOT_LIKE,
			self::OP_LIKE,
			self::OP_IN,
			self::OP_NOT_NULL,
			self::OP_NULL,
		];
	}

	public static function Init() : self {
		return new DatasetWhere();
	}
}