<?php
/**
 * File for class DatasetWhereClause definition in the coreplus project
 * 
 * @package Core\Datamodel
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20131022.1742
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Datamodel;


/**
 * The full WHERE clause for a dataset or Model.
 * 
 * This can be named, has a separator, and can contain multiple DatasetWhere items 
 * as well as multiple sub clauses.
 * 
 * @package Core\Datamodel
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class DatasetWhereClause{

	/**
	 * If multiple statements are contained herein, this is the separator of all statements.
	 *
	 * @var string
	 */
	private $_separator = 'AND';

	/**
	 * The array of statements (of groups) contained herein
	 *
	 * @var DatasetWhereClause[]|DatasetWhere[]
	 */
	private $_statements = array();


	/**
	 * The name of this group/clause.  Completely meaningless other than external lookups.
	 * (FUTURE FEATURE)
	 * @var string
	 */
	private $_name;

	/**
	 * @param string $name The name of this group/clause.  Completely meaningless other than external lookups.
	 */
	public function __construct($name = '_unnamed_'){
		$this->_name = $name;
	}

	/**
	 * Add a where statement by the three components.
	 * Only supports one where at a time, but useful for some of the tricky statements.
	 *
	 * @since 2.4.0
	 *
	 * @param $field
	 * @param $operation
	 * @param $value
	 */
	public function addWhereParts($field, $operation, $value){
		$c = new DatasetWhere();
		$c->field = $field;
		$c->op = $operation;
		$c->value = $value;
		$this->_statements[] = $c;

	}

	/**
	 * Add a where statement to this clause.
	 *
	 * @param $arguments
	 */
	public function addWhere($arguments){

		if(!is_array($arguments)) {
			$arguments = [ $arguments ];
		}

		foreach($arguments as $key => $argument) {
			if($argument instanceof DatasetWhereClause) {
				// Allow another clause to be sent in, that will be set as a child of this one.
				$this->_statements[] = $argument;
			}
			elseif($argument instanceof DatasetWhere) {
				// Allow a child statement to be passed in.
				$this->_statements[] = $argument;
			}
			elseif(!is_numeric($key) && is_array($argument)) {
				// Allow ['key' => [1, 2, 3]] to be passed in
				$statement = new DatasetWhere();
				$statement->field = $key;
				$statement->op    = 'IN';
				$statement->value = $argument;

				$this->_statements[] = $statement;
			}
			elseif(!is_numeric($key)) {
				// Allow ['key' => 'val', 'key2' => 'val2'] to be passed in too.
				$statement = new DatasetWhere();
				$statement->field = $key;
				$statement->op    = '=';
				$statement->value = $argument;

				$this->_statements[] = $statement;
			}
			else {
				// Allow just a plain ol string to be passed in too
				$this->_statements[] = new DatasetWhere($argument);
			}
		}
	}

	/**
	 * Shortcut function to add a subgroup to an existing group.
	 *
	 * @param $sep
	 * @param $arguments
	 */
	public function addWhereSub($sep, $arguments){
		$subgroup = new DatasetWhereClause();
		$subgroup->setSeparator($sep);
		$subgroup->addWhere($arguments);

		$this->addWhere($subgroup);
	}

	public function getStatements(){
		return $this->_statements;
	}

	public function setSeparator($sep){
		$sep = trim(strtoupper($sep));
		switch($sep){
			case 'AND':
			case 'OR':
				$this->_separator = $sep;
				break;
			default:
				throw new DMIException('Invalid separator, [' . $sep . ']');
		}
	}

	public function getSeparator(){
		return $this->_separator;
	}

	/**
	 * Sometimes you just want a good'ol "flat" representation.
	 */
	public function getAsArray(){
		$children = array();
		foreach($this->_statements as $s){
			if($s instanceof DatasetWhereClause){
				$children[] = $s->getAsArray();
			}
			elseif($s instanceof DatasetWhere){
				if($s->preformatted) {
					// Allow preformatted WHERE strings to be supported
					$children[] = $s->preformatted;
				}
				elseif($s->field !== null) {
					// Regular OO approach; there field is set.
					$children[] = $s->field . ' ' . $s->op . ' ' . $s->value;
				}
			}
		}
		return array('sep' => $this->_separator, 'children' => $children);
	}

	/**
	 * Get any/all statements that have a field set to that which is requested.
	 *
	 * Useful for looking up to see if a specific column has been set in a where statement.
	 *
	 * @param string $fieldname The field to search for
	 * @return array
	 */
	public function findByField($fieldname){
		$matches = array();
		foreach($this->_statements as $s){
			if($s instanceof DatasetWhereClause){
				$matches = array_merge($matches, $s->findByField($fieldname));
			}
			elseif($s instanceof DatasetWhere){
				if($s->field == $fieldname) $matches[] = $s;
			}
		}

		return $matches;
	}

}