<?php
/**
 * Email wrapper around the system mail utility in Core.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2018  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\EmailAPI;

use Exception;
use ReflectionClass;
use UserModel;
use Core\GPG\GPG;
use Core\Templates\Template;
use Core\Templates\TemplateInterface;
use Core\Core;
use Core\ConfigHandler;
use Core\View;
use Core\HookHandler;
use Core\Logger\LogEntry;
use Core\Cache;
use Core\Filestore\Factory as FileFactory;
use HTMLToMD\Converter;

/**
 * Description of Email
 *
 * @author charlie
 */
abstract class Email implements EmailInterface {
	
	/**
	 * The template to render this view with.
	 * Should be the partial path of the template, including emails/
	 *
	 * @example emails/mycomponent/thishappened.tpl
	 * @var string
	 */
	public $templatename;
	
	/**
	 * @var Template The template to render this email with.
	 */
	private $_template = null;
	
	/**
	 * @var string HTML-version of the body set from the application or template
	 */
	private $_htmlBody = null;
	
	/**
	 * @var string An optional source to track this email in the logging system.
	 */
	private $_source = '';
	
	/**
	 * @var bool Set to true to request GPG encryption when available.
	 */
	public $encrypted = false;
	
	/**
	 * @var string Cached key that can be wiped out on postSend.
	 */
	private $_stashed = null;
	
	public function __construct() {
		// Load in some default options for this email based on the configuration options.
		
		$user = \Core\user();
		
		// Previous version
		/*
		// Provide a default From: header.
		$fromEmail = ConfigHandler::Get('/core/email/from');
		// Ensure the email is set to something.
		if(!$fromEmail && isset($_SERVER['HTTP_HOST'])){
			$fromEmail = 'website@' . $_SERVER['HTTP_HOST'];
		}
		$fromName = ConfigHandler::Get('/core/email/from_name');
		$this->setFrom($fromEmail, $fromName);
		*/
		
		
		// New version w/Sender and From as different entries.
		// Provide a default From: header.
		$senderEmail = ConfigHandler::Get('/core/email/from');
		// Ensure the email is set to something.
		if(!$senderEmail && isset($_SERVER['HTTP_HOST'])){
			$domain = $_SERVER['HTTP_HOST'];
			if($domain === 'localhost') {
				// PHPMailer 6.x will complain about this.
				$domain = 'localhost.localdomain';
			}

			$senderEmail = 'website@' . $domain;
		}
		$this->setSender($senderEmail);
		
		if (ConfigHandler::Get('/core/email/dynamic_sender') && $user->exists() && $user->get('email')) {
			// Provide a default sender for when it's initiated by an authenticated user.
			if($user->get('first_name')) {
				// Prefer real names if available.
				$this->setFrom($user->get('email'), $user->get('first_name') . ' ' . $user->get('last_name'));
			}
			else{
				$this->setFrom($user->get('email'));
			}
		}
		else {
			// User not available or guest, send as the default web user.
			$this->setFrom($senderEmail, ConfigHandler::Get('/core/email/from_name'));
		}

		
		// Tack on some anti-abuse and tracking headers.
		$this->setHeader('X-AntiAbuse', 'This header was added to track abuse, please include it with any abuse report');
		if ($user->exists()) {
			$this->setHeader('X-AntiAbuse', 'User_id - ' . $user->get('id'));
			$this->setHeader('X-AntiAbuse', 'User_name - ' . $user->getDisplayName());
		}
		$this->setHeader('X-AntiAbuse', 'Original Domain - ' . SERVERNAME);
		$this->setHeader('X-AntiAbuse', 'Sitename - ' . SITENAME);
		$this->setHeader('MimeOLE', 'Core Plus');
		$this->setHeader('X-Content-Encoded-By', 'Core Plus ' . Core::GetComponent()->getVersion());
		$this->setHeader('X-Mailer', 'Core Plus ' . Core::GetComponent()->getVersion() . ' (https://corepl.us)');
	}

	/**
	 * Assign a value to this email template.
	 *
	 * Just serves as a pass-through for the Template::assign() method.
	 *
	 * @param string $key
	 * @param mixed  $val
	 */
	public function assign(string $key, $val) {
		$this->getTemplate()->assign($key, $val);
	}
	
	/**
	 * Get the main body, probably as HTML
	 * 
	 * @return string
	 */
	public function getBody() : string {
		if($this->_htmlBody !== null) {
			return $this->_htmlBody;
		}
		elseif($this->templatename && ($tpl = $this->getTemplate())) {
			return $tpl->fetch();
		}
		else {
			return '';
		}
	}

	/**
	 * Get the "source" for this email
	 * 
	 * @return string
	 */
	public function getSource(): string {
		return $this->_source;
	}
	
	/**
	 * Get the template responsible for rendering this email.
	 *
	 * @return TemplateInterface
	 */
	public function getTemplate() : TemplateInterface {
		if (!$this->_template) {
			$this->_template = Template::Factory($this->templatename);
		}

		return $this->_template;
	}

	/**
	 * Enable GPG encryption on this email.
	 * 
	 * The parameter is expected to be the public key fingerprint of the recipient.
	 *
	 * @param string $fingerprint
	 * 
	 * @return EmailInterface
	 */
	public function setEncryption(string $fingerprint): EmailInterface {
		$this->encrypted = true;
		
		return $this;
	}
	
	/**
	 * Set the "source" for this email to be recorded in the system logger.
	 * 
	 * @param string $source
	 * 
	 * @return EmailInterface
	 */
	public function setSource(string $source): EmailInterface {
		$this->_source = $source;
		
		return $this;
	}
	
	/**
	 * Set the main body as HTML and allow for auto-conversions 
	 * 
	 * Will auto-include all skin-level templates required.
	 * Will auto-convert from HTML to plain text.
	 * 
	 * @param $body string
	 */
	public function setBody(string $body) {
		$this->_htmlBody = $body;
	}

	/**
	 * Convenience function for sending an email to a User account
	 *
	 * @param \UserModel $user
	 *
	 * @return EmailInterface
	 */
	public function setToUser(\UserModel $user) : EmailInterface {
		$fname = $user->first_name;
		$lname = $user->last_name;
		$dname = $user->getDisplayName();
		$email = $user->email;

		if($fname && $lname) {
			$name = $fname . ' ' . $lname;
		}
		elseif($fname) {
			$name = $fname;
		}
		else {
			$name = $dname;
		}

		if($name === $email) {
			$name = '';
		}

		$this->setTo($email, $name);

		// Allow chaining
		return $this;
	}
	
	/**
	 * Stash this entire Email object in a cached object for future use
	 * 
	 * @return string The cache key that can be used to reload this Email
	 */
	public function stash() : string {
		if(!$this->_stashed) {
			// Generatea new stashed key if it's not already set; otherwise just re-use the same one.
			$this->_stashed = \Core\random_hex(24, true);
		}
		
		// Flatten the templating data so that it's not being pushed around various page loads.
		// This helps reduce the size and ensures a consistant variable state.
		if(
			$this->templatename && 
			($template = $this->getTemplate())
		){
			$this->_htmlBody = $template->fetch();
		}
		$this->templatename = '';
		$this->_template = null;
		//$this->_view = null;
		
		Cache::Set('email-' . $this->_stashed, serialize($this));
		
		return $this->_stashed;
	}
	
	/**
	 * Apply any pre-send logic required from Core
	 * 
	 * @return bool True/False if the email should be sent from the calling interface.
	 */
	protected function preSend() : bool {
		if(!ConfigHandler::Get('/core/email/enable_sending')){
			// Allow a config option to disable sending entirely.
			LogEntry::Init()
				->type('info')
				->icon('envelope-o')
				->code('/email/disabled')
				->source($this->_source)
				->message('Email sending is disabled, not sending email ' . $this->getSubject() . '!')
				->send();
			
			return false;
		}

		if(ConfigHandler::Get('/core/email/sandbox_to')){
			// Allow a sandbox to be declared, where all outbound emails are re-routed to this address.
			// This is useful in development.
			
			foreach($this->getTo() as $address){
				$this->setHeader('X-Original-To', $address->asString());
			}
			
			foreach($this->getCC() as $address){
				$this->setHeader('X-Original-CC', $address->asString());
			}
			
			foreach($this->getBCC() as $address){
				$this->setHeader('X-Original-BCC', $address->asString());
			}
			
			$this->clearRecipients();
			$this->setTo(ConfigHandler::Get('/core/email/sandbox_to'));
		}
		
		// If there is body content at the Core level of the email, (probable if there are templates used),
		// Ensure that is wrapped in the necessary template and populate the plain-text version in the underlying emailer.
		if(($htmlBody = $this->getBody())) {
			
			// Use the templating system to drop the body into the HTML template for the full HTML body version
			$view1 = new View();
			$view1->mode = View::MODE_EMAILORPRINT;
			$view1->contenttype = View::CTYPE_HTML;
			$view1->mastertemplate = 'email';
			$view1->setBody($htmlBody);
			$this->setMainBody($view1->fetch());
			
			// Now generate a plain text version of this template.
			// Use markdown for conversion;
			// it produces better results that phpMailer's built-in system!
			$converter = new Converter();

			$textBody = $converter->convert($htmlBody);
			
			$view2 = new View();
			$view2->mode = View::MODE_EMAILORPRINT;
			$view2->contenttype = View::CTYPE_PLAIN;
			$view2->mastertemplate = 'email';
			$view2->setBody($textBody);
			$this->setAltBody($view2->fetch());
		}
		
		return true;
	}
	
	/**
	 * Core logic to be applied after the subsystem has sent the message.
	 */
	protected function postSend() {
		
		// Allow external systems to be notified that an email has been sent.
		HookHandler::DispatchHook('/core/email/sent', $this);
		
		
		// Record this in the system log.
		$to = [];
		foreach($this->getTo() as $address){
			$to[] = $address->address;
		}
		$to = implode(',', $to);
		
		LogEntry::Init()
			->icon('envelope-o')
			->code('/email/sent')
			->source($this->_source)
			->type('info')
			->message('Sent ' . $this->getSubject() . ' to ' . $to)
			->send();
		
		
		// If this was stached in the Cache, clean that up
		if($this->_stashed) {
			Cache::Delete('email-' . $this->_stashed);
		}
		
		// If there are temporary attachments, clear them out now.
		$attachments = $this->getAttachments();
		foreach($attachments as $attachment) {
			$file = FileFactory::File($attachment->filepath);
			if($file->_type === 'tmp') {
				$file->delete();
			}
		}
	}

	
	/**
	 * Get an array with classname => title of all registered email providers available.
	 * 
	 * @return array
	 */
	public static function GetBackends(){
		$components = Core::GetComponents();
		$backends = [];
		foreach($components as $c){
			$backends = array_merge($backends, $c->getEmailBackends());
		}
		
		return $backends;
	}
	
	/**
	 * Create a new Email object based on the configured backend that can be used to send.
	 * 
	 * @param string $templatename Optional template to use for rendering this email
	 * 
	 * @return Email
	 * 
	 * @throws Exception
	 */
	public static function Factory(string $templatename = '') : Email {
		$backend = ConfigHandler::Get('/core/email/mailer');
		
		if($backend === '') {
			throw new Exception('No email provider configured, unable to send emails.');
		}
		
		if(class_exists($backend)){
			$ref = new ReflectionClass($backend);
			$instance = $ref->newInstance();
			
			if($templatename) {
				$instance->templatename = $templatename;
			}
			
			return $instance;
		}
		else{
			throw new Exception('Email backend provider ' . $backend . ' not available.');
		}
	}
	
	/**
	 * Unstash an Email from system cache and reload the Email object
	 * 
	 * @param string $key Cached key to reload
	 * 
	 * @return Email
	 * 
	 * @throws Exception
	 */
	public static function Unstash(string $key) : Email {
		$stashed = Cache::Get('email-' . $key);
		
		if($stashed) {
			return unserialize($stashed);
		}
		else{
			// *shrugs*
			throw new Exception('Email not found');
		}
	}
}

