<?php
/**
 * File for class Factory definition in the coreplus project
 * 
 * @package Core\Filestore
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20130604.2130
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Filestore;

use Core\Filestore\Backends;
use Core\Filestore\Contents\ContentSVG;
use Exception;


/**
 * A short teaser of what Factory does.
 *
 * More lengthy description of what Factory does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for Factory
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @package Core\Filestore
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
abstract class Factory {
	/**
	 * Array of file objects that have been instantiated to act as a cache.
	 *
	 * @var array
	 */
	protected static $_Files = array();

	/**
	 * Array of file objects that have been instantiated to act as a cache.
	 *
	 * @var array
	 */
	protected static $_Directories = array();

	/**
	 * Cache of incoming URIs to the fully resolved version.
	 *
	 * @var array
	 */
	protected static $_ResolveCache = array();

	/**
	 * Static function to act as Factory for the underlying Filestore system.
	 * This will parse the incoming URI and return the appropriate type based on Core settings and filetype.
	 *
	 * @param $uri
	 *
	 * @return File
	 */
	public static function File($uri) {

		// GOGO caching ;)
		if(isset(self::$_ResolveCache[$uri])){
			$resolved = self::$_ResolveCache[$uri]->getFilename();
			if(isset(self::$_Files[$resolved])){
				return self::$_Files[$resolved];
			}
		}

		// base64 comes first.  If the filename is encoded in that, decode it first.
		//if (strpos($uri, 'base64:') === 0){
		//	$uri = base64_decode(substr($uri, 7));
		//}

		// Allow FTP files to be requested here!
		// This needs to be before the :// check, because technically FTP can be a remote file,
		// but it has extra functionality, (namely being able to write or perform other operations through FTP)
		if(strpos($uri, 'ftp://') === 0){
			// Don't cache remote files.
			return new Backends\FileFTP($uri);
		}


		if(ROOT_URL_NOSSL && strpos($uri, ROOT_URL_NOSSL) === 0){
			// If this is a local file, just the URL version.... allow that remap too!
			$uri = ROOT_PDIR . substr($uri, strlen(ROOT_URL_NOSSL));

			if(($pos = strpos($uri, '?')) !== false) {
				// Trim query string suffixes from the local file.
				$uri = substr($uri, 0, $pos);
			}
		}
		elseif(ROOT_URL_SSL && strpos($uri, ROOT_URL_SSL) === 0){
			// If this is a local file, just the URL version.... allow that remap too!
			$uri = ROOT_PDIR . substr($uri, strlen(ROOT_URL_SSL));

			if(($pos = strpos($uri, '?')) !== false) {
				// Trim query string suffixes from the local file.
				$uri = substr($uri, 0, $pos);
			}
		}

		// Allow remote files to be requested here too!
		if(strpos($uri, '://') !== false){
			// Don't cache remote files.
			return new Backends\FileRemote($uri);
		}


		$isCDN = false;
		if(strpos($uri, 'asset/') === 0){
			$isCDN = true;
			$uri = ASSET_PDIR . substr($uri, 6);
		}
		elseif(strpos($uri, 'assets/') === 0){
			$isCDN = true;
			$uri = ASSET_PDIR . substr($uri, 7);
		}
		elseif(strpos($uri, ASSET_PDIR) === 0){
			$isCDN = true;
		}
		elseif(strpos($uri, 'public/') === 0){
			$isCDN = true;
			$uri = PUBLIC_PDIR . substr($uri, 7);
		}
		elseif(strpos($uri, PUBLIC_PDIR) === 0){
			$isCDN = true;
		}
		elseif(strpos($uri, 'private/') === 0){
			$isCDN = true;
			$uri = PRIVATE_PDIR . substr($uri, 8);
		}
		elseif(strpos($uri, PRIVATE_PDIR) === 0){
			$isCDN = true;
		}
		elseif(strpos($uri, 'tmp/') === 0){
			$uri = TMP_DIR . substr($uri, 4);
		}
		
		
		if($isCDN){
			// The CDN type configured will determine the backend.
			if(CDN_TYPE === 'local'){
				$file = new Backends\FileLocal($uri);
			}
			elseif(CDN_TYPE === 'ftp'){
				$file = new Backends\FileFTP($uri);
			}
			else{
				throw new Exception('File "' . $uri . '" appears to be a CDN-based file, but no CDN source available');
			}
		}
		else{
			$file = new Backends\FileLocal($uri);
		}

		// Cache this for future calls on this page load.
		self::$_Files[$file->getFilename()] = $file;
		return $file;
	}

	/**
	 * Static function to act as Factory for the underlying Filestore system.
	 * This will parse the incoming URI and return the appropriate type based on Core settings and filetype.
	 *
	 * @param $uri
	 *
	 * @return Directory
	 */
	public static function Directory($uri){
		//var_dump($uri);

		// base64 comes first.  If the filename is encoded in that, decode it first.
		if (strpos($uri, 'base64:') === 0){
			$uri = base64_decode(substr($uri, 7));
		}

		// Allow FTP files to be requested here!
		// This needs to be before the :// check, because technically FTP can be a remote file,
		// but it has extra functionality, (namely being able to write or perform other operations through FTP)
		if(strpos($uri, 'ftp://') === 0){
			return new Backends\DirectoryFTP($uri);
		}

		// Allow remote files to be requested here too!
		//if(strpos($uri, '://') !== false){
		//	return new Backends\FileRemote($uri);
		//}
		
		
		
		$isCDN = false;
		if(strpos($uri, 'asset/') === 0){
			$isCDN = true;
			$uri = ASSET_PDIR . substr($uri, 6);
		}
		elseif(strpos($uri, 'assets/') === 0){
			$isCDN = true;
			$uri = ASSET_PDIR . substr($uri, 7);
		}
		elseif(strpos($uri, ASSET_PDIR) === 0){
			$isCDN = true;
		}
		elseif(strpos($uri, 'public/') === 0){
			$isCDN = true;
			$uri = PUBLIC_PDIR . substr($uri, 7);
		}
		elseif(strpos($uri, PUBLIC_PDIR) === 0){
			$isCDN = true;
		}
		elseif(strpos($uri, 'private/') === 0){
			$isCDN = true;
			$uri = PRIVATE_PDIR . substr($uri, 8);
		}
		elseif(strpos($uri, PRIVATE_PDIR) === 0){
			$isCDN = true;
		}
		elseif(strpos($uri, 'tmp/') === 0){
			$uri = TMP_DIR . substr($uri, 4);
		}
		
		
		if($isCDN){
			// The CDN type configured will determine the backend.
			if(CDN_TYPE === 'local'){
				return new Backends\DirectoryLocal($uri);
			}
			elseif(CDN_TYPE === 'ftp'){
				return new Backends\DirectoryFTP($uri);
			}
			else{
				throw new Exception('Directory "' . $uri . '" appears to be a CDN-based location, but no CDN source available');
			}
		}
		else{
			return new Backends\DirectoryLocal($uri);
		}
	}

	
	/**
	 * Get the resolved Contents object for the given file.
	 * 
	 * This returns the specific Object type based on the mimetype of the File,
	 * so each different type of file may have different functionality.
	 * 
	 * @param \Core\Filestore\File $file
	 * 
	 * @return ContentsInterface
	 */
	public static function ContentsFromFile(File $file) : ContentsInterface {
		$ext = $file->getExtension();
		$mime = $file->getMimetype();
		
		switch ($mime) {
			case 'application/gzip':
			case 'application/x-gzip':
				// gzip can be a wrapper around a lot of things.
				// Some of them even have their own content functions.
				if ($ext === 'tgz' || $ext === 'tar.gz'){
					return new Contents\ContentTARGZ($file);
				}
				else{
					return new Contents\ContentGZ($file);
				}

			case 'application/octet-stream':
				// These are fun... basically I'm relying on the extension here.
				if($ext == 'zip'){
					return new Contents\ContentZIP($file);
				}
				else{
					error_log('@fixme Unknown extension for application/octet-stream mimetype [' . $ext . ']');
					return new Contents\ContentUnknown($file);
				}
			
			case 'application/pgp-signature':
			case 'application/pgp':
				return new Contents\ContentASC($file);

			case 'application/zip':
				return new Contents\ContentZIP($file);

			case 'image/svg+xml':
				return new ContentSVG($file);
				
			case 'image/png':
			case 'image/jpeg':
			case 'image/gif':
				return new Contents\ContentImage($file);

			case 'text/plain':
				// Sometimes these are actually other files based on the extension.
				if ($ext === 'asc'){
					return new Contents\ContentASC($file);
				}
				elseif($ext === 'php'){
					return new Contents\ContentPHP($file);
				}
				else{
					return new Contents\ContentUnknown($file);
				}
			
			case 'text/x-php':
				return new Contents\ContentPHP($file);

			case 'text/xml':
			case 'application/xml':
				return new Contents\ContentXML($file);

			case 'text/csv':
				return new Contents\ContentCSV($file);

			default:
				error_log('@fixme Unknown file mimetype [' . $mime . '] with extension [' . $ext . ']');
				return new Contents\ContentUnknown($file);
		}
	}

	/**
	 * If a file needs to be removed from cache, (ie it was renamed, deleted, etc)
	 * this method should be called to ensure that a future call doesn't use a corrupt/incorrect file!
	 *
	 * @param $file string|File
	 */
	public static function RemoveFromCache($file) {
		if($file instanceof File){
			$filename = $file->getFilename();
		}
		else{
			$filename = $file;
		}

		// Is this file resolved already?
		if(isset(self::$_Files[$filename])){
			// Note, unsetting the object will not purge it from memory!
			// If another method is using that memory space, then it'll remain as a valid object.
			unset(self::$_Files[$filename]);
		}

		// And lookup the lookup cache.
		$keys = array_keys(self::$_ResolveCache, $filename);
		foreach($keys as $k){
			unset(self::$_ResolveCache[$k]);
		}
	}
}