<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\Filestore;
use Core\I18N\I18NLoader;

/**
 * Description of FileHelper
 *
 * @author charlie
 */
final class FileHelper {
	private static $_MapExtToMime = [
		'atom'  => 'application/atom+xml',
		'css'   => 'text/css',
		'csv'   => 'text/csv',
		'eml'   => 'message/rfc822',
		'fgl'   => 'application/fgl+text',
		'gif'   => 'image/gif',
		'html'  => 'text/html',
		'htm'   => 'text/html',
		'ics'   => 'text/calendar',
		'jpg'   => 'image/jpeg',
		'jpeg'  => 'image/jpeg',
		'js'    => 'text/javascript',
		'json'  => 'application/json',
		'md'    => 'text/x-markdown',
		'otf'   => 'font/otf',
		'phtml' => 'text/html+php',
		'png'   => 'image/png',
		'rss'   => 'application/rss+xml',
		'svg'   => 'image/svg+xml',
		'tpl'   => 'text/html+smarty',
		'ttf'   => 'font/ttf',
		'txt'   => 'text/plain',
		'xhtml' => 'application/xhtml+xml',
		'xml'   => 'application/xml',
	];
	
	private static $_MapMimeToExt = [
		'application/atom+xml'  => 'atom',
		'application/fgl+text'  => 'fgl',
		'application/json'      => 'json',
		'application/rss+xml'   => 'rss',
		'application/xhtml+xml' => 'xhtml',
		'application/xml'       => 'xml',
		'font/otf'              => 'otf',
		'font/ttf'              => 'ttf',
		'image/gif'             => 'gif',
		'image/jpeg'            => 'jpeg',
		'image/png'             => 'png',
		'image/svg'             => 'svg',
		'image/svg+xml'         => 'svg',
		'message/rfc822'        => 'eml',
		'text/css'              => 'css',
		'text/csv'              => 'csv',
		'text/html'             => 'html',
		'text/calendar'         => 'ics',
		'text/javascript'       => 'js',
		'text/plain'            => 'txt',
		'text/x-markdown'       => 'md',
	];
	
	private static $_MimetypeIconAliases;
	
	/**
	 * Get an extension from a given filename.
	 *
	 * Will return just the extension itself without the ".", or a blank string if empty.
	 *
	 * @param $str
	 *
	 * @return string
	 */
	public static function GetExtension(string $filename) : string {
		// File doesn't have any extension... easy enough!
		if (strpos($filename, '.') === false){
			return '';
		}

		// strtolower for simplicity.
		$filename = strtolower($filename);

		$ext = substr($filename, strrpos($filename, '.') + 1);

		if($ext == 'gz' && substr($filename, -7) == '.tar.gz'){
			return 'tar.gz';
		}

		return $ext;
	}
	
	/**
	 * Get the mimetype icon for this file.
	 *
	 * @return File
	 */
	public static function GetIconFileForMimetype(string $mimetype) : File {
		$filemime = str_replace('/', '-', $mimetype);

		$file = Factory::File('assets/images/mimetypes/' . $filemime . '.svg');
		if(!$file->exists()){
			// Lookup if this mimetype has a generic icon alias
			$aliases = self::_GetMimetypeIconAliases();
			
			if(isset($aliases[ $mimetype ])){
				$file = Factory::File('assets/images/mimetypes/' . $aliases[ $mimetype ] . '.svg');
			}
			else{
				// Inform the developer, otherwise it's not a huge issue.
				\Core\log_debug('Unable to locate mimetype icon [' . $mimetype . '], resorting to "unknown"');
				$file = Factory::File('assets/images/mimetypes/unknown.svg');
			}
		}
		
		return $file;
	}
	
	/**
	 * Utility function to translate a filesize in bytes into a human-readable version.
	 *
	 * @param int $filesize Filesize in bytes
	 * @param int $round Precision to round to
	 *
	 * @return string
	 */
	public static function FormatSize($filesize, $round = 2) : string {
		$suf = array('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB');
		$c   = 0;
		while ($filesize >= 1024) {
			$c++;
			$filesize = $filesize / 1024;
		}

		return I18NLoader::FormatNumber($filesize, $round) . ' ' . $suf[$c];
	}
	
	/**
	 * Simple function to try to guess the mimetype for a given file extension.
	 * This is not guaranteed to be the actual mimetype of the file, but just the most common for a given extension.
	 *
	 * @param string $ext
	 *
	 * @return string
	 */
	public static function GuessMimeFromExt(string $ext) : string {
		
		return isset(self::$_MapExtToMime[ $ext ]) ? 
			self::$_MapExtToMime[ $ext ] : 
			'application/octet-stream';
	}

	/**
	 * Convert a common mimetype to its extension.
	 *
	 * @param $mimetype
	 *
	 * @return string
	 */
	public static function GuessExtFromMime(string $mimetype) : string {
		
		return isset(self::$_MapMimeToExt[ $mimetype ]) ? 
			self::$_MapMimeToExt[ $mimetype ] : 
			'';
	}
	
	private static function _GetMimetypeIconAliases() : array {
		if(self::$_MimetypeIconAliases === null){
			self::$_MimetypeIconAliases = yaml_parse_file(ROOT_PDIR . 'core/resources/mimetype_icon_aliases.yml');
		}
		
		return self::$_MimetypeIconAliases;
	}
}
