<?php
/**
 * Remote file backend.
 *
 * Provides standard API access to remote HTTP files.
 *
 * @package Core\Filestore\Backends
 * @since 2.5.6
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Filestore\Backends;

use Core\Utilities\Profiler\Profiler;
use Core\Utilities\Profiler\ProfilerEvent;
use Exception;
use Core\Cache;
use Core\Filestore\File;
use Core\Filestore\FileHelper;
use Core\Filestore\FileMetaHelper;
use Core\Filestore\Factory;
use Core\Filestore\ContentsInterface;


class FileRemote implements File {

	/**
	 * The username to use if basic authentication is required.
	 *
	 * @var string
	 */
	public $username = null;

	/**
	 * The password to use if basic authentication is required.
	 *
	 * @var string
	 */
	public $password = null;

	/**
	 * Set to false to require the remote file to be downloaded every request.
	 * @var bool
	 */
	public $cacheable = true;

	/**
	 * The fully resolved filename of this file.
	 *
	 * @var string
	 */
	private $_url = null;

	/**
	 * A key/value paired array of headers for this given URL.
	 * Useful for determining if a file exists without downloading it.
	 *
	 * @var array
	 */
	private $_headers = null;

	/**
	 * The response code for this file.
	 * Generally 200, 302, or 404.
	 *
	 * @var int
	 */
	private $_response = null;

	/**
	 * Temporary local version of the file.
	 * This is necessary for some operations such as "copyFrom" and "identicalTo"
	 *
	 * @var FileLocal
	 */
	private $_tmplocal = null;
	
	private $_requestHeaders = [];
	
	private $_method;
	
	private $_payload = null;

	/** @var string MD5 of the filename key, just used to save a calculation */
	private $_filenameKey;

	/**
	 * If the file was a 302, this is the temporary redirect placeholder.
	 *
	 * This is a separate file because according to the RFC2616 spec,
	 * requests that fall under a 302 are independent of each other and should be cached independently.
	 *
	 * @var null|FileRemote
	 */
	protected $_redirectFile = null;

	/**
	 * Level of redirect counts this file request is under.
	 *
	 * Used to prevent infinite redirect loops.
	 *
	 * @var int
	 */
	protected $_redirectCount = 0;

	public function __construct($filename = null) {
		if ($filename) {
			$this->setFilename($filename);
		}
	}

	/**
	 * Get the title of this file, either generated from the filename or pulled from the meta data as appropriate.
	 *
	 * @return string
	 */
	public function getTitle() : string {
		$metas = new FileMetaHelper($this);

		// If no title was set, I need to pick one by default.
		if(($t = $metas->getMetaTitle('title'))){
			return $t;
		}
		else{
			// Generate a moderately meaningful title from the filename.
			$title = $this->getBasename(true);
			$title = preg_replace('/[^a-zA-Z0-9 ]/', ' ', $title);
			$title = trim(preg_replace('/[ ]+/', ' ', $title));
			$title = ucwords($title);
			return $title;
		}
	}

	public function getFilesize($formatted = false) {
		$h = $this->_getHeaders();

		if(isset($h['Content-Length'])){
			// yay!
			$size = $h['Content-Length'];
		}
		else{
			// Well damn, download the file then...
			$tmp = $this->_getTmpLocal();
			$size = $tmp->getFilesize(false);
		}

		return ($formatted) ? FileHelper::FormatSize($size, 2) : $size;
	}

	public function getMimetype() : string {
		if (!$this->exists()) {
			return '';
		}

		$h    = $this->_getHeaders();
		$type = (isset($h['Content-Type'])) ? $h['Content-Type'] : '';

		return $type;
	}

	public function getExtension() : string {
		return FileHelper::GetExtension($this->getBasename());
	}

	/**
	 * Get the URL of this file
	 *
	 * @return string | null
	 */
	public function getURL() {

		// If basic authentication is required, this must be done in the URL.
		if ($this->username && $this->password) {
			$url = str_replace('://', '://' . $this->username . ':' . $this->password . '@', $this->_url);
		}
		elseif ($this->username) {
			$url = str_replace('://', '://' . $this->username . '@', $this->_url);
		}
		else {
			$url = $this->_url;
		}

		return $url;
	}

	/**
	 * Get the filename of this remote file.
	 */
	public function getFilename($prefix = ROOT_PDIR) : string {
		return $this->_url;
	}

	/**
	 * Get the base filename of this file.
	 */
	public function getBaseFilename($withoutext = false) : string {
		return $this->getBasename($withoutext);
	}

	/**
	 * Get the directory name of this file
	 *
	 * Will return the parent directory name, ending with a trailing slash.
	 *
	 * @return string
	 */
	public function getDirectoryName() : string {
		return dirname($this->getFilename()) . '/';
	}

	/**
	 * Get the filename for a local clone of this file.
	 * For local files, it's the same thing, but remote files will be copied to a temporary local location first.
	 *
	 * @return string
	 */
	public function getLocalFilename() : string {
		return $this->_getTmpLocal()->getFilename();
	}

	/**
	 * Get the File for a local clone of this file.
	 * For local files, it's the same thing, but remote files will be copied to a temporary local location first.
	 *
	 * @return File
	 */
	public function getLocalFile() : File {
		return $this->_getTmpLocal();
	}

	/**
	 * Get an ascii hash of the filename.
	 * useful for transposing this file to another page call.
	 *
	 * @return string The encoded string
	 */
	public function getFilenameHash() : string {
		return 'base64:' . base64_encode($this->_url);
	}

	/**
	 * Get the hash for this file.
	 */
	public function getHash() : string {
		// In order for this to work, I need to apply the operation on a local copy.
		$local = $this->_getTmpLocal();
		return $local->getHash();
	}

	public function delete() : bool {
		throw new Exception('Cannot delete a remote file!');
	}

	/**
	 * Copies the file to the requested destination.
	 * If the destination is a directory (ends with a '/'), the same filename is used, (if possible).
	 * If the destination is relative, ('.' or 'subdir/'), it is assumed relative to the current file.
	 *
	 * @param string|Filestore\File $dest
	 * @param boolean     $overwrite
	 *
	 * @return Filestore\File
	 */
	public function copyTo($dest, $overwrite = false) : File {
		//echo "Copying " . $this->_filename . " to " . $dest . "\n"; // DEBUG //

		if (!$dest instanceof File) {
			// Well it should be damnit!....
			$file = $dest;

			// Get the location of the destination, be it relative or absolute.
			// If the file does not start with a "/", assume it's relative to this current file.
			//if($file[0] != '/'){
			//	$file = dirname($this->_filename) . '/' . $file;
			//}

			// Is the destination a directory or filename?
			// If it's a directory just tack on this current file's basename.
			if (substr($file, -1) == '/') {
				$file .= $this->getBaseFilename();
			}

			// Now dest can be instantiated as a valid file object!
			$dest = Factory::File($file);
		}

		if ($this->identicalTo($dest)) return $dest;

		// GO!
		// The receiving function's logic will handle the rest.
		$dest->copyFrom($this, $overwrite);

		return $dest;
	}

	public function copyFrom(File $src, $overwrite = false) : bool {
		throw new Exception('Unable to write to remote files!');
	}

	public function getContents() {
		return $this->_getTmpLocal()->getContents();
	}

	public function putContents($data) : bool {
		throw new Exception('Unable to write to remote files!');
	}

	public function getContentsObject() : ContentsInterface {
		return Factory::ContentsFromFile($this);
	}

	public function isImage() : bool {
		$m = $this->getMimetype();
		return (preg_match('/image\/jpeg|image\/png|image\/gif/', $m) ? true : false);
	}

	public function isText() : bool {
		$m = $this->getMimetype();
		return (preg_match('/text\/.*|application\/x-shellscript/', $m) ? true : false);
	}

	/**
	 * Get if this file can be previewed in the web browser.
	 *
	 * @return boolean
	 */
	public function isPreviewable() : bool {
		return ($this->isImage());
	}

	/**
	 * Display a preview of this file to the browser.  Must be an image.
	 *
	 * @param string|int $dimensions A string of the dimensions to create the image at, widthxheight.
	 *                               Also supports the previous version of simply "dimension", as an int.
	 * @param boolean    $includeHeader Include the correct mimetype header or no.
	 */
	public function displayPreview($dimensions = "300x300", $includeHeader = true) {
		if (!$this->exists()) {
			// Log it so the admin knows that the file is missing, otherwise nothing is shown.
			error_log('File not found [ ' . $this->_url . ' ]', E_USER_NOTICE);

			// Return a 404 image.
			$file = Factory::File('asset/images/mimetypes/notfound.png');
			$file->displayPreview($dimensions, $includeHeader);
		}
		else{
			// Yay, just use the local copy to get the preview url :p
			$file = $this->_getTmpLocal();
			$file->displayPreview($dimensions, $includeHeader);
		}
	}

	public function getPreviewURL($dimensions = "300x300") : string {
		if (!$this->exists()) {
			// Log it so the admin knows that the file is missing, otherwise nothing is shown.
			error_log('File not found [ ' . $this->_url . ' ]', E_USER_NOTICE);

			// Return a 404 image.
			$file = Factory::File('asset/images/mimetypes/notfound.png');
			return $file->getPreviewURL($dimensions);
		}
		else{
			// Yay, just use the local copy to get the preview url :p
			$file = $this->_getTmpLocal();
			return $file->getPreviewURL($dimensions);
		}
	}

	/**
	 * See if this file is in the requested directory.
	 *
	 * @param $path string
	 *
	 * @return boolean
	 */
	public function inDirectory($path) : bool {
		// Just a simple strpos shortcut...
		return (strpos($this->_url, $path) !== false);
	}

	/**
	 * @param Filestore\File|string $otherfile
	 *
	 * @return bool
	 */
	public function identicalTo($otherfile) : bool {

		if ($otherfile instanceof File) {
			// Just compare the hashes.
			//var_dump($this->getHash(), $this, $otherfile->getHash(), $otherfile); die();
			return ($this->_getTmpLocal()->getHash() == $otherfile->getHash());
		}
		else {
			// Can't be the same if it doesn't exist!
			if (!file_exists($otherfile)){
				return false;
			}

			$result = exec('diff -q "' . $this->_getTmpLocal()->getFilename() . '" "' . $otherfile . '"', $array, $return);
			return ($return == 0);
		}
	}

	public function exists() : bool {
		$this->_getHeaders();

		return ($this->_response != 404);
	}

	public function isReadable() : bool {
		$this->_getHeaders();

		return ($this->_response != 404);
	}

	public function isOK(){
		$this->_getHeaders();

		return ($this->_response == 200 || $this->_response == 301 || $this->_response == 302);
	}

	public function requiresAuthentication(){
		$this->_getHeaders();

		return ($this->_response == 401 || $this->_response == 403);
	}

	/**
	 * Get the HTTP status code for this file.
	 *
	 * @return int
	 */
	public function getStatus(){
		$this->_getHeaders();
		return $this->_response;
	}

	public function isLocal() : bool {
		// Simple function that indicates if the file is on a local filesystem
		// Please note, even mounted filesystems are considered local for this matter.
		// Amazon S3 and other CDN services.... are not.

		return false;
	}

	public function getMTime() : int {
		// Remote files don't support modified time.
		// @todo Maybe they should! ;)
		return 0;
	}

	/**
	 * Get the type of resource, one of File::TYPE_* constants
	 *
	 * @return string
	 */
	public function getType() : string {
		return File::TYPE_OTHER;
	}

	/**
	 * Get the base filename of this file.
	 *
	 * This makes use of the Content-Disposition header as per RFC2616 Section 19.5 Item 1
	 * @url http://www.w3.org/Protocols/rfc2616/rfc2616-sec19.html
	 *
	 * For example:
	 * <pre>
	 *  Content-Type: image/jpeg
	 *  Content-Disposition: attachment; filename=genome.jpeg;
	 *      modification-date="Wed, 12 Feb 1997 16:29:51 -0500";
	 *  Content-Description: a complete map of the human genome
	 *
	 *  &lt;jpeg data&gt;
	 * </pre>
	 *
	 * @param boolean $withoutext Set to true to drop the extension.
	 *
	 * @return string
	 */
	public function getBasename($withoutext = false) : string {
		$basename = null;

		// First, check the Content-Disposition Header.
		$d = $this->_getHeader('Content-Disposition');
		if($d !== null) {
			// Content-Disposition WAS provided by the server!
			// This means that we can read that filename and know what the server wanted the file to be named.
			$dParts = explode(';', $d);
			foreach($dParts as $p) {
				if(strpos($p, 'filename=') !== false) {
					$value = trim(substr($p, strpos($p, '=') + 1), " '\"");

					// Any directories requested? (Directories here are NOT supported by the spec)
					$value = str_replace('/', '-', $value);

					$basename = $value;
				}
			}
		}

		if($basename === null && ($l = $this->_getHeader('Location'))){
			// This will also pull from the Location header if it's set.
			$basename = $l;
		}

		// Still no?  Just use the URL.
		if($basename === null){
			$basename = $this->getFilename();
		}

		if (strpos($basename, '?') !== false) {
			// Take off everything after the '?'.
			$basename = substr($basename, 0, strpos($basename, '?'));
		}

		// Drop off the directory and everything else.
		// (yes, basename somehow works on URLs too!)
		$basename = basename($basename);

		if ($withoutext) {
			$ext = $this->getExtension();
			if($ext != '') {
				return substr($basename, 0, (-1 - strlen($ext)));
			}
		}

		return $basename;
	}

	/**
	 * Rename this file to a new name
	 *
	 * @param $newname
	 *
	 * @return bool
	 */
	public function rename($newname) : bool {
		return false;
	}

	/**
	 * Get the mimetype icon for this file.
	 *
	 * @param string $dimensions
	 *
	 * @return string
	 */
	public function getMimetypeIconURL($dimensions = '32x32') : string {
		$filemime = str_replace('/', '-', $this->getMimetype());

		$file = Factory::File('assets/images/mimetypes/' . $filemime . '.png');
		if(!$file->exists()){
			if(DEVELOPMENT_MODE){
				// Inform the developer, otherwise it's not a huge issue.
				error_log('Unable to locate mimetype icon [' . $filemime . '], resorting to "unknown" (filename: ' . $this->getFilename('') . ')');
			}
			$file = Factory::File('assets/images/mimetypes/unknown.png');
		}
		return $file->getPreviewURL($dimensions);
	}

	/**
	 * Get the preview file object without actually populating the sources.
	 * This is useful for checking to see if the file exists before resizing it over.
	 *
	 * WARNING, this will NOT check if the file exists and/or copy data over!
	 *
	 * @param string $dimensions
	 *
	 * @return File
	 */
	public function getQuickPreviewFile($dimensions = '300x300') : File {
		return $this->_getTmpLocal()->getQuickPreviewFile($dimensions);
	}

	/**
	 * Get the preview file with the contents copied over resized/previewed.
	 *
	 * @param string $dimensions
	 *
	 * @return File
	 */
	public function getPreviewFile($dimensions = '300x300') : File {
		return $this->_getTmpLocal()->getPreviewFile($dimensions);
	}

	/**
	 * Check if this file is writable.
	 *
	 * @return boolean
	 */
	public function isWritable() : bool {
		return false;
	}
	
	public function setFilename($filename) {
		$this->_url = $filename;
	}

	/**
	 * Set the request method for this remote file
	 * 
	 * @param string $method
	 *
	 * @throws Exception
	 */
	public function setMethod(string $method){
		$method = strtoupper($method);
		$supported = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH'];

		if(!in_array($method, $supported)) {
			throw new Exception('Unsupported method: ' . $method);
		}

		$this->_method = $method;
	}

	/**
	 * Set the POST, PATCH, PUSH, or DELETE payload with an associative array
	 *
	 * Will be encoded appropriately on transport send.
	 *
	 * @param $data
	 * @throws Exception
	 */
	public function setPayload(array $data){
		if(!is_array($data)) {
			throw new Exception('Payloads must be an associative array.');
		}
		
		$this->_payload = $data;
	}

	/**
	 * Set a particular REQUEST header to this file.
	 *
	 * @param string $value The header key to set before the ':')
	 * @param string $key   The header value to set (after the ':')
	 */
	public function setRequestHeader(string $value, string $key){
		$this->_requestHeaders[ $value ] = $key;
	}

	/**
	 * Get the headers for this given file.
	 * This will go out and query the server with a HEAD request if no headers set otherwise.
	 * 
	 * Not applicable payloads
	 */
	protected function _getHeaders() {
		if ($this->_headers === null) {

			if($this->_payload) {
				return [];
			}

			$this->_headers = [];

			$response = $this->_performCURLRequest(true, false);
			$h = explode("\n", $response);

			// Read and process all the headers provided by the remote server
			foreach ($h as $line) {
				if (strpos($line, 'HTTP/1.') !== false) {
					$this->_response = substr($line, 9, 3);
				}
				elseif (strpos($line, ':') !== false) {
					$k                  = substr($line, 0, strpos($line, ':'));
					$v                  = trim(substr($line, strpos($line, ':') + 1));
					// Content-Type can have an embedded charset request.
					if($k == 'Content-Type' && strpos($v, 'charset=') !== false){
						$this->_headers['Charset'] = substr($v, strpos($v, 'charset=') + 8);
						$v = substr($v, 0, strpos($v, 'charset=') - 2);
					}
					$this->_headers[$k] = $v;
				}
			}

			if(($this->_response == '302' || $this->_response == '301') && isset($this->_headers['Location'])){
				/*
				From: http://www.ietf.org/rfc/rfc2616.txt and http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html

				10.3.3 302 Found

				The requested resource resides temporarily under a different URI.
				Since the redirection might be altered on occasion, the client SHOULD continue to use the Request-URI for future requests.
				This response is only cacheable if indicated by a Cache-Control or Expires header field.

				The temporary URI SHOULD be given by the Location field in the response.
				Unless the request method was HEAD,
				the entity of the response SHOULD contain a short hypertext note with a hyperlink to the new URI(s).

				If the 302 status code is received in response to a request other than GET or HEAD,
				the user agent MUST NOT automatically redirect the request unless it can be confirmed by the user,
				since this might change the conditions under which the request was issued.
				*/
				$newcount = $this->_redirectCount + 1;
				if($newcount <= 5) {
					$this->_redirectFile = new FileRemote();
					$this->_redirectFile->_redirectCount = $newcount;
					$this->_redirectFile->setFilename($this->_headers['Location']);
					$this->_redirectFile->_getHeaders();
				}
				else{
					trigger_error('Too many redirects when requesting ' . $this->getURL(), E_USER_WARNING);
				}
			}
		}

		if(($this->_response == '302' || $this->_response == '301') && $this->_redirectFile !== null){
			return $this->_redirectFile->_headers;
		}
		else{
			return $this->_headers;
		}

	}

	protected function _getHeader($header) {
		$h = $this->_getHeaders();
		return (isset($h[$header])) ? $h[$header] : null;
	}

	/**
	 * Get the temporary local version of the file.
	 * This is useful for doing operations such as hash and identicalto.
	 *
	 * @return FileLocal
	 */
	protected function _getTmpLocal() {
		if ($this->_tmplocal === null) {
			// Calculate the hash of the filename (URL), to have something to reference
			$this->_filenameKey = md5($this->getFilename());
			// The header cache is a cache of just the response headers
			$headerCache = Cache::Get('remotefile-cache-header-' . $this->_filenameKey);
			// Gets set to false if all cache ends up being good and we don't need to re-pull the file
			$downloadNeeded = true;


			if ($this->cacheable && $headerCache) {
				// If this file is marked as cacheable, (by default they ARE),
				// and there are response headers available, we can try to resolve if the cache is fresh enough.

				if(
					isset($headerCache['headers']['Expires']) &&
					strtotime($headerCache['headers']['Expires']) > time()
				) {
					// If the Expires is set and still valid, we can safely skip even checking for new headers,
					// as the remote server told us not to check until after the expiration date.
					// As such, preload the headers and response information
					$downloadNeeded = false;
					$this->_tmplocal = Factory::File($this->_calculateFilename());
					$this->_headers = $headerCache['headers'];
					$this->_response = $headerCache['response'];
				}
				elseif (
					isset($headerCache['headers']['ETag']) &&
					$this->_getHeader('ETag') &&
					$this->_getHeader('ETag') == $headerCache['headers']['ETag']
				) {
					// The ETag performs a HEAD lookup via getHeader, so it'll populate the headers and response
					$downloadNeeded = false;
					$this->_tmplocal = Factory::File($this->_calculateFilename());
				}
				elseif (
					isset($headerCache['headers']['Last-Modified']) &&
					$this->_getHeader('Last-Modified') &&
					$this->_getHeader('Last-Modified') == $headerCache['headers']['Last-Modified']
				) {
					// The Last-Modified performs a HEAD lookup via getHeader, so it'll populate the headers and response
					$downloadNeeded = false;
					$this->_tmplocal = Factory::File($this->_calculateFilename());
				}
			}

			if ($downloadNeeded) {
				// Make sure that the headers are updated, this is a requirement to use the 302 tag.
				// Since we're using caching, this doesn't actually repreform a HEAD request, but is
				// needed as a failsafe in case $this->cacheable is set to false.
				$this->_getHeaders();

				if(($this->_response == '302' || $this->_response == '301') && $this->_redirectFile !== null){
					// The headers detected an HTTP redirect, so return that instead.
					$this->_tmplocal = $this->_redirectFile->_getTmpLocal();
				}
				else{
					$this->_tmplocal = Factory::File($this->_calculateFilename());
					// Ensure this directory exists
					Factory::Directory($this->_tmplocal->getDirectoryName())->mkdir();
					// Use CURL to retrieve the body of the remote file.
					$this->_performCURLRequest(false, true, $this->_tmplocal->getFilename());
				}


				// And remember this header data for nexttime.
				Cache::Set(
					'remotefile-cache-header-' . $this->_filenameKey,
					[
						'headers'  => $this->_getHeaders(),
						'response' => $this->_response,
					]
				);
			}
		}

		return $this->_tmplocal;
	}

	/**
	 * Calculate a temporary filename for this request,
	 * using the file key and (if possible), provided filename.
	 *
	 * This is important because if we download a JPEG and pass it to an image processor,
	 * it generally needs to know what type of file it's dealing with, and file extensions are a quick
	 * way of providing hinting about the file contents.
	 *
	 * @return string
	 */
	private function _calculateFilename() : string {
		$contentDisposition = $this->_getHeader('Content-Disposition');
		$contentType = $this->_getHeader('Content-Type');
		$filename = null;
		$extension = null;

		if($contentDisposition && strpos($contentDisposition, 'filename=') !== false) {
			// A filename was requested in the response, use that!
			$filename = preg_replace('/.*filename="([^"]*)".*/', '$1', $contentDisposition);
		}

		if(!$filename && $contentType) {
			// No filename detected, try to at least calculate the extension.
			$extension = FileHelper::GuessExtFromMime($contentType);
		}

		// Now we can create a decently accurate filename to work with locally
		$filebase = $this->_filenameKey;
		if($filename) {
			// Filename was found, use that.
			$filebase .= '-' . basename($filename);
		}
		if($extension) {
			// No filename, but we do have an extension to work with.
			$filebase .= '.' . $extension;
		}

		return 'tmp/remotefile-cache/' . $filebase;
	}

	/**
	 * Perform the actual CURL request, optionally redirecting output directly to a file
	 *
	 * While technically both headers and body can be requested, it usually doesn't make sense to do so,
	 * as both are returned in a combined string.  Only set both to TRUE if you actually want a
	 * raw dump of the entire response.
	 *
	 * To save memory on large files, we have the option to dump directly to a file instead of
	 * loading the contents into memory.  Since the most common use case of this is to create a local
	 * clone of the file for processing, this is recommended.
	 *
	 * @param bool $headers            TRUE to pull header information via a HEAD request
	 * @param bool $body               TRUE to pull body information
	 * @param string|null $destination Filename of local file to directly write to
	 *
	 * @return bool|string
	 */
	private function _performCURLRequest(bool $headers, bool $body, string $destination = null) {
		$requestHeaders = \Core\get_standard_http_headers();
		if(count($this->_requestHeaders)) {
			$requestHeaders = array_merge($requestHeaders, $this->_requestHeaders);
		}

		// Flatten the request headers to a CURL-friendly string
		array_walk($requestHeaders, function(&$val, $key) {
			$val = $key . ': ' . $val;
		});

		// Calculate the request type
		if($headers && !$body) {
			// Headers requested but no body, this means a HEAD request.
			$method = 'HEAD';
		}
		elseif($this->_method) {
			// If method was explictly set, allow that to be used.
			$method = $this->_method;
		}
		elseif($this->_payload) {
			// Payloads default to a POST, but can be overridden via setMethod.
			$method = 'POST';
		}
		else {
			// Default method
			$method = 'GET';
		}

		// Start a profiler to keep track of this request
		$debugMessage = 'Performing CURL ' . $method . ' request' .
			' to ' . $this->getURL() .
			' (' . implode('; ', $requestHeaders) . ')';

		if($destination) {
			$debugMessage .= ' > ' . $destination;
		}

		$event = new ProfilerEvent($debugMessage);
		$event->start();

		$curl = curl_init();
		curl_setopt_array(
			$curl,
			array(
				CURLOPT_HEADER         => $headers,
				CURLOPT_NOBODY         => !$body,
				CURLOPT_RETURNTRANSFER => true,
				CURLOPT_CUSTOMREQUEST  => $method,
				CURLOPT_URL            => $this->getURL(),
				CURLOPT_HTTPHEADER     => $requestHeaders,
			)
		);

		if($destination) {
			// If a destination was requested, save response directly to save memory.
			$fp = fopen($destination, 'w+');
			curl_setopt($curl, CURLOPT_FILE, $fp);
		}

		if($this->_payload) {
			curl_setopt($curl, CURLOPT_POSTFIELDS, $this->_payload);
		}

		$result = curl_exec($curl);
		if($result === false){
			switch(curl_errno($curl)){
				// @todo Implement all 95+ error codes eventually.
				case CURLE_UNSUPPORTED_PROTOCOL:
				case CURLE_FAILED_INIT:
					$this->_response = 400;
					break;
				case CURLE_COULDNT_CONNECT:
				case CURLE_COULDNT_RESOLVE_HOST:
				case CURLE_COULDNT_RESOLVE_PROXY:
					$this->_response = 404;
					break;
				default:
					$this->_response = 500;
					break;
			}
		}

		curl_close($curl);

		if($destination && $fp) {
			// Don't forget to close the socket connection if we're saving the file directly.
			fclose($fp);
		}

		$event->stop();

		return $result;
	}

	/**
	 * Send a file to the user agent
	 *
	 * @param bool $forcedownload Set to true to force download instead of just sending the file.
	 *
	 * @throws \Exception
	 *
	 * @return void
	 */
	public function sendToUserAgent($forcedownload = false) {
		$view = \Core\view();
		$request = \Core\page_request();

		$view->mode = \View::MODE_NOOUTPUT;
		$view->contenttype = $this->getMimetype();
		$view->updated = $this->getMTime();
		if($forcedownload){
			$view->headers['Content-Disposition'] = 'attachment; filename="' . $this->getBasename() . '"';
			$view->headers['Cache-Control'] = 'no-cache, must-revalidate';
			$view->headers['Content-Transfer-Encoding'] = 'binary';
		}
		$view->headers['Content-Length'] = $this->getFilesize();

		// Send all the view headers
		$view->render();

		// And now the actual content if it's not a HEAD request.
		if($request->method != \PageRequest::METHOD_HEAD){
			echo $this->getContents();
		}
	}
}

