<?php
/**
 * All core Form objects in the system
 *
 * @package Core\Forms
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Forms;

use Core\BadRequestException;
use Core\Date\DateTime;
use Core\Model;
use Core\ModelValidationException;
use Core\Request;
use Core\Router;
use Core\Session;
use Core\View;
use Core\ViewStatusExceptionInterface;
use Exception;

/**
 * The main Form object.
 *
 * @package Core\Forms
 */
class Form extends FormGroup {

	/** @var string The original URL of the page this form was rendered on.  Used for security. */
	public $originalurl = '';

	/** @var string The referring page from this form.  Used for redirect purposes. */
	public $referrer = '';

	/** @var bool Set to true if this form is pulled from cached session data */
	public $cached = false;

	public $testmode = false;

	/**
	 * Standard mappings for 'text' to class of the FormElement.
	 * This can be extended, ie: wysiwyg or captcha.
	 *
	 * @var array
	 */
	public static $Mappings = array(
		'access'           => '\\Core\\Forms\\AccessStringInput',
		'button'           => '\\Core\\Forms\\ButtonInput',
		'buttoncreator'    => '\\Core\\Forms\\ButtonCreator',
		'checkbox'         => '\\Core\\Forms\\CheckboxInput',
		'checkboxes'       => '\\Core\\Forms\\CheckboxesInput',
		'color'            => '\\Core\\Forms\\ColorInput',
		'date'             => '\\Core\\Forms\\DateInput',
		'datetime'         => '\\Core\\Forms\\DateTimeInput',
		'email'            => '\\Core\\Forms\\EmailInput',
		'file'             => '\\Core\\Forms\\FileInput',
		'hidden'           => '\\Core\\Forms\\HiddenInput',
		'license'          => '\\Core\\Forms\\LicenseInput',
		'markdown'         => '\\Core\\Forms\\MarkdownInput',
		'message'          => '\\Core\\Forms\\MessageBox',
		'number'           => '\\Core\\Forms\\NumberInput',
		'pagemeta'         => '\\Core\\Forms\\PageMeta',
		'pagemetas'        => '\\Core\\Forms\\PageMetasInput',
		'pagemetaauthor'   => '\\Core\\Forms\\PageMetaAuthorInput',
		'pagemetakeywords' => '\\Core\\Forms\\PageMetaKeywordsInput',
		'pageparentselect' => '\\Core\\Forms\\PageParentSelectInput',
		'pagerewriteurl'   => '\\Core\\Forms\\PageRewriteURLInput',
		'pagepageselect'   => '\\Core\\Forms\\PagePageSelectInput',
		'pageinsertablegroup' => '\\Core\\Forms\\PageInsertableGroup',
		'password'         => '\\Core\\Forms\\PasswordInput',
		'radio'            => '\\Core\\Forms\\RadioInput',
		'reset'            => '\\Core\\Forms\\ResetInput',
		'select'           => '\\Core\\Forms\\SelectInput',
		'state'            => '\\Core\\Forms\\StateInput',
		'submit'           => '\\Core\\Forms\\SubmitInput',
		'system'           => '\\Core\\Forms\\SystemInput',
		'text'             => '\\Core\\Forms\\TextInput',
		'textarea'         => '\\Core\\Forms\\TextareaInput',
		'time'             => '\\Core\\Forms\\TimeInput',
		'user'             => '\\Core\\Forms\\UserInput',
		'wysiwyg'          => '\\Core\\Forms\\TextareaInput',
	);

	public static $GroupMappings = array(
		'default'          => '\\Core\\Forms\\FormGroup',
		'tabs'             => '\\Core\\Forms\\TabsGroup',
	);


	/**
	 * A cache of the actual models attached via addModel().
	 *
	 * @var array
	 */
	private $_models = array();


	/**
	 * Construct a new Form object
	 *
	 * @param array $atts Array of attribute to assign to this form off the bat.
	 */
	public function  __construct($atts = null) {

		if($atts === null){
			$atts = [];
		}
		// Some defaults
		if(!isset($atts['method'])) $atts['method'] = 'POST';
		if(!isset($atts['orientation'])) $atts['orientation'] = 'horizontal';

		parent::__construct($atts);

		$this->_validattributes = array('accept', 'accept-charset', 'action', 'enctype', 'id', 'method', 'name', 'target', 'style');

		// Will get set back to true on form submission for preserving the input values.
		$this->persistent = false;

		// Referrer is required to know where go to "back" to
		$this->referrer = Router::$DefaultRequest->referrer;
		$this->originalurl = ROOT_URL . substr(CUR_CALL, 1);

		// Create placeholders for validation and lookup keys.
		$this->addElement('hidden', ['name' => '__formid']);
		$this->addElement('hidden', ['name' => '__nonce']);
	}

	public function getTemplateName() {
		return 'forms/form.tpl';
	}

	/**
	 * Generate a unique hash for this form and return it as a flattened string.
	 *
	 * This is an MD5 hash of the callsmethod, any system inputs, the URL, and the user's private session key.
	 *
	 * @return string
	 */
	public function generateUniqueHash() : string {
		$bits = [];

		if($this->get('callsmethod')){
			$bits[] = $this->get('callsmethod');
		}

		foreach($this->getElements(true) as $el){
			if($el instanceof SystemInput ){
				$bits[] = $el->get('name') . ':' . $el->get('value');
			}
		}

		// Forms on a given page should be unique to that specific page.
		// This allows a from to be used across different pages without conflicting with each other.
		$bits[] = $this->originalurl;

		// Tack on the user's private session key to ensure this hash is difficult to repeat
		$bits[] = Session::GetPrivateKey();

		return md5(implode(';', $bits));
	}

	/**
	 * Render this form and all inside elements to valid HTML.
	 *
	 * This will also save the form to the session data for post-submission validation.
	 *  (if called with null or "foot")
	 *
	 * @param mixed $part "body|head|foot| or null
	 *        Render just a specific part of the form.  Useful for advanced usage.
	 *        null:     Render all of the form and its element.
	 *        "head":   Render just the beginning of the form, including the <form> opening tag.
	 *        "body":   Render just the body of the form, specifically the elements.
	 *        "foot":   Render just the end of the form, including the </form> closing tag.
	 *        "hidden": Render the hidden elements, these are still sent to the user agent for processing/submission.
	 *
	 * @return string (valid HTML)
	 */
	public function  render($part = null) {

		switch($part) {
			case null:
				$renderHead = true;
				$renderBody = true;
				$renderFoot = true;
				$renderCustom = false;
				break;
			case 'head':
				$renderHead = true;
				$renderBody = false;
				$renderFoot = false;
				$renderCustom = false;
				break;
			case 'body':
				$renderHead = false;
				$renderBody = true;
				$renderFoot = false;
				$renderCustom = false;
				break;
			case 'foot':
				$renderHead = false;
				$renderBody = false;
				$renderFoot = true;
				$renderCustom = false;
				break;
			case 'system':
				$renderHead = false;
				$renderBody = false;
				$renderFoot = false;
				$renderCustom = false;
				break;
			default:
				$renderHead = false;
				$renderBody = false;
				$renderFoot = false;
				$renderCustom = true;
		}


		$out = '';

		// Standard render option - header
		if($renderHead) {
			foreach ($this->getElements() as $e) {
				// Check and see if there are any elements in this form that require a fileupload.
				if ($e->requiresupload) {
					$this->set('enctype', 'multipart/form-data');
				}

				// Ensure the child element knows this parent form, in case it needs something here.
				$e->form = $this;
			}

			if($this->get('callsmethod')) {
				// Remap this submission to the automatic handler in Core's FormController.
				// This is required because as of Core 6.3, this is no longer an automatic check.
				$this->set('action', '/form/submit');
			}


			if($this->get('uniqueid')) {
				$formid = $this->get('uniqueid');
			}
			else {
				// Calculate the hash for this form so that it has a unique token!
				// As of Core API 6.3, all forms require a hash,
				// (whether or not the application uses them is another matter)
				//
				// This form ID is used as a unique key on the form to validate the submission
				// and to retrieve the form state once a submission is performed.
				$formid = $this->generateUniqueHash();
				$this->set('uniqueid', $formid);
			}


			// Set the value for the unique ID so it's preserved in the form.
			if($this->getElement('__formid')) {
				$this->getElement('__formid')->setValue($formid);
			}


			// Try to load any persistent form from the session before proceeding.
			// If this fails, it'll just continue as normal.
			$this->reloadFromSession();

			if($this->getElement('__nonce')) {
				// Generate the ID, but do not save anything, that will happen in the footer.
				// This way, the ID of the new nonce will at least be available in the form at the time of rendering.
				$nonce = \NonceModel::GenerateKey();
				$this->getElement('__nonce')->setValue($nonce);
			}


			$tpl = \Core\Templates\Template::Factory('forms/form-head');
			$tpl->assign('group', $this);
			$out .= $tpl->fetch();
		}

		// Standard render option - body
		if($renderBody) {
			$tpl = \Core\Templates\Template::Factory('forms/form-body');
			$els = '';
			// Fill in the elements
			foreach ($this->_elements as $e) {
				try {
					$els .= $e->render();
				}
				catch(\Exception $ex) {
					\Core\log_exception($ex);

					if(DEVELOPMENT_MODE) {
						$els .= $ex->getMessage();
					}
				}
			}

			$tpl->assign('group', $this);
			$tpl->assign('elements', $els);

			$out .= $tpl->fetch();
		}

		// Standard render option - foot
		if($renderFoot) {
			// Save the Nonce data
			if($this->getElement('__nonce')) {
				$this->saveToNonce();
			}

			$tpl = \Core\Templates\Template::Factory('forms/form-foot');
			$tpl->assign('group', $this);
			$out .= $tpl->fetch();
		}

		// Custom rendering option - system inputs
		if($part === 'system') {
			$el = $this->getElement('__formid');
			if($el) {
				$out .= $el->render();
			}

			$el = $this->getElement('__nonce');
			if($el) {
				$out .= $el->render();
			}
		}
		elseif($renderCustom) {
			$el = $this->getElement($part);
			if($el) {
				$out .= $el->render();
			}
		}

		return $out;
	}

	/**
	 * Render a given model's element based on its column name.
	 *
	 * Will expect the Model column to have a valid renderable input element.
	 *
	 * @param Model  $model      Model object to retrieve the information
	 * @param string $column     Column name to render as a form element
	 * @param array  $attributes Optional override attributes
	 *
	 * @return string
	 */
	public function renderModelElement(Model $model, string $column, array $attributes = []) : string {
		$element = $this->getElement($column);

		if(!$element) {
			// If the element is not already available, auto create it.

			try {
				$col = $model->getColumn($column);
			}
			catch(Exception $ex) {
				return $ex->getMessage();
			}

			$element = $col->getAsFormElement();

			if(!$element) {
				// Some columns will not be renderable as form elements.
				return '';
			}

			$this->addElement($element);
		}

		if(count($attributes)) {
			// There are override elements, set those before rendering.
			$element->setFromArray($attributes);
		}

		return $element->render();
	}

	/**
	 * Render an existing element
	 *
	 * @param string $name       Element name to render as a form element
	 * @param array  $attributes Optional override attributes
	 *
	 * @return string
	 */
	public function renderElement(string $name, array $attributes = []) : string {
		$element = $this->getElement($name);

		if(!$element) {
			return $name . ' element not created yet, please use renderNewElement instead.';
		}

		if(count($attributes)) {
			// There are override elements, set those before rendering.
			$element->setFromArray($attributes);
		}

		return $element->render();
	}

	/**
	 * Render a new element
	 *
	 * As fallback, can also render an existing element.
	 *
	 * @param string $type       Element type to render as a form element
	 * @param array  $attributes Optional override attributes
	 *
	 * @return string
	 */
	public function renderNewElement(string $type, array $attributes) : string {

		if(!isset($attributes['name'])) {
			return 'Missing required attribute, name';
		}

		$name = $attributes['name'];
		$element = $this->getElement($name);

		if(!$element) {
			// Expected behaviour.
			$element = FormElement::Factory($type, $attributes);
			$this->addElement($element);
		}
		else {
			if(count($attributes)) {
				// There are override elements, set those before rendering.
				$element->setFromArray($attributes);
			}
		}

		return $element->render();
	}

	/**
	 * Get a group by its name/title.
	 * Will create the group if it does not exist.
	 *
	 * @param string $name Name of the group to find/create
	 * @param string $type Type of group, used in conjunction with the GroupMappings array
	 * @return FormGroup
	 */
	public function getGroup($name, $type = 'default'){
		$element = $this->getElement($name);
		if(!$element){
			// Determine the type type.
			if(isset(self::$GroupMappings[$type])) $class = self::$GroupMappings[$type];
			else $class = '\\Core\\Forms\\FormGroup'; // Default.

			$ref = new \ReflectionClass($class);
			$element = $ref->newInstance(['name' => $name, 'title' => $name]);
			$this->addElement($element);
		}

		return $element;
	}

	/**
	 * Get the associated model for this form, if there is one.
	 * This model will also be populated automatically with all the data submitted.
	 *
	 * @param string $prefix The prefix name to lookup the model with.
	 *
	 * @return Model
	 */
	public function getModel($prefix = 'model') {

		// A model needs to be defined first of all...
		if(!isset($this->_models[$prefix])){
			return null;
		}
		/** @var $model Model */
		$model = $this->_models[$prefix];

		$model->setFromForm($this, $prefix);

		return $model;
	}

	/**
	 * Get all associated models for this form, if there are any.
	 *
	 * These models will also be populated automatically with all the data submitted.
	 *
	 * @return Model[]
	 */
	public function getModels() : array {
		$ret = [];

		foreach($this->_models as $prefix => $model) {
			$model->setFromForm($this, $prefix);
			$ret[ $prefix ] = $model;
		}

		return $ret;
	}

	/**
	 * Load this form's values from the provided array, usually GET or POST.
	 * This is really an internal function that should not be called externally.
	 *
	 * @param array   $src
	 * @param boolean $quiet Set to true to squelch errors.
	 */
	public function loadFrom($src, $quiet = false) {
		$els = $this->getElements(true, false);
		foreach ($els as $e) {
			/** @var $e FormElement */
			// Be sure to clear any errors from the previous page load....
			$e->clearError();

			if($e->get('disabled')){
				// Readonly elements cannot get written from the UA.
				continue;
			}

			if($e instanceof SystemInput) {
				// Skip system inputs, these cannot be remapped from incoming data.
				continue;
			}

			if(!$e->rendered) {
				// Skip any element not rendered,
				// these need to be ignored as the user never had the opportunity to modify them.
				continue;
			}

			$value = $e->lookupValueFrom($src);
			$e->set('value', $value);

			if ($e->hasError() && !$quiet){
				\Core\set_message($e->getError(), 'error');
			}
		}
	}

	/**
	 * Add a model's rendered elements to this form.
	 *
	 * All models must have a common prefix, generally this is "model", but if multiple models are on one form,
	 *  then different prefixes can be used.
	 *
	 * If $overrideElements is used, that array is used as the source of elements instead of
	 * querying the Model for.  This is useful if you need to change the default behaviour
	 * of the Form elements in the Controller prior to rendering.
	 *
	 * @param Model $model            Model to populate elements from
	 * @param string $prefix           Prefix to create elements as
	 * @param array  $overrideElements Array of Elements to override the Model's defined attributes.
	 */
	public function addModel(Model $model, string $prefix = 'model', array $overrideElements = null) {

		// Is this model already attached?
		if(isset($this->_models[$prefix])){
			return;
		}

		$this->_models[$prefix] = $model;

		if($overrideElements !== null){
			$elements = $overrideElements;
		}
		else{
			// The model can handle giving an array of form elements.
			$elements = $model->asFormArray();
		}

		foreach($elements as $k => $el){
			// Update the name as it will need to be prefixed with this model's prefix.
			if($prefix){
				$name = $el->get('name');
				if(preg_match('/^[a-zA-Z_]*\[/', $name)){
					// Name already contains blah[foo]
					// Translate that to $prefix[blah][foo].
					$name = $prefix . '[' . preg_replace('/^([a-zA-Z_]*)\[/', '$1][', $name);
				}
				else{
					$name = $prefix . '[' . $name . ']';
				}
				$el->set('name', $name);
			}

			if($el instanceof FormElement) {
				// I need to give the model a chance to act on this new element too.
				// Sometimes models may have a few special things to update on the element.
				// $model->setFromForm($this, $prefix);
				$model->setToFormElement($k, $el);
			}
			elseif($el instanceof FormGroup) {
				// Groups behave a little differently, but same idea.
				foreach($el->getElements() as $subel) {
					// Update the name as it will need to be prefixed with this model's prefix.
					if($prefix){
						$name = $subel->get('name');
						if(preg_match('/^[a-zA-Z_]*\[/', $name)){
							// Name already contains blah[foo]
							// Translate that to $prefix[blah][foo].
							$name = $prefix . '[' . preg_replace('/^([a-zA-Z_]*)\[/', '$1][', $name);
						}
						else{
							$name = $prefix . '[' . $name . ']';
						}
						$subel->set('name', $name);
					}
				}
			}


			$this->addElement($el);
		}

		// Anything else?
		$model->addToFormPost($this, $prefix);
	}

	/**
	 * Replace a Model with another version of the same type of Model.
	 *
	 * This is usually unnecessary, but some Models, (like the PageModel),
	 * may have different elements based on the settings on that Page.
	 *
	 * @param Model $model
	 * @param string $prefix
	 */
	public function replaceModel(Model $model, string $prefix = 'model') {
		if(isset($this->_models[ $prefix ])) {
			// Model already set, (that's good in this case)
			// Remove all elements that match this prefix, they should get recreated by add.
			$this->removeElementsByName('^' . $prefix . '\[');

			unset($this->_models[ $prefix ]);
		}

		// Now that the model has been removed, re-add it.
		$this->addModel($model, $prefix);

	}

	/**
	 * Add a given element to this form, (or group in this form).
	 * If the element as the "group" property, it will automatically be added to that respective group.
	 *
	 * @param       $element
	 * @param array $atts
	 */
	public function addElement($element, $atts = []){
		if($element instanceof FormElement) {
			// Ensure the child element knows this parent form, in case it needs something here.
			$element->form = $this;
		}
		elseif($element instanceof FormGroup) {
			// Ensure all child objects have the parent form set correctly.
			foreach($element->getElements(true, false) as $el) {
				$el->form = $this;
			}
		}

		// Group support! :)
		if(isset($atts['group'])){
			$grouptype = isset($atts['grouptype']) ? $atts['grouptype'] : 'default';

			$this->getGroup( $atts['group'], $grouptype )->addElement($element, $atts);
		}
		elseif($element instanceof FormElement && $element->get('group')){
			$grouptype = $element->get('grouptype') ? $element->get('grouptype') : 'default';

			$this->getGroup( $element->get('group'), $grouptype )->addElement($element, $atts);
		}
		elseif($element instanceof FormGroup && $element->get('group')) {
			$grouptype = $element->get('grouptype') ? $element->get('grouptype') : 'default';

			$this->getGroup( $element->get('group'), $grouptype )->addElement($element, $atts);
		}
		else{
			parent::addElement($element, $atts);
		}
	}

	/**
	 * Switch an element type from one to another.
	 * This is useful for doing some fine tuning on a pre-generated form, ie
	 *  a "string" field in the Model should be interperuted as an image upload.
	 *
	 * @param string $elementname The name of the element to switch
	 * @param string $newtype The standard name of the new element type
	 *
	 * @return boolean Return true on success, false on failure.
	 */
	public function switchElementType($elementname, $newtype) {
		$el = $this->getElement($elementname);
		if (!$el) return false;

		// Default.
		if (!isset(self::$Mappings[$newtype])) $newtype = 'text';

		$cls = self::$Mappings[$newtype];

		// If it's already the newtype, no change required.
		if (get_class($el) == $cls) return false;

		$atts = $el->getAsArray();

		// Don't need this one
		unset($atts['__class']);
		$newel = new $cls();
		$newel->setFromArray($atts);
		//var_dump($el, $atts, $newel, $newel->getInputAttributes());
		$this->switchElement($el, $newel);
		return true;
	}

	/**
	 * Save this Form as a short-lived Nonce and record the ID in the session under the form ID
	 * so it can be retrieved on the next GET request.
	 *
	 * Used when errors are incurred on the submit logic to redisplay the form back to the user
	 *
	 * @return void
	 */
	public function saveToSession() {
		$id = $this->get('uniqueid');

		if (!$id) {
			// If there's no unique ID available or generated, nothing can be saved.
			// This is alright if you're handling the form submission manually.
			return;
		}


		// Since this is meant to be a short lived token to persist just between page loads,
		// a short lifetime is sufficient for the persistent information.
		$nonce = \NonceModel::Generate('2 minutes', $id . '-' . REMOTE_IP, $this);
		Session::Set('FormData/' . $id, $nonce, 120);

		error_log('saveToSession Saving to Form: ' . $id . ' (Nonce: ' . $nonce . ')');
	}

	/**
	 * Save this Form to a Nonce object and return the ID of the saved Model
	 *
	 * This is the traditional retention method, (since Session proved unreliable),
	 * and is the method used on initial loading of Forms.
	 *
	 */
	public function saveToNonce() {
		$id = $this->get('uniqueid');

		if (!$id) {
			// If there's no unique ID available or generated, nothing can be saved.
			// This is alright if you're handling the form submission manually.
			return;
		}

		if($this->getElement('__nonce')) {
			$key = $this->getElementValue('__nonce');

			$nonce = new \NonceModel();
			$nonce->key = $key;
			$nonce->setExpiration('1 hour');
			$nonce->setHash($id . '-' . REMOTE_IP);
			$nonce->set('data', $this);

			$nonce->save();
		}
	}

	/**
	 * Reload a saved Form from the session data, useful for providing persistence across page views.
	 *
	 * @throws \Exception
	 */
	public function reloadFromSession() {
		$formid = $this->get('uniqueid');
		$savedID = Session::Get('FormData/' . $formid);
		if($savedID === null) {
			// No form saved, cool!
			// This happens on the first page view, and is perfectly acceptable.
			return;
		}

		// Attempt to deserialize the saved Form object.
		$nonce = new \NonceModel($savedID);
		if(!$nonce->isValid($formid . '-' . REMOTE_IP)) {
			// Nonce is not valid, complain and skip out.
			// This shouldn't happen, but who knows!
			\Core\log()
				->level(LOG_LEVEL_INFO)
				->type('error')
				->message('Form ' . $formid . ' was saved in the session but Nonce was not located!')
				->send();
			return;
		}

		// This should be the saved form state, which we should be able to use to provide a preserved state.
		$savedForm = $nonce->get('data');
		if(!($savedForm && $savedForm instanceof Form)) {
			\Core\log()
				->level(LOG_LEVEL_INFO)
				->type('error')
				->message('Form ' . $formid . ' had a Nonce key, but data saved was not a valid Form object or was otherwise corrupt')
				->send();
			return;
		}


		// Pre-process some fields that need to behave differently when remapped.
		foreach($this->getElements() as $el) {
			if(
				$el->get('name') &&
				$el->get('value') &&
				($el instanceof SubmitInput) &&
				$savedForm->getElement($el->get('name'))
			) {
				// Patch to keep the submit button's value, this gets blown away on dynamic submissions because
				// it wasn't actually used to submit the form, so its value will be NULL.
				$savedForm->getElement($el->get('name'))->set('value', $el->get('value'));
			}
			elseif(
				$el instanceof HiddenInput &&
				$el->get('name') === '__formid' &&
				$savedForm->getElement('__formid')
			) {
				// Form IDs need to be updated correctly as their ID may change.
				$savedForm->getElement('__formid')->set('value', $el->get('value'));
			}
		}


		// Overwrite some persistent variables from the stored data.
		$this->_elements = $savedForm->_elements;
		$this->referrer = $savedForm->referrer;
		$this->originalurl = $savedForm->originalurl;


		// Once loaded from the Nonce system, we can safely delete this key.
		// Once loaded, mark this form as used.
		// Probably not required, but might as well do cleanup.
		$nonce->markUsed();
		Session::UnsetKey('FormData/' . $formid);

		error_log('reloadFromSession Removed from Form: ' . $formid);
	}

	public function clearFromSession(){
		// If the unique hash has already been set, use that.
		// otherwise, generate it from the set elements.
		$hash = $this->get('uniqueid') ? $this->get('uniqueid') : $this->generateUniqueHash();

		error_log('clearFromSession Removing form from Form: ' . $hash);

		Session::UnsetKey('FormData/' . $hash);
	}

	/**
	 * Retrieve the submitted Form object populated with the values submitted
	 *
	 * This is called automatically when a Controller method has a 'Form' parameter,
	 * and gets populated with the contents of the saved Form.
	 *
	 * @param string $method The request method to check for the form ID
	 *
	 * @return Form
	 *
	 * @throws \Exception
	 */
	public static function GetSubmittedForm(string $method) : Form {
		// Where shall we check for the Form?
		switch($method) {
			case Request::METHOD_POST:
				$formid = filter_input(INPUT_POST, '__formid', FILTER_SANITIZE_STRING);
				$savedID = filter_input(INPUT_POST, '__nonce', FILTER_SANITIZE_STRING);
				break;
			case Request::METHOD_GET:
				$formid = filter_input(INPUT_GET, '__formid', FILTER_SANITIZE_STRING);
				$savedID = filter_input(INPUT_GET, '__nonce', FILTER_SANITIZE_STRING);
				break;
			default:
				throw new BadRequestException('Unsupported REQUEST method - ' . $method . ', please use POST or GET only!');
		}

		if(!$formid) {
			// If no formID token was provided, no form to lookup!
			throw new BadRequestException('No form token located from ' . $method . ' data');
		}

		if(!$savedID) {
			throw new BadRequestException('No Nonce ID was provided from ' . $method . ' data');
		}

		// There has to be data in the session.
		//$savedID = Session::Get('FormData/' . $formid);

		// Auto-cleanup the Session; it'll get put back if from the Form render method if it's required again.
		//Session::UnsetKey('FormData/' . $formid);

		/*if($savedID === null) {
			var_dump('formid', $formid);
			var_dump('Nonce ID', $savedID);
			var_dump('Session Data', $_SESSION['FormData']);
			// Requested form was not found in the session data.
			throw new BadRequestException('Form token provided, but not registered in server session.  Please go back and try again.');
		}*/

		// Load the NONCE that should contain this Form
		$nonce = new \NonceModel($savedID);
		if(!$nonce->isValid($formid . '-' . REMOTE_IP)) {

			\Core\log()
				->type('error')
				->level(LOG_LEVEL_INFO)
				->message('Submitted Form ID does not match Nonce')
				->send();
			
			throw new BadRequestException('Invalid form token submitted, form timed out, or your IP address changed.  Please go back and try again.');
		}

		// Attempt to deserialize the saved Form object.
		$form = $nonce->get('data');
		if(!($form && $form instanceof Form)) {

			\Core\log()
				->type('error')
				->level(LOG_LEVEL_HIGH)
				->message('Serialized Form object was corrupt!')
				->send();

			throw new BadRequestException('Stored form data located, but was not a valid object.  Please go back and try again; if you continue getting this issue contact the system administrator.');
		}

		// Ensure the submission types match up.
		if (strtoupper($form->get('method')) !== $method) {
			throw new BadRequestException('Request method does not match up expected form method');
		}

		if($method === Request::METHOD_POST) {
			//$form->loadFrom(filter_input_array(INPUT_POST, FILTER_SANITIZE_STRING));
			$form->loadFrom(filter_input_array(INPUT_POST, FILTER_UNSAFE_RAW), true);
		}
		else{
			//$form->loadFrom(filter_input_array(INPUT_GET, FILTER_SANITIZE_STRING));
			$form->loadFrom(filter_input_array(INPUT_GET, FILTER_UNSAFE_RAW), true);
		}

		if(!$form->testmode) {
			// Once loaded, mark this form as used.
			// Probably not required, but might as well do cleanup.
			// Only perform this if the form is not requested to be in test mode.
			$nonce->markUsed();
		}

		//error_log('GetSubmittedForm Loaded and removed from Form: ' . $formid);

		// Allow this form to be re-saved back to the session immediately if testmode is enabled on this form.
		// This allows developers to spam refresh on the submission logic without re-rendering the form.
		//if(DEVELOPMENT_MODE && $form->testmode) {
		//	Session::Set('FormData/' . $formid, serialize($form));
		//}

		return $form;
	}

	/**
	 * Scan through a standard Model object and populate elements with the correct fields and information.
	 *
	 * @param Model $model
	 *
	 * @return Form
	 */
	public static function BuildFromModel(Model $model) : Form {
		$f = new Form();
		$f->addModel($model);
		return $f;
	}
}
