<?php
/**
 * Collection of useful utilities that don't quite fit into any class.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core;

use Core\Datamodel\BackendInterface;
use Core\Datamodel\DMI;
use Core\Filestore\FTP\FTPConnection;
use Core\Filestore\Factory;
use Core\Filestore\FileMetaHelper;
use Core\Logger\LogEntry;
use Core\Filestore\File;
use Core\Renderer\ImageRenderer;
use Core\Filestore\FileHelper;
use Core\I18N\I18NLoader;
use Exception;
use MultiSiteHelper;
use PageModel;
use SimpleXMLElement;
use SystemLogModel;
use UserModel;
use Error;


/**
 * Shortcut function to get the current system database/datamodel interface.
 * @return \Core\Datamodel\BackendInterface;
 */
function db() : BackendInterface {
	return DMI::GetSystemDMI()->connection();
}

/**
 * Get the global FTP connection.
 *
 * This is used for local assets and public files by some sites.
 *
 * Returns the FTP resource or false on failure.
 *
 * @return FTPConnection | false
 */
function ftp() {
	static $ftp = null;

	if($ftp === null){

		if(!defined('FTP_USERNAME')){
			// Prevent the installer from freaking out.
			$ftp = false;
			return false;
		}

		if(!defined('FTP_PASSWORD')){
			// Prevent the installer from freaking out.
			$ftp = false;
			return false;
		}

		if(!defined('FTP_PATH')){
			// Prevent the installer from freaking out.
			$ftp = false;
			return false;
		}

		if(!FTP_USERNAME){
			// If there is no username for the FTP server, don't try to connect either.
			$ftp = false;
			return false;
		}

		$ftp = new FTPConnection();
		$ftp->host = '127.0.0.1';
		$ftp->username = FTP_USERNAME;
		$ftp->password = FTP_PASSWORD;
		$ftp->root = FTP_PATH;
		$ftp->url = ROOT_WDIR;

		try{
			$ftp->connect();
		}
		catch(Exception $e){
			\Core\ErrorManagement\exception_handler($e);
			$ftp = false;
			return false;
		}
	}

	if($ftp && $ftp instanceof FTPConnection){
		try{
			$ftp->reset();
		}
		catch(Exception $e){
			\Core\ErrorManagement\exception_handler($e);
			$ftp = false;
			return false;
		}
	}

	return $ftp;
}

/**
 * Get the current user model that is logged in.
 *
 * To support legacy systems, this will also return the User object if it's available instead.
 * This support is for < 2.8.x Core installations and will be removed after some amount of time TBD.
 *
 * If no user systems are currently available, null is returned.
 *
 * @throws \Exception
 *
 * @return UserModel
 */
function user() : UserModel {
	static $_CurrentUserAccount = null;

	if(!class_exists('UserModel')){
		throw new \Exception('UserModel is not available, cannot instantiate a User.');
	}

	if($_CurrentUserAccount !== null){
		// Cache this for the page load.
		return $_CurrentUserAccount;
	}

	// This is disabled until a secure version can be implemented.
	/*if(isset($_SERVER['HTTP_X_CORE_AUTH_KEY'])){
		// Allow an auth key to be used to authentication the requested user instead!
		$user = UserModel::Find(['apikey = ' . $_SERVER['HTTP_X_CORE_AUTH_KEY']], 1);
		if($user){
			$_CurrentUserAccount = $user;
		}
	}
	else*/
	if(Session::Get('user') instanceof UserModel){
		// There is a valid user account in the session!
		// But check if this user is forced to be resynced first.
		if(isset(Session::$Externals['user_forcesync'])){
			// A force sync was requested by something that modified the original UserModel object.
			// Keep the user logged in, but reload the data from the database.
			$_CurrentUserAccount = UserModel::Construct(Session::Get('user')->get('id'));
			// And cache this updated user model back to the session.
			Session::Set('user', $_CurrentUserAccount);
			unset(Session::$Externals['user_forcesync']);
		}
		else{
			$_CurrentUserAccount = Session::Get('user');
		}
	}

	if($_CurrentUserAccount === null){
		// No valid user found.
		$_CurrentUserAccount = new UserModel();
	}

	// If this is in multisite mode, blank out the access string cache too!
	// This is because siteA may have some groups, while siteB may have another.
	// We don't want a user going to a site they have full access to, hopping to another and having cached permissions!
	if(Core::IsComponentAvailable('multisite') && class_exists('MultiSiteHelper') && MultiSiteHelper::IsEnabled()){
		$_CurrentUserAccount->clearAccessStringCache();
	}

	// Did this user request sudo access for another user?
	if(Session::Get('user_sudo') !== null){
		$sudo = Session::Get('user_sudo');

		if($sudo instanceof UserModel){
			// It's a valid user!

			if($_CurrentUserAccount->checkAccess('p:/user/users/sudo')){
				// This user can SUDO!
				// (only if the other user is < SA or current == SA).
				if($sudo->checkAccess('g:admin') && !$_CurrentUserAccount->checkAccess('g:admin')){
					Session::UnsetKey('user_sudo');
					log()
						->level(LOG_LEVEL_INFO)
						->type('security')
						->code('/user/sudo')
						->message('Authorized sudo non-admin user tried to access an admin account!')
						->user($sudo)
						->send();
					// SystemLogModel::LogSecurityEvent('/user/sudo', 'Authorized but non-SA user requested sudo access to a system admin!', null, $sudo->get('id'));
				}
				else{
					// Ok, everything is good.
					// Remap the current user over to this sudo'd account!
					$_CurrentUserAccount = $sudo;
				}
			}
			else{
				// This user can NOT sudo!!!
				Session::UnsetKey('user_sudo');
				SystemLogModel::LogSecurityEvent('/user/sudo', 'Unauthorized user requested sudo access to another user!', null, $sudo->get('id'));
			}
		}
		else{
			Session::UnsetKey('user_sudo');
		}
	}

	return $_CurrentUserAccount;
}

/**
 * Get the brightness value of a given color, between 0 - 255.
 *
 * 0 - 122 is dark
 * 123 - 255 is bright
 *
 * @param string $color
 *
 * @return int
 */
function color_brightness_level(string $color) : int {
	// Allow / expect hex values, they may have the hex indicator.
	$color = trim(str_replace('#', '', $color));

	$red = hexdec(substr($color, 0, 2));
	$green = hexdec(substr($color, 2, 2));
	$blue = hexdec(substr($color, 4, 2));

	// Convert each value to their brightness level based on standard color values.
	$red *= 0.2126;
	$green *= 0.7152;
	$blue *= 0.0722;

	// Each gets added together to form the final value.
	$brightness = $red + $blue + $green;

	// The returning value expects an int, so do that.
	return (int)round($brightness, 0);
}


/**
 * Create a colored badge start snippet with the provided attributes
 *
 * @param string $color
 * @param array $attributes
 * @return string
 */
function badge_colored_background_snippet(string $color, array $attributes = []) : string {
	if(!isset($attributes['class'])) {
		$attributes['class'] = [];
	}
	$attributes['class'][] = 'badge';
	$attributes['class'][] = color_brightness_level($color) > 122 ? 'light-background' : 'dark-background';

	if(!isset($attributes['style'])) {
		$attributes['style'] = [];
	}
	$attributes['style']['background-color'] = $color;

	return '<span ' . array_to_html_attributes_string($attributes) . '>';
}


/**
 * Get the system View
 *
 * @return View
 */
function view() : View {
	return Router::$DefaultView;
}


/**
 * Get the standard HTTP request headers for retrieving remote files.
 *
 * @return array
 */
function get_standard_http_headers() : array {
	return [
		'User-Agent' => 'Core Plus ' . Core::GetComponent('core')->getVersion() . ' (https://corepl.us)',
		'Referer' => SERVERNAME,
	];
}


/**
 * Resolve an asset to a fully-resolved URL.
 *
 * @param string $asset
 *
 * @return string The full url of the asset, including the http://...
 */
function resolve_asset($asset) : string {
	// Allow already-resolved links to be returned verbatim.
	if(strpos($asset, '://') !== false){
		return $asset;
	}

	// Since an asset is just a file, I'll use the builtin file store system.
	// (although every file coming in should be assumed to be an asset, so
	//  allow for a partial path name to come in, assuming asset/).
	if(strpos($asset, 'assets/') !== 0 && strpos($asset, 'asset/') !== 0) {
		$asset = 'assets/' . $asset;
	}
	
	try{
		$file = Factory::File($asset);
	}
	catch (Exception $ex){
		\Core\log_exception($ex);
		return '';
	}
	
	return $file->getURL();
}

/**
 * Resolve a url or application path to a fully-resolved URL.
 *
 * This can also be an already-resolved link.  If so, no action is taken
 *  and the original URL is returned unchanged.
 * 
 * ### Examples
 * 
 * ```{.syntax-php}
 * // Resolves to the current page, useful for no-action and javascript links.
 * resolve_link('#');
 * 
 * // No change, already resolved.
 * resolve_link('http://domain.tld/blah');
 * 
 * // Resolve to the current page, only with the requested argument
 * resolve_link('?argument=something');
 * 
 * // Resolve to the previous page, useful for "go-back" functionality
 * resolve_link('BACK');
 * ```
 *
 * @param string $url
 *
 * @return string The full url of the link, including the http://...
 */
function resolve_link($url) : string {
	// Simple checks for some various inbound links.
	// These generally don't require translation.

	if($url === '' || $url === null) {
		// Empty URL... WHAT?
		return EXEC_MODE === 'WEB' ? ROOT_URL : '';
	}

	if ($url[0] === '#'){
		// Allow "#" to be verbatim without translation.
		return $url;
	}

	if($url === 'BACK'){
		// Special use-case to resolve the previous page, (aka referrer).
		$request = Router::$DefaultRequest;
		return $request->referrer ?? '';
	}
	
	if (strpos($url, '://') !== false){
		// Allow already-resolved links to be returned verbatim.
		return $url;
	}

	if(strpos($url, 'tel:') === 0) {
		// tel: links should not be processed.
		return $url;
	}

	if(strpos($url, 'mailto:') === 0) {
		// Mail links are allowed as-is
		return $url;
	}

	
	if($url[0] === '?') {
		// Allow links starting with ? to be read as the current page.
		$url = REL_REQUEST_PATH . $url;
	}
	
	
	if($url[0] === '.') {
		// Allow requesting just a specific mimetype for the current request, eg: ".csv" to request the CSV version of the current page.
		if(strpos(REL_REQUEST_PATH, '.') !== false) {
			$url = substr(REL_REQUEST_PATH, 0, strpos(REL_REQUEST_PATH, '.')) . $url;
		}
		else{
			$url = REL_REQUEST_PATH . $url;
		}
	}


	// Use the URL mapper to do the bulk of the work here.
	$map = new BaseURLMap($url);
	try {
		// We substr the baseurl because the ROOTURL will already have a trailing slash.
		return ROOT_URL . substr($map->getRewriteURL(), 1);
	}
	catch(Exception $ex) {
		log_warning($ex->getMessage());
		return '';
	}

	// Allow multisite URLs to be passed in natively.
	/*if(stripos($url, 'site:') === 0){
		$slashpos = strpos($url, '/');
		$site = substr($url, 5, $slashpos-5);
		$url = substr($url, $slashpos);
	}
	else{
		$site = null;
	}*/
	
	try{
		return PageModel::GetByURI($url)->getResolvedURL();
	}
	catch (Exception $ex) {
		// This is acceptable, as not all Controllers will have a registered Page.
		// Allow multisite URLs to be passed in natively.
		if(stripos($url, 'site:') === 0){
			$slashpos = strpos($url, '/');
			$site = substr($url, 5, $slashpos-5);
			$url = substr($url, $slashpos + 1);
			
			return PageModel::GetRootURL($site) . $url;
		}
		else{
			return ROOT_URL . substr($url, 1);
		}
	}
}

/**
 * Reverse resolve a link back to its base URL
 *
 * @param string $url
 *
 * @return string
 */
function reverse_resolve_link(string $url) : string {
	// Pre-process the URL for quick checks and automatic replacements
	if($url === '' || $url === '#') {
		return $url;
	}
	elseif(strpos($url, ROOT_URL_NOSSL) === 0) {
		$url = '/' . substr($url, strlen(ROOT_URL_NOSSL));
	}
	elseif(strpos($url, ROOT_URL_SSL) === 0) {
		$url = '/' . substr($url, strlen(ROOT_URL_SSL));
	}
	elseif(ROOT_URL_NOSSL === $url . '/') {
		// If the url is the plain domain without a trailing slash, it's still just the top-level page
		$url = '/';
	}
	elseif(ROOT_URL_SSL === $url . '/') {
		// If the url is the plain domain without a trailing slash, it's still just the top-level page
		$url = '/';
	}

	// Resolve internal URLs, passthru others.
	if(strpos($url, '/') === 0) {
		$map = new BaseURLMap($url);
		try {
			return $map->getBaseURL();
		}
		catch(Exception $ex) {
			log_warning($ex->getMessage());
			return $url;
		}
	}
	else {
		return $url;
	}
}

/**
 * Redirect the user to another page via sending the Location header.
 *    Prevents any POST data from being reloaded.
 *
 * @param  string $page The page URL to redirect to
 * @param  int    $code  The HTTP status code to send to the browser, MUST be 301 or 302.
 *
 * @throws \Exception
 */
function redirect($page, $code = 302){
	if(!($code == 301 || $code == 302)){
		throw new \Exception('Invalid response code requested for redirect, [' . $code . '].  Please ensure it is either a 301 (permanent), or 302 (temporary) redirect!');
	}
	
	$view = new View();
	$view->status = $code;
	$view->addHeader('Location', resolve_link($page));
	$view->render();
	die();
}

/**
 * Utility function to reload the current page
 */
function reload(){
	$view = new View();
	$view->status = View::STATUS_FOUND;
	$view->addHeader('Location', resolve_link(CUR_CALL));
	$view->render();
	die();
}

/**
 * Utility function to just go back to a page before this one.
 */
function go_back() {
	$request = Router::$DefaultRequest;
	$history = $request->getReferrer();

	if($history != CUR_CALL){
		redirect($history);
	}
	else{
		reload();
	}
}


/**
 * Convenience function to render out a list of controls from a Model
 *
 * Common Options:
 *
 * * hover       TRUE or FALSE if hover should be enabled.
 * * proxy-force TRUE or FALSE if the proxy should be forced to be used.
 * * proxy-text  string The label of the proxy text.
 *
 * @param Model|string $object
 * @param array        $options List of options to provide for the rendering.
 *
 * @return string
 */
function controls_as_html($object, array $options = []) : string {
	
	// Default options and their values
	$subject    = $options['subject'] ?? null;
	$hover      = $options['hover'] ?? true;
	$proxyForce = $options['proxy-force'] ?? false;
	$proxyText  = $options['proxy-text'] ?? 'Controls';
	
	
	if($object instanceof Model){
		$controls = ViewControls::DispatchModel($object);
	}
	elseif(is_string($object)) {
		$controls = ViewControls::Dispatch($object, $subject);
	}
	else {
		return 'Either a Model or a baseurl is required for controls!';
	}
	
	
	$controls->hovercontext = $hover;
	$controls->setProxyForce($proxyForce);
	$controls->setProxyText($proxyText);
	
	
	// Render out controls.
	return $controls->fetch();
}


/**
 * Render markup for an image tag based on the file, URL, or path snippet.
 *
 * ## Options
 *
 *  * assign
 *    * string
 *    * Set to a string to assign that variable name instead of returning the output.
 *  * dimensions
 *    * Provide both width and height in pixels, along with special instructions
 *    * Structure is "widthxheight" with no spaces between the "x" and the two integers.
 *    * Special modes available are:
 *    * Carat "`^`" at the end of the string fits the smallest dimension instead of the largest.
 *    * Exclamation mark "`!`" at the end forces size regardless of aspect ratio.
 *    * Greater than "`>`" at the end will only increase image sizes.
 *    * Less than "`<`" at the end will only decrease image sizes.
 *  * height
 *    * int
 *    * Maximum image height (in pixels).
 *    * If both width and height are provided, the image will be constrained to both without any distortion.
 *  * inline
 *    * int "1" or "0", (default "0")
 *    * New in Core 4.2.0
 *    * Request that the resized image be encoded as base64 and inserted inline in the markdown instead of returned as the URL.
 *  * includemeta
 *    * Set to "1" to include extra markup for the metadata
 *  * placeholder
 *    * string
 *    * placeholder image if the requested image is blank or not found.  Useful for optional fields that should still display something.
 *    * Current values: "building", "generic", "person", "person-tall", "person-wide", "photo"
 *  * width
 *    * int
 *    * Maximum image width (in pixels).
 *    * If both width and height are provided, the image will be constrained to both without any distortion.
 *
 * @param string|File  $path    Path, URL, or file for this image
 * @param array        $options List of options
 *
 * @return string
 */
function img_as_html($path, $options = []) : string {
	
	global $__Core_Asset_Inline;
	
	$renderer = new ImageRenderer();
	
	if($path instanceof File) {
		$renderer->setFile($path);
	}
	else {
		$renderer->setSource($path);
	}
	
	$renderer->setAttributes($options);
	
	if(isset($__Core_Asset_Inline) && $__Core_Asset_Inline) {
		$renderer->setInline();
	}
	
	return $renderer->fetch();
}


/**
 * Format a filesize in bytes as a string
 * 
 * @param string|int $size
 * @param number     $round
 * 
 * @return string
 */
function filesize_as_string($size, $round = 2) : string {
	return FileHelper::FormatSize($size, $round);
}


function filespeed_as_string($speed, $round = 1) : string {
	
	$suf = ['bps', 'Kbps', 'Mbps', 'Gbps', 'Tbps', 'Pbps', 'Ebps', 'Zbps', 'Ybps'];
	$c   = 0;
	while ($speed >= 1024) {
		$c++;
		$speed = $speed / 1024;
	}
	
	return I18NLoader::FormatNumber($speed, $round) . ' ' . $suf[$c];
}


/**
 * Utility to Core-ify a given HTML string.
 *
 * Will use a tokenizer to scan for <img /> tags and <a /> tags.
 *
 * @param $html
 * @return string
 */
function parse_html($html){
	// Counter for the current position of the tokenizer.
	$x = 0;
	// Set to the position of the current image
	$imagestart = null;

	// @todo a rel=nofollow tags for external/untrusted links
	//       This can make use of an external utility to allow the admin to set which links are allowed.

	// @todo a tags that are absolutely resolved or have a core prefix such as core:///about-us or what not.

	while($x < strlen($html)){
		// Replace images with the optimized version.
		if(substr($html, $x, 4) == '<img'){
			$imagestart = $x;
			$x+= 3;
			continue;
		}

		$fullimagetag = null;

		if($imagestart !== null && $html[$x] == '>'){
			// This will equal the full image HTML tag, ie: <img src="blah"/>...
			$fullimagetag = substr($html, $imagestart, $x-$imagestart+1);
		}
		elseif($imagestart !== null && substr($html, $x, 2) == '/>'){
			// This will equal the full image HTML tag, ie: <img src="blah"/>...
			$fullimagetag = substr($html, $imagestart, $x-$imagestart+2);
		}

		if($imagestart !== null && $fullimagetag){
			// Convert it to a DOM element so I can process it.
			$simple = new SimpleXMLElement($fullimagetag);
			$attributes = array();
			foreach($simple->attributes() as $k => $v){
				$attributes[$k] = (string)$v;
			}

			$renderer = new ImageRenderer();
			$renderer->setAttributes($attributes);
			$img = $renderer->fetch();

			// Figure out the offset for X.  I'll need to modify this after I merge it in.
			$x += strlen($img) - strlen($fullimagetag);
			// Split this string back in.
			$html = substr_replace($html, $img, $imagestart, strlen($fullimagetag));

			// Reset...
			$imagestart = null;
		}
		$x++;
	}
	//var_dump($html);

	return $html;
}

/**
 * Add a message to the user's stack.
 *	It will be displayed the next time the user (or session) renders the page.
 * 
 * If t:MESSAGE_[SOMETHING] is passed in as the first argument,
 * the i18n system will take effect and translate the string if possible.
 * 
 * Additionally, MESSAGE_SUCCESS_, MESSAGE_ERROR_, MESSAGE_TUTORIAL_, MESSAGE_WARNING_, and
 * MESSAGE_INFO_ will trigger that type of message code to be returned.
 * 
 * Also when t: is used, any arguments passed in after the first string will be sent to the i18n system for replacement.
 *
 * @param string $message_text The message text or the MESSAGE_ string constant for i18n and automatic type detection!
 * @param string $message_type Message type or arguments to pass into the translate method.
 *
 * @return boolean (on success)
 */
function set_message($messageText, $messageType = 'info'){
	if(strpos($messageText, 't:MESSAGE_') === 0){
		// It's an i18n message!  Retrieve the locale version of text and the message type.
		$messageText = substr($messageText, 2);
		$args = func_get_args();
		
		if(strpos($messageText, 'MESSAGE_SUCCESS_') === 0){
			$messageType = 'success';
		}
		elseif(strpos($messageText, 'MESSAGE_ERROR_') === 0){
			$messageType = 'error';
		}
		elseif(strpos($messageText, 'MESSAGE_TUTORIAL_') === 0){
			$messageType = 'tutorial';
		}
		elseif(strpos($messageText, 'MESSAGE_WARNING_') === 0){
			$messageType = 'warning';
		}
		elseif(strpos($messageText, 'MESSAGE_INFO_') === 0){
			$messageType = 'info';
		}
		else{
			$messageType = 'info';
		}
		
		if(func_num_args() > 1){
			// Use func_call to call 1, as I need to pass in the other options too!
			$messageText = call_user_func_array('t', $args);
		}
		else{
			$messageText = t($messageText);
		}
	}

	// CLI doesn't use sessions, echo directly to stdout instead.
	if(EXEC_MODE == 'CLI'){
		$messageText = preg_replace('/<br[^>]*>/i', "\n", $messageText);
		echo "[" . $messageType . "] - " . $messageText . "\n";
	}
	else{
		$stack = Session::Get('message_stack', []);

		$stack[] = array(
			'mtext' => $messageText,
			'mtype' => $messageType,
		);
		Session::Set('message_stack', $stack);
	}
}


/**
 * Retrieve the messages and optionally clear the message stack.
 *
 * @param bool $returnSorted Set to true to sort the message by message type
 * @param bool $clearStack   Set to false to NOT clear the message stack
 *
 * @return array
 */
function get_messages($returnSorted = FALSE, $clearStack = TRUE){
	return Core::GetMessages($returnSorted, $clearStack);
}


/**
 * Generate a random hex-decimal value of a given length.
 * 
 * This uses the new PasswordGenerator in HEX mode as of Core 6.3
 *
 * @param int  $length
 * @param bool $casesensitive [false] Set to true to return a case-sensitive string.
 *                            Otherwise the resulting string will simply be all uppercase.
 *
 * @return string
 */
function random_hex($length = 1, $casesensitive = false) : string {
	$generator = new Utilities\PasswordGenerator();
	$generator->minLength = $length;
	return $generator->generateHex($casesensitive);
}


/**
 * Simple method to compare two values with each other in a more restrictive manner
 * than == but not quite fully typecasted.
 *
 * This is useful for the scenarios that involve needing to check that "3" == 3, but "" != 0.
 * 
 * Returns true if they seem to be the same, false if they differ.
 *
 * @param mixed $val1
 * @param mixed $val2
 *
 * @return bool
 */
function compare_values($val1, $val2) : bool {
	if($val1 === $val2){
		// Exact same values and exact same typecasts.  They're the same!
		return true;
	}
	elseif(is_numeric($val1) && is_numeric($val2) && $val1 == $val2){
		// Both values are numeric and seem to be the same value, ie: "3" and 3.
		return true;
	}
	elseif(is_scalar($val1) && is_scalar($val2) && strlen($val1) == strlen($val2) && $val1 == $val2){
		// If they're both strings of the same length and equal to each other... same value.
		return true;
	}

	return false;
}

/**
 * Compare two values as strings explictly.
 * This is useful for numbers that need to behave like strings, ie: postal codes with their leading zeros.
 *
 * @param $val1
 * @param $val2
 *
 * @return bool
 */
function compare_strings($val1, $val2) : bool {
	if($val1 === $val2){
		// Exact same values and exact same typecasts.  They're the same!
		return true;
	}
	if(strlen($val1) == strlen($val2) && $val1 == $val2){
		// If they're both strings of the same length and equal to each other... same value.
		return true;
	}

	return false;
}

/**
 * Log a deprecated notice to the logger.
 * 
 * Use the following code to get the calling location automatically
 * ```
 * log_deprecated(
 *   'this_thing', 
 *   debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS, 2)[1])
 * );
 * ```
 * 
 * @param string $function
 * @param array $caller
 */
function log_deprecated(string $function, array $caller){
	
	$callerStr = '';
	if(
		isset($caller['function']) && 
		isset($caller['class']) &&
		isset($caller['type'])
	){
		$callerStr = '  Called from ' . $caller['class'] . $caller['type'] . $caller['function'] . '()';
	}
	
	$entry = new LogEntry();
	$entry->level   = LOG_LEVEL_DEBUG;
	$entry->type    = 'debug';
	$entry->message = 'Use of ' . $function . ' is deprecated and will be removed in a future version.' . $callerStr;
	$entry->send();
}

/**
 * Convenience function to record a "trace" entry in Core's loggers.
 * 
 * These are used for low-level development and should not really be used by general public.
 * 
 * @param string $message The message to log
 */
function log_trace(string $message){
	LogEntry::Init()
		->level(LOG_LEVEL_TRACE)
		->type('debug')
		->message($message)
		->trace(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS))
		->send();
}

/**
 * Convenience function to record a debug message.
 * 
 * These are useful for the bulk of development and command line debugging.
 * 
 * @param string $message The message to log
 */
function log_debug(string $message){
	LogEntry::Init()
		->level(LOG_LEVEL_DEBUG)
		->type('debug')
		->message($message)
		->trace(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS))
		->send();
}

/**
 * Convenience function to record an informative message
 * 
 * This is the standard log level for day-to-day operations such as user registrations.
 * 
 * @param string $message The message to log
 */
function log_info(string $message){
	LogEntry::Init()
		->level(LOG_LEVEL_INFO)
		->type('info')
		->message($message)
		->trace(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS))
		->send();
}

/**
 * Convenience function to record a notice message
 * 
 * Useful for missing images or configuration problems that aren't critical.
 * This message gets filed as an error but only if INFO level logging is enabled.
 * 
 * @param string $message The message to log
 */
function log_notice(string $message){
	LogEntry::Init()
		->level(LOG_LEVEL_INFO)
		->type('error')
		->message($message)
		->trace(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS))
		->send();
}

/**
 * Convenience function to record an important informative message
 * 
 * @param string $message The message to log
 */
function log_important(string $message){
	LogEntry::Init()
		->level(LOG_LEVEL_HIGH)
		->type('info')
		->message($message)
		->trace(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS))
		->send();
}

/**
 * Convenience function to record a warning (a relatively important error, but not critical).
 * 
 * @param string $message The message to log
 */
function log_warning(string $message){
	LogEntry::Init()
		->level(LOG_LEVEL_HIGH)
		->type('error')
		->message($message)
		->trace(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS))
		->send();
}

/**
 * Convenience function to record an error in Core, this is the highest level of log message and gets saved as an error.
 * 
 * @param string $message The message to log
 */
function log_error(string $message){
	LogEntry::Init()
		->level(LOG_LEVEL_CRITICAL)
		->type('error')
		->message($message)
		->trace(debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS))
		->send();
}

function log_exception(Exception $exception){
	LogEntry::Init()
		->level(LOG_LEVEL_CRITICAL)
		->exception($exception)
		->send();
}

function log_error_exception(Error $exception){
	LogEntry::Init()
		->level(LOG_LEVEL_CRITICAL)
		->type('error')
		->message($exception->getMessage())
		->code(get_class($exception))
		->trace($exception->getTrace())
		->send();
}

/**
 * Shorthand to generate a new LogEntry that can be used for logging with minimal code.
 * 
 * @return LogEntry
 */
function log() : LogEntry {
	return new LogEntry();
}


/**
 * Validate an email address.
 * Provide email address (raw input)
 * Returns true if the email address has the email
 * address format and the domain exists.
 *
 * Copied (almost) verbatim from http://www.linuxjournal.com/article/9585?page=0,3
 * @author Douglas Lovell @ Linux Journal
 *
 * @return boolean|string
 */
function check_email_validity($email) {
	$atIndex = strrpos($email, "@");
	if (is_bool($atIndex) && !$atIndex) {
		return 'Email is missing @ symbol.';
		//return false;
	}

	$domain    = substr($email, $atIndex + 1);
	$local     = substr($email, 0, $atIndex);
	$localLen  = strlen($local);
	$domainLen = strlen($domain);
	if ($localLen < 1 || $localLen > 64) {
		// local part length exceeded
		return 'Email user is too long.';
		//return false;
	}

	if ($domainLen < 1 || $domainLen > 255) {
		// domain part length exceeded
		//return false;
		return 'Email domain is too long.';
	}

	if ($local[0] == '.' || $local[$localLen - 1] == '.') {
		// local part starts or ends with '.'
		//return false;
		return 'Email user can not start or end with a period.';
	}

	if (preg_match('/\\.\\./', $local)) {
		// local part has two consecutive dots
		//return false;
		return 'Email user can not have two consecutive periods.';
	}
	if (!preg_match('/^[A-Za-z0-9\\-\\.]+$/', $domain)) {
		// character not valid in domain part
		//return false;
		return 'Email domain contains invalid characters.';
	}

	if (preg_match('/\\.\\./', $domain)) {
		// domain part has two consecutive dots
		//return false;
		return 'Email domain can not have two consecutive periods.';
	}

	if (!preg_match('/^(\\\\.|[A-Za-z0-9!#%&`_=\\/$\'*+?^{}|~.-])+$/', str_replace("\\\\", "", $local))) {
		// character not valid in local part unless local part is quoted
		if (!preg_match('/^"(\\\\"|[^"])+"$/', str_replace("\\\\", "", $local))) {
			//return false;
			return 'Email user contains special characters and must be quoted.';
		}
	}

	// Allow the admin to skip DNS checks via config.
	if (ConfigHandler::Get('/core/email/verify_with_dns') && !(checkdnsrr($domain, "MX") || checkdnsrr($domain, "A"))) {
		// domain not found in DNS
		//return false;
		return 'Email domain does not seem to exist.';
	}

	// All checks passed?
	return true;
}

function str_to_latin($string){

	$internationalmappings = array(
		'À' => 'A',
		'Á' => 'A',
		'Â' => 'A',
		'Ã' => 'A',
		'Ä' => 'A',
		'Å' => 'AA',
		'Æ' => 'AE',
		'Ç' => 'C',
		'È' => 'E',
		'É' => 'E',
		'Ê' => 'E',
		'Ë' => 'E',
		'Ì' => 'I',
		'Í' => 'I',
		'Î' => 'I',
		'Ï' => 'I',
		'Ð' => 'D',
		'Ł' => 'L',
		'Ñ' => 'N',
		'Ò' => 'O',
		'Ó' => 'O',
		'Ô' => 'O',
		'Õ' => 'O',
		'Ö' => 'O',
		'Ø' => 'OE',
		'Ù' => 'U',
		'Ú' => 'U',
		'Ü' => 'U',
		'Û' => 'U',
		'Ý' => 'Y',
		'Þ' => 'Th',
		'ß' => 'ss',
		'à' => 'a',
		'á' => 'a',
		'â' => 'a',
		'ã' => 'a',
		'ä' => 'a',
		'å' => 'aa',
		'æ' => 'ae',
		'ç' => 'c',
		'è' => 'e',
		'é' => 'e',
		'ê' => 'e',
		'ë' => 'e',
		'ì' => 'i',
		'í' => 'i',
		'î' => 'i',
		'ï' => 'i',
		'ð' => 'd',
		'ł' => 'l',
		'ñ' => 'n',
		'ń' => 'n',
		'ò' => 'o',
		'ó' => 'o',
		'ô' => 'o',
		'õ' => 'o',
		'ō' => 'o',
		'ö' => 'o',
		'ø' => 'oe',
		'ś' => 's',
		'ù' => 'u',
		'ú' => 'u',
		'û' => 'u',
		'ū' => 'u',
		'ü' => 'u',
		'ý' => 'y',
		'þ' => 'th',
		'ÿ' => 'y',
		'ż' => 'z',
		'Œ' => 'OE',
		'œ' => 'oe',
		'&' => 'and'
	);

	// This is slightly more simple than the javascript version.
	return str_replace(array_keys($internationalmappings), array_values($internationalmappings), $string);
}

/**
 * Cleanup a string and ensure it can make a valid URL component.
 *
 * Note, this is only meant for an individual directory level of a URL, such as the filename or a directory name.
 * This is because it removes slashes "/".
 *
 * <h3>Usage</h3>
 *
 * <p>Typical Usage</p>
 * <code>
 * // Will print "something-foo"
 * echo \Core\str_to_url('Something Foo!');
 *
 * // International characters are also handled gracefully
 * // Will print "thors hammer"
 * echo \Core\str_to_url('Þors hammer');
 * </code>
 *
 * <p>If you are dealing with a filename instead of a url, you may want to preserve the extensions.</p>
 * <code>
 * // Will print awesome-hot-image.jpg
 * echo \Core\str_to_url('AWESOME höt Image!!!!!!!.JPG');
 * </code>
 *
 * @param string $string  Incoming string to convert
 * @param bool   $keepdots Set to true if you want to keep dots "." and file extensions.
 *
 * @return string
 */
function str_to_url($string, $keepdots = false){
	// URLs should only be in latin.
	$string = str_to_latin($string);

	if(ConfigHandler::Get('/core/page/url_remove_stop_words')){
		$stopwords = get_stop_words();

		$exploded = explode(' ', $string);
		$nt = '';
		foreach($exploded as $w){
			$lw = strtolower($w);
			if(!in_array($lw, $stopwords)){
				$nt .= ' ' . $w;
			}
		}
		$string = trim($string);
	}

	// Spaces and slashes get replaced with a separator
	$string = str_replace([' ', '/'], '-', $string);

	// Anything else I missed?  Get rid of it!
	if($keepdots){
		$string = preg_replace('/[^a-z0-9\-\._]/i', '', $string);
	}
	else{
		$string = preg_replace('/[^a-z0-9\-_]/i', '', $string);
	}

	// Multiple separators should get truncated, along with beginning and trailing ones.
	$string = preg_replace('/[-]+/', '-', $string);
	$string = preg_replace('/^-/', '', $string);
	$string = preg_replace('/-$/', '', $string);

	// And lowercase it.
	$string = strtolower($string);

	return $string;
}

/**
 * Check if a given string starts with a given match.
 * 
 * @param string $haystack
 * @param string $needle
 * 
 * @return bool
 */
function str_starts_with(string $haystack, string $needle) : bool {
	return substr($haystack, 0, strlen($needle)) === $needle;
}

/**
 * Check if a given string ends with a given match.
 * 
 * @param string $haystack
 * @param string $needle
 * 
 * @return bool
 */
function str_ends_with(string $haystack, string $needle) : bool {
	$length = strlen($needle);

	return 
		$length === 0 || 
		substr($haystack, -$length) === $needle;
}


/**
 * Function to translate a PHP upload error into a human-readable string.
 *
 * If no error is seen, a blank string is returned.
 *
 * @param int
 * @return string
 */
function translate_upload_error($errno){
	return \Core\Filestore\translate_upload_error($errno);
}

/**
 * Convert a numeric dollar value (in 2-decimal places), to the full cents value
 *
 * This addresses all rounding issues that comes about from handling float precision.
 *
 * Best case is the BC Math extension is available, otherwise a fallback is provided.
 *
 * If a fully formatted string is provided with the currency symbol, it's automatically stripped out.
 *
 * @param string|float|int $value
 *
 * @return int
 */
function dollars_to_cents($value) : int {

	if(is_string($value)) {
		// If there's a currency formatter, remove that.
		$value = str_replace(['$', '€', '£', '¥'], '', $value);
	}

	if(function_exists('bcmul')) {
		return (int)bcmul($value, 100);
	}
	else {
		return (int) ( (string) ($value * 100) );
	}
}

/**
 * Convert a numeric dollar value (in 2-decimal places), to the full cents value
 *
 * This addresses all rounding issues that comes about from handling float precision.
 *
 * Best case is the BC Math extension is available, otherwise a fallback is provided.
 *
 * If a fully formatted string is provided with the currency symbol, it's automatically stripped out.
 *
 * @param string|float|int $value
 *
 * @return int
 */
function cents_to_dollars($value) : float {

	if(function_exists('bcdiv')) {
		return bcdiv($value, 100, 2);
	}
	else {
		return $value / 100;
	}
}

/**
 * Check a file mimetype against a base.
 *
 * Useful for upload scripts.
 *
 * @param $acceptlist string List of accept types to accept.
 *                    Can be wildcard types too, for example image/*, or simply *.
 * @param $mimetype string File Mimetype to check against
 * @param $extension string Extension of the file to check against.
 *                  (not required, but provides better true positives, [and potentially more false positives] ).
 *
 * @return string Empty string if passed, else the error.
 */
function check_file_mimetype($acceptlist, $mimetype, $extension = null){
	return \Core\Filestore\check_file_mimetype($acceptlist, $mimetype, $extension);
}

/**
 * Check if an array is a plain numerically indexed array, or not.
 *
 * It's useful for checking if an array's keys are meant to be used, or simply ignored.
 * This is important because sometimes a selectbox will have options set like array('foo', 'bar', 'baz')
 * and other times it's set as array(id => blah, id => foo, id => mep), where the id is important.
 *
 * @param $array array
 *
 * @return bool
 */
function is_numeric_array($array){
	if(!is_array($array)) return false;

	reset($array);
//var_dump($array);
	// First check.. if the first index a 0?  If not... it's not a numeric array!
	if(key($array) !== 0) return false;

	// Numerically indexed arrays will have a size == to the final key.
	$c = count($array) - 1;
	end($array);
	if(key($array) !== $c) return false;

	// Hopefully.... checking the first and last keys.
	return true;
}

/**
 * Get an array of English "stop" words.
 *
 * These are words that are insignificant to search engines and text parsing scripts.
 *
 * @return array
 */
function get_stop_words(){
	$stopwords = array('a', 'about', 'above', 'above', 'across', 'after', 'afterwards', 'again', 'against', 'all', 'almost', 'alone', 'along', 'already', 'also','although','always','am','among', 'amongst', 'amoungst', 'amount',  'an', 'and', 'another', 'any','anyhow','anyone','anything','anyway', 'anywhere', 'are', 'around', 'as',  'at', 'back','be','became', 'because','become','becomes', 'becoming', 'been', 'before', 'beforehand', 'behind', 'being', 'below', 'beside', 'besides', 'between', 'beyond', 'bill', 'both', 'bottom','but', 'by', 'call', 'can', 'cannot', 'cant', 'co', 'con', 'could', 'couldnt', 'cry', 'de', 'describe', 'detail', 'do', 'done', 'down', 'due', 'during', 'each', 'eg', 'eight', 'either', 'eleven','else', 'elsewhere', 'empty', 'enough', 'etc', 'even', 'ever', 'every', 'everyone', 'everything', 'everywhere', 'except', 'few', 'fifteen', 'fify', 'fill', 'find', 'fire', 'first', 'five', 'for', 'former', 'formerly', 'forty', 'found', 'four', 'from', 'front', 'full', 'further', 'get', 'give', 'go', 'had', 'has', 'hasnt', 'have', 'he', 'hence', 'her', 'here', 'hereafter', 'hereby', 'herein', 'hereupon', 'hers', 'herself', 'him', 'himself', 'his', 'how', 'however', 'hundred', 'ie', 'if', 'in', 'inc', 'indeed', 'interest', 'into', 'is', 'it', 'its', 'itself', 'keep', 'last', 'latter', 'latterly', 'least', 'less', 'ltd', 'made', 'many', 'may', 'me', 'meanwhile', 'might', 'mill', 'mine', 'more', 'moreover', 'most', 'mostly', 'move', 'much', 'must', 'my', 'myself', 'name', 'namely', 'neither', 'never', 'nevertheless', 'next', 'nine', 'no', 'nobody', 'none', 'noone', 'nor', 'not', 'nothing', 'now', 'nowhere', 'of', 'off', 'often', 'on', 'once', 'one', 'only', 'onto', 'or', 'other', 'others', 'otherwise', 'our', 'ours', 'ourselves', 'out', 'over', 'own','part', 'per', 'perhaps', 'please', 'put', 'rather', 're', 'same', 'see', 'seem', 'seemed', 'seeming', 'seems', 'serious', 'several', 'she', 'should', 'show', 'side', 'since', 'sincere', 'six', 'sixty', 'so', 'some', 'somehow', 'someone', 'something', 'sometime', 'sometimes', 'somewhere', 'still', 'such', 'system', 'take', 'ten', 'than', 'that', 'the', 'their', 'them', 'themselves', 'then', 'thence', 'there', 'thereafter', 'thereby', 'therefore', 'therein', 'thereupon', 'these', 'they', 'thickv', 'thin', 'third', 'this', 'those', 'though', 'three', 'through', 'throughout', 'thru', 'thus', 'to', 'together', 'too', 'top', 'toward', 'towards', 'twelve', 'twenty', 'two', 'un', 'under', 'until', 'up', 'upon', 'us', 'very', 'via', 'was', 'we', 'well', 'were', 'what', 'whatever', 'when', 'whence', 'whenever', 'where', 'whereafter', 'whereas', 'whereby', 'wherein', 'whereupon', 'wherever', 'whether', 'which', 'while', 'whither', 'who', 'whoever', 'whole', 'whom', 'whose', 'why', 'will', 'with', 'within', 'without', 'would', 'yet', 'you', 'your', 'yours', 'yourself', 'yourselves', 'the');
	return $stopwords;
}

/**
 * Generate a globally unique identifier that can be used as a replacement for an autoinc or similar.
 *
 * This method IS compatible with multiple servers on a single codebase!
 *
 * An example of a UUID returned by this function would be: "1a3f-c5dbcaaf9db-8d77"
 *
 * @since 2.4.2
 *
 * @return string
 */
function generate_uuid(){
	static $__serverid = null;
	
	if($__serverid === null){
		$serverid = defined('SERVER_ID') ? SERVER_ID : '0001';

		if($serverid == '1'){
			// Catch for legacy servers that had '1' as the server ID.
			// after 5.0.1, this has been migrated to a global UUID of its own.
			$serverid = '0001';
		}
		
		if(strlen($serverid) > 4){
			// It should be honestly!  It's 32-digits long.
			$serverid = substr($serverid, -4);
		}
		
		$__serverid = $serverid;
	}
	else{
		$serverid = $__serverid;
	}
	
	//var_dump($serverid); die();
	
	return strtolower(
		$serverid . '-' . 
		dechex(microtime(true) * 10000) . '-' .
		random_hex(4)	
	);
}

/**
 * Clone of the php version_compare function, with the exception that it treats
 * version numbers the same that Debian treats them.
 *
 * @param string $version1 Version to compare
 * @param string $version2 Version to compare against
 * @param string $operation Operation to use or null
 *
 * @return bool | int Boolean if $operation is provided, int if omited.
 */
function version_compare($version1, $version2, $operation = null) {
	$version1 = new VersionString($version1);
	return $version1->compare($version2, $operation);
}

/**
 * Format a time duration in a human-readable format.
 * 
 * The formatting returned is in the format of:
 *
 * * 0.0000001 = "100 ns"
 * * 0.0000010 = "1 µs" (10E-6)
 * * 0.0000100 = "10 µs"
 * * 0.0001000 = "100 µs"
 * * 0.0010000 = "1 ms" (10E-3)
 * * 0.0100000 = "10 ms"
 * * 0.1000000 = "100 ms"
 * * 1.0000000 = "1 s"
 * * 60.000000 = "1 m"
 * * 70.000000 = "1 m 10 s"
 * * 3600.0000 = "1 h"
 * * 3720.0000 = "1 h 2 m"
 * 
 * @param float $time_in_seconds Time (in seconds) to format
 * @param int   $round           Optionally rounding precision
 *
 * @return string
 */
function time_duration_format(float $time_in_seconds, int $round = 4) : string {
	// Parts of the result.
	$parts = [];
	
	if($time_in_seconds < 0.000001){
		$parts[] = [
			'suffix' => 'ns',
			'round' => true,
			'precision' => true,
			'value' => $time_in_seconds * 1000000000,
		];
	}
	elseif($time_in_seconds < 0.001){
		$parts[] = [
			'suffix' => 'µs',
			'round' => true,
			'precision' => true,
			'value' => $time_in_seconds * 1000000,
		];
	}
	elseif($time_in_seconds < 1){
		$parts[] = [
			'suffix' => 'ms',
			'round' => true,
			'precision' => true,
			'value' => $time_in_seconds * 1000,
		];
	}
	elseif($time_in_seconds < SECONDS_ONE_MINUTE){
		$parts[] = [
			'suffix' => 's',
			'round' => true,
			'precision' => true,
			'value' => $time_in_seconds,
		];
	}
	elseif($time_in_seconds < SECONDS_ONE_HOUR){
		$m = floor($time_in_seconds / 60);
		$s = round($time_in_seconds - $m*60, 0);
		if($s > 0){
			$parts[] = [
				'formatted' => t('STRING_N_MINUTE_ABBREVIATED', $m),
			];
			$parts[] = [
				'formatted' => t('STRING_N_SECOND_ABBREVIATED', $s),
			];
		}
		else{
			$parts[] = [
				'formatted' => t('STRING_N_MINUTE', $m),
			];
		}
	}
	elseif($time_in_seconds < SECONDS_ONE_DAY){
		$h = floor($time_in_seconds / SECONDS_ONE_HOUR);
		$m = round(($time_in_seconds - $h * SECONDS_ONE_HOUR)/60, 0);
		if($m > 0){
			$parts[] = [
				'formatted' => t('STRING_N_HOUR_ABBREVIATED', $h),
			];
			$parts[] = [
				'formatted' => t('STRING_N_MINUTE_ABBREVIATED', $m),
			];
		}
		else{
			$parts[] = [
				'formatted' => t('STRING_N_HOUR', $h),
			];
		}
	}
	else{
		$d = floor($time_in_seconds / SECONDS_ONE_DAY);
		$h = round(($time_in_seconds - $d*SECONDS_ONE_DAY)/SECONDS_ONE_HOUR, 0);
		if($h > 0){
			$parts[] = [
				'formatted' => t('STRING_N_DAY', $d),
			];
			$parts[] = [
				'formatted' => t('STRING_N_HOUR_ABBREVIATED', $h),
			];
		}
		else{
			$parts[] = [
				'formatted' => t('STRING_N_DAY', $d),
			];
		}
	}
	
	$out = [];
	foreach($parts as $p){
		if(isset($p['formatted'])){
			// Already formatted, nothing to do.
			$out[] = $p['formatted'];
		}
		else{
			$f = $p['value'];
			if($p['round']){
				$f = round($f, $round);
			}
			if($p['precision']){
				$f = number_format($f, $round);
			}
			$out[] = $f . ' ' . $p['suffix'];
		}
	}
	
	return implode(' ', $out);
}

function is_ip_private($ip){
	
	$ipBin = inet_pton($ip);
	
	$privates = [
		[
			// 10.0.0.0/8 - Private IPv4 addresses (rfc1918)
			'net'   => '10.0.0.0',
			'mask'  => '255.0.0.0',
		],
		[
			// 100.64.0.0/10 - Carrier NAT (rfc6598)
			'net'   => '100.64.0.0',
			'mask'  => '255.192.0.0',
		],
		[
			// Loop-back Only
			'net'   => '127.0.0.0',
			'mask'  => '255.0.0.0',
		],
		[
			// Private local-only
			'net'   => '169.254.0.0',
			'mask'  => '255.255.0.0',
		],
		[
			// 172.16.0.0/12 - Private IPv4 addresses (rfc1918)
			'net'   => '172.16.0.0',
			'mask'  => '255.240.0.0',
		],
		[
			// Test Networks (rfc5735)
			'net'   => '192.0.2.0',
			'mask'  => '255.255.255.0',
		],
		[
			// Private IPv4 addresses (rfc1918)
			'net'   => '192.168.0.0',
			'mask'  => '255.255.0.0',
		],
		[
			// 198.18.0.0/15 - Private IPv4 addresses (rfc1918)
			'net'   => '198.18.0.0',
			'mask'  => '198.18.0.0',
		],
		[
			// Test Networks (rfc5735)
			'net'   => '198.51.100.0',
			'mask'  => '255.255.255.0',
		],
		[
			// Test Networks (rfc5735)
			'net'   => '203.0.113.0',
			'mask'  => '255.255.255.0',
		],
		[
			// Private IPv6 (rfc4193)
			'net'   => 'fd00::',
			'mask'  => 'ff00::',
		],
		[
			// Localhost
			'net'   => '::1',
			'mask'  => 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff',
		]
	];
	
	foreach($privates as $p){
		
		$netBin = inet_pton($p['net']);
		$maskBin = inet_pton($p['mask']);
		
		if(($ipBin & $maskBin) === $netBin){
			return true;
		}
	}
	
	return false;
}

function inet_ntoh($bytes){
	$hexAsString = '0x';
	foreach(str_split($bytes) as $char){
		$hexAsString .= str_pad(dechex(ord($char)), 2, '0', STR_PAD_LEFT);
	}
	return $hexAsString;
}

/**
 * Calculate the mean in a given numeric array.
 * 
 * The array is expected to be sorted first!
 * 
 * Will return false if a an out-of-range percentile is provided.
 * 
 * The array is passed in by-ref to save memory.
 * 
 * @param $array
 * @param $percentile
 * 
 * @return mixed|false
 */
function mean(&$array, $percentile){
	if($percentile < 0 || $percentile > 100){
		return false;
	}
	
	if($percentile == 0){
		// Use reset+current to not modify the original array.
		reset($array);
		return current($array);
	}
	
	if($percentile == 100){
		// Use end+current to not modify the original array.
		end($array);
		return current($array);
	}
	
	// Calculate the key position of the requested percentile.
	$s = sizeof($array);
	$idx = ceil($s * $percentile / 100) - 1;
	if($idx > $s){
		$idx = $s;
	}
	return $array[$idx];
}

/**
 * Convert an associative array of attributes and their values to an HTML string
 *
 * If a key has a value of NULL, just the attribute is rendered.
 * If a key has an array of items, the items are merged into a space-separated string
 *
 * @param array $attributes
 * @return string
 */
function array_to_html_attributes_string(array $attributes) : string {
	$sets = [];
	foreach($attributes as $k => $v) {
		if($v === null) {
			$sets[] = $k;
		}
		elseif($k === 'style') {
			// Style differs slightly from the traditional attributes.
			$styles = [];
			foreach($v as $sk => $sv) {
				$styles[] = $sk . ':' . htmlentities($sv) . ';';
			}
			$sets[] = 'style="' . implode(' ', $styles) . '"';
		}
		elseif(is_array($v)) {
			// Remove any duplicates, usually located in class names.
			$v = array_unique($v);
			$sets[] = $k . '="' . htmlentities(implode(' ', $v), ENT_COMPAT) . '"';
		}
		else {
			$sets[] = $k . '="' . htmlentities($v, ENT_COMPAT) . '"';
		}
	}

	return implode(' ', $sets);
}