<?php
/**
 * File for class Column definition in the coreplus project
 * 
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20140406.2005
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\ListingTable;


/**
 * A short teaser of what Column does.
 *
 * More lengthy description of what Column does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for Column
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @package Core\ListingTable
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class Column {
	/** @var string The title of this column */
	public $title;
	
	/** @var string The model column name used for sorting */
	public $sortkey;
	
	/** @var boolean T/F if this column is hidden by default. */
	public $visible = true;
	
	/** @var string The model's view name for rendering output */
	public $renderkey;
	
	/** @var string Abbreviation for this column */
	public $abbr = '';
	
	/** @var string|null Set to a string to group this column with other like-named columns */
	public $group = null;
	
	/** @var string System name of this column, useful for CSS overrides and javascript calls. */
	public $name = null;
	
	/** @var string Icon to display along with the text, optionally. */
	public $icon = null;

	/** @var null|string 'ASC', 'DESC', or null */
	public $currentSort = null;
	
	public $iconmode = self::DISPLAY_TEXTONLY;
	
	private $_title = null;
	private $_titleProcessed = null;
	private $_titleEscaped = null;
	
	
	const DISPLAY_TEXTONLY = 1;
	const DISPLAY_ICONONLY = 2;
	const DISPLAY_ICONTEXT = 3;

	public function getClass(){
		$classes = [];
		
		// Everything here is a cell.
		$classes[] = 'listing-table-cell';
		
		// Include the name for styling from CSS and JS calls!
		$classes[] = 'column-name-' . $this->name;

		if(!$this->visible){
			$classes[] = 'column-optional';
		}
		
		if($this->iconmode === self::DISPLAY_ICONONLY) {
			$classes[] = 'column-display-icon-only';
		}
		
		if($this->group){
			$classes[] = 'column-group-' . $this->group;
		}

		return implode(' ', $classes);
	}

	public function getTH(){
		return $this->_render('th');
	}
	
	public function getDIV(){
		return $this->_render('div');
	}
	
	/**
	 * Get the title attribute for this column, optionally escaped for HTML tags.
	 * 
	 * @param bool $escaped
	 * @return string
	 */
	public function getTitle($escaped = false){
		if($this->_title !== $this->title){
			// Calculate if this title needs some I18N work.
			// This is done to save execution calls for strpos, substr, and t.
			$this->_title = $this->title;
			// The title supports I18N!
			if(strpos($this->_title, 't:') === 0){
				$this->_titleProcessed = t(substr($this->_title, 2));
			}
			else{
				$this->_titleProcessed = $this->_title;
			}
			
			$this->_titleEscaped = str_replace('"', '&quot;', $this->_titleProcessed);
		}
		
		return $escaped ? $this->_titleEscaped : $this->_titleProcessed;
	}
	
	/**
	 * Set the sort and render keys
	 * 
	 * @param string $key
	 * @return Column
	 */
	public function key(string $key) : Column {
		$this->renderkey = $key;
		$this->sortkey = $key;
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Set the icon name to display and mode to BOTH text and icon.
	 * 
	 * @param string $icon
	 * @return Column
	 */
	public function icon(string $icon) : Column {
		$this->icon = $icon;
		$this->iconmode = self::DISPLAY_ICONTEXT;
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Set the icon name to display and hide the text label
	 * 
	 * @param string $icon
	 * @return Column
	 */
	public function icononly(string $icon) : Column {
		$this->icon = $icon;
		$this->iconmode = self::DISPLAY_ICONONLY;
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Set the name
	 * 
	 * @param string $name
	 * @return Column
	 */
	public function name(string $name) : Column {
		$this->name = \Core\str_to_url($name);
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Set the title
	 * 
	 * @param string $title
	 * @return Column
	 */
	public function title(string $title) : Column {
		$this->title = $title;
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Hide this column by default
	 * 
	 * @return Column
	 */
	public function hidden() : Column {
		$this->visible = false;
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Show this column by default
	 *
	 * @return Column
	 */
	public function show() : Column {
		$this->visible = true;
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Disable sorting for this column
	 *
	 * @return Column
	 */
	public function disableSortable() : Column {
		$this->sortkey = null;
		
		// Allow chaining
		return $this;
	}
	
	private function _render($tag){
		$out = '';
		
		if($this->abbr){
			$label = '<abbr title="' . $this->getTitle(true) . '">' . $this->abbr . '</abbr>';
		}
		elseif($this->iconmode === self::DISPLAY_ICONONLY && $this->icon) {
			$label = '<i class="icon icon-' . $this->icon . '" title="' . $this->getTitle(true) . '"></i>';
		}
		elseif($this->iconmode === self::DISPLAY_ICONTEXT && $this->icon) {
			$label = '<i class="icon icon-' . $this->icon . '" title="' . $this->getTitle(true) . '"></i> ' . $this->getTitle();
		}
		else{
			$label = $this->getTitle();
		}

		$atts = [];
		if($this->sortkey){
			$atts['data-sortkey'] = $this->sortkey;
			// The title for sort by is no longer needed; that is added to the icon itself.
			//$atts['title'] = 'Sort By ' . str_replace('"', '&quot;', $title);
		}
		if($this->renderkey){
			$atts['data-renderkey'] = $this->renderkey;
		}
		$atts['class'] = 'listing-table-cell-header ' . $this->getClass();
		$atts['data-viewkey'] = 'column-name-' . $this->name;
		$atts['data-viewtitle'] = $this->getTitle();
		
		
		$icons = '';
		if($this->sortkey) {
			$icons = '<div class="sort-icons">';

			$class = $this->currentSort === 'ASC' ? ' active' : '';
			$icons .= '<i title="' . t('STRING_SORT_BY_S_ASC', $this->getTitle(true)) . '"' .
				' data-key="' . $this->sortkey . '"' .
				' data-dir="up" class="icon sort-icon icon-sort-up' . $class . '"></i>';

			$class = $this->currentSort === 'DESC' ? ' active' : '';
			$icons .= '<i title="' . t('STRING_SORT_BY_S_DESC', $this->getTitle(true)) . '"' .
				' data-key="' . $this->sortkey . '"' .
				' data-dir="down" class="icon sort-icon icon-sort-down' . $class . '"></i>';

			$icons .= '</div>';
		}

		
		$out .= '<' . $tag;
		foreach($atts as $k => $v){
			$out .= ' ' . $k . '="' . $v . '"';
		}
		$out .= '>' . $label . $icons . '</' . $tag . '>';

		return $out;
	}

	public static function Factory() : Column {
		return new Column();
	}
}