<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace Core\ListingTable;

use Core\Date\DateTime;
use Core\ModelFactory;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\File;
use Core\FilterForm;
use Core\Forms\FormElement;
use Core\Router;
use Core\View;
use HTMLtoPDF\Converter as PDFConverter;

/**
 * Scaffolding class for a listing table Report, 
 * used to provide a downloadable report of the given data subset
 *
 * @author charlie
 */
abstract class Report {
	
	/**
	 * @var ModelFactory
	 */
	protected $_factory;
	
	/**
	 * @var Filter[]
	 */
	protected $_filters;
	
	public $testmode = false;

	/**
	 * @var array Key/Value pair of values set for this report
	 */
	public $optionValues = [];
	
	public function getEntries() : array {
		return $this->_factory->get();
	}

	/**
	 * Get the machine name of this report
	 *
	 * Used in the PDF filename generation
	 *
	 * @return string
	 */
	public function getName() : string {
		return 'generic-report';
	}

	/**
	 * Get the human-friendly name of this report
	 *
	 * @return string
	 */
	public function getTitle() : string {
		return 'Generic Report';
	}

	/**
	 * Get a short description of what this report provides
	 *
	 * @return string
	 */
	public function getDescription() : string {
		return 'Generic report';
	}
	
	public function getOptions() : array {
		return [];
	}

	/**
	 * Get all the options for this report as an array of FormElements
	 *
	 * @return FormElement[]
	 */
	public function getOptionsAsFormElements() : array {
		$ret = [];
		
		foreach($this->getOptions() as $name => $dat) {
			$type = $dat['type'];
			$dat['name'] = 'report_option[' . $name . ']';
			$dat['id'] = 'report-option-' . $name;

			if(isset($this->optionValues[ $name ])) {
				$dat['value'] = $this->optionValues[ $name ];
			}
			
			$ret[ $name ] = FormElement::Factory($type, $dat);
		}
		
		return $ret;
	}

	/**
	 * Retrieve the options for this report based on GET values
	 *
	 * @return array
	 */
	public function getOptionValues() : array {
		$values = [];

		foreach($this->getOptions() as $name => $dat) {
			if(isset($this->optionValues[ $name ])) {
				$values[ $name ] = $this->optionValues[ $name ];
			}
			elseif(isset($dat['value'])) {
				$values[ $name ] = $dat['value'];
			}
			else {
				$values[ $name ] = null;
			}
		}

		return $values;
	}

	/**
	 * Retrieve a single option value, or null if not available
	 *
	 * @param string $option
	 *
	 * @return mixed
	 */
	public function getOptionValue(string $option) {
		$opts = $this->getOptionValues();

		return $opts[$option] ?? null;
	}
	
	public function getTemplateName() : string {
		return '';
	}
	
	public function setFactory(ModelFactory $fac) {
		$this->_factory = $fac;
	}

	public function setFilters(array $fil) {
		$this->_filters = $fil;
	}

	public function renderAsPDF() {
		$file = $this->generate(true);
		$file->sendToUserAgent(!$this->testmode);
		$file->delete();
	}

	public function renderAsHTML() {
		$file = $this->generate(false);
		$file->sendToUserAgent(!$this->testmode);
		$file->delete();
	}
	
	public function generate(bool $asPDF = true) : File {
		// Setup
		$request = Router::$DefaultRequest;
		$this->_factory->limit(9999);
		// Reset the data so the streamer works.
		$this->_factory->getDataset()->_data = null;

		$view = new View();
		$view->mastertemplate = $asPDF ? 'skins/_print' : false;
		$view->templatename = $this->getTemplateName();
		

		// Populate the view
		$view->assign('report', $this);
		$view->assign('entries', $this->getEntries());
		$view->assign('filters', $this->_filters);
		$view->assign('options', $this->getOptionValues());
		$view->assign('as_pdf', $asPDF);
		$view->assign('generated_date', DateTime::NowGMT());
		
		// Create the report HTML
		$tmp = FileFactory::File('/tmp/' . $this->getName() . '-' . DateTime::Now('Ymd') . '.html');
		$tmp->putContents($view->fetch());
		
		if(!$asPDF || $this->testmode) {
			// If the HTML version is requested or testmode is enabled, simply return that.
			return $tmp;
		}

		// Build the PDF
		$conv = new PDFConverter();
		$conv->sourceFile = $tmp->getFilename();
		$pdf = $conv->convert();

		// Cleanup
		$tmp->delete();
		
		return $pdf;
	}
}
