<?php


namespace Core\ListingTable\Renderer;


use Core\ListingTable\Column;
use Core\ListingTable\Table;
use Core\ListingTable\TableInterface;
use Core\Model;

class CollectionRenderer  implements RendererInterface {

	private $_page;
	private $_total;
	private $_limit;
	private $_records;

	/**
	 * Constructor
	 *
	 * @param Column[] $columns
	 */
	public function __construct(TableInterface $table) {
		// Set the necessary information from the Table

		$this->_page    = $table->getPage();
		$this->_total   = $table->getTotalCount();
		$this->_limit   = $table->getLimit();
		$this->_records = $table->getRecords();
	}

	/**
	 * Render the contents of this interface with all the requested data
	 *
	 * @return array
	 *         +pager   bool   REQUIRED true/false if the pager is enabled
	 *         +options string REQUIRED HTML output of options for this interface
	 *         +body    string REQUIRED HTML output of the body
	 *         +page    int    Current page number, required if pager is set to true
	 *         +total   int    Total number of results, required if pager is set to true
	 *         +pages   int    Total number of pages, required if pager is set to true
	 */
	public function render() : array {
		$pages = (int) ceil($this->_total / $this->_limit);
		$body  = $this->_renderHeader();

		foreach($this->_records as $record) {
			$body .= $this->_renderModel($record);
		}

		$body .= $this->_renderFooter();

		return [
			'pager'   => true,
			'options' => '',
			'body'    => $body,
			'page'    => $this->_page,
			'total'   => $this->_total,
			'pages'   => $pages,
		];
	}

	/**
	 * Render the header for this listing
	 *
	 * @return string
	 */
	private function _renderHeader(): string {
		return '<ul class="listing-collection">';
	}

	/**
	 * Render an individual record
	 *
	 * @param Model $record
	 * @return string
	 */
	private function _renderModel(Model $record): string {
		// Here we just use the Model's built-in rendering method.
		return $record->render('collection-item');
	}

	/**
	 * Render the footer for this listing
	 *
	 * @return string
	 */
	private function _renderFooter(): string {
		return '</ul>';
	}
}