<?php

namespace Core\ListingTable\Renderer;

use Core\ListingTable\Column;
use Core\ListingTable\Table;
use Core\ListingTable\TableInterface;
use Core\Model;
use Core\Templates\Template;

/**
 * Class TableRenderer
 *
 * Renders data in a "table" like format, with rows and cells.
 *
 * @package Core\ListingTable\Renderer
 */
class TableRenderer implements RendererInterface {
	private $_allColumns;
	private $_columns;
	private $_bulkActions = false;
	private $_page;
	private $_total;
	private $_limit;
	private $_records;

	private $_tplModel;
	private $_tplRecord;

	/**
	 * TableRenderer constructor.
	 *
	 * @param TableInterface $table
	 */
	public function __construct(TableInterface $table) {

		// Set the necessary information from the Table
		$this->_allColumns  = $table->getColumns();
		$this->_columns     = $table->getColumnsVisible();
		$this->_bulkActions = $table->checkHasBulkActions();
		$this->_page        = $table->getPage();
		$this->_total       = $table->getTotalCount();
		$this->_limit       = $table->getLimit();
		$this->_records     = $table->getRecords();
	}

	/**
	 * Render the contents of this interface with all the requested data
	 *
	 * @return array
	 *         +pager   bool   REQUIRED true/false if the pager is enabled
	 *         +options string REQUIRED HTML output of options for this interface
	 *         +body    string REQUIRED HTML output of the body
	 *         +page    int    Current page number, required if pager is set to true
	 *         +total   int    Total number of results, required if pager is set to true
	 *         +pages   int    Total number of pages, required if pager is set to true
	 */
	public function render() : array {
		$pages = (int) ceil($this->_total / $this->_limit);
		$body  = $this->_renderHeader();

		foreach($this->_records as $record) {
			// The record may be a Model or a generic array.
			// The latter is useful for manual data management such as from multiple tables
			// or non Core-DMI sources.
			if($record instanceof Model) {
				$body .= $this->_renderModel($record);
			}
			else {
				$body .= $this->_renderRecord($record);
			}
		}

		$body .= $this->_renderFooter();

		return [
			'pager'   => true,
			'options' => $this->_renderOptions(),
			'body'    => $body,
			'page'    => $this->_page,
			'total'   => $this->_total,
			'pages'   => $pages,
		];
	}

	private function _renderHeader() :  string {
		$header = '<div class="listing-table-record listing-table-header">';

		if($this->_bulkActions) {
			$header .= <<<EOD
	<div class="listing-table-cell-header listing-table-cell column-bulk-select screen">
		<input type="checkbox" class="bulk-select-all"/>
	</div>
EOD;
		}

		foreach($this->_columns as $column) {
			$header .= $column->getDIV();
		}

		$header .= '<div class="listing-table-cell-header listing-table-cell column-controls screen">&nbsp;</div>';
		$header .= '</div>';

		return $header;
	}

	private function _renderModel(Model $record) : string {
		if($this->_tplModel === null) {
			$this->_tplModel = Template::Factory('includes/listingtable/table-model');
			$this->_tplModel->assign('columns', $this->_columns);
			$this->_tplModel->assign('bulk_actions', $this->_bulkActions);
		}

		$this->_tplModel->assign('record', $record);
		return $this->_tplModel->fetch();
	}

	private function _renderRecord(array $record): string {
		if($this->_tplRecord === null) {
			$this->_tplRecord = Template::Factory('includes/listingtable/table-record');
			$this->_tplRecord->assign('columns', $this->_columns);
			$this->_tplRecord->assign('bulk_actions', $this->_bulkActions);
		}

		$this->_tplRecord->assign('record', $record);
		return $this->_tplRecord->fetch();
	}

	private function _renderFooter() : string {
		return '';
	}

	private function _renderOptions() : string {
		$tpl = Template::Factory('includes/listingtable/table-options');
		$tpl->assign('limit', $this->_limit);
		$tpl->assign('columns', $this->_allColumns);
		return $tpl->fetch();
	}
}