<?php

/*
 * @copyright Copyright (C) 2009-2017  charlie
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Logger;

use Core\Datamodel\DMIQueryException;
use Exception;

/**
 * Description of LogEntry
 *
 * @author charlie
 */
class LogEntry {
	
	/**
	 * @var string Caller information that may help developers to track down where this entry was triggered
	 */
	public $caller = null;
	
	/**
	 * @var string An optional identifier for this log entry
	 */
	public $code = null;
	
	/**
	 * @var string An optional lengthy description for this log entry
	 */
	public $details = null;
	
	/**
	 * @var string An optional icon to display along with this log message
	 */
	public $icon = null;
	
	/**
	 * @var int The level of this log entry.
	 * 
	 * LOG_LEVEL_CRITICAL (1)
	 * LOG_LEVEL_HIGH (2)
	 * LOG_LEVEL_INFO (3)
	 * LOG_LEVEL_DEBUG (4)
	 * LOG_LEVEL_TRACE (5)
	 */
	public $level = \LOG_LEVEL_INFO;
	
	/**
	 * @var string The actual message to log
	 */
	public $message = null;
	
	/**
	 * @var string An optional source to keep track of what type of message this comes from
	 */
	public $source = null;

	/**
	 * @var array Hash of custom tags for this log entry
	 */
	public $tags = [];
	
	/**
	 * @var array Trace report for this LogEntry, when available.
	 */
	public $trace = null;
	
	/**
	 * @var string Type of log entry: general, security, or some custom string
	 */
	public $type = 'application';

	/**
	 * @var string An optional user ID if this log entry affects a given user
	 */
	public $user = null;

	public function __construct() {

	}

	/**
	 * Chainable method to set the caller file location
	 * 
	 * @param string $caller
	 * 
	 * @return $this
	 */
	public function caller(string $caller) : LogEntry {
		$this->caller = $caller;
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param string $code The code of the log entry
	 * 
	 * @return $this
	 */
	public function code(string $code) : LogEntry {
		$this->code = $code;
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param string $details Any supplemental details for this entry
	 * 
	 * @return $this
	 */
	public function details(string $details) : LogEntry {
		$this->details = $details;
		return $this;
	}
	
	/**
	 * Chainable method to set the log details from an Exception
	 * 
	 * @param Exception $ex
	 */
	public function exception(Exception $ex) : LogEntry {
		$this
			->caller($ex->getFile() . ':' . $ex->getLine())
			->code(get_class($ex))
			->message($ex->getMessage())
			->trace($ex->getTrace())
			->type('application')
			->level(LOG_LEVEL_HIGH);

		if($ex instanceof DMIQueryException && $ex->query) {
			// These have extra pieces of information to log.
			$this->details($ex->query);
		}
		
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param string $icon You can even fucking do an icon here!
	 * 
	 * @return $this
	 */
	public function icon(string $icon) : LogEntry {
		$this->icon = $icon;
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param int $level The level of this message, should be one of the LOG_LEVEL constants
	 * 
	 * @return $this
	 */
	public function level(int $level) : LogEntry {
		$this->level = $level;
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param string $message The message of the log entry
	 * 
	 * @return $this
	 */
	public function message(string $message) : LogEntry {
		$this->message = $message;
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * The "source" of a log entry should be an object in Core as opposed to a file and line.
	 * 
	 * The expected use of this is have the application being able to query any entries that
	 * pertain to "mything-1234" and display those entries along with MyThingModel(1234).
	 * 
	 * @param string $source
	 * 
	 * @return $this
	 */
	public function source(string $source) : LogEntry {
		$this->source = $source;
		return $this;
	}
	
	/**
	 * Chainable method to set a custom tag along with this log entry
	 * 
	 * @param string $key
	 * @param string $value
	 * @return \Core\Logger\LogEntry
	 */
	public function tag(string $key, string $value) : LogEntry {
		$this->tags[ $key ] = $value;
		
		// Allow chaining
		return $this;
	}
	
	/**
	 * Chainable method to set the trace stack for this LogEntry.
	 * 
	 * This is useful when used along with `debug_backtrace(DEBUG_BACKTRACE_IGNORE_ARGS)` as the parameter.
	 * 
	 * @param array $trace
	 * 
	 * @return $this
	 */
	public function trace(array $trace = null) : LogEntry {
		if($trace === null){
			// Ignore empty traces.
			return $this;
		}
		
		$this->trace = [];
		
		// Use this to trim off unnecessary text in the trace.
		$relative = dirname(__FILE__, 6);

		$counter = 0;
		foreach($trace as $t){
			$counter++;
			
			if(isset($t['class']) && isset($t['type'])){
				$call = $t['class'] . $t['type'] . $t['function'] . '()';
			}
			else{
				$call = $t['function'] . '()';
			}
			
			$line = isset($t['line']) ? $t['line'] : null;
			
			$file = isset($t['file']) ? $t['file'] : null;
			
			$rel = isset($t['file']) && strpos($t['file'], $relative) === 0 ?
				substr($t['file'], strlen($relative) + 1) :
				$file;
			
			$this->trace[] = [
				'counter'  => $counter,
				'call'     => $call,
				'line'     => $line,
				'file'     => $file,
				'relative' => $rel,
				'formatted' => ' #' . $counter . ': ' . $call . ($rel ? ' called from ' . $rel : '') . ($rel && $line ? ':' . $line : ''),
			];
		}
		
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param string $type The type of the log, should be lowercase and all one word
	 * 
	 * @return $this
	 */
	public function type(string $type) : LogEntry {
		if($type === 'error') {
			$this->type = 'application';
			$this->level = LOG_LEVEL_HIGH;
		}
		elseif($type === 'info') {
			$this->type = 'application';
			$this->level = LOG_LEVEL_INFO;
		}
		else {
			// Type should just be a string, no more than 32 characters.
			// To be safe, (since it could be sent to external systems),
			// strip non-string characters out.
			$this->type = substr(filter_var($type, FILTER_SANITIZE_STRING), 0, 32);
		}

		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param string $user
	 * 
	 * @return $this
	 */
	public function userid(string $user) : LogEntry {
		$this->user = $user;
		return $this;
	}
	
	/**
	 * Chainable method to set the value because PHP doesn't fucking have named arguments.
	 * 
	 * @param \UserModel $user
	 * 
	 * @return $this
	 */
	public function user(\UserModel $user) : LogEntry {
		$this->user = $user->get('id');
		return $this;
	}
	
	/**
	 * Send this log entry to the various Log Sinks configured on the system.
	 */
	public function send(){
		Logger::Log($this);
	}
	
	/**
	 * Shorthand to initialize a new LogEntry object
	 * 
	 * @return \Core\Logger\LogEntry
	 */
	public static function Init() : LogEntry {
		return new self();
	}
}
