<?php

namespace Core\Renderer;

use Core\Filestore\File;
use Core\Filestore\Factory as FileFactory;
use Core\Filestore\Backends\FileRemote;
use Core\Filestore\FileHelper;
use Core\Filestore\FileMetaHelper;
use Core\Utilities\Profiler\ProfilerEvent;
use Exception;

class ImageRenderer {
	
	/**
	 * @var array HTML attributes to tack onto the <img> tag.
	 */
	private $_attributes = [];
	
	/**
	 * @var bool Set to true if the image should be loaded inline
	 */
	private $_isInline = false;
	
	/**
	 * @var bool Set to true if this image should just be a transparent gif
	 */
	private $_isBlank = false;
	
	/**
	 * @var string[] Dimensions to render the image as
	 */
	private $_dimensions = [];
	
	/**
	 * @var File Source file to render
	 */
	private $_file;
	
	/**
	 * @var string Placeholder image to render if no file available
	 */
	private $_placeholder;

	/**
	 * @var FileMetaHelper Helper to retrieve file metadata for an image
	 */
	private $_meta;
	
	/**
	 * @var string Blank GIF base64 code
	 */
	const BlankGif = 'data:image/gif;base64,R0lGODlhAQABAIAAAP///wAAACH5BAEAA‌​AAALAAAAAABAAEAAAICR‌​AEAOw==';
	
	/**
	 * Set the source for this image renderer as a File
	 * 
	 * @param File $file
	 */
	public function setFile(File $file) {
		$this->_file = $file;
		$this->_isBlank = false;
	}
	
	/**
	 * Set the source for this image as a src or pathname
	 * 
	 * @param string $src
	 */
	public function setSource(string $src) {
		if($src === '' || $src === 'BLANK') {
			$this->setBlank();
		}
		else {
			$this->setFile(FileFactory::File($src));
		}
	}
	
	/**
	 * Set to render just a blank image
	 */
	public function setBlank() {
		$this->_file = null;
		$this->_isBlank = true;
	}
	
	/**
	 * Set this image to attempt to render inline
	 * 
	 * @param bool $inline 
	 */
	public function setInline(bool $inline = true) {
		$this->_isInline = $inline;
	}
	
	/**
	 * Set the placeholder for this image if the original image was not located
	 * 
	 * @param string $placeholder
	 */
	public function setPlaceholder(string $placeholder) {
		$this->_placeholder = $placeholder;
	}
	
	/**
	 * Set the base dimensions for this image
	 * 
	 * @param string $dimensions
	 */
	public function setDimensions(string $dimensions) {
		$this->_dimensions['_BASE_'] = $dimensions;
	}
	
	/**
	 * Set the array of attributes to pass into this Image
	 * 
	 * @param array $attribute
	 */
	public function setAttributes(array $attributes) {
		foreach($attributes as $k => $v) {
			switch($k) {
				case 'inline':
					$this->setInline();
					break;
				case 'src':
					$this->setSource($v);
					break;
				case 'file':
					$this->setFile($v);
					break;
				case 'dimensions':
					$this->setDimensions($v);
					break;
				case 'placeholder':
					$this->setPlaceholder($v);
					break;
				case 'sizes':
					$sizes = explode(';', $v);
					foreach($sizes as $s) {
						list($media, $dimensions) = explode('=', $s);
						$this->addMediaSize($media, $dimensions);
					}
					break;
				case 'width':
				case 'height':
					// Skip these and process them seperately
					break;
				default:
					// All other attributes get mapped directly as HTML attributes.
					$this->setAttribute($k, $v);
			}
		}
		
		if(isset($attributes['width']) && isset($attributes['height'])) {
			// Width and Height are requested, map these to dimensions.
			$this->setDimensions($attributes['width'] . 'x' . $attributes['height']);
		}
		elseif(isset($attributes['width'])) {
			// Only width was requested, assume a square based on that value.
			$this->setDimensions($$attributes['width'] . 'x' . $attributes['width']);
		}
	}
	
	/**
	 * Set a single attribute onto this image
	 * 
	 * @param string $key
	 * @param mixed  $value
	 */
	public function setAttribute(string $key, $value) {
		$this->_attributes[$key] = $value;
	}
	
	/**
	 * Add a media query size for a differently scaled image
	 * 
	 * @param string $media
	 * @param string $dimensions
	 */
	public function addMediaSize(string $media, string $dimensions) {
		$this->_dimensions[ $media ] = $dimensions;
	}
	
	/**
	 * Fetch the resulting HTML for this image
	 */
	public function fetch() : string {

		$baseDimensions = isset($this->_dimensions['_BASE_']) ? $this->_dimensions['_BASE_'] : '';
		
		if(
			$this->_file &&
			$this->_file->exists() &&
			$this->_file->isImage()
		) {
			$profiler = new ProfilerEvent('Rendering image ' . $this->_file->getFilename());
			$profiler->start();


			// Ensure some basic attributes for local images.
			if(empty($this->_attributes['alt'])) {
				// If the alt tag is either omitted or empty, generate it from the metadata.
				try {
					$alt = $this->_getMeta()->getMetaTitle('title');
				}
				catch(Exception $ex) {
					$alt = null;
				}

				if(!$alt) {
					$alt = $this->_file->getTitle();
				}

				$this->setAttribute('alt', $alt);
			}


			$hasMediaDimensions = count($this->_dimensions) > 1;
			
			// Allow for additional media sizes.
			if($hasMediaDimensions) {
				$html = $this->_getImage($this->_file, $baseDimensions, true);
				foreach($this->_dimensions as $media => $size) {
					if($media !== '_BASE_') {
						$html = $this->_getSource($this->_file, $media, $size) . $html;
					}
				}
				
				$html = '<picture ' . $this->_getAttributes() . '>' . $html . '</picture>';
			}
			else {
				$html = $this->_getImage($this->_file, $baseDimensions);
			}
		}
		elseif($this->_file && $this->_file->exists()) {
			// It exists, but is not an image, we can still render a mimetype.
			$profiler = new ProfilerEvent('Rendering mimetype image for ' . $this->_file->getFilename());
			$profiler->start();

			$html = $this->_getImage($this->_file, $baseDimensions);
		}
		elseif($this->_placeholder) {
			$profiler = new ProfilerEvent('Rendering placeholder image ' . $this->_placeholder);
			$profiler->start();

			$placeholder = FileFactory::File('assets/images/placeholders/' . $this->_placeholder . '.svg');

			if($placeholder) {
				$html = $this->_getImage($placeholder, $baseDimensions);
			}
			else {
				$html = $this->_placeholder . ' not found';
			}

		}
		elseif($this->_isBlank) {
			$profiler = new ProfilerEvent('Rendering blank image');
			$profiler->start();

			$html = $this->_getBlankImage($baseDimensions);
		}
		else {
			$profiler = new ProfilerEvent('Rendering unavailable image');
			$profiler->start();

			$html = 'No image available';
		}
		
		
		/*
		// If the extended metadata was requested... look that up too!
		if(isset($params['includemeta']) && $params['includemeta'] && $f){
			
			$metahelper  = new \Core\Filestore\FileMetaHelper($f);
			$metacontent = $metahelper->getAsHTML();
			if($metacontent){
				$html = '<div class="image-metadata-wrapper">' . $html . $metacontent . '</div>';
			}
		}
		*/
		

		$profiler->stop();
		return $html;
	}
	
	/**
	 * Get the <img> tag for a given File
	 * 
	 * @param File $src
	 * @param string $dimensions
	 * 
	 * @return string
	 */
	private function _getImage(File $src, string $dimensions, bool $includeAttributes = true) : string {
		$attributes = $includeAttributes ? $this->_getAttributes() : '';
		
		if(
			$this->_isInline && 
			strpos($src->getMimetype(), 'image/svg') === 0 && 
			$src->getType() === File::TYPE_ASSET
		) {
			// Inline asset SVGs can be rendered directly to the browser.
			return $src->getContents();
		}
		elseif(
			$this->_isInline &&
			strpos($src->getMimetype(), 'image/svg') === 0
		) {
			// Inline asset SVGs are encapsulated as base64 code, just like raster images.
			$attributes .= ' ' . $this->_getWidthHeight($dimensions);
			$url = $this->_getSrc($src, $dimensions);
		}
		elseif(
			$src instanceof FileRemote || 
			strpos($src->getMimetype(), 'image/svg') === 0
		) {
			// Remote images and SVGs already use an explicit width and height attributes, (at least when provided)
			$attributes .= ' ' . $this->_getWidthHeight($dimensions);
			$url = $src->getURL();
		}
		elseif($src->isImage()) {
			// Otherwise, raster and local images are resized on the server and provided smaller.
			$url = $this->_getSrc($src, $dimensions);
		}
		else {
			// Other file type, render the src of the mimetype icon.
			$mimetype = FileHelper::GetIconFileForMimetype($src->getMimetype());

			$url = $this->_getSrc($mimetype, $dimensions);
		}
		
		return '<img src="' . $url . '" ' . $attributes . '/>';
	}
	
	/**
	 * Get the <source> tag for a given File when multiple media sizes are provided
	 *
	 * @param File $src
	 * @param string $media
	 * @param string $dimensions
	 *
	 * @return string
	 */
	private function _getSource(File $src, string $media, string $dimensions) : string {
		if(
			$src instanceof FileRemote ||
			strpos($src->getMimetype(), 'image/svg') === 0
		) {
			// Remote images and SVGs already use an explicit width and height attributes, (at least when provided)
			$url = $src->getURL();
		}
		else {
			// Otherwise, raster and local images are resized on the server and provided smaller.
			$url = $this->_getSrc($src, $dimensions);
		}
		
		return '<source media="' . $media . '" srcset="' . $url . '"/>';
	}
	
	/**
	 * Get the value for the src attribute based on the given File
	 * 
	 * @param File $src
	 * @param string $dimensions
	 * 
	 * @return string
	 */
	private function _getSrc(File $src, string $dimensions) : string {
		if($src instanceof FileRemote) {
			// Remote files always just get the URL.
			return $src->getURL();
		}
		elseif($this->_isInline && $src->getFilesize() < 1048576*4 && $dimensions) {
			// Local file and inline was requested, retrieve the actual contents of the image.
			// Only resize when dimensions are requested.
			$preview = $src->getPreviewFile($dimensions);
			return 'data:' . $preview->getMimetype() . ';base64,' . base64_encode($preview->getContents());
		}
		elseif($this->_isInline && $src->getFilesize() < 1048576) {
			// Inline was requested, but no dimensions were requested, just try the base image itself.
			return 'data:' . $src->getMimetype() . ';base64,' . base64_encode($src->getContents());
		}
		elseif($dimensions) {
			// Dimensions were requested, so resize to those sizes
			$preview = $src->getQuickPreviewFile($dimensions);
			return $preview->exists() ? $preview->getURL() : \Core\resolve_link('/file/preview') . '?f=' . $src->getFilenameHash() . '&d=' . $dimensions;
		}
		else {
			// No dimensions were requested, so just retrieve the base image URL itself.
			return $src->getURL();
		}
	}
	
	private function _getBlankImage(string $dimensions) : string {
		return '<img src="' . self::BlankGif . '" ' . $this->_getAttributes() . ' ' . $this->_getWidthHeight($dimensions) . '/>';
	}
	
	private function _getAttributes() : string {
		$atts = [];
		foreach($this->_attributes as $k => $v) {
			$atts[] = $k . '="' . htmlentities($v) . '"';
		}
		
		return implode(' ', $atts);
	}

	/**
	 * Get the file metahelper associated with this image.
	 *
	 * @return FileMetaHelper
	 *
	 * @throws \Exception
	 */
	private function _getMeta() : FileMetaHelper {
		if($this->_meta === null) {
			$this->_meta = FileMetaHelper::ConstructFromFile($this->_file);
		}

		return $this->_meta;
	}
	
	private function _getWidthHeight(string $dimensions) : string {
		if($dimensions === '') {
			return '';
		}
		$width = preg_replace('#[^0-9]?([0-9]*)x.*#', '$1', $dimensions);
		$height = preg_replace('#.*x([0-9]*)[^0-9]?#', '$1', $dimensions);
		
		return 'width="' . $width . '" height="' . $height . '"';
	}
}

