<?php
/**
 * File for class ModelHelper definition in the tenant-visitor project
 * 
 * @package Core\Search
 * @author Charlie Powell <charlie@evalagency.com>
 * @date 20131126.1744
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Search;
use Core\Datamodel\Dataset;
use Core\Datamodel\DatasetSelect;
use Core\Datamodel\DatasetWhereClause;
use Core\Model;


/**
 * A short teaser of what ModelHelper does.
 *
 * More lengthy description of what ModelHelper does and why it's fantastic.
 *
 * <h3>Usage Examples</h3>
 *
 *
 * @todo Write documentation for ModelHelper
 * <h4>Example 1</h4>
 * <p>Description 1</p>
 * <code>
 * // Some code for example 1
 * $a = $b;
 * </code>
 *
 *
 * <h4>Example 2</h4>
 * <p>Description 2</p>
 * <code>
 * // Some code for example 2
 * $b = $a;
 * </code>
 *
 * 
 * @package Core\Search
 * @author Charlie Powell <charlie@evalagency.com>
 *
 */
class Helper {
	/**
	 * Get an array of words to skip in search and indexing.
	 *
	 * @return array
	 */
	public static function GetSkipWords(){
		return ['a', 'an', 'but', 'for', 'nor', 'of', 'or', 'so', 'the', 'to', 'yet'];
	}

	/**
	 * Hook handler to save the index data for a given model record.
	 *
	 * @param \Model $model
	 *
	 * @return bool
	 */
	public static function ModelPreSaveHandler(Model $model){
		// Only process models that have a search index on them.
		if(!$model::$HasSearch) return true;

		$str = $model->getSearchIndexString();

		// Lowercase it.
		$str = strtolower($str);

		// Convert this string to latin.
		$str = \Core\str_to_latin($str);

		// Contractions get dropped with no space.
		$str = str_replace("'", '', $str);

		// Skip punctuation.
		$str = preg_replace('/[^a-z0-9 ]/', ' ', $str);

		$parts = explode(' ', $str);

		// Conjunction words to skip
		$skips = self::GetSkipWords();
		
		// Generate the plain text index and only run the doublemetaphone conversion if it changes.
		// This is because that function is painfully slow, so the less it runs the better!
		foreach($parts as $k => $word){
			if(!$word){
				// Skip blank words
				unset($parts[$k]);
			}
			elseif(in_array($word, $skips)){
				// Skip "skip" words
				unset($parts[$k]);	
			}
			// OK to add, no else needed here.
		}
		
		// Strip uniques
		$parts = array_unique($parts);
		
		$model->set('search_index_str', implode(' ', $parts));
		
		if($model->changed('search_index_str')){
			// It changed, NOW generate the doublemetaphone conversion strings.
			$primary = [];
			$secondary = [];
			foreach($parts as $k => $word){
				$it = new DoubleMetaPhone($word);
				$primary[] = $it->primary;
				$secondary[] = $it->secondary;
			}
			
			$model->set('search_index_pri', implode(' ', $primary));
			$model->set('search_index_sec', implode(' ', $secondary));
		}

		return true;
	}

	public static function AddSearchToDataset(string $query, Dataset $ds, float $relevancy = 0.5) {
		/*
		 * (
		LEAST((LENGTH(search_index_str) - LENGTH(REPLACE(search_index_str, 'thersa', ''))), 1) +
		LEAST((LENGTH(search_index_pri) - LENGTH(REPLACE(search_index_pri, '0RS', ''))), 1) +
		LEAST((LENGTH(search_index_sec) - LENGTH(REPLACE(search_index_sec, 'TRS', ''))), 1)
	) / 3 __relevancy,

		HAVING __search_relevancy > .6
		 */

		list($indexes, $primaries, $secondaries) = self::ConvertStringToKeywords($query);

		$subwhere = new DatasetWhereClause('search');
		$subwhere->setSeparator('or');

		$selectCount    = 0;
		$selectFormulas = [];
		$selectParts = [];

		// And add a where clause for each one.
		foreach($indexes as $word){
			$subwhere->addWhere('search_index_str LIKE %' . $word . '%');
			$selectFormulas[] = 'LEAST((LENGTH(search_index_str) - LENGTH(REPLACE(search_index_str, %s, ""))), 1) * 3';
			$selectParts[] = $word;
			$selectCount += 3;
		}
		foreach($primaries as $word){
			if(strlen($word) > 2) {
				// We need at least 3 syllables for this lookup to really work.
				$subwhere->addWhere('search_index_pri LIKE %' . $word . '%');
				$selectFormulas[] = 'LEAST((LENGTH(search_index_pri) - LENGTH(REPLACE(search_index_pri, %s, ""))), 1) * 2';
				$selectParts[] = $word;
				$selectCount += 2;
			}
		}
		foreach($secondaries as $word){
			if(strlen($word) > 2) {
				// We need at least 3 syllables for this lookup to really work.
				$subwhere->addWhere('search_index_sec LIKE %' . $word . '%');
				$selectFormulas[] = 'LEAST((LENGTH(search_index_sec) - LENGTH(REPLACE(search_index_sec, %s, ""))), 1) * 1';
				$selectParts[] = $word;
				$selectCount += 1;
			}
		}

		$searchRelevancySelect = new DatasetSelect();
		$searchRelevancySelect->formula = '( ' . implode(' + ', $selectFormulas) . ' ) / ' . $selectCount;
		$searchRelevancySelect->alias = '__search_relevancy';
		$searchRelevancySelect->formulaParams = $selectParts;

		$ds->where($subwhere);
		$ds->select($searchRelevancySelect);
		$ds->having('__search_relevancy >= ' . $relevancy);
	}

	/**
	 * Translate a query string to a populated where clause based on the search index criteria.
	 *
	 * @param string $query
	 *
	 * @return DatasetWhereClause
	 */
	public static function GetWhereClause($query){
		$subwhere = new DatasetWhereClause('search');
		$subwhere->setSeparator('or');

		// Lowercase it.
		$query       = strtolower($query);
		// Convert this string to latin.
		$query       = \Core\str_to_latin($query);
		// Skip punctuation.
		$query       = preg_replace('/[^a-z0-9 ]/', '', $query);
		// Split out words.
		$parts       = explode(' ', $query);
		$skips       = self::GetSkipWords();
		$indexes     = [];
		$primaries   = [];
		$secondaries = [];

		foreach($parts as $word){
			if(in_array($word, $skips)) continue;
			if(!$word) continue;

			$it = new DoubleMetaPhone($word);

			$indexes[]     = $word;
			$primaries[]   = $it->primary;
			$secondaries[] = $it->secondary;
		}

		// Remove duplicates
		$indexes = array_unique($indexes);
		$primaries = array_unique($primaries);
		$secondaries = array_unique($secondaries);

		// And add a where clause for each one.
		foreach($indexes as $word){
			if($word) {
				// Required to skip spaces.
				$subwhere->addWhere('search_index_str LIKE %' . $word . '%');
			}
		}
		foreach($primaries as $word){
			if($word){
				// Required to skip spaces.
				$subwhere->addWhere('search_index_pri LIKE %' . $word . '%');
			}
		}
		foreach($secondaries as $word){
			if($word){
				// Required to skip spaces.
				$subwhere->addWhere('search_index_sec LIKE %' . $word . '%');
			}
		}

		return $subwhere;
	}

	/**
	 * Sanitize and convert a query string to the set of queries necessary for searching using double metaphone.
	 *
	 * Will return an array consisting of 3 keys, 0 is the index, 1 is primary, and 2 is secondary.
	 *
	 * @param string $query
	 *
	 * @return array
	 */
	public static function ConvertStringToKeywords(string $query) : array {
		// Lowercase it.
		$query       = strtolower($query);
		// Convert this string to latin.
		$query       = \Core\str_to_latin($query);
		// Skip punctuation.
		$query       = preg_replace('/[^a-z0-9 ]/', '', $query);
		// Split out words.
		$parts       = explode(' ', $query);
		$skips       = self::GetSkipWords();
		$indexes     = [];
		$primaries   = [];
		$secondaries = [];

		foreach($parts as $word){
			if(in_array($word, $skips)) {
				// Skip short or common words
				continue;
			}
			if(!$word) {
				// Skip empty words
				continue;
			}

			$it = new DoubleMetaPhone($word);

			$indexes[]     = $word;
			$primaries[]   = $it->primary;
			$secondaries[] = $it->secondary;
		}

		// Remove duplicates
		$indexes = array_unique($indexes);
		$primaries = array_unique($primaries);
		$secondaries = array_unique($secondaries);

		return [$indexes, $primaries, $secondaries];
	}
	
	public static function CalculateSearchRelevance(Model $model, string $query) : float {
		// Lowercase it.
		$query       = strtolower($query);
		// Convert this string to latin.
		$query       = \Core\str_to_latin($query);
		// Skip punctuation.
		$query       = preg_replace('/[^a-z0-9 ]/', '', $query);
		// Split out words.
		$parts       = explode(' ', $query);
		
		$ignore = self::GetSkipWords();
		foreach($parts as $k => $word){
			// Skip blank words.
			if(!$word){
				unset($parts[$k]);
				continue;
			}
			// Unset any to-be-skipped word.
			if(in_array($word, $ignore)){
				unset($parts[$k]);
				continue;
			}
		}
		
		// All query words in the result mean 100% relevancy.
		$size = sizeof($parts);
		if(!$size){
			return 0.0;
		}
		
		$wordweight = 100 / $size;
		// And each word has 3 parts to it.
		$wordweight /= 3;
		
		$rel = 0.0;
		
		$str = explode(' ', $model->get('search_index_str'));
		$pri = explode(' ', $model->get('search_index_pri'));
		$sec = explode(' ', $model->get('search_index_sec'));
		
		foreach($parts as $word){
			$it = new DoubleMetaPhone($word);
			
			if(in_array($word, $str)){
				// Exact matches here get an automatic boost!
				$rel += ($wordweight * 3);
			}
			else{
				foreach($str as $w){
					if(strpos($w, $word) !== false){
						// If a partial match is located, add a fraction of the word weight, (since it wasn't a complete match).
						$rel += $wordweight * (strlen($word) / strlen($w));
						break;
					}
				}
			}
			
			if(in_array($it->primary, $pri)) {
				// This word was matched in the primary search, add the standard weight
				$rel += $wordweight;
			}
			if(in_array($it->secondary, $sec)) {
				// This word was matched in the secondary search, add the standard weight
				$rel += $wordweight;
			}
		}
		
		return floatval( min($rel, 100) );
	}
} 
