<?php
/**
 * [PAGE DESCRIPTION HERE]
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Templates;

use Core\ConfigHandler;
use Core\Core;
use Core\Filestore\Factory as FileFactory;
use Error;
use Core\Utilities\Profiler\ProfilerEvent;


/**
 * Class Template description
 * 
 * @package Core\Templates
 */
class Template {
	
	private $_variables = [];
	private $_template;
	private $_renderer;
	private $_filename;

	public static $Renderers = [
		'tpl' => 'Core\\Templates\\Backends\\Smarty',
		'phtml' => 'Core\\Templates\\Backends\\PHTML',
	];
	
	private static $_Paths = null;
	private static $_Files = null;
	private static $_Templates = null;

	/**
	 * Template constructor.
	 *
	 * @param string|null $template
	 *
	 * @throws TemplateException If the template could not be loaded
	 */
	public function __construct(string $template = null) {
		if($template !== null && $template !== '') {
			// Auto set the template filename, for convenience.
			$this->setFilename($template);
		}
	}
	
	/**
	 * Assign a variable to be accessible inside the template.
	 * 
	 * @param string $key   Variable name to make available 
	 * @param mixed  $value Contents of the variable
	 */
	public function assign(string $key, $value) {
		$this->_variables[ $key ] = $value;
	}
	
	/**
	 * Get an assigned variable or NULL if not set
	 * 
	 * @param string $key
	 * 
	 * @return NULL|mixed
	 */
	public function getVariable(string $key) {
		return $this->_variables[ $key ] ?? null;
	}
	
	/**
	 * Set the filename or template name for this Template
	 * 
	 * Can be fully resolved or shorthand
	 * 
	 * @param string $template
	 *
	 * @throws TemplateException If the template could not be loaded
	 */
	public function setFilename(string $template) {
		if(self::$_Paths === null) {
			self::RequeryPaths();
		}
		
		$this->_filename = self::ResolveFile($template);
		
		// Resolve the renderer based on the discovered extension.
		$ext = pathinfo($this->_filename, PATHINFO_EXTENSION);
		if(isset(self::$Renderers[ $ext ])) {
			$this->_renderer = self::$Renderers[ $ext ];
		}
	}
	
	/**
	 * Check if this template exists and is available on the filesystem
	 * 
	 * @return bool
	 */
	public function checkExists() : bool {
		return $this->_filename !== null && $this->_filename !== '' && is_file($this->_filename);
	}


	/**
	 * Render this Template with the underlying Renderer and return the data
	 *
	 * @param string $template Optional filename to render, if not set already.
	 *
	 * @return string
	 *
	 * @throws \ReflectionException
	 * @throws TemplateException
	 */
	public function fetch(string $template = null) : string {
		
		if($template !== null) {
			$this->setFilename($template);
		}
		
		if(!$this->_filename) {
			throw new TemplateException('No filename requested for Template rendering.');
		}
		
		if(!$this->checkExists()) {
			throw new TemplateException($this->_filename . ' does not exist and the Template cannot be rendered.');
		}
		
		if(!class_exists($this->_renderer)) {
			throw new TemplateException('Template renderer ' . $this->_renderer . ' was not located for ' . $this->_filename);
		}

		$ref = new \ReflectionClass($this->_renderer);
		$renderer = $ref->newInstance();
		foreach($this->_variables as $key => $value) {
			$renderer->assign($key, $value);
		}

		try {
			return $renderer->fetch($this->_filename);
		}
		catch(Error $er) {
			// Convert this to a standard Exception
			throw new TemplateException($er->getMessage());
		}
	}

	/**
	 * @param string $filename Filename of the template
	 * 
	 * @return TemplateInterface
	 */
	public static function Factory(string $filename) : TemplateInterface {
		try{
			// As a convenience, attempt to auto-resolve it.
			// This is not required and may fail if the user did not specify a resolveable path
			// for the template instance.  This is allowed as the template can exist without being
			// linked to an underlying file!
			$filename = self::ResolveFile($filename);
		}
		catch (TemplateException $ex) {
			// Ignore.
		}
		
		// The template backend will depend on the extension.
		$ext = pathinfo($filename, PATHINFO_EXTENSION);

		switch($ext){
			case 'phtml':
				$template = new Backends\PHTML();
				break;
			case 'tpl':
				$template =  new Backends\Smarty();
				break;
			default:
				$template =  new Backends\Smarty();
				break;
		}

		if($filename) {
			$template->setFilename($filename);
		}
		
		return $template;
	}

	/**
	 * Resolve a filename stub to a fully resolved path.
	 *
	 * @param string $filename Filename to resolve
	 *
	 * @throws TemplateException
	 * 
	 * @return string
	 */
	public static function ResolveFile(string $filename) : string {
		if(strpos($filename, ROOT_PDIR) === 0){
			// If the template starts with the root pdir, then it must already be fully resolved!
			return $filename;
		}
		
		if(self::$_Paths === null) {
			self::RequeryPaths();
		}
		
		if(!is_array(self::$_Files)) {
			self::$_Files = [];
		}
		
		// Allow caching to speed up checks.
		$cachekey = $filename;
		if(is_array(self::$_Files) && isset(self::$_Files[ $cachekey ] ) ) {
			return self::$_Files[ $cachekey ];
		}

		// If a path with no extension is requested, allow that to be auto-resolved based on the list of supported templates.
		$ext = pathinfo($filename, PATHINFO_EXTENSION);
		$renderers = array_keys(self::$Renderers);

		if(!in_array($ext, $renderers)) {
			// If the extension was requested, but it's not a valid renderer type,
			// assume the user meant something like "skins/_stream.html.phtml".
			// This will revert the extension back to "AUTO"
			$ext = '';
		}
		
		$dirs = self::GetPaths();

		// Trim off the beginning '/' if there is one;  All directories end with a '/'.
		if ($filename[0] == '/'){
			$filename = substr($filename, 1);
		}

		foreach ($dirs as $d) {
			if($ext === '') {
				// Automatic
				foreach($renderers as $r) {
					if (file_exists($d . $filename . '.' . $r) && is_file($d . $filename . '.' . $r)){
						self::$_Files[ $cachekey ] = $d . $filename . '.' . $r;
						return self::$_Files[ $cachekey ];
					}
				}
			}
			elseif(file_exists($d . $filename) && is_file($d . $filename)) {
				self::$_Files[ $cachekey ] = $d . $filename;
				return $d . $filename;
			}
		}

		throw new TemplateException('Requested template ' . $filename . ' not found in system.');
	}

	/**
	 * Get an array of all the registered Template paths.
	 *
	 * @return array
	 */
	public static function GetPaths(){
		if(self::$_Paths === null){
			self::RequeryPaths();
		}

		return self::$_Paths;

	}

	/**
	 * Requery the application's template paths
	 * 
	 * Does not return anything, but can be useful if changes are 
	 * made to the application and the internal cache needs updated.
	 */
	public static function RequeryPaths() {
		$profiler = new ProfilerEvent('Template Path Scan');
		$profiler->start();
		
		self::$_Paths = [];
		self::$_Files = [];
		self::$_Templates = [];
		
		// Tack on the custom directory.
		// This needs to be before the current theme because it takes precedence.
		if(is_dir(ROOT_PDIR . 'themes/custom/')){
			self::$_Paths[] = ROOT_PDIR . 'themes/custom/';
		}

		// Tack on the current theme's directory.
		if(ConfigHandler::Get('/theme/selected')){
			self::$_Paths[] = ROOT_PDIR . 'themes/' . ConfigHandler::Get('/theme/selected') . '/';
		}

		// Tack on the search directories from the loaded components.
		// Also handle the plugins directory search.
		foreach (Core::GetComponents() as $c) {
			if(is_dir($c->getBaseDir() . 'templates/')){
				self::$_Paths[] = $c->getBaseDir() . 'templates/';
			}
		}
		
		$profiler->stop();
	}
}
