<?php
/**
 * PHP-Based template system.
 *
 * These tempalates are nothing more than a php file that contains HTML.  They're not recommended, but can be used when
 * performance or compatibility conflicts arise with the Smarty-based templates.
 *
 * Javascript template files are a perfect example; they tend to use { ... } as delimiters too, so trying to escape those
 * in a smarty template would be insane.
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @since 2.2.0
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

namespace Core\Templates\Backends;

use Core\Router;
use Core\Templates\Template;
use Core\View;
use Core\Templates\TemplateInterface;
use Core\Templates\TemplateException;
use Core\Utilities\Profiler\ProfilerEvent;
use Exception;
use WidgetModel;

class PHTML implements TemplateInterface {
	/**
	 * An array of scoped variables for this template.
	 * @var array
	 */
	private $_scope = array();

	/**
	 * The filename of this template
	 * @var String
	 */
	protected $_filename;

	/** @var View|null View that is responsible for this template, optional */
	private $_view = null;

	/**
	 * Assign a variable to this template
	 *
	 * @param array|string $key The variable name or array to set
	 * @param mixed|null   $var The value of the single variable to set
	 */
	public function assign($key, $var = null){
		if(is_array($key)){
			foreach($key as $k => $v){
				$this->_scope[$k] = $v;
			}
		}
		else{
			$this->_scope[$key] = $var;
		}
	}

	/**
	 * Returns a single or all template variables
	 *
	 * @param string|null  $varname        variable name or null
	 * @return string|array|null variable value or or array of variables
	 */
	public function getTemplateVars($varname = null){
		if($varname === null){
			return $this->_scope;
		}
		elseif(isset($this->_scope[$varname])){
			return $this->_scope[$varname];
		}
		else{
			return null;
		}
	}

	public function fetch($template = null){
		// Since the only way to capture php output is to use buffers...
		ob_start();
		$this->render($template);
		$contents = ob_get_clean();
		return $contents;
	}

	public function render($template = null){

		// Resolve this template.
		if($template === null){
			$file = $this->_filename;
		}
		else{
			$file = Template::ResolveFile($template);
			if($file === null){
				throw new TemplateException('Template ' . $template . ' could not be found!');
			}
		}

		if(!file_exists($file)){
			throw new TemplateException('Unable to read template ' . $file . ', file does not exist');
		}
		if(!is_readable($file)){
			throw new TemplateException('Unable to read template ' . $file . ', file is not readable');
		}
		
		$profiler = new ProfilerEvent('Core Template Render (PHTML) - ' . $file);
		$profiler->start();

		// I need to make the assigned variables visible in the scope.
		foreach($this->_scope as $key => $var){
			${$key} = $var;
		}

		try{
			require($file);
		}
		catch(Exception $e){
			$profiler->stop();
			throw new TemplateException($e->getMessage(), $e->getCode(), $e->getPrevious());
		}
		$profiler->stop();
	}

	/**
	 * Get a single variable from the template variables.
	 *
	 * @param string $varname The name of the variable
	 *
	 * @return mixed
	 */
	public function getVariable($varname) {
		return $this->getTemplateVars($varname);
	}

	/**
	 * Set a template filename to be remembered if fetch or render are called with null parameters.
	 *
	 * @param string $template Filename to remember for this template.
	 *
	 * @return void
	 */
	public function setFilename($template) {
		// Make sure it's resolved first.
		$this->_filename = Template::ResolveFile($template);
	}

	/**
	 * Get the basename of this template.
	 *
	 * @return string
	 */
	public function getBasename(){
		return basename($this->_filename);
	}

	/**
	 * Get the full filename of this template
	 *
	 * @return string|null
	 */
	public function getFilename(){
		return $this->_filename;
	}

	/**
	 * Scan through this template file and see if it has optional stylesheets that the admin can select to enable.
	 *
	 * @return boolean
	 */
	public function hasOptionalStylesheets() {
		return false;
	}

	/**
	 * Scan through this template file and see if it has widgetareas contained within.
	 *
	 * @return boolean
	 */
	public function hasWidgetAreas() {
		return false;
	}

	/**
	 * Get an array of widget areas defined on this template.
	 *
	 * The returning array is associative with the widgetarea name as the key,
	 * and each value is an array of name and installable.
	 *
	 * @return array
	 */
	public function getWidgetAreas(){
		// PHTML templates do not support widget areas yet.
		return [];
	}

	/**
	 * Get the list of optional stylesheets in this template.
	 *
	 * The returned array will be an array of the attributes on the declaration, with at minimum 'src' and 'title'.
	 *
	 * @return array
	 */
	public function getOptionalStylesheets() {
		// TODO: Implement getOptionalStylesheets() method.
	}

	/**
	 * Get an array of the insertables in this template.
	 *
	 * Should have "name", "type", "title", "value", and "description" in each array.
	 * Should also have any formelement-specific key necessary for operation, ie: "basedir", "accept", etc.
	 *
	 * @return array
	 */
	public function getInsertables() {
		// TODO: Implement getInsertables() method.
	}

	/**
	 * Load another template within this one with the same scope and render the results.
	 *
	 * @param string $template
	 * @return string
	 */
	public function load(string $template) : string {
		$template = Template::Factory($template);
		foreach($this->_scope as $k => $v) {
			$template->assign($k, $v);
		}
		return $template->fetch();
	}

	public function css(string $src, string $media = 'all') {
		$view = $this->_view ? $this->_view : Router::$DefaultView;

		$view->addStylesheet($src, $media);
	}

	/**
	 * Load a script on the View responsible for displaying this template.
	 *
	 * As a convenience factor, allow a library to be passed in.
	 *
	 * @param string $src
	 * @param string $location ["head", "foot", "library"]
	 */
	public function script(string $src, string $location) {
		$view = $this->_view ? $this->_view : Router::$DefaultView;

		if($location === 'library') {
			$this->scriptLibrary($src);
		}
		else {
			$view->addScript($src, $location);
		}
	}

	/**
	 * Load a script library by its name
	 *
	 * @param string $libraryName
	 */
	public function scriptLibrary(string $libraryName) {
		try {
			\Core\Core::LoadScriptLibrary($libraryName);
		}
		catch(Exception $e) {
			echo $e->getMessage();
		}
	}

	/**
	 * Load and return a widget with the given baseurl
	 *
	 * Only supports API 2.0 version of widgets and above.
	 *
	 * @param string $baseurl
	 *
	 * @return string
	 */
	public function widget(string $baseurl, array $parameters = []) : string {
		$view = $this->_view ? $this->_view : Router::$DefaultView;

		// Version 2.0 uses baseurl as the defining call.
		$api = 2.0;
		$parts = WidgetModel::SplitBaseURL($baseurl);
		$name = $parts['controller'];
		$method = $parts['method'];
		if($parts['parameters']) {
			$parameters = array_merge($parts['parameters'], $parameters);
		}

		if(!class_exists($name)) {
			if(DEVELOPMENT_MODE) {
				return '<p class="message-error">ERROR, Class for ' . $baseurl . ' (' . $name . ') not found on system, widget disabled.</p>';
			}
			else{
				return '';
			}
		}

		/** @var $w \Core\Widget */
		$w = new $name();
		$w->_params = $parameters;

		// Ensure that the widget's View knows it's linked to a parent!
		$w->getView()->parent = $view;

		// Populate the request with the inbound data too.
		$request = $w->getRequest();

		if($parameters) {
			$request->parameters = $parameters;
		}

		$return = call_user_func(array($w, $method));
		$dat = null;

		if(is_int($return)) {
			if(DEVELOPMENT_MODE) {
				return '<p class="message-error">ERROR, Widget ' . $name . '/' . $method . ' returned error code ' . $return . '</p>';
			}
			else{
				return '';
			}
		}
		elseif($return === null) {
			// This is typical for widgets that have an execute and render step, which some do.
			$widgetView = $w->getView();

			if(!$widgetView->templatename && $widgetView->status === View::STATUS_OK && $widgetView->contenttype === View::CTYPE_HTML) {
				// Allow the template to be omitted in the Widget logic, this can be inferred.
				$cnameshort = (strpos($name, 'Widget') == strlen($name) - 6) ? substr($name, 0, -6) : $name;
				$widgetView->templatename = strtolower('widgets/' . $cnameshort . '/' . $method);
			}

			return $widgetView->fetch();
		}
		// If it's just a string, return that.
		elseif(is_string($return)) {
			return $return;
		}
	}
}
