<?php
/**
 * Defines the schema for the Config table
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\ConfigHandler;
use Core\Core;
use Core\Dict;
use Core\Forms\FormElement;
use Core\I18N\I18NLoader;
use Core\Model;


/**
 * Class ConfigModel
 *
 * @property string $component       The component that registered this config, useful for uninstalling and cleanups
 * @property int    $created         Creation timestamp, populated automatically
 * @property string $default_value
 * @property bool   $encrypted
 * @property array  $form_attributes Set from the content of the form-attributes XML parameter.
 * @property string $key
 * @property string $mapto           The define constant to map the value to on system load.
 * @property array  $options
 * @property bool   $overrideable    If children sites can override this configuration option
 * @property string $type
 * @property int    $updated         Last update timestamp, populated automatically
 * @property string $value
 */
class ConfigModel extends Model {

	public static $ExportGroup = 't:STRING_SYSTEM_CONFIGURATION';

	public static $Schema = array(
		'key'           => [
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 255,
			'required'  => true,
			'null'      => false,
		],
		'type'          => [
			'type'    => Model::ATT_TYPE_ENUM,
			'options' => ['string', 'text', 'int', 'boolean', 'enum', 'set'],
			'default' => 'string',
			'null'    => false,
		],
		'encrypted' => [
			'type' => Model::ATT_TYPE_BOOL,
			'default' => 0,
		],
		// Added on 2015.11 to keep track of which config options need to be cleaned up on component removal.
		// Do not enforce this until all components support it.
		'component'    => [
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 48,
			'required'  => false,
			'default'   => '',
			'null'      => false,
			'form' => array(
				'type' => 'disabled',
			),
			'comment'   => 'The component that registered this config, useful for uninstalling and cleanups',
		],
		'default_value' => [
			'type'    => Model::ATT_TYPE_TEXT,
			'default' => null,
			'null'    => true,
		],
		'value'         => [
			'type'    => Model::ATT_TYPE_TEXT,
			'default' => null,
			'null'    => true,
		],
		'options'       => [
			'type'      => Model::ATT_TYPE_DATA,
			'default'   => null,
			'null'      => true,
			'encoding'  => Model::ATT_ENCODING_JSON,
		],
		'mapto'         => [
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 64,
			'default'   => null,
			'comment'   => 'The define constant to map the value to on system load.',
			'null'      => true,
		],
		'overrideable' => [
			'type' => Model::ATT_TYPE_BOOL,
			'default' => false,
			'comment' => 'If children sites can override this configuration option',
		],
		'form_attributes' => [
			'type' => Model::ATT_TYPE_DATA,
		    'comment' => 'Set from the content of the form-attributes XML parameter.',
			'encoding'  => Model::ATT_ENCODING_JSON,
		],
		'created'       => [
			'type' => Model::ATT_TYPE_CREATED
		],
		'updated'       => [
			'type' => Model::ATT_TYPE_UPDATED
		]
	);

	public static $Indexes = array(
		'primary' => array('key'),
	);

	/**
	 * Get either the set value or the default value if that is null.
	 *
	 * This value will also be typecasted to the correct type.
	 *
	 * @return mixed
	 */
	public function getValue() {
		$v = $this->get('value');

		// If it's null, it just get the default value.
		if ($v === null){
			$v = $this->get('default');
		}
		// If it's encrypted, decrypt it first!
		// This must be done outside the regular model encryption logic because it's dynamic for each config.
		elseif($this->get('encrypted') && $v !== ''){
			preg_match('/^\$([^$]*)\$([0-9]*)\$(.*)$/m', $v, $matches);

			$cipher = $matches[1];
			$passes = $matches[2];
			$size = openssl_cipher_iv_length($cipher);
			// Now I can trim off the beginning crap from the encrypted string.
			$dec = substr($v, strlen($cipher) + 5, 0-$size);
			$iv = substr($v, 0-$size);

			for($i=0; $i<$passes; $i++){
				$dec = openssl_decrypt($dec, $cipher, SECRET_ENCRYPTION_PASSPHRASE, true, $iv);
			}

			$v = $dec;
		}

		return self::TranslateValue($this->get('type'), $v);
	}

	public function setValue($value){
		if($this->get('encrypted')){
			$cipher = 'AES-256-CBC';
			$passes = 10;
			$size = openssl_cipher_iv_length($cipher);
			$iv = mcrypt_create_iv($size, MCRYPT_RAND);

			$enc = $value;
			for($i=0; $i<$passes; $i++){
				$enc = openssl_encrypt($enc, $cipher, SECRET_ENCRYPTION_PASSPHRASE, true, $iv);
			}

			$payload = '$' . $cipher . '$' . str_pad($passes, 2, '0', STR_PAD_LEFT) . '$' . $enc . $iv;

			return parent::set('value', $payload);
		}
		else{
			return parent::set('value', $value);
		}
	}

	/**
	 * Translate the database value to a strict datatype.
	 * This will ensure that a "boolean" config with value of "0" actually returns (===) false.
	 *
	 * @param $type
	 * @param $value
	 *
	 * @return array|bool|int
	 */
	public static function TranslateValue($type, $value){
		switch ($type) {
			case 'int':
				return (int)$value;
			case 'boolean':
				return ($value == '1' || $value == 'true') ? true : false;
			case 'set':
				return array_map('trim', explode('|', $value));
			default:
				return $value;
		}
	}

	/**
	 * 
	 * @return array
	 */
	public function getFormAttributes() : array {
		$opts = $this->get('form_attributes');

		if(!is_array($opts)) {
			// Ensure the retrieved data is actually data.
			$opts = [];
		}
		
		if(!isset($opts['type'])){
			// Determine this automatically by the config data type.
			switch ($this->get('type')) {
				case 'text':
					$opts['type'] = 'textarea';
					break;
				case 'enum':
					$opts['type'] = 'select';
					break;
				case 'boolean':
					$opts['type'] = 'radio';
					break;
				case 'set':
					$opts['type'] ='checkboxes';
					break;
				default:
					// Also string and int
					$opts['type'] = 'text';
					break;
			}
		}
		
		$key = $this->get('key');

		$gname = substr($key, 1);
		$gname = ucwords(substr($gname, 0, strpos($gname, '/')));

		// NEW i18n support for config options!
		$i18nKey = I18NLoader::KeyifyString($key);
		$opts['title'] = 't:STRING_CONFIG_' . $i18nKey;
		$opts['description'] = 't:MESSAGE_CONFIG_' . $i18nKey;

		// SELECT
		if(
			($opts['type'] === 'select' || $opts['type'] === 'checkboxes') && 
			$this->get('options')
		){
			// This is set from the main option set.
			$opts['options'] = new Dict();
			foreach($this->get('options') as $val){
				$opts['options'][ $val ] = 't:VALUE_CONFIG_' . $i18nKey . '_' . I18NLoader::KeyifyString($val);
			}
		}

		// RADIO
		if($opts['type'] == 'radio'){
			$opts['options'] = Dict::FromAssoc(['false' => 't:STRING_NO', 'true'  => 't:STRING_YES']);
		}

		if(!isset($opts['group'])){
			$opts['group'] = $gname;
		}

		// The name can't be set by the XML metadata, but it is based on the type of form element, slightly.
		if($opts['type'] === 'checkboxes'){
			// Append "[]" to the name as there are many checkboxes!
			$opts['name'] = 'config[' . $key . '][]';
		}
		else{
			$opts['name'] = 'config[' . $key . ']';
		}

		return $opts;
	}

	/**
	 * Transpose a populated form element from the underlying ConfigModel object.
	 * Will populate the name, options, validation, etc.
	 *
	 * @return FormElement
	 *
	 * @throws \Exception
	 */
	public function getAsFormElement() : FormElement {
		// key is in the format of:
		// /user/displayname/displayoptions

		$key        = $this->get('key');
		$attributes = $this->getFormAttributes();
		$val        = ConfigHandler::Get($key);
		$type       = $attributes['type'];
		$el         = FormElement::Factory($type, $attributes);

		if($type == 'radio'){
			// Ensure that this matches what the radios will have.
			if ($val == '1' || $val == 'true' || $val == 'yes') $val = 'true';
			else $val = 'false';
		}

		if($this->get('type') == 'int' && $type == 'text'){
			$el->validation        = '/^[0-9]*$/';
			$el->validationmessage = $attributes['group'] . ' - ' . $attributes['title'] . ' expects only whole numbers with no punctuation.';
		}

		if($type == 'checkboxes' && !is_array($val)){
			// Convert the found value to an array so it matches what checkboxes are expecting.
			$val  = array_map('trim', explode('|', $val));
		}

		$el->set('value', $val);
		
		// If multisite is enabled and this config is NOT set to overrideable, then set the field as read only!
		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled() && MultiSiteHelper::GetCurrentSiteID()){
			if(!$this->get('overrideable')){
				$el->set('readonly', true);
				$el->set('disabled', true);
			}
		}

		return $el;
	}

	/**
	 * Alias of getAsFormElement
	 *
	 * @return \FormElement
	 *
	 * @throws \Exception
	 */
	public function asFormElement(){
		return self::getAsFormElement();
	}
} // END class ConfigModel extends Model

