<?php
/**
 * Defines the schema for the Page table
 *
 * @package Core
 * @author Charlie Powell <charlie@evalagency.com>
 * @copyright Copyright (C) 2009-2017  Charlie Powell
 * @license GNU Affero General Public License v3 <http://www.gnu.org/licenses/agpl-3.0.txt>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/agpl-3.0.txt.
 */

use Core\BaseURLMap;
use Core\ConfigHandler;
use Core\Core;
use Core\Datamodel\Dataset;
use Core\Dict;
use Core\Forms\Form;
use Core\Forms\FormElement;
use Core\Model;
use Core\ModelFactory;
use Core\ModelValidationException;
use Core\Router;
use Core\Templates\Template;
use Core\Templates\TemplateException;
use Core\View;
use Core\ViewMeta\Title;
use HTMLToMD\Converter;


/**
 * Class PageModel
 *
 * @property string $access             Access string of the page
 * @property bool   $admin              If this page is an administration page
 * @property string $admin_group        Admin pages can be grouped together.  This is the name.
 * @property string $baseurl
 * @property string $body               [Cached] Body content of this page
 * @property string $component          The component that registered this page, useful for uninstalling and cleanups
 * @property int    $created            Creation timestamp, populated automatically
 * @property string $deleteurl          The URL to perform the POST on to delete this page
 * @property string $editurl            The edit URL for this page, set by the creating application.
 * @property int    $expires
 * @property bool   $fuzzy              If this url is fuzzy or an exact match
 * @property bool   $indexable          Page is displayed on the sitemap, search, and search crawlers
 * @property string $last_template      The last page template used to render this page, useful in edit pages.
 * @property string $page_template      Allows the specific page template to be overridden.
 * @property int    $pageviews          Number of page views
 * @property string $parenturl
 * @property string $password_protected Password or phrase to protect this page
 * @property float  $popularity         Cache of the popularity score of this page
 * @property int    $published          The published date
 * @property string $published_expires
 * @property string $published_status
 * @property string $rewriteurl
 * @property string $search_index_pri   Flattened DMP-PRI representation for the search index
 * @property string $search_index_sec   Flattened DMP-SEC representation for the search index
 * @property string $search_index_str   Flattened string representation for the search index
 * @property bool   $selectable         Selectable as a parent url
 * @property int    $site               The site id in multisite mode, (or -1 if global)
 * @property string $theme_template     Allows the page to define its own theme and widget information.
 * @property string $title              [Cached] Title of the page
 * @property int    $updated            Last update timestamp, populated automatically
 */
class PageModel extends Model {

	public static $ExportGroup = 't:STRING_SITE_CONTENT';

	public static $Schema = array(
		'title' => array(
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'default'   => null,
			'comment'   => '[Cached] Title of the page',
			'null'      => true,
			'form'      => array(
				'type' => 'text',
				'group' => 't:STRING_BASIC',
				'grouptype' => 'tabs',
			),
			'formatter' => ['this', 'getTitle'],
		),
		'parenturl' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'null' => true,
			'form' => array(
				'type' => 'select',
				'group' => 't:STRING_META_URL_SEO_INFORMATION',
				'grouptype' => 'tabs',
				'source' => 'this::getParentsAsOptions',
			),
		),
		'site' => array(
			'type' => Model::ATT_TYPE_SITE,
			'form' => [
				'type' => 'system',
				'group' => 't:STRING_ACCESS_AND_ADVANCED',
				'grouptype' => 'tabs',
			],
			'comment' => 'The site id in multisite mode, (or -1 if global)',
		),
		'baseurl' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'required' => true,
			'null' => false,
			'form' => array(
				'type' => 'system',
			),
		),
		'rewriteurl' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'null' => false,
			'validation' => array('this', 'validateRewriteURL'),
			'form' => array(
				'type' => 'pagerewriteurl',
				'group' => 't:STRING_META_URL_SEO_INFORMATION',
				'grouptype' => 'tabs',
			),
			'formatter' => ['this', 'getRewriteURL'],
		),
		'editurl' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'default' => '',
			'required' => false,
			'null' => false,
			'form' => array(
				'type' => 'disabled',
			),
			'comment' => 'The edit URL for this page, set by the creating application.',
		),
		'deleteurl' => array(
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'default'   => '',
			'required'  => false,
			'null'      => false,
			'form' => array(
				'type' => 'disabled',
			),
			'comment'   => 'The URL to perform the POST on to delete this page',
		),
		'component'    => array(
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 48,
			'required'  => false,
			'default'   => '',
			'null'      => false,
			'form' => array(
				'type' => 'disabled',
			),
			'comment'   => 'The component that registered this page, useful for uninstalling and cleanups',
		),
		'theme_template' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'default' => null,
			'null' => true,
			'comment' => 'Allows the page to define its own theme and widget information.',
			'form' => array(
				'type' => 'select',
				'source' => '\\Core\\Theme::GetSkinsAsOptions',
				'group' => 't:STRING_ACCESS_AND_ADVANCED',
				'grouptype' => 'tabs',
			)
		),
		'page_template' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 64,
			'default' => null,
			'null' => true,
			'comment' => 'Allows the specific page template to be overridden.',
			'form' => array(
				'type' => 'pagepageselect',
				'group' => 't:STRING_BASIC',
				'grouptype' => 'tabs',
			)
		),
		'last_template' => array(
			'type'      => Model::ATT_TYPE_STRING,
			'maxlength' => 64,
			'default'   => null,
			'null'      => true,
			'formtype'  => 'disabled',
			'comment'   => 'The last page template used to render this page, useful in edit pages.',
		),
		'expires' => array(
			'type' => Model::ATT_TYPE_INT,
			'default' => 3600,
			'form' => [
				'type' => 'select',
				'options' => [
					'0'       => 'No Cache Allowed',
					'30'      => '30 seconds',
					'60'      => '1 minute',
					'120'     => '2 minutes',
					'300'     => '5 minutes',
					'600'     => '10 minutes',
					'1800'    => '30 minutes',
					'3600'    => '1 hour',
					'7200'    => '2 hours',
					'14400'   => '4 hours',
					'21600'   => '6 hours',
					'28800'   => '8 hours',
					'43200'   => '12 hours',
					'64800'   => '18 hours',
					'86400'   => '24 hours',
					'172800'  => '2 days',
					'604800'  => '1 week',
					'2462400' => '1 month',
				],
				'group' => 't:STRING_ACCESS_AND_ADVANCED',
				'grouptype' => 'tabs',
			],
			'formatter' => '\Core\Formatter\GeneralFormatter::TimeDuration',
		),
		'access' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 512,
			'comment' => 'Access string of the page',
			'null' => false,
			'default' => '*',
			'form' => array(
				'type' => 'access',
				'group' => 't:STRING_ACCESS_AND_ADVANCED',
				'grouptype' => 'tabs',
			),
			'formatter' => '\Core\Formatter\GeneralFormatter::AccessString',
		),
		'password_protected' => array(
			'type' => Model::ATT_TYPE_STRING,
			'maxlength' => 128,
			'comment' => 'Password or phrase to protect this page',
			'null' => false,
			'default' => '',
			'form' => array(
				'type' => 'text',
				'group' => 't:STRING_ACCESS_AND_ADVANCED',
				'grouptype' => 'tabs',
			),
		),
		'fuzzy' => array(
			'type' => Model::ATT_TYPE_BOOL,
			'comment' => 'If this url is fuzzy or an exact match',
			'null' => false,
			'default' => '0',
			'formtype' => 'system'
		),
		'admin' => array(
			'type' => Model::ATT_TYPE_BOOL,
			'comment' => 'If this page is an administration page',
			'null' => false,
			'default' => '0',
			'formtype' => 'system'
		),
		'admin_group' => array(
			'type' => Model::ATT_TYPE_STRING,
			'comment' => 'Admin pages can be grouped together.  This is the name.',
			'null' => false,
			'default' => '',
			'formtype' => 'disabled',
		),
		'pageviews' => array(
			'type' => Model::ATT_TYPE_INT,
			'formtype' => 'disabled',
			'default' => 0,
			'comment' => 'Number of page views',
			'model_audit_ignore' => true, // Custom key for the component "Model Audit".
		),
		'selectable' => array(
			'type' => Model::ATT_TYPE_BOOL,
			'default' => 1,
			'comment' => 'Selectable as a parent url',
			'formtype' => 'disabled',
		),
		'indexable' => array(
			'type' => Model::ATT_TYPE_BOOL,
			'default' => 1,
			'comment' => 'Page is displayed on the sitemap, search, and search crawlers',
			'form' => [
				'group' => 't:STRING_META_URL_SEO_INFORMATION',
				'grouptype' => 'tabs',
			],
		),
		'popularity' => array(
			'type' => Model::ATT_TYPE_FLOAT,
			'default' => 0.000,
			'precision' => '10,8',
			'comment' => 'Cache of the popularity score of this page',
			'formtype' => 'disabled',
		),
		'published_status'      => array(
			'type'    => Model::ATT_TYPE_ENUM,
			'options' => ['published', 'draft'],
			'default' => 'published',
			'form' => array(
				'group' => 't:STRING_PUBLISH_SETTINGS',
				'grouptype' => 'tabs',
			),
			'formatter' => ['this', 'getPublishedStatus'],
		),
		'published' => array(
			'type' => Model::ATT_TYPE_INT,
			'form' => array(
				'type' => 'datetime',
				'group' => 't:STRING_PUBLISH_SETTINGS',
				'grouptype' => 'tabs',
			),
			'comment' => 'The published date',
			'formatter' => '\Core\Formatter\GeneralFormatter::DateStringSD',
		),
		'published_expires' => array(
			'type' => Model::ATT_TYPE_STRING,
			'null' => true,
			'default' => null,
			'form' => array(
				'type' => 'datetime',
				'group' => 't:STRING_PUBLISH_SETTINGS',
				'grouptype' => 'tabs',
				'datetimepicker_dateformat' => 'yy-mm-dd',
				'datetimepicker_timeformat' => 'HH:mm',
			),
			'formatter' => '\Core\Formatter\GeneralFormatter::DateStringSDT',
		),
		'body' => array(
			'type'      => Model::ATT_TYPE_TEXT,
			'default'   => '',
			'comment'   => '[Cached] Body content of this page',
			'null'      => false,
			'form'      => array(
				'type' => 'disabled',
			),
		),
		'seotitle' => [
			'type' => Model::ATT_TYPE_META,
			'formatter' => ['this', 'getSEOTitle'],
		],
		'teaser' => [
			'type' => Model::ATT_TYPE_META,
			'formatter' => ['this', 'getTeaser'],
		],
	);

	public static $Indexes = array(
		'primary'            => ['site', 'baseurl'],
		'unique:rewrite_url' => ['site', 'rewriteurl'],
		'baseurlidx'         => ['baseurl'],
		'adminidx'           => ['admin'],
		'rewritefuzzy'       => ['rewriteurl', 'fuzzy'],
		'baseurlfuzzy'       => ['baseurl', 'fuzzy'],
	);

	public static $HasCreated = true;
	public static $HasUpdated = true;
	public static $HasSearch  = true;

	/**
	 * Set this to the full templatename path to enable resolution of the template and any optional subtemplates.
	 * @var null|string
	 */
	public $templatename = null;

	private $_class;
	private $_method;
	private $_params;

	/**
	 * The View component for this page.
	 * @var View
	 */
	private $_view;

	/** @var null|array Cache of Rewrite URL -> Base URL (each base can have multiple rewrites) */
	private static $_RewriteCache = null;

	/** @var null|array A Cache of fuzzy pages, (and their rewrite URLs), to serve as a quick lookup. */
	private static $_FuzzyCache = null;
	
	/** @var null|array Cache of Base URL -> Primary Rewrite URL */
	private static $_BaseCache = null;


	public function  __construct(...$keys) {
		$this->_linked = array(
			'Insertable' => array(
				'link' => Model::LINK_HASMANY,
				'on' => 'baseurl'
			),
			'PageMeta' => array(
				'link' => Model::LINK_HASMANY,
				'on' => array('site' => 'site', 'baseurl' => 'baseurl'),
			),
			'RewriteMap' => array(
				'link' => Model::LINK_HASMANY,
				'on' => array('site' => 'site', 'baseurl' => 'baseurl', 'fuzzy' => 'fuzzy'),
			)
		);

		parent::__construct(...$keys);
	}

	/**
	 * Check if the given User can access this particular object based on their access permissions.
	 *
	 * By default this is a simple 'return true' check, but can be extended for custom functionality.
	 *
	 * Please note, THIS DOES NOT PREVENT ANY ACTIONS, simply is a convenience function for controllers
	 * to utilize to check the object ACL.
	 *
	 * @param UserModel|null $user The user to check access.
	 *
	 * @return bool
	 */
	public function checkCanAccess(UserModel $user = null) : bool {
		if($user === null) {
			$user = \Core\user();
		}

		if(!$this->isPublished() && !$user->checkAccess('g:admin')) {
			// The page is not published and the user is not an admin, skip!
			return false;
		}

		// Default is to check against this page's access setting.
		return $user->checkAccess($this->access);
	}
	

	/**
	 * Get the logo URL for this page, if available.
	 * 
	 * @return null|string
	 */
	public function getLogoURL(){
		$logo = $this->getLogo();
		return $logo !== null ? $logo->getPreviewURL('24x24') : null;
	}
	
	/**
	 * Get the logo File for this page, if available.
	 * 
	 * @return null|\Core\Filestore\File
	 */
	public function getLogo(){
		if(($img = $this->getImage())){
			// Best way, this page has a main image set!
			return $img;
		}
		elseif($this->get('component')){
			// Does this page have a registered component?
			// If so, I can pull that logo!
			$c = Core::GetComponent($this->get('component'));
			if(($logo = $c->getLogo())){
				return \Core\Filestore\Factory::File($logo);
			}
		}
		
		return null;
	}


	/**
	 * Validate a rewrite URL for a page
	 *
	 * Will return true if successful, or throw an exception otherwise
	 *
	 * @param string $v
	 * @return bool
	 *
	 * @throws ModelValidationException
	 */
	public function validateRewriteURL($v) {


		if (!$v) {
			// If it's empty, that's fine, it'll get reset to the baseurl.
			return true;
		}

		try {
			$baseURLColumn = $this->getColumn('baseurl');
			if($v === $baseURLColumn->value) {
				// If it's the same as the baseurl, that's fine.
				return true;
			}
		}
		catch(Exception $ex) {
			throw new ModelValidationException('Page does not contain a baseurl column, unable to validate URL!');
		}

		if ($v[0] != '/') {
			throw new ModelValidationException('Rewrite URL must start with a "/"');
		}

		if(strpos($v, '#') !== false) {
			// Pound signs may be common to implement, provide a custom message for them.
			throw new ModelValidationException('Rewrite URL must not contain a pound sign (#)');
		}

		if(strpos($v, ' ') !== false) {
			// Spaces may be common to implement, provide a custom message for them.
			throw new ModelValidationException('Rewrite URL must not contain spaces, please use underscores or dashes instead');
		}

		// Use regex to validate a URL fragment for simplicity for everything else
		if(!preg_match('#^[a-zA-Z0-9/_\-]*$#', $v)) {
			throw new ModelValidationException('Rewrite URL contains invalid characters, please use only Latin letters, numbers, and dashes.');
		}

		// See if the controller segment matches an existing controller.  This cannot happen because
		// that controller would no longer be accessible, example:
		// Blog::View, Blog::Create, etc and /blog.

		$controller = substr($v, 1, ( (strpos($v, '/', 1) !== false) ? strpos($v, '/', 1) : strlen($v)) );
		if($controller && class_exists($controller . 'Controller')){
			throw new ModelValidationException('Rewrite URL cannot start with /' . $controller . ', that is a reserved URL.');
		}

		// Lookup if there is a conflicting URL.
		$ds = Dataset::Init()
			->table('page')
			->select('*')
			->whereGroup('OR', 'baseurl = ' . $v, 'rewriteurl = ' . $v);

		// If this page exists, I don't want to include this page in the count.
		if ($this->checkExists()){
			$ds->where('baseurl != ' . parent::get('baseurl'));
		}

		// Enterprise/multisite mode anyone?
		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			$ds->whereGroup('OR', 'site = -1', 'site = ' . MultiSiteHelper::GetCurrentSiteID());
		}

		$ds->execute();

		if ($ds->num_rows > 0) {
			if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
				foreach($ds as $row){
					if($row['site'] == $this->get('site') || $row['site'] == '-1'){
						// Only trigger the URL taken error if it's on the same site or it's a global page.
						throw new ModelValidationException('Rewrite URL already taken');
					}
				}
			}
			else{
				throw new ModelValidationException('Rewrite URL already taken');
			}
		}

		// All good?
		return true;
	}

	/**
	 * Get the base template name for this page based strictly on its baseurl.
	 * 
	 * This utilizes the BaseURL system for determining the template.
	 *
	 * @return string
	 */
	public function getBaseTemplateName() : string {
		$url = new BaseURLMap($this->get('baseurl'));
		return $url->getDefaultTemplateName();
	}

	/**
	 * Get the template name, taking the page_template into consideration.
	 *
	 * @return string
	 */
	public function getTemplateName() : string {
		// can I just cheat and return the last displayed template?
		//if($this->get('last_template')){
		//	return $this->get('last_template');
		//}

		$t = $this->getBaseTemplateName();
		
		// Allow the specific template to be overridden.
		if (($override = $this->get('page_template'))){
			$t = $t . '/' . $override;
		}

		return $t;
	}

	/**
	 * Get an array of the metadata for the metadata.
	 * This is useful for constructing form elements for a given page.
	 *
	 * @return array
	 */
	public function getMetasArray() {

		$fullmetas = array(
			// The SEO title.  This isn't quite a meta tag, but part of that system regardless, so might as well.
			'title' => array(
				'title'       => 'Search-Optimized Title',
				'description' => 'If a value is entered here, the &lt;title&gt; tag of the page will be replaced with this value.  Useful for making the page more indexable by search bots.',
				'type'        => 'text',
				'value'       => (($meta = $this->getMeta('title')) ? $meta->get('meta_value_title') : null),
			),
			// Image
			'image' => array(
				'title'       => 'Image',
				'description' => 'Optional image to showcase this page',
				'type'        => 'file',
				'basedir'     => 'public/page/image/',
				'browseable'  => true,
				'value'       => (($meta = $this->getMeta('image')) ? $meta->get('meta_value_title') : null),
			),
			// Author
			'author' => array(
				'title'       => 'Author',
				'description' => 'Completely optional, but feel free to include it if relevant',
				'type'        => 'pagemetaauthor',
				'value'       => (($meta = $this->getMeta('author')) ? $meta->get('meta_value_title') : null),
			),
			// The author id
			'authorid' => array(
				'type'        => 'hidden',
				'value'       => (($meta = $this->getMeta('author')) ? $meta->get('meta_value') : null),
			),
			// Keywords, (the human friendly text of them)
			'keywords' => array(
				'title'       => 'Keywords',
				'description' => 'Provides taxonomy data for this page, separate different keywords with a comma.',
				'type'        => 'pagemetakeywords',
				'model'       => $this,
			),
			'description' => array(
				'title'       => 'Description/Teaser',
				'description' => 'Teaser text that displays on search engine and social network preview links',
				'maxlength'   => 256,
				'type'        => 'textarea',
				'value'       => (($meta = $this->getMeta('description')) ? $meta->get('meta_value_title') : null),
			)
		);

		return $fullmetas;
	}

	/**
	 * Get a specific meta tag, or null if it doesn't exist.
	 *
	 * There are a couple exceptions that will return an array of results.  Currently it is simply keywords.
	 *
	 * @param string $name
	 *
	 * @return PageMetaModel | null | array
	 */
	public function getMeta($name) {
		$metas = $this->getLink('PageMeta');
		if($name == 'keywords'){
			$keywords = array();
			foreach($metas as $meta){
				/** @var $meta PageMetaModel */
				if($meta->get('meta_key') == 'keyword') $keywords[] = $meta;
			}
			return $keywords;
		}
		else{
			foreach($metas as $meta){
				/** @var $meta PageMetaModel */
				if($meta->get('meta_key') == $name) return $meta;
			}
		}

		return null;
	}

	public function getMetaValue($name){
		$m = $this->getMeta($name);

		return $m ? $m->get('meta_value_title') : '';
	}
	
	public function asFormArray() : array {
		$ret = parent::asFormArray();
		$site = $this->get('site');
		$baseurl = $this->get('baseurl');
		
		// If the user has access to manage sites globally, then enable that option!
		if(
			Core::IsComponentAvailable('multisite') && 
			MultiSiteHelper::IsEnabled() && 
			\Core\user()->checkAccess('g:admin') &&
			isset($ret['site'])
		){
			$opts = $ret['site']->getAsArray();
			unset($opts['__class']);
			
			$ret['site'] = \Core\Forms\FormElement::Factory( 'select', $opts );
			
			// Ensure that the form element's "parent" is the same as this column's parent.
			// If it's null, then it'll be null there! (which is fine.)
			// Remember since this is an object, only the REFERENCE will be used.
			$ret['site']->parent = $this;
		}
		
		// Add in some extra fields that are associated with Pages too!
		// I need to add the rewrite options, (I need to get them too).
		$ret['rewrites'] = \Core\Forms\FormElement::Factory('textarea',
			[
				'name' => 'rewrites',
				'group' => 't:STRING_META_URL_SEO_INFORMATION',
				'title' => 'Rewrite Aliases',
				'value' => $this->getRewriteURLs(),
				'description' => 'Enter rewrite aliases that point to this page, one per line.  You may use the fully resolved path or simply the part after the ".com".',
			]
		);


		// Load in the currently set insertable content for this page
		// This will generally pull from the database, but will use the local cache of elements
		// in the case of a temporary form submission.
		// Also provide a lookup map for all insertables.
		$insertableModels = $this->getLink('Insertable');
		$insertableMap = [];
		foreach($insertableModels as $model) {
			$insertableMap[ $model->name ] = $model;
		}

		// And the page insertables.
		try {
			$tpl = Template::Factory($this->getTemplateName());
		}
		catch(Exception $ex) {
			$tpl = null;
			Core::SetMessage('Unable to load ' . $this->getTemplateName() . ', please ensure a valid template is set.', 'error');
		}

		if($tpl){
			// All the insertables from this template
			$insertables = $tpl->getInsertables();
			// Used to keep track of grouped insertables, as they're pulled out of order.
			$groups = [];

			foreach($insertables as $key => $dat){
				$type = $dat['type'];
				unset($dat['type']);
				$dat['grouptype'] = 'tabs';

				if($type == 'group') {
					$dat['name'] = 'insertablegroups[' . $key . ']';
					$dat['data-group'] = $key;
					$dat['data-descriptor'] = $dat['descriptor'];
					$groupKey = 'insertablegroups_' . $key;
					$group = \Core\Forms\FormGroup::Factory($type, $dat);

					$elementDat = [
						'name' => 'insertables[' . $key . ']',
						'data-group' => $key,
						'value' => [],
					];
					// Look for any already-set value and create new elements for that set.
					$insertable = $insertableMap[ $key ] ?? null;

					if($insertable && $insertable->value !== null) {
						$elementDat['value'] = $insertable->get('value');
					}

					$elementKey = 'insertables_' . $key;
					$element = \Core\Forms\FormElement::Factory('pageinsertablegroup', $elementDat);
					$element->set('group_parent', $group);

					$ret[$groupKey] = $group;
					$ret[$elementKey] = $element;
					$groups[$groupKey] = [];
				}
				elseif(isset($dat['grouping']) && $dat['grouping']) {
					// Handle any insertable group differently, namely these get attached to the group if set.
					$groupKey = 'insertablegroups_' . $dat['grouping'];

					if(!isset($ret[$groupKey])) {
						\Core\log_notice('Insertable ' . $key . ' belongs to a group, but the group was not defined!');
						continue;
					}

					$lookupKey = $dat['grouping'] . '[' . $dat['name'] . ']';
					/** @var \Core\Forms\FormGroup $group */
					$group = $ret[$groupKey];

					$dat['name'] = 'insertables[' . $dat['grouping'] . '][__INSERTABLEGROUPTEMPLATE][' . $dat['name'] . ']';

					// Grouped insertables cannot contain multiple values!
					//$dat['multiple'] = false;

					// Add the base template to the group
					$templateDat = $dat;
					//$templateDat['data-group-index'] = 0;
					//$templateDat['class'] .= ' insertablegroup-template';
					$group->addElement($type, $templateDat);

					// Look for any already-set value and create new elements for that set.
					/*
					$insertable = $insertableMap[ $lookupKey ] ?? null;
					if($insertable && $insertable->value !== null) {
						$val = json_decode($insertable->get('value'), true);
						if(!$val) {
							$val = [];
						}

						// The first element is the template's value, it can safely be dropped off.
						array_shift($val);

						foreach($val as $k => $v) {
							if(!isset($groups[$groupKey][$k])) {
								$groups[$groupKey][$k] = [];
							}

							$dat['value'] = $v;
							$dat['data-group-index'] = $k + 1; // Set the index to the position (+1!)  This +1 skips the index 0 template
							$groups[$groupKey][$k][] = FormElement::Factory($type, $dat);
						}
					}
					*/
				}
				else {
					$dat['name'] = 'insertables[' . $key . ']';

					// Look for any already-set value and create new elements for that set.
					$insertable = $insertableMap[ $key ] ?? null;

					if($insertable && $insertable->value !== null) {
						$dat['value'] = $insertable->get('value');
					}

					$ret['insertables_' . $key] = \Core\Forms\FormElement::Factory($type, $dat);
				}
			}

			foreach($groups as $groupKey => $rows) {
				$group = $ret[$groupKey];
				foreach($rows as $row) {
					foreach($row as $el) {
						$group->addElement($el);
					}
				}
			}
		}
		
		// This system allows templates to manipulate some of the meta fields.
		$metaRegroups = [];
		if($tpl){
			$tmpMetas = $tpl->getMetas();
			
			if(isset($tmpMetas['page-edit-meta-regroup']) && is_array($tmpMetas['page-edit-meta-regroup'])){
				foreach($tmpMetas['page-edit-meta-regroup'] as $set){
					$pos = strpos($set, ' ');
					$key = substr($set, 0, $pos);
					$dst = substr($set, $pos+1);
					$metaRegroups[$key] = $dst;
				}
			}
			elseif(isset($tmpMetas['page-edit-meta-regroup'])){
				$set = $tmpMetas['page-edit-meta-regroup'];
				$pos = strpos($set, ' ');
				$key = substr($set, 0, $pos);
				$dst = substr($set, $pos+1);
				$metaRegroups[$key] = $dst;
			}
		}
		
		// I need to add the pagemetas!
		foreach($this->getMetasArray() as $key => $dat){
			$type = $dat['type'];
			unset($dat['type']);
			$dat['name'] = 'metas[' . $key . ']';
			$dat['group'] = isset($metaRegroups[ $dat['name'] ]) ? $metaRegroups[ $dat['name'] ] : 't:STRING_META_URL_SEO_INFORMATION'; 
			
			$ret['metas_' . $key] = \Core\Forms\FormElement::Factory($type, $dat);
		}
		
		return $ret;
	}
	
	/*
	public function render($key){
		if($key == 'site'){
			$s = $this->get('site');
			if($s == -1){
				return 'Global';
			}
			elseif($s == 0){
				return 'Root-Only';
			}
			else{
				return 'Local (' . $s . ')';
			}
		}
		elseif($key == 'title'){
			$p = $this->getParent();
			return ($p ? $p->get('title') . ' &raquo;' : '') . $this->get('title');
		}
		elseif($key == 'expires'){
			$e = $this->get('expires');
			
			if($e == 0){
				return t('STRING_DISABLED');
			}
			else{
				return \Core\time_duration_format($e);
			}
		}
		elseif($key == 'created' || $key == 'updated'){
			return \Core\Date\DateTime::FormatString($this->get($key), 'SD');
		}
		elseif($key == 'status'){
			return $this->getPublishedStatus();
		}
		elseif($key == 'published'){
			$d = $this->get('published');
			if($d){
				return \Core\Date\DateTime::FormatString($d, 'SD');
			}
			else{
				return t('STRING_NOT_PUBLISHED');
			}
		}
		elseif($key == 'published_expires'){
			$d = $this->get('published_expires');
			if($d){
				return \Core\Date\DateTime::FormatString($d, 'SD');
			}
			else{
				return t('STRING_NO_EXPIRATION');
			}
		}
		elseif($key == 'seotitle'){
			return $this->getSEOTitle();
		}
		elseif($key == 'teaser'){
			return $this->getTeaser();
		}
		else{
			return parent::render($key);
		}
	}
	*/

	public function set(string $k, $v){
		if($k == 'site'){
			// I need to run through and ensure that insertables remain linked up to the same site ID as their parent page.
			// This is a bit atypical on page creation because when the insertable is created, the page doesn't exist yet,
			// and potentially has an invalid site ID, that is only corrected after the insertables are assigned.
			//
			// Please note, this only applies when in multi-site mode.
			$insertables = $this->getLink('Insertable');

			foreach($insertables as $ins){
				/** @var $ins InsertableModel */
				$ins->set('site', $v);
			}
		}

		parent::set($k, $v);
	}

	/**
	 * Set all meta data for this page
	 *
	 * @param $metaarray array Associated key/value paired array of data to set.
	 *
	 * @return bool
	 */
	public function setMetas($metaarray) {
		if (is_array($metaarray) && count($metaarray)){
			foreach($metaarray as $k => $v){
				$this->setMeta($k, $v);
			}
			return true;
		}

		return false;
	}

	/**
	 * Set a specific meta property or name for this page.
	 *
	 * @param $name string
	 * @param $value string|array
	 */
	public function setMeta($name, $value) {
		// Get, all of the metas for this model!
		$metas = $this->getLink('PageMeta');

		// keywords behave slightly differently here.
		if($name == 'keywords'){
			if(!is_array($value)) $value = array($value => $value);

			// I need to make sure that each value is a key/value pair.
			foreach($value as $valueidx => $valueval){
				if(is_numeric($valueidx)){
					// This will replace any numeric based key with the url version of the value.
					// This may have an odd side effect of transposing numeric values like 2013,
					// but since the url version of a number is just the number itself, it should be ok.
					unset($value[$valueidx]);
					$value[ \Core\str_to_url($valueval) ] = $valueval;
				}
			}

			foreach($metas as $idx => $meta){
				/** @var $meta PageMetaModel */

				// I'm only interested in these!
				if($meta->get('meta_key') != 'keyword') continue;

				if(isset($value[ $meta->get('meta_value') ])){
					// Yay, update the value title
					$meta->set('meta_value_title', $value[ $meta->get('meta_value') ]);
					unset($value[ $meta->get('meta_value') ]);
				}
				else{
					// Nope?  Delete it!
					$this->deleteLink($meta);
				}
			}

			// Any new incoming keywords left?
			foreach($value as $metavalue => $metavaluetitle){
				if(!$metavaluetitle) continue;

				$meta = new PageMetaModel($this->get('site'), $this->get('baseurl'), 'keyword', $metavalue);
				$meta->set('meta_value_title', $metavaluetitle);

				// And append it so it'll get saved on save!
				$this->setLink('PageMeta', $meta);
			}
		}
		elseif($name == 'authorid'){
			// This affects the author tag instead, look for that!
			foreach($metas as $idx => $meta){
				/** @var $meta PageMetaModel */

				// I'm only interested in this one!
				if($meta->get('meta_key') != 'author') continue;

				// It must be the one I'm looking for.
				$meta->set('meta_value', $value);
				return; // :)
			}

			// Doesn't exist?
			$meta = new PageMetaModel($this->get('baseurl'), 'author', $value);

			// And append it so it'll get saved on save!
			$this->setLink('PageMeta', $meta);
		}
		// Default action, one to one!
		else{
			// Look for this key to see if it exists.
			foreach($metas as $idx => $meta){
				/** @var $meta PageMetaModel */

				// I'm only interested in this one!
				if($meta->get('meta_key') != $name) continue;

				// It must be the one I'm looking for.
				if($value){
					// Does it have a value?
					$meta->set('meta_value_title', $value);
				}
				else{
					// It's blank but exists... I can fix that :p
					$this->deleteLink($meta);
				}
				return; // :)
			}

			// Doesn't exist?
			if($value !== null){
				$meta = new PageMetaModel($this->get('site'), $this->get('baseurl'), $name, '');
				$meta->set('meta_value_title', $value);

				// And append it so it'll get saved on save!
				$this->setLink('PageMeta', $meta);
			}
		}
	}

	/**
	 * Set the insertable's value for this page.
	 * Will automatically create the InsertableModel and attach it to this page if it doesn't exist.
	 *
	 * Please note, will NOT save the models automatically.
	 *
	 * @param $name
	 * @param $value
	 */
	public function setInsertable($name, $value){
		// Get, all of the insertables for this model!
		$insertables = $this->getLink('Insertable');


		if(preg_match('#[a-zA-Z_][a-zA-Z_0-9]*\[[a-zA-Z_][a-zA-Z_0-9]*\]#', $name)) {
			// This represents an insertable group, convert the values to JSON before saving.
			$value = json_encode($value);
		}

		foreach($insertables as $ins){
			/** @var $ins InsertableModel */

			// Look for this key to see if it exists.

			// I'm only interested in this one!
			if($ins->get('name') == $name){
				$ins->set('site', $this->get('site'));
				$ins->set('value', $value);
				return; // :)
			}
		}

		// Doesn't exist?
		$ins = new InsertableModel($this->get('site'), $this->get('baseurl'), $name);
		$ins->set('value', $value);

		// And append it so it'll get saved on save!
		$this->setLink('Insertable', $ins);
	}

	/**
	 * Get the rewrite URLs of this page as a "\n" separated string, suitable for forms.
	 *
	 * @return string
	 */
	public function getRewriteURLs(){
		$rewrites = $this->getLink('RewriteMap');

		$out = '';
		foreach($rewrites as $r){
			/** @var $r RewriteMapModel */
			$v = $r->get('rewriteurl');
			if($v[0] == '/') $out .= ROOT_URL . substr($v, 1) . "\n";
			else $out .= $v . "\n";
		}

		return trim($out);
	}

	/**
	 * Get the primary rewrite URL for this page, formatted as a link in HTML mode
	 *
	 * @param string|null $format
	 * @return string
	 */
	public function getRewriteURL(string $format = null) : string {
		if($format === 'text/html') {
			return '<a href="' . $this->getResolvedURL() . '" title="' . htmlentities($this->getTitle()) . '">' . $this->rewriteurl . '</a>';
		}
		else {
			return $this->rewriteurl;
		}
	}

	/**
	 * Set the available rewrites for this page.
	 * Will automatically create the RewriteMapModel and attach it to this page if it doesn't exist.
	 *
	 * Please note, will NOT save the models automatically.
	 *
	 * @param string|array $urls newline, comma, or pipe delimited string of urls, also arrays are accepted.
	 */
	public function setRewriteURLs($urls){
		// make sure the URLs are in a format I can actually use.
		if(!is_array($urls)){
			$string = $urls;
			$urls = array();

			// Convert the various accepted delimiters to a standard single one so I can explode easily.
			$string = str_replace([',', '|', "\r"], "\n", $string);
			$urls = array_map('trim', explode("\n", $string));
		}

		// Next step now that I have an array, I need to go through each one and standardize it.
		// ie: if ROOT_URL is present, I need to trim that off since it's redundant.
		// Same thing for any other URL in fact....
		// The mapping doesn't care about the hostname.
		foreach($urls as $k => $v){
			if(!$v){
				unset($urls[$k]);
			}
			elseif(strpos($v, ROOT_URL_NOSSL) === 0){
				$urls[$k] = '/' . substr($v, strlen(ROOT_URL_NOSSL));
			}
			elseif(strpos($v, ROOT_URL_SSL) === 0){
				$urls[$k] = '/' . substr($v, strlen(ROOT_URL_SSL));
			}
			elseif(strpos($v, '://') !== false){
				// Trim whatever is before the ://
				$v = substr($v, strpos($v, '://') + 3);
				// Trim whatever is left after the first '/'.t
				$urls[$k] = substr($v, strpos($v, '/'));
			}
			else{
				// It just needs to start with a '/'.
				if($v[0] != '/'){
					$urls[$k] = '/' . $v;
				}
			}
		}


		// Get, all of the rewrites for this model currently.
		// This is because if there is a deletion, I want to be able to delete the old link.
		$rewrites = $this->getLink('RewriteMap');

		// Queue up any that need to be deleted first.
		foreach($rewrites as $rewrite){
			/** @var $rewrite RewriteMapModel */
			if(!in_array($rewrite->get('rewriteurl'), $urls)){
				$this->deleteLink($rewrite);
			}
		}

		// Now I can add any new one.

		foreach($urls as $url){
			// Since RewriteURLs can be set from anywhere and they're pretty in-flux,
			// I need to lookup this url and force it to be what I need it to be.
			$rewrite = RewriteMapModel::Find(['rewriteurl = ' . $url], 1);
			if(!$rewrite){
				$rewrite = new RewriteMapModel();
				$rewrite->set('rewriteurl', $url);
			}

			$this->setLink('RewriteMap', $rewrite);
		}
	}

	/**
	 * Set properties on this model from a form object, optionally with a specific prefix.
	 *
	 * @param Form        $form   Form object to pull data from
	 * @param string|null $prefix Prefix that all keys should be matched to, (optional)
	 */
	public function setFromForm(\Core\Forms\Form $form, $prefix = null){

		// Pull the meta and insertables to try to minimize data loss when changing the site id.
		$this->getLink('Insertable');
		$this->getLink('PageMeta');

		// This will take care of all the standard elements.
		parent::setFromForm($form, $prefix);

		if($form->getElement($prefix . '[rewrites]')){
			// And this will take care of the rewrites
			$rewrites = $form->getElement($prefix . '[rewrites]')->get('value');
			$this->setRewriteURLs($rewrites);
		}

		// And this will take care of the meta elements.
		$baselen = strlen($prefix . '[metas]');
		foreach($form->getElementsByName($prefix . '\[metas\]') as $el) {
			/** @var FormElement $el */
			$name = $el->get('name');
			$key = substr($name, $baselen+1, -1);

			$this->setMeta($key, $el->get('value'));
		}

		// And this will take care of the insertable elements.
		$baselen = strlen($prefix . '[insertables]');
		foreach($form->getElementsByName($prefix . '\[insertables\]') as $el) {
			$name = $el->get('name');
			// Trim the prefix along with insertables off the beginning.
			$key = substr($name, $baselen+1);

			if(preg_match('#\[\]$#', $key)) {
				// If the key ends with [], it's a multiple key, trim those off.
				$key = substr($key, 0, -2);
			}

			if(preg_match('#^[a-zA-Z0-9_]*\]$#', $key)) {
				// If the key ends with ] due to not being completely trimmed; fix that.
				$key = substr($key, 0, -1);
			}

			if(strpos($key, '__INSERTABLEGROUPTEMPLATE') !== false) {
				// The page system creates insertable templates for grouped inserts.
				// These are handled by the dedicated insertablegroup element, so they can be safely skipped.
				continue;
			}

			$this->setInsertable($key, $el->get('value'));
		}
	}

	public function setToFormElement($key, \Core\Forms\FormElement $element){
		if($key == 'page_template'){
			// Make sure to set the element's templatename.
			$element->set('templatename', $this->getBaseTemplateName());
		}
	}

	public function getResolvedURL() {
		if(strpos($this->get('baseurl'), '://') !== false){
			return $this->get('baseurl');
		}

		// If enterprise // multisite mode is enabled and this page model is NOT the current site...
		// I need to lookup THAT site's root url and use that instead.
		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			if($this->get('site') == -1){
				// "-1" pages exist on every site as a relative page.
				$base = ROOT_URL;
			}
			elseif($this->get('site') != MultiSiteHelper::GetCurrentSiteID()){
				// The site isn't the same as the current.... I need to do a little bit of work.
				// Please note, this only happens about 0.5% of the time.
				// primarily in the sitemap plugin :/
				$site = MultiSiteModel::Construct($this->get('site'));
				$base = 'http://' . $site->get('url') . '/';
			}
			else{
				// Who cares in this case, just use the root url.
				$base = ROOT_URL;
			}
		}
		else{
			// Enterprise/MM isn't enabled... no biggie!
			$base = ROOT_URL;
		}

		return $base . substr($this->get('rewriteurl'), 1);
	}

	public function save($defer = false) {
		// Ensure some helper variables are set.
		if (!$this->rewriteurl) {
			$this->rewriteurl = $this->baseurl;
		}


		try {
			$rewriteColumn = $this->getColumn('rewriteurl');

			if($rewriteColumn->changed()) {
				// If the rewrite URL was changed, I need to invalidate the cache.
				// This is because many components that may change the url, will immediately want to reload to that new url.

				// Append this page to the rewrite cache instead of invalidating the entire thing to save a few DB queries.
				$siteid  = ($this->site == -1) ? '_GLOBAL_' : $this->site;
				self::$_RewriteCache[$siteid][ $this->rewriteurl ] = $this->baseurl;

				// And set the authoritative rewrite URL to the new one.
				self::$_BaseCache[$siteid][ $this->baseurl ] = $this->rewriteurl;

				if($this->checkExists()) {
					// If this model existed before and the URL has changed, update the lookup table!
					// This will act as a basis of rewrite rules for changed URLs, allowing users to change
					// their pages rewriteurls without adversely affecting inbounding links.

					$map = new RewriteMapModel($this->getOriginalValue('rewriteurl'));
					$map->site = $this->site;
					$map->baseurl = $this->baseurl;
					$map->fuzzy = $this->fuzzy;
					$map->save();
				}
			}
		}
		catch(Exception $ex) {
			// Log and continue; this is not a critical step.
			\Core\log_exception($ex);
		}

		// Make sure the page's published date is correct.
		if($this->get('published_status') == 'published' && !$this->get('published')){
			// If this is set to published, but the date hasn't been set yet, (ie: the user just didn't fill in a date),
			// set the published date to right now!
			$this->set('published', \Core\Date\DateTime::NowGMT());
		}
		elseif($this->get('published_status') == 'draft'){
			// Draft pages are not allowed to have a published date at all.
			// This doesn't have any technical reason, simply to keep the data clean.
			$this->set('published', 0);
		}

		// Update this page's popularity score, just for freshness.
		$this->set('popularity', $this->getPopularityScore());

		return parent::save($defer);
	}

	/**
	 * Get the immediate parent page of this page, based on its parenturl.
	 *
	 * Will return null if this page has no parent.
	 *
	 * @return PageModel|null
	 */
	public function getParent(){

		if(!$this->exists()){
			return null;
		}

		$tree = $this->getParentTree();
		if(!sizeof($tree)){
			return null;
		}

		$last = sizeof($tree) - 1;

		return $tree[$last];
	}

	/**
	 * Get the listing of parent pages up from this page
	 *
	 * As of Core 7.x, this will NO LONGER prepend /admin automatically!
	 *
	 * @return array
	 */
	public function getParentTree() : array {
		return $this->_getParentTree();
	}

	/**
	 * Calculate the popularity score of this page based on views and age.
	 *
	 * This will actually perform the calculation, not just use the cached version.
	 *
	 * @return float
	 */
	public function getPopularityScore(){
		$score = $this->get('pageviews');
		$created = $this->get('published');

		if(!$created){
			// If this page is not published yet, then it shouldn't have a score.
			return 0.000;
		}

		// Pages that are not indexable never have a score either.
		// This helps keep down on the admin pages being updated.
		if(!$this->get('indexable')){
			return 0.000;
		}

		if($score == 0){
			return 0.000;
		}

		// log(10)   = 1
		// log(100)  = 2
		// log(1000) = 3
		$order = log10($score);

		// Number of seconds that have elapsed since creation and now.
		$seconds = time() - $created;

		// Convert the seconds age to months.  Each month that lapses is another magnitude required to remain popular.
		// (scratch that, 57, [or two months], as a date basis is better.)
		$months = $seconds / SECONDS_TWO_MONTH;

		// If the age is less than a week, then average the age up to a week to pad new pages a little.
		// This helps prevent new pages from severely out-ranking actually-popular pages.
		$months = max($months, 0.5);

		// This effectively makes it so
		// 10 views     now         = 0.5 (with the above skewing)
		// 32 views     now         = 1.
		// 100 views    in 2 months = 1.
		// 1000 views   in 4 months = 1.
		// 10000 views  in 6 months = 1.
		// 100000 views in 8 months = 1.
		$long_number = $order - $months;

		// Increment score by 10 to help keep them above non-ranked pages.
		$long_number += 10;

		return round($long_number, 5);
	}
	
	/**
	 * Get the title of this page, (with automatic i18n translation)
	 *
	 * @return string
	 */
	public function getTitle() : string {
		$t = $this->get('title');
		
		if(strpos($t, 't:') === 0){
			return t(substr($t, 2));
		}
		else{
			return $t ?: '';
		}
	}

	/**
	 * Get the *automatic* SEO title for this page.
	 *
	 * This can be overridden by setting the meta title attribute explicitly!
	 * 
	 * Uses the ViewMeta\Title class for rendering the value, and will use the base page's title value if not set
	 *
	 * @return string
	 */
	public function getSEOTitle() : string {
		$meta = $this->getMeta('title');
		
		if(!$meta){
			$meta = new Title();
			$meta->setContent($this->get('title'));
		}
		
		return $meta->__toString();
	}

	/**
	 * Get the teaser of this page, aka meta description
	 *
	 * @param boolean $require_something Set to true if you want to require *something* to be returned.
	 * @return string
	 */
	public function getTeaser($require_something = false){
		$meta = $this->getMeta('description');
		$config = ConfigHandler::Get('/core/page/teaser_template');

		if($meta){
			// An explicit meta tag always overrides any template settings.
			return $meta->get('meta_value_title');
		}
		elseif($config){
			// Next up is the configuration template.
			return $this->_parseTemplateString($config);
		}
		elseif($require_something){
			// Return the body text.
			return substr(strip_tags($this->get('body')), 0, 150);
		}
		else{
			return '';
		}
	}
	/**
	 * Get the image object or null
	 *
	 * @return Core\Filestore\File|null
	 */
	public function getImage(){
		$meta = $this->getMeta('image');
		if(!$meta){
			return null;
		}

		$file = $meta->get('meta_value_title');

		if(!$file){
			return null;
		}
		$f = \Core\Filestore\Factory::File($file);
		return $f;
	}

	/**
	 * Get the image object or null
	 *
	 * @return UserModel|null
	 */
	public function getAuthor(){
		$meta = $this->getMeta('author');
		if(!$meta) return null;

		$uid = $meta->get('meta_value');

		if(!$uid) return null;

		$u = UserModel::Construct($uid);
		return $u;
	}

	/**
	 * Get the cache key for this page's index that is acceptable for use on caching systems.
	 *
	 * @return string
	 */
	public function getIndexCacheKey(){
		return 'page-cache-index-' . $this->get('site') . '-' . md5($this->get('baseurl'));
	}
	
	public function getControlLinks() : array {
		$accessAdmin = \Core\user()->checkAccess('g:admin');
		$accessView = \Core\user()->checkAccess($this->get('access'));
		$accessWidgets = \Core\user()->checkAccess('p:/core/widgets/manage');
		$baseurl = $this->get('baseurl');
		$ret = [];

		$view     = Router::$DefaultView;
		$page     = Router::$DefaultPageModel;

		
		if($accessView) {
			$ret[] = [
				'title' => 't:STRING_VIEW',
				'icon' => 'view',
				'link' => $baseurl,
			];
		}
		
		if($accessAdmin) {
			if($this->editurl) {
				$ret[] = [
					'title' => 't:STRING_EDIT',
					'icon' => 'edit',
					'link' => $this->editurl,
					'group'   => \Core\ViewControl::GROUP_MANAGE,
				];
			}
			
			switch($this->getPublishedStatus()){
				case 'draft':
					$ret[] = [
						'title'   => 't:STRING_PUBLISH_PAGE',
						'icon'    => 'thumbs-up',
						'link'    => '/admin/page/publish?baseurl=' . $baseurl,
						'group'   => \Core\ViewControl::GROUP_MANAGE,
						'confirm' => '',
					];
					break;
				case 'expired':
					$ret[] = [
						'title'   => 't:STRING_REPUBLISH_PAGE',
						'icon'    => 'thumbs-up',
						'link'    => '/admin/page/publish?baseurl=' . $baseurl,
						'confirm' => '',
						'group'   => \Core\ViewControl::GROUP_MANAGE,
					];
					break;
				case 'published':
					$ret[] = [
						'title'   => 't:STRING_UNPUBLISH_PAGE',
						'icon'    => 'thumbs-down',
						'link'    => '/admin/page/unpublish?baseurl=' . $baseurl,
						'confirm' => '',
						'group'   => \Core\ViewControl::GROUP_MANAGE,
					];
					break;
			}
			
			if($this->deleteurl) {
				$ret[] = [
					'title'   => 't:STRING_DELETE',
					'icon'    => 'times',
					'link'    => $this->deleteurl,
					'group'   => \Core\ViewControl::GROUP_ADMIN,
					'confirm' => 't:MESSAGE_ASK_COMPLETEY_DELETE_PAGE',
				];
			}
		}

		if($accessWidgets) {
			$selected = $this->theme_template;
			if(!$selected) {
				$selected = ConfigHandler::Get('/theme/default_template');
			}
			$selected = '/skins/' . $selected;
			$ret[] = [
				'title' => 't:STRING_SKIN_WIDGETS',
				'link' => '/widget/admin?template=' . $selected,
				'icon' => 'cubes',
				'group'   => \Core\ViewControl::GROUP_ADMIN,
			];
		}
		
		$parent = parent::getControlLinks();
		
		return array_merge($ret, $parent);
	}

	/**
	 * Get a textual representation of this Model as a flat string.
	 *
	 * Used by the search systems to index the model, (or multiple models into one).
	 * 
	 * This is the PageModel specific version as Pages behave differently than most other systems.
	 *
	 * @return string
	 */
	public function getSearchIndexString() : string {
		// The default behaviour is to sift through the records on this model itself.
		$strs = [];
		
		// First string to pull in are the various URLs!
		$strs[] = $this->getResolvedURL();
		$strs[] = $this->getRewriteURLs();
		
		// Title is important.
		$strs[] = $this->get('title');
		// As is the SEO title
		$strs[] = $this->getSEOTitle();
		
		// Gimme some metadata.
		$strs[] = $this->getTeaser(true);
		
		// Lastly the body!
		// This is probably going to be HTML, (one would hope at least!)
		// So convert it to text before saving.
		// Otherwise there would be a million "p" entries in the string!
		$body = $this->get('body');
		$converter = new Converter();
		$strs[] = $converter->convert($body);

		return implode(' ', $strs);
	}

	/**
	 * Get a string representation of the published status of this article.
	 * This is primarily an internal and administrative function, since from the user's perspective,
	 * pages are simply available or not.
	 *
	 * @return string
	 */
	public function getPublishedStatus() {
		if($this->get('published_status') == 'draft'){
			// The page is set as "draft".  Not published.
			return 'draft';
		}

		if($this->get('published') > \Core\Date\DateTime::NowGMT()){
			// The publish date is in the future. Not published.
			return 'pending';
		}

		if($this->get('published_expires') && $this->get('published_expires') <= \Core\Date\DateTime::NowGMT()){
			// The publish expire date is set but that date has already passed.  Not published.
			return 'expired';
		}

		return 'published';
	}

	/**
	 * Get all pages on the system that are not "this page", useful for the parent select.
	 * 
	 * It's public because it's called by the Form system via the model schema defined above.
	 * 
	 * @return Dict
	 */
	public function getParentsAsOptions() : \Core\Dict {
		$f = new ModelFactory('PageModel');
		if ($this->get('baseurl')){
			$f->where('baseurl != ' . $this->get('baseurl'));
		}
		$opts = PageModel::GetPagesAsOptions($f, 't:STRING_NO_PARENT_PAGE');

		return $opts;
	}

	/**
	 * Get if this page is published and the published date is at least now or earlier.
	 *
	 * @return bool
	 */
	public function isPublished(){
		return ($this->getPublishedStatus() === 'published');
	}

	/**
	 * Purge the entire page cache for this given page.
	 */
	public function purgePageCache(){
		$indexkey = $this->getIndexCacheKey();
		$index = \Core\Cache::Get($indexkey);
		if($index && is_array($index)){
			foreach($index as $key){
				\Core\Cache::Delete($key);
			}
		}
		\Core\Cache::Delete($indexkey);
	}

	/**
	 * Get the parent tree for this page
	 *
	 * As of Core 7.x, this will NO LONGER prepend /admin automatically!
	 *
	 * @param int $antiinfiniteloopcounter
	 * @return array
	 */
	private function _getParentTree($antiinfiniteloopcounter = 6) : array {
		if ($antiinfiniteloopcounter <= 0){
			return [];
		}
		
		$p = false;
//echo "Running _getParentTree for " . $this->get('baseurl') . '<br/>';
//echo '<pre>'; debug_print_backtrace();
		
//echo '<pre>'; debug_print_backtrace(); var_dump($this); die();

		if($this->get('parenturl')){
			$parentURL = $this->get('parenturl');
		}
		else{
			// No parent URL available, simply return nothing.
			return [];
		}
		
		$parent = PageModel::Construct($parentURL);
		
		return array_merge($parent->_getParentTree(--$antiinfiniteloopcounter), [ $parent ]);
	}

	private function _populateView() {
		// Transpose some useful data for it.
		$this->_view->error = View::ERROR_NOERROR;
		$this->_view->baseurl = $this->get('baseurl');
		$this->_view->setParameters($this->getParameters());
		$this->_view->templatename = $this->getTemplateName();
		$this->_view->mastertemplate = ($this->get('template')) ? $this->get('template') : ConfigHandler::Get('/theme/default_template');

		$this->_view->setBreadcrumbs($this->getParentTree());
	}

	/**
	 * Parse one of the SEO template options that are configurable and return the result.
	 *
	 * @param $string_template
	 *
	 * @return string
	 */
	private function _parseTemplateString($string_template){

		$parent = $this->getParent();

		$metadescription = $this->getMeta('description');
		$bodysnippet = substr(strip_tags($this->get('body')), 0, 150);
		$author = $this->getAuthor();

		$rep = [
			//Replaced with the published date of the page
			'%%date%%' => \Core\Date\DateTime::FormatString($this->get('published'), \Core\Date\DateTime::SHORTDATE),
			//Replaced with the title of the page
			'%%title%%' => $this->getTitle(),
			//Replaced with the title of the parent page of the current page
			'%%parent_title%%' => ($parent ? $parent->getTitle() : ''),
			//The site's name
			'%%sitename%%' => SITENAME,
			//Replaced with the page excerpt (or auto-generated if it does not exist)
			'%%excerpt%%' => ($metadescription ? $metadescription->get('meta_value_title') : $bodysnippet),
			//Replaced with the current tag/tags
			'%%tag%%' => '',
			//Replaced with the current search phrase
			'%%searchphrase%%' => '',
			//Replaced with the page modified time
			'%%modified%%' => \Core\Date\DateTime::FormatString($this->get('updated'), \Core\Date\DateTime::SHORTDATE),
			//Replaced with the page author's username
			'%%name%%' => ($author ? $author->getDisplayName() : ''),
			//Replaced with the current time
			'%%currenttime%%' => \Core\Date\DateTime::Now(\Core\Date\DateTime::TIME),
			//Replaced with the current date
			'%%currentdate%%' => \Core\Date\DateTime::Now(\Core\Date\DateTime::SHORTDATE),
			//Replaced with the current day
			'%%currentday%%' => \Core\Date\DateTime::Now('d'),
			//Replaced with the current month
			'%%currentmonth%%' => \Core\Date\DateTime::Now('m'),
			//Replaced with the current year
			'%%currentyear%%' => \Core\Date\DateTime::Now('Y'),
			//Replaced with the current page number (i.e. page 2 of 4)
			'%%page%%' => '1', // @todo Support for this.
			//Replaced with the current page total
			'%%pagetotal%%' => '1', // @todo Support for this.
		];

		return str_ireplace(array_keys($rep), array_values($rep), $string_template);
	}


	/****************** Helper Static Functions **************************/

	/**
	 * Split a base url into its corresponding parts, controller method and parameters.
	 * Also supports the rewriteurl.
	 *
	 * @param string   $base The URL base to lookup, aka baseurl.
	 * @param int|null $site The Site ID to lookup, only functional in multi-site mode.
	 *
	 * @return array
	 */
	public static function SplitBaseURL($base, $site = null) {

		// Example of incoming data at this point:
		// "/"
		// ""
		// "/repo.xml.gz"
		// "/controllerfoo/methodblah/param1/param2?qstr=foo"

		if (!$base) return null;

		// Support additional arguments
		// This should be done prior to checking for aliases and what not because
		// the query string parameters should NOT effect the rewrite URL and aliases.
		$args = null;
		$argstring = '';
		if (($qpos = strpos($base, '?')) !== false) {
			$argstring = urldecode(substr($base, $qpos + 1));
			preg_match_all('/([^=&]*)={0,1}([^&]*)/', $argstring, $matches);
			$args = array();
			foreach ($matches[1] as $idx => $v) {
				if (!$v){
					// This preg_match_all creates duplicate keys when there are multiple, so a=a&b=b will produce
					// [ 'a', '', 'b', '' ]
					// for some reason.
					// This continue is a quick hack to ignore that erroneous blank key.
					continue;
				}
				$a =& $args;
				while(($paranpos = strpos($v, '[')) !== false){
					$k1 = strtolower(substr($v, 0, $paranpos));
					$v = substr($v, $paranpos+1, strpos($v, ']')-$paranpos-1);

					if(!isset($a[$k1])){
						$a[$k1] = [];
					}
					$a =& $a[$k1];
				}
				$a[strtolower($v)] = $matches[2][$idx];
			}
			$base = substr($base, 0, $qpos);
		}

		// Example of incoming data at this point:
		// "/"
		//     base: "/"
		//     args: NULL
		// ""
		//     RETURNED NULL
		// "/repo.xml.gz"
		//     base: "/repo.xml.gz"
		//     args: NULL
		// "/controllerfoo/methodblah/param1/param2?qstr=foo"
		//     base: "/controllerfoo/methodblah/param1/param2"
		//     args: [qstr] => foo


		// Content Types should not dictate the controller and method either!
		$ext = 'html';
		$posofdot = strpos($base, '.');
		if($posofdot){
			// Or, it can also be the first part up until the first '.'.
			// Also if there was a dot at this level, then the URL must be something like
			// foo.ext
			// This needs to be remapped to FooController->index with the extension of 'ext'.
			$ext  = substr($base, $posofdot+1);
			$base = substr($base, 0, $posofdot);
		}

		$ctype = \Core\Filestore\extension_to_mimetype($ext);
		// Invalid mimetype?  Default to an HTML file.
		if(!$ctype){
			$ctype = 'text/html';
		}

		// Example of incoming data at this point:
		// "/"
		//     base:  "/"
		//     args:  NULL
		//     ext:   'html'
		//     ctype: 'text/html'
		//
		// ""
		//     RETURNED NULL
		//
		// "/repo.xml.gz"
		//     base:  "/repo"
		//     args:  NULL
		//     ext:   xml.gz
		//     ctype: 'application/octet-stream'
		//
		// "/controllerfoo/methodblah/param1/param2?qstr=foo"
		//     base:  "/controllerfoo/methodblah/param1/param2"
		//     args:  [qstr] => foo
		//     ext:   'html'
		//     ctype: 'text/html'


		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			if($site === null){
				$site = MultiSiteHelper::GetCurrentSiteID();
			}
		}
		else{
			$site = null;
		}

		// This will ensure that the cache is updated and that any rewrite URL is remapped to its baseurl if possible.
		$lookup = self::_LookupUrl($base, $site);

		if($lookup['found']){
			$base = $lookup['url'];
		}
		else{
			$try = $lookup['url'];

			if($site === null){
				$tries = [];
				if(isset(self::$_FuzzyCache) && is_array(self::$_FuzzyCache)){
					foreach(self::$_FuzzyCache as $dat){
						$tries = array_merge($tries, $dat);
					}
				}
			}
			elseif(isset(self::$_FuzzyCache[$site])){
				$tries = array_merge(self::$_FuzzyCache['_GLOBAL_'], self::$_FuzzyCache[$site]);
			}
			else{
				$tries = self::$_FuzzyCache['_GLOBAL_'];
			}

			while($try != '' && $try != '/') {
				if(isset($tries[$try])) {
					// The fuzzy page must have the requested arguments, they just need to be tacked onto the end of the base.
					$base = $tries[$try] . substr($base, strlen($try));
					break;
				}
				elseif(in_array($try, $tries)) {
					$base = $tries[array_search($try, $tries)] . substr($base, strlen($try));
					break;
				}
				$try = substr($try, 0, strrpos($try, '/'));
			}

			// This is a weird exception for when the root page is set to fuzzy.
			// It's not checked in the above loop because it thinks the check is over.
			//if(!$fuzzyfound && isset(self::$_FuzzyCache['/'])){
			//	$base = self::$_FuzzyCache['/'] . $base;
			//}
		}

		// Trim off both beginning and trailing slashes.
		$base = trim($base, '/');

		// Logic for the Controller.
		$posofslash = strpos($base, '/');


		if ($posofslash){
			// The controller is usually the first part up until the first '/'.
			$controller = substr($base, 0, $posofslash);
			$base = substr($base, $posofslash+1);
		}
		else{
			$controller = $base;
			$base = false;
		}

		// Preferred way of handling controller names.
		if (class_exists($controller . 'Controller')) {
			switch (true) {
				case is_subclass_of($controller . 'Controller', 'Controller'):
					$controller = $controller . 'Controller';
					break;
				default:
					// Not a valid controller
					return null;
			}
		}
		// Not quite preferred way, but still works.
		elseif (class_exists($controller)) {
			if(!
				(is_subclass_of($controller, 'Controller'))
			){
				return null;
			}
		}
		else {
			// Not even found!
			return null;
		}


		// Logic for the Method.
		//if(substr_count($base, '/') >= 1){
		if ($base) {

			$posofslash = strpos($base, '/');

			// The method can be extended.
			// This means that a method can be in the format of Sites/Edit, which should resolve to Sites_Edit.
			// This only takes effect if the method exists on the controller.
			if ($posofslash) {
				$method = str_replace('/', '_', $base);
				while (!method_exists($controller, $method) && strpos($method, '_')) {
					$method = substr($method, 0, strrpos($method, '_'));
				}
			}
			else {
				$method = $base;
			}

			// Now trim the base again based on the length of the method.
			$base = substr($base, strlen($method) + 1);
		}
		else {
			// The controller may have an "Index" controller.  That doesn't need to be explictly called.
			$method = 'index';
		}

		// One last check that the method exists, (because there's only 1 scenario that checks above)
		if (!method_exists($controller, $method)) {
			return null;
		}


		// Provide some logic for security.
		// Keep any method starting with a '_' private by preventing
		// direct access from the browser.
		if ($method[0] == '_') return null;

		// Logic for the parameters.
		$params = ($base !== false) ? explode('/', $base) : null;


		// Build these onto a base for a standardized callable URL.
		$baseurl = '/' . ((strpos($controller, 'Controller') == strlen($controller) - 10) ? substr($controller, 0, -10) : $controller);
		// No need to add a method if it's the index.
		if (!($method == 'index' && !$params)) $baseurl .= '/' . str_replace('_', '/', $method);
		$baseurl .= ($params) ? '/' . implode('/', $params) : '';


		// Rewrite URL may be useful too!
		$rewriteurl = self::_LookupReverseUrl($baseurl, $site);


		// Keep the original mimetype extension if set.
		if($ext != 'html'){
			//$rewriteurl .= '.' . \Core\Filestore\mimetype_to_extension($ctype);
			// Some binary types of data just return a generic "application/octet-stream",
			// which is useless for figuring out which extension it original was.
			$rewriteurl .= '.' . $ext;
		}

		// Keep the arguments on the rewrite version.
		if ($args) {
			$rewriteurl .= '?' . $argstring;
			if ($params) $params = array_merge($params, $args);
			else $params = $args;
		}

		if($site === null){
			$rooturl = ROOT_URL;
		}
		else{
			$rooturl = MultiSiteModel::Construct($site)->getResolvedURL();
		}

		// Reassemble everything too!
		$fullurl = trim($rooturl, '/') . '/' . trim($rewriteurl, '/');
		/*
		if($params && sizeof($params) > 0){
			foreach($params as $k => $v){
				if(is_numeric($k)){
					$fullurl .= '/' . $v;
				}
				else{
					if(strpos($fullurl, '?') === false){
						$fullurl .= '?' . $k . '=' . $v;
					}
					else{
						$fullurl .= '&' . $k . '=' . $v;
					}
				}
			}
		}
		*/
		
		
		// Load up the aliases, just in case that's needed too!
		$aliases = [];
		if($site === null){
			foreach(self::$_RewriteCache as $set){
				$aliases = array_merge($aliases, array_keys($set, $baseurl));
			}
		}
		else{
			if(isset(self::$_RewriteCache['_GLOBAL_'])){
				$aliases = array_merge($aliases, array_keys(self::$_RewriteCache['_GLOBAL_'], $baseurl));
			}
			if(isset(self::$_RewriteCache[$site])){
				$aliases = array_merge($aliases, array_keys(self::$_RewriteCache[$site], $baseurl));
			}
		}
		
		

		// Tack on the "arguments" too, these are

		return array(
			'controller' => $controller,
			'method'     => $method,
			'parameters' => $params,
			'rooturl'    => $rooturl,
			'baseurl'    => $baseurl,
			'rewriteurl' => $rewriteurl,
			'ctype'      => $ctype,
			'extension'  => $ext,
			'fullurl'    => $fullurl,
			'rewritemap' => $aliases,
		);
	}

	/**
	 * Get all pages, (with an optional where clause), as a valid option array
	 *
	 * This array contains key of "baseurl", value of "parent &raquo; title ( url )"
	 * that is directly pluggable into the Form system or a manual foreach loop.
	 *
	 * @param mixed $where Either a ModelFactory (usually with custom-crafted where clauses),
	 *                     or a string of the where clause
	 *                     or false to omit the where clause.
	 * @param mixed $blanktext The text to include with the blank entry
	 *                         If false, no blank field is included.
	 *
	 * @return Dict
	 */
	public static function GetPagesAsOptions($where = false, $blanktext = false) : Dict {
		if ($where instanceof ModelFactory) {
			$f = $where;
		}
		elseif (!$where) {
			$f = new ModelFactory('PageModel');
		}
		else {
			$f = new ModelFactory('PageModel');
			$f->where($where);
		}

		if(Core::IsComponentAvailable('multisite') && MultiSiteHelper::IsEnabled()){
			$g = new Core\Datamodel\DatasetWhereClause();
			$g->setSeparator('OR');
			$g->addWhere('site = -1');
			$g->addWhere('site = ' . MultiSiteHelper::GetCurrentSiteID());
			$f->where($g);
		}

		// They must be selectable!
		$f->where('selectable = 1');

		// Get the pages
		$pages = $f->get();

		// Now I can assemble the list of options with useful labels
		$opts = array();
		$results = new Dict();
		$results->autoI18N = true;
			
		foreach ($pages as $p) {
			$baseurl = $p->get('baseurl');

			$t = '';
			foreach ($p->getParentTree() as $subp) {
				$title = $subp->get('title');
				if(strpos($title, 't:') === 0){
					$title = t(substr($title, 2));
				}
				
				$t .= $title . ' &raquo; ';
			}
			
			$title = $p->get('title');
			if(strpos($title, 't:') === 0){
				$title = t(substr($title, 2));
			}
			
			$t .= $title;
			$t .= ' ( ' . $p->get('rewriteurl') . ' )';
			$tlen = strlen(html_entity_decode($t));
			if($tlen > 80){
				// This needs to take into account for the html characters.
				$t = substr($t, 0, (77 + (strlen($t) - $tlen)) ) . '&hellip;';
			}
			$opts[$baseurl] = $t;
		}

		// Sort'em
		asort($opts);

		// Default should always be at the top (if requested).
		if ($blanktext){
			$results->addItem('', $blanktext);
		}
		foreach($opts as $k => $v){
			$results->addItem($k, $v);
		}
		
		// And here ya go!
		return $results;
	}

	/**
	 * Update all pages' popularity ranking as part of a hook.
	 *
	 * @return bool
	 */
	public static function PopularityMassUpdateHook(){
		$pages = PageModel::Find();
		foreach($pages as $page){
			/** @var PageModel $page */
			$page->save();
			// Yup, all that is needed to be done is save the page.
			// The PageModel will auto-update the popularity ranking on saves.
		}
		return true;
	}
	
	/**
	 * Perform a model search on the records of this PageModel.
	 * 
	 * This has extra support for key:var tags, where key can be
	 * 
	 * * keyword
	 *
	 * @param string $query The base query to search
	 * @param array $where  Any additional where parameters to add onto the factory
	 *
	 * @return PageModel[]
	 */
	public static function Search(string $query, array $where = []) : array {
		$ret = [];
		$tag = null;

		// If this object does not support searching, simply return an empty array.
		$ref = new ReflectionClass(get_called_class());

		if(!$ref->getProperty('HasSearch')->getValue()){
			return $ret;
		}

		$fac = new ModelFactory(get_called_class());

		if(sizeof($where)){
			$fac->where($where);
		}

		if($ref->getProperty('HasDeleted')->getValue()){
			$fac->where('deleted = 0');
		}
		
		// Used in the relavency check, as tags do not get calculated into the base query.
		$relAdd = 0;
		
		// Check if this query has some of the advanced query strings.
		if(($pos = strpos($query, 'tag:')) !== false){
			$tag = preg_replace('/.*tag:([a-zA-Z0-9\-]*).*/', '$1', $query);
			// And drop this query from the original query.
			$query = trim(preg_replace('/tag:([a-zA-Z0-9\-]*)/', '', $query));
			
			$pageMetas = PageMetaModel::FindRaw(['meta_value = ' . $tag, 'meta_key = keyword']);
			$pageURLs = [];
			foreach($pageMetas as $row){
				$pageURLs[] = $row['baseurl'];
			}
			$fac->where('baseurl IN ' . implode(',', $pageURLs));
			$relAdd += ($query === '' ? 100 : 50);
		}

		if($query){
			$fac->search($query);
		}
		
		$ret = $fac->get();
		
		if($tag !== null) {
			// Add in the tag weight for this search
			foreach($ret as $m) {
				$m->set(
					'__search_relevancy', min($m->get('__search_relevancy') + $relAdd, 100)
				);
			}
		}
		
		// Sort the results before returning them.
		// Because otherwise, what's the point of a search algorithm?!?
		usort($ret, function($a, $b) {
			/** @var $a Core\Search\ModelResult */
			/** @var $b Core\Search\ModelResult */
			return $a->__search_relevancy < $b->__search_relevancy;
		});

		return $ret;
	}
	
	/**
	 * Helper method to get the resolved ROOT_URL for a given site
	 * 
	 * This has no function when not in MM mode.
	 * 
	 * @param string $site Site to check against
	 * @return string      Corresponding ROOT_URL
	 */
	public static function GetRootURL($site = null) : string {
		// If enterprise // multisite mode is enabled and this page model is NOT the current site...
		// I need to lookup THAT site's root url and use that instead.
		if(
			$site != -1 &&
			$site !== null &&
			Core::IsComponentAvailable('multisite') && 
			MultiSiteHelper::IsEnabled() &&
			$site != MultiSiteHelper::GetCurrentSiteID()
		){
			return 'http://' . MultiSiteModel::Construct($site)->get('url') . '/';
		}
		else{
			// Enterprise/MM isn't enabled... no biggie!
			return ROOT_URL;
		}
	}

	/**
	 * Retrieve the page routing data based on a URI
	 *
	 * This path can be a rewriteurl or pageurl.
	 *
	 * If no page is listed that matches the query, an exception is thrown.
	 *
	 * @param string $uri
	 *
	 * @throws Exception
	 *
	 * @return array
	 */
	public static function GetRoutingDataByUri(string $uri) : array {
		self::_LoadRoutingFromDB();

		if(strpos($uri, 'site:') === 0){
			// Requested URI contains a site: prefix; extract out the requested site and url components.
			$slashpos = strpos($uri, '/');
			$site = substr($uri, 5, $slashpos-5);
			$uri = substr($uri, $slashpos);
		}
		elseif(
			Core::IsComponentAvailable('multisite') &&
			MultiSiteHelper::IsEnabled()
		){
			// Site is in MM, but URL does not contain the site: prefix, defaults to the current site.
			$site = MultiSiteHelper::GetCurrentSiteID();
		}
		else{
			// Default is 0 when MM is not available.
			$site = '0';
		}


		// Search for exact matches from the base and rewrite caches
		// For both GLOBAL site and the specific site that was requested, (or current).
		if(isset(self::$_BaseCache[ 'site:' . $site . $uri ])){
			return self::$_BaseCache[ 'site:' . $site . $uri ] + ['match' => $uri];
		}
		elseif(isset(self::$_BaseCache[ 'site:-1' . $uri ])){
			return self::$_BaseCache[ 'site:-1' . $uri ] + ['match' => $uri];
		}
		elseif(isset(self::$_RewriteCache[ 'site:' . $site . $uri ])){
			return self::$_RewriteCache[ 'site:' . $site . $uri ] + ['match' => $uri];
		}
		elseif(isset(self::$_RewriteCache[ 'site:-1' . $uri ])){
			return self::$_RewriteCache[ 'site:-1' . $uri ] + ['match' => $uri];
		}


		// Try the fuzzy cache, this requires a little more work as it needs to strpos the matches.
		$try = $uri;
		do {
			$try = substr($try, 0, strrpos($try, '/'));
			if(isset(self::$_FuzzyCache[ 'site:' . $site . $try ])){
				return self::$_FuzzyCache[ 'site:' . $site . $try ] + ['match' => $try];
			}
			elseif(isset(self::$_FuzzyCache[ 'site:-1' . $try ])){
				return self::$_FuzzyCache[ 'site:-1' . $try ] + ['match' => $try];
			}

		} while($try != '');


		throw new Exception($uri . ' is not registered as a Page');

	}

	/**
	 * Search for the given PageModel based on its URI path.
	 * 
	 * This path can be a rewriteurl or pageurl.
	 * 
	 * @param string $uri
	 * 
	 * @throws Exception
	 *
	 * @return PageModel
	 */
	public static function GetByURI($uri) : PageModel{
		$dat = self::GetRoutingDataByUri($uri);
		return PageModel::Construct($dat['site'], $dat['baseurl']);
	}
	
	public static function PostViewSaveHandler() {
		$page = Router::$DefaultPageModel;
		$view = Router::$DefaultView;
		
		if(!$page) {
			// If there's no page, then nothing to save.
			// This is common with dynamic pages that do not have registered Pages.
			return true;
		}
		
		// Increase the page view count and the last template, this is used by some subsystems like Widget management.
		$page->pageviews += 1;
		$page->last_template = $view->templatename;
		
		$page->save();
		return true;
	}

	/**
	 * Retrieve results for /page/search
	 *
	 * @param string $query
	 * @return array
	 */
	public static function PageSearchHook(string $query) : array {
		$results = PageModel::Search(
			$query,
			[
				'indexable = 1',
				'baseurl != /page/sitemap'
			]
		);

		foreach($results as $idx => $model) {
			if(!$model->checkCanAccess()) {
				// Skip any page the user cannot access,
				// this will handle published status and the access string.
				unset($results[ $idx ]);
				continue;
			}
		}

		return $results;
	}
	
	/**
	 * Load the routing caches (rewrite, fuzzy, and base), into memory.
	 * 
	 * If already loaded, nothing happens.
	 * 
	 * @return void
	 */
	private static function _LoadRoutingFromDB(){
		if (self::$_RewriteCache === null) {
			try{
				$pages = \Core\Datamodel\Dataset::Init()
					->select('site', 'rewriteurl', 'baseurl', 'fuzzy')
					->table('page')
					->executeAndGet();
				
				$rewrites = \Core\Datamodel\Dataset::Init()
					->select('site', 'rewriteurl', 'baseurl', 'fuzzy')
					->table('rewrite_map')
					->executeAndGet();
			} catch (Exception $ex) {
				// If this fails, the Database probably isn't setup yet.
				// This just means that there will be no rewrite/base mapping,
				// but this system should otherwise resume to work in some capacity.
				$pages = [];
				$rewrites = [];
			}

			self::$_RewriteCache = [];
			self::$_BaseCache    = [];
			self::$_FuzzyCache   = [];
			
			foreach($pages as $row){

				// Remap the fuzzy indicator to boolean
				$fuzzy = ($row['fuzzy'] === '1' || $row['fuzzy'] === 1);

				if($row['baseurl'] !== $row['rewriteurl']){
					// Rewrites only get updated if the base URL is not the same as the rewrite URL,
					// eg: the registered page has an alias for it.
					self::$_RewriteCache[ 'site:' . $row['site'] . $row['rewriteurl'] ] = [
						'site'       => $row['site'],
						'baseurl'    => $row['baseurl'],
						'rewriteurl' => $row['rewriteurl'],
						'fuzzy'      => $fuzzy,
					];
				}
				
				self::$_BaseCache[ 'site:' . $row['site'] . $row['baseurl'] ] = [
					'site'       => $row['site'],
					'baseurl'    => $row['baseurl'],
					'rewriteurl' => $row['rewriteurl'],
					'fuzzy'      => $fuzzy,
				];

				if($row['fuzzy']) {
					self::$_FuzzyCache[ 'site:' . $row['site'] . $row['rewriteurl'] ] = [
						'site'       => $row['site'],
						'baseurl'    => $row['baseurl'],
						'rewriteurl' => $row['rewriteurl'],
						'fuzzy'      => $fuzzy,
					];
				}
			}


			foreach($rewrites as $row) {
				// This behaves simiularly to pages, but instead of remapping to the rewrite, it uses the existing page
				// data to retrieve the rewrite.
				if(
					isset(self::$_BaseCache[ 'site:' . $row['site'] . $row['baseurl'] ]) &&
					!isset(self::$_RewriteCache[ 'site:' . $row['site'] . $row['rewriteurl'] ])
				) {
					$map = self::$_BaseCache[ 'site:' . $row['site'] . $row['baseurl'] ];

					// Remap the fuzzy indicator to boolean
					$fuzzy = ($row['fuzzy'] === '1' || $row['fuzzy'] === 1);

					self::$_RewriteCache[ 'site:' . $row['site'] . $row['rewriteurl'] ] = [
						'site'       => $row['site'],
						'baseurl'    => $map['baseurl'],
						'rewriteurl' => $map['rewriteurl'],
						'fuzzy'      => $fuzzy,
					];
				}
			}
		}
	}
}
